/* -*- Mode: C++ -*-
 * $Id: Controller.h,v 1.2 2000/09/04 15:22:58 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Controller.h
 * Unitname: BasicSystem
 * $Revision: 1.2 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/04 15:22:58 $
 *
 *
 * HISTORY
 *
 * $Log: Controller.h,v $
 * Revision 1.2  2000/09/04 15:22:58  frehe
 * Added an #include statement to make the documentation better.
 *
 * Revision 1.1  2000/09/03 20:41:08  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file Controller.h
 * Include file for the class Controller.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2000/09/04 15:22:58 $
 */


#ifndef CONTROLLER_H
#define CONTROLLER_H


#include <signal.h>
#include <string>

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif


#include "basic_system_compiler_directives.h"
#include "Time.h"
#include "Timer.h"
#include "RSException.h"


RS_BEGIN_NAMESPACE


class Actuators;
class Sensors;
class ControlInterface;
class ServerInterface;
class ServerCommand;
class ControlCommand;


/** Used to represent signal handling functions. */
typedef RETSIGTYPE Sigfunc(int);


/** Represent the different results the controller can return. */
enum ControllerResult {
  CONTROLLER_OK,
  CONTROLLER_ERROR
};


/**
 * @class Controller Controller.h
 * This class is the controller of the agent.
 * It will take care of all synchronization between the different
 * parts of the agent, including sending commands to the soccer
 * server and generating events to the decision maker.
 * <!--Controller is a model of the design pattern Singleton. The only
 * instance is accessible through Controller* Controller::instance(); -->
 *
 * <!-- @invariant The class invariant is... -->
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2000/09/04 15:22:58 $
 */
class Controller {
public:
  Controller(ServerInterface* const, Sensors* const,
	     Actuators* const, ControlInterface* const,
	     const UInt16 cycle_length, const UInt16 recv_step,
	     const UInt16 sense_step, const UInt16 visual_step,
	     const UInt16 max_sigalrms, const UInt16 min_ticks,
	     const UInt16 ticks_before_cmd_warning,
	     const UInt16 max_ticks) throw();

  // Not used since the constructor is nice.
  // The only access point to the controller object,
  // since it is a singleton.
  // static Controller* instance();
  
  ~Controller() throw();
  
  bool serverAlive() const throw();

  bool init() throw(RSException);
  bool start() throw(RSException);

#ifdef RS_ENABLE_COACH_DEBUGGING
  bool startCoach(const std::string& host, const UInt16 port,
		  const UInt32 server_buffer_size) throw(RSException);
  bool sendCoachMessage(const std::string& msg) throw(RSException);
#endif
  
  bool updateTime(const UInt16 new_time) throw();
  
  bool sendCommand(ServerCommand*) throw(RSException);
  bool sendCommand(ControlCommand*) throw();
  bool receiveInput(char* const input) throw(RSException);

  void delayedActions(Int16 dashes, Int16 kicks, Int16 says,
		      Int16 turns, Int16 turn_necks) throw();
  void missingActions(Int16 dashes, Int16 kicks, Int16 says,
		      Int16 turns, Int16 turn_necks) throw();

  static Controller* currentControllerObject;
  void sigAlarmHandler(int);
  void sigIOHandler(int);

private:
  ServerInterface* const theServerInterface;
  Sensors* const theSensors;
  Actuators* const theActuators;
  ControlInterface* const theControlInterface;

#ifdef RS_ENABLE_COACH_DEBUGGING
  ServerInterface* theCoachInterface;
#endif

  UInt16 maxCyclesSinceIOBeforeShutdown;
  UInt16 minTicksBeforeSendingCommand;
  UInt16 ticksBeforeCommandWarning;
  UInt16 maxTicksBeforeSendingCommand;

  UInt16 cycleLength;
  UInt16 tickLength;
  UInt16 ticksBetweenCycles;
  UInt16 ticksBetweenSenses;
  UInt16 ticksBetweenVisuals;

  /** The timer used for the internal clock. */
  Timer commandTimer;

  /** Flag used to start the internal clock. */
  bool startClock;

  /** Count the number of ticks this cycle. */
  UInt16 ticks;
  /** Count the number of ticks since last sensory sensor data. */
  UInt16 senseTicks;
  /** Count the number of ticks since last visual sensor data. */
  UInt16 visualTicks;
  /** The time the last cycle started. */
  Time lastCycleTime;
  /** The time of the last aural sensor data. */
  Time lastAuralTime;
  /** The time of the last sensory sensor data. */
  Time lastSenseTime;
  /** The time of the last visual sensor data. */
  Time lastVisualTime;
  /** The time last reported from the server. */
  UInt16 lastKnownServerTime;
  /** The current time, according to the agent. Stored in CTime. */
  Time& currentTime;
  /** The number of cycles added to the last known server
    time to get the current time, aCording to the agent. */
  UInt16 estimatedTime;
  /** The number of stopped cycles. */
  UInt16 stoppedCycles;
  
  char* sensorData;
  UInt16 sigalarmsSinceSigio;
  Time lastActTime;

  sigset_t initHandlers();

  sigset_t sigalarmask;
  sigset_t sigiomask;
  sigset_t allSigsMask;

  //Sigfunc* Signal(int signo, Sigfunc* func, sigset_t* mask = NULL);

  Controller(const Controller&);
  Controller& operator= (const Controller&);
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "Controller.icc"
#endif


#endif // !CONTROLLER_H
