/* -*- Mode: C++ -*-
 * $Id: SensorInterface.h,v 1.2 2000/09/13 00:02:02 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: SensorInterface.h
 * Unitname: BasicSystem
 * $Revision: 1.2 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/13 00:02:02 $
 *
 *
 * HISTORY
 *
 * $Log: SensorInterface.h,v $
 * Revision 1.2  2000/09/13 00:02:02  frehe
 * Made minor changes to the forEach methods.
 *
 * Revision 1.1  2000/09/03 20:41:09  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file SensorInterface.h
 * Include file for the class SensorInterface.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2000/09/13 00:02:02 $
 */


#ifndef SENSOR_INTERFACE_H
#define SENSOR_INTERFACE_H


#include <queue>
#include <vector>
#include <string>

#include "basic_system_types.h"
#include "soccer_types.h"
#include "ServerCommand.h"
#include "Time.h"


RS_BEGIN_NAMESPACE


/**
 * @class SensorInterface SensorInterface.h
 * This class is the interface between the sensors
 * and the rest of the agent.
 * But it is not connected to the view manager.
 * To use the view manager you have to use the
 * FrameworkSensorInterface instead.
 *
 * <!-- @invariant The class invariant is... -->
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2000/09/13 00:02:02 $
 */
class SensorInterface {
public:
  SensorInterface(const std::string& my_team_name);
  virtual ~SensorInterface();

  virtual void newCycle();
  virtual void forceNewCycle();
  
  Time getCurrentTime() const;

  void forEachMarker(void (*fn)(const ServerMarkerObject&)) const;
  UInt getNumberOfSeenMarkers() const;
  void forEachLine(void (*fn)(const ServerLineObject&)) const;
  UInt getNumberOfSeenLines() const;
  void forEachPlayer(void (*fn)(const ServerPlayerObject&)) const;
  UInt getNumberOfSeenPlayers() const;
  
  bool sawBall() const; // Necessary? If direction isn't unknown then I saw the ball...
  FloatUnknown getBallDistance() const;
  AngleDegUnknown getBallDirection() const;
  FloatUnknown getBallDistanceChange() const;
  FloatUnknown getBallDirectionChange() const;

  ViewQuality getViewQuality() const;
  ViewWidth getViewWidth() const;
  Float getStamina() const;
  Float getEffort() const;
  Float getSpeed() const;
  AngleDeg getHeadDirection() const;
  UInt getKicks() const;
  UInt getDashes() const;
  UInt getTurns() const;
  UInt getSays() const;
  UInt getTurnNecks() const;
  
  const std::string& getOurTeamName() const;
  const std::string& getTheirTeamName() const;
  void setTheirTeamName(const std::string& team_name);
  Side getOurSide() const;
  Side getTheirSide() const;
  UInt getUniformNumber() const;
  
  PlayMode getPlayMode() const;
  UInt getOurScore() const;
  UInt getTheirScore() const;
  
  std::string getCoachMessage() const;
  std::string getPlayerMessage() const;

  const ServerCommand* const getLastCommand() const;


  // These are only available if you are using a coach!!!
  UInt getCoachMessageTime() const;

  PointUnknown getBallPosition() const;
  VectorUnknown getBallSpeed() const;

  PointUnknown getAgentPosition() const;
  VectorUnknown getAgentSpeed() const;
  AngleDegUnknown getAgentDirection() const;
  AngleDegUnknown getAgentNeckDirection() const;

  void forEachOpponent(void (*fn)(const CoachPlayerObject&)) const;
  void forEachTeammate(void (*fn)(const CoachPlayerObject&)) const;

  std::vector<CoachPlayerObject> getOpponents() const;
  std::vector<CoachPlayerObject> getTeammates() const;

  
  //virtual void setTime(const UInt time);
  // This method is called *after* the server message is parsed.
  // Except for the init message.
  virtual void setLastSensorType(const SensorType sensor_type);
  virtual void seeMarker(const MarkerName marker,
			 const Float32U dist,
			 const Float32U ang,
			 const bool behind);
  virtual void seeLine(const SideLineName line,
		       const Float32U dist,
		       const Float32U ang);
  virtual void seeBall(const Float32U dist,
		       const Float32U ang,
		       const Float32U dist_chng,
		       const Float32U dir_chng);
  virtual void seePlayer(const Side player_side,
			 const Team player_team,
			 const UInt16U player_number,
			 const Float32U dist,
			 const Float32U ang,
			 const Float32U dist_chng,
			 const Float32U dir_chng,
			 const Float32U body_dir,
			 const Float32U face_dir);
  virtual void senseBody(const ViewQuality view_quality,
			 const ViewWidth view_width,
			 const Float32 stamina,
			 const Float32 effort,
			 const Float32 speed,
			 const Float32 head_angle,
			 const UInt16 kicks,
			 const UInt16 dashes,
			 const UInt16 turns,
			 const UInt16 says,
			 const UInt16 turn_necks);
  virtual void hearInitMessage(const char my_side,
			       const UInt my_number,
			       const PlayMode play_mode,
			       const bool reconnect);
  virtual void hearReferee(const PlayMode play_mode,
			   const Side side);
  virtual void hearScore(const Side side,
			 const UInt new_score);
  virtual void hearCoach(const char* const message);
  virtual void hearPlayer(const char* const message);
  virtual void setLastCommand(ServerCommand* command);

  virtual void setCoachTime(Int time);
  virtual void coachSeeMarker(MarkerName marker, Float32U posX, Float32U posY);
  virtual void coachSeeBall(Float32U posX, Float32U posY,
			    Float32U speedX, Float32U speedY);
  virtual void coachSeePlayer(Side player_side, Team player_team,
			      UInt16U player_number,
			      Float32U posX, Float32U posY,
			      Float32U speedX, Float32U speedY,
			      Float32U bodyDir, Float32U neckDir);

protected:
  bool forcedNewCycle;
  Time forcedNewCycleTime;
  
  // This is a reference to the Time::current_time
  // and Time::clockStopped class variables. The reason is that I rather
  // want the user to use this class than Time directly.
  Time& currentTime;
  bool& clockStopped;

  Time lastCycleTime;

  std::priority_queue<ServerMarkerObject> markers;

  std::priority_queue<ServerLineObject> lines;

  bool didSeeBall;
  FloatUnknown ballDistance;
  AngleDegUnknown ballDirection;
  FloatUnknown ballDistanceChange;
  FloatUnknown ballDirectionChange;

  std::vector<ServerPlayerObject> players;
  
  ViewQuality viewQuality;
  ViewWidth viewWidth;
  Float stamina;
  Float effort;
  Float speed;
  AngleDeg headDirection;
  UInt kicks;
  UInt dashes;
  UInt turns;
  UInt says;
  UInt turnNecks;

  std::string ourTeamName;
  std::string theirTeamName;
  Side ourSide;
  Side theirSide;
  UInt uniformNumber;
  
  PlayMode playMode;
  UInt ourScore;
  UInt theirScore;
  
  std::string coachMessage;
  std::string playerMessage;

  ServerCommand* lastPeriodicCommand;
  std::vector<ServerCommand*> lastImmediateCommands;

  // Coach related stuff
  UInt coachMessageTime;
  
  PointUnknown ballPosition;
  VectorUnknown ballSpeed;

  PointUnknown agentPosition;
  VectorUnknown agentSpeed;
  AngleDegUnknown agentDirection;
  AngleDegUnknown agentNeckDirection;

  std::vector<CoachPlayerObject> opponents;
  std::vector<CoachPlayerObject> teammates;

  virtual void updateBeforeNewCycle();

  FloatUnknown getFloatUnknown(const Float32U& f);

  //Hide copy-constructor and assignment
  SensorInterface(const SensorInterface&);
  SensorInterface& operator= (const SensorInterface&);
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "SensorInterface.icc"
#endif


#endif // !SENSOR_INTERFACE_H
