/* -*- Mode: C++ -*-
 * $Id: Controller.cc,v 1.5 2000/10/24 15:15:17 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Controller.cc
 * Unitname: BasicSystem
 * $Revision: 1.5 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/10/24 15:15:17 $
 *
 *
 * HISTORY
 *
 * $Log: Controller.cc,v $
 * Revision 1.5  2000/10/24 15:15:17  frehe
 * Fixed the generation of the onNewCycle event.
 *
 * Revision 1.4  2000/10/23 12:05:29  frehe
 * Removed all explicit output to std::cout.
 *
 * Revision 1.3  2000/09/05 21:15:52  frehe
 * Made minor changes.
 *
 * Revision 1.2  2000/09/04 15:23:46  frehe
 * Fixed a naming miss.
 *
 * Revision 1.1  2000/09/03 20:41:13  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file Controller.cc
 * Implementation file for the class Controller.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2000/10/24 15:15:17 $
 */


#include <cstdlib>

#include "Controller.h"
#include "Actuators.h"
#include "Sensors.h"
#include "ServerInterface.h"
#include "ControlInterface.h"
#include "ServerCommand.h"
#include "ControlCommand.h"
#include "SetCyclesSinceIOBeforeShutdown.h"
#include "SetMinTicksBeforeSendingCommand.h"
#include "SetMaxTicksBeforeSendingCommand.h"
#include "SetTicksBeforeCommandWarning.h"
#include "SetTicksBetweenCycles.h"
#include "SetTicksBetweenSenses.h"
#include "SetTicksBetweenVisuals.h"

#include "rs_error.h"
#include "rs_log.h"


RS_BEGIN_NAMESPACE


Controller* Controller::currentControllerObject = NULL;


/** Signal handler for sigalrm. It is declared as an external
  C function since the Unix signal handler requires this. */
extern "C" RETSIGTYPE externSigAlarmHandler(int sig);
/** Signal handler for sigio. It is declared as an external
  C function since the Unix signal handler requires this. */
extern "C" RETSIGTYPE externSigIOHandler(int sig);


Controller::Controller(ServerInterface* const server_interface,
		       Sensors* const sensors,
		       Actuators* const actuators,
		       ControlInterface* const control_interface,
		       const UInt16 cycle_length,
		       const UInt16 tick_length,
		       const UInt16 sense_step,
		       const UInt16 visual_step,
		       const UInt16 max_sigalrms,
		       const UInt16 min_ticks,
		       const UInt16 ticks_before_cmd_warning,
		       const UInt16 max_ticks)
  throw()
  : theServerInterface(server_interface),
    theSensors(sensors),
    theActuators(actuators),
    theControlInterface(control_interface),
#ifdef RS_ENABLE_COACH_DEBUGGING
    theCoachInterface(NULL),
#endif
    maxCyclesSinceIOBeforeShutdown(max_sigalrms),
    minTicksBeforeSendingCommand(min_ticks),
    ticksBeforeCommandWarning(ticks_before_cmd_warning),
    maxTicksBeforeSendingCommand(max_ticks),
    cycleLength(cycle_length),
    tickLength(tick_length),
    ticksBetweenCycles(cycle_length / tick_length),
    ticksBetweenSenses(sense_step / tick_length),
    ticksBetweenVisuals(visual_step / tick_length),
    commandTimer(),
    startClock(false),
    ticks(0),
    senseTicks(0),
    visualTicks(0),
    lastCycleTime(Time(0,0)),
    lastAuralTime(0),
    lastSenseTime(0),
    lastVisualTime(0),
    lastKnownServerTime(0),
    currentTime(Time::currentTime),
    estimatedTime(0),
    stoppedCycles(0),
    sensorData(NULL),
    sigalarmsSinceSigio(0),
    lastActTime(Time(0,0))
  //sigalarmask(),
  //sigiomask(),
  //allSigsMask()
{
  sensorData = new char[theServerInterface->getServerBufferSize()];

  /*
  std::cout << cycleLength << "; " << tickLength << "; "
	    << ticksBetweenCycles << "; "
	    << ticksBetweenSenses << "; "
	    << ticksBetweenVisuals << std::endl;
   */
  
  // Set this controller to be the current controller.
  currentControllerObject = this;
}


Controller::~Controller()
  throw()
{
  delete[] sensorData;
#ifdef RS_ENABLE_COACH_DEBUGGING
  if ( theCoachInterface != NULL )
    delete theCoachInterface;
#endif
}


bool Controller::init()
  throw(RSException)
{
  // Init the sensors and actuators.
  theSensors->init(this);
  theActuators->init(this);

  // Init the signal handlers.
  allSigsMask = initHandlers();

  // Open the connection to the server and return the result.
  return theServerInterface->connect();
}


bool Controller::start()
  throw(RSException)
{
  // Set the start flag, the internal timer will be started
  // after the first sense_body message.
  startClock = true;

  sigprocmask(SIG_UNBLOCK, &sigiomask, NULL);
  sigprocmask(SIG_UNBLOCK, &sigalarmask, NULL);

  // Initialize the decision maker, send an init message to the server
  ServerCommand* cmd = theControlInterface->initAgent();
  if ( cmd == NULL )
    RS_ERROR("[Start]: no init command generated by the decision!");
  else {
    RS_LOG( LA_CONTROLLER, "Sent init message: " << cmd->serverString() );
    sendCommand(cmd);
  }
  // and start the sense-decide-act loop.
  theControlInterface->start();

  return true;
}


#ifdef RS_ENABLE_COACH_DEBUGGING
bool Controller::startCoach(const std::string& host, const UInt16 port,
			    const UInt32 server_buffer_size)
  throw(RSException)
{
  theCoachInterface = new ServerInterface(host, port, server_buffer_size);
  
  return theCoachInterface->connect();
}
#endif


#ifdef RS_ENABLE_COACH_DEBUGGING
bool Controller::sendCoachMessage(const std::string& msg)
  throw(RSException)
{
  return theCoachInterface->sendCommand(msg);
}
#endif


/*
Sigfunc* Controller::Signal(int signo, Sigfunc *func, sigset_t *mask)
{
  //Adapted from Stevens "Network programming"

  struct sigaction act, oact;

  act.sa_handler = func;
  sigemptyset(&act.sa_mask);
  if (mask != NULL)
    act.sa_mask = *mask;
  act.sa_flags = 0;

  if (signo == SIGALRM) {
#ifdef SA_INTERRUPT
    act.sa_flags |= SA_INTERRUPT;   // SunOS 4.x
#endif
  } else {
#ifdef SA_RESTART
    act.sa_flags |= SA_RESTART;     // SVR4, 4.4BSD
#endif
  }

  if (sigaction(signo, &act, &oact) < 0)
    return (SIG_ERR);

  return (oact.sa_handler);
}
*/

sigset_t Controller::initHandlers()
{
  /*
  // From RoboLog
  sigset_t mask;
  struct sigaction vec;

  sigfillset(&mask);             // block all signals during timer event
  sigdelset(&mask, SIGINT);      // except INT
  sigdelset(&mask, SIGTERM);     // and TERM

  vec.sa_handler = (RETSIGTYPE (*) (int)) sigalrm_handler;
  vec.sa_mask = mask;
  vec.sa_flags = 0;
  if (sigaction(SIGALRM, &vec, 0) == -1) {
    ROBOBUG(ROBOBUG_FATAL,ROBOLOG_INIT,
	    "sigaction error (SIGALRM)");
  }

  vec.sa_handler = (RETSIGTYPE (*) (int)) sigio_handler;
  if (sigaction(SIGIO, &vec, 0) == -1) {
    ROBOBUG(ROBOBUG_FATAL,ROBOLOG_INIT,
	    "sigaction error (SIGIO)");
  }
  
  // unblock
  sigemptyset(&mask);
  sigaddset(&mask, SIGALRM);
  sigaddset(&mask, SIGIO);
  sigprocmask(SIG_UNBLOCK, &mask, 0);
  */

  // Adopted from CMUnited'98 by Peter Stone et al.
  sigemptyset(&sigalarmask);
  sigaddset(&sigalarmask, SIGALRM);
  sigemptyset(&sigiomask);
  sigaddset(&sigiomask, SIGIO);

  struct sigaction act, oact;

  //Signal(SIGALRM, externSigAlarmHandler, &sigiomask);
  act.sa_handler = externSigAlarmHandler;
  sigemptyset(&act.sa_mask);
  act.sa_mask = sigiomask;
  act.sa_flags = 0;
#ifdef SA_INTERRUPT
    act.sa_flags |= SA_INTERRUPT;   // SunOS 4.x
#endif
  sigaction(SIGALRM, &act, &oact);

  //Signal(SIGIO, externSigIOHandler, &sigalarmask);
  act.sa_handler = externSigIOHandler;
  sigemptyset(&act.sa_mask);
  act.sa_mask = sigalarmask;
  act.sa_flags = 0;
#ifdef SA_RESTART
    act.sa_flags |= SA_RESTART;     // SVR4, 4.4BSD
#endif
  sigaction(SIGIO, &act, &oact);


  sigprocmask(SIG_UNBLOCK, &sigiomask, NULL);
  sigprocmask(SIG_UNBLOCK, &sigalarmask, NULL);

  sigset_t sigsetmask;
  sigprocmask(SIG_BLOCK, NULL, &sigsetmask);   // get's the currently unblocked signals
  return sigsetmask;   
}


void Controller::sigAlarmHandler(int)
{
  RS_LOG( LA_CONTROLLER, "Entering SigAlarm..." );

  //sigprocmask(SIG_BLOCK, &allSigsMask, NULL);
  sigprocmask(SIG_BLOCK, &sigiomask, NULL);  

  // Notify the decision maker that a new tick is to be processed.
  theControlInterface->onBeforeTick();
  
  // For each SIGALRM add a tick to each counter.
  ticks++;
  visualTicks++;
  senseTicks++;

  // If ticks is greater than ticksBetweenCycles, then a new cycle should've begun.
  // Therefore increase the estimated time and the current time.
  if (ticks > ticksBetweenCycles) {
    estimatedTime++;

    // Update the current time depending on whether the game is stopped or not.
    if ( currentTime.isClockStopped() )
      currentTime.setS(stoppedCycles + estimatedTime);
    else {
      currentTime.setT(lastKnownServerTime + estimatedTime);
      currentTime.setS(0);
    }

    // Since it is ticksBetweenCycles+1 we should remove ticksBetweenCycles ticks,
    // which leaves us with 1 tick left.
    ticks = 1;
  }

  // Since we might have lost messages from the server we need to notify
  // the decision maker that there should've been a new cycle now.
  //theControlInterface->OnEstimatedNewCycle(estimatedTime, ticks);
  if ( estimatedTime > 0 ) {
    if ( theControlInterface->onEstimatedNewCycle(estimatedTime, ticks) ) {
      lastCycleTime = currentTime;
      theControlInterface->onNewCycle(true);
    }
  }
  
  /*
  std::cout << "ALRM: " << lastKnownServerTime
	    << "; " << last_calc_serverTime_sense_body
	    << "; " << last_calc_serverTimeTimer
	    << ", " << ticks
	    << std::endl;
	    */

  // Check whether the agent sent a future action already.
  if ( currentTime < lastActTime ) {
    RS_WARNING(currentTime << ": Sent future action(s) already! Last action time = " << lastActTime);
    // Assume it depends on some timing problem and set lastActTime to currentTime
    // to prevent losses of action oppertunities.
    lastActTime = currentTime;
  }
  
  // Check whether the agent sent an action last cycle or not.
  if ( ticks == 1 &&
       ( currentTime.getT() - lastActTime.getT() > 1 ||
	 ( currentTime.isClockStopped() &&
	   currentTime.getS() - lastActTime.getS() > 1) ) )
    RS_WARNING(currentTime << ": Missed action(s)! Last action time = "
	       << lastActTime);

  // If the actuators want to send a periodic action and one haven't been
  // sent this cycle then send it.
  if ( lastActTime != currentTime &&
       theActuators->waitingPeriodicAction() &&
       ticks >= minTicksBeforeSendingCommand &&
       ticks <= maxTicksBeforeSendingCommand ) {
    theActuators->executeNextPeriodicAction();
    lastActTime = currentTime;
  } else {
    if ( lastActTime != currentTime &&
	 !theActuators->waitingPeriodicAction() &&
	 ticks >= ticksBeforeCommandWarning ) {
      // Warn the decision maker about the fact that no command has been sent yet!
      theControlInterface->onCommandWarning();
    }
    // If the actuators want to send an immediate action then send it.
    if ( theActuators->waitingImmediateAction() )
      theActuators->executeNextImmediateAction();
  }

  // If there is a control command waiting execute it.
  if ( theActuators->waitingControlCommand() )
    theActuators->executeNextControlCommand();  
  
  // Check if the currentTime differs from the lastKnownServerTime.
  if ( false && currentTime.getT() != lastKnownServerTime )
    RS_WARNING("Diff: " << lastKnownServerTime 
	       << " != " << currentTime 
	       << " (" << lastSenseTime << ", " << lastVisualTime
	       << ", " << lastAuralTime << ") (" 
	       << estimatedTime << ", " << ticks << ")");
  
  ++sigalarmsSinceSigio;

  // Notify the decision maker that a new tick was processed.
  theControlInterface->onAfterTick();

  //sigprocmask(SIG_UNBLOCK, &allSigsMask, NULL);
  sigprocmask(SIG_UNBLOCK, &sigiomask, NULL);

  RS_LOG( LA_CONTROLLER, "Leaving SigAlarm..."  );
}


void Controller::sigIOHandler(int)
{
  RS_LOG( LA_CONTROLLER, "Entering SigIO..." );

  //sigprocmask(SIG_BLOCK, &allSigsMask, NULL);
  sigprocmask(SIG_BLOCK, &sigalarmask, NULL);  

  // Notify the decision maker that a new sensors data is to be processed.
  theControlInterface->onBeforeSensorData();
  
  /*
  // Simulate loss of messages.
  if ( int((double (rand())/RAND_MAX)*50) <= 2 ) {
    std::cout << "SKIPPED SENSOR DATA!" << std::endl;
    while ( ReceiveInput(sensorData) ) {}
    return;
  }
  */
  
  bool first = true;
  UInt16 time_of_first_data = 0;
  
  while ( receiveInput(sensorData) ) {
    RS_LOG( LA_CONTROLLER,
	    "NEW INFO! " << theSensors->getCurrentTime() << ": "
	    << sensorData );

    theSensors->analyze(sensorData);

    if ( first ) {
      time_of_first_data = lastKnownServerTime;
      first = false;
    }
    
    // This cycle doesn't work 100%.
    // If you get two sense_bodies in the same cycle
    // but they belong to two different cycles, then
    // this algorithm will not find that.
    // It will instead start the cycle twice, first
    // when it recieves the second sense_body and then
    // again when a message with the new time arrives.
    // Which can be long enough so that a cycle will
    // be missed... Not good!
    // It should be solvable by keeping track of how
    // long it was since the time changed and set
    // a limit for how short a stopped clock interval
    // can be. Eg. less than three cycles is not a
    // stopped clock. Ie. (x sb) (x sb) (x+1 see)
    // Is not a stopped clock but:
    // (x sb) ... (x sb) ... (x sb) ... (x+1 see)
    // is interpreted as the clock has been stopped,
    // and it is a new cycle.
    // A more aCurate algorithm can be made when you
    // have aCess to the referee calls, since they
    // determine if the clock has been stopped or not.

    switch ( theSensors->getLastSensorType() ) {
    case SENSOR_SENSE:
      RS_LOG( LA_CONTROLLER, "SENSE_BODY " << lastKnownServerTime );

      if ( currentTime.getT() == lastSenseTime.getT() &&
	   currentTime.getS() == 0 && !currentTime.isClockStopped() )
 	RS_WARNING(currentTime << ": Two sensebodies in one cycle!");

      // If more than one cycle between senses sense_body messages are lost.
      if ( currentTime.getT() - lastSenseTime.getT() > 1 ||
	   ( currentTime.isClockStopped() &&
	     currentTime.getS() - lastSenseTime.getS() > 1) )
	RS_WARNING(currentTime << ": More than one cycle between sense_body's!"
		   << " lastSenseTime = " << lastSenseTime);
	
      lastSenseTime = currentTime;
      theControlInterface->onPhysicalSensorData();
      break;
    case SENSOR_VISUAL:
      lastVisualTime = currentTime;
      theControlInterface->onVisualSensorData();
      break;
    case SENSOR_AURAL:
      lastAuralTime = currentTime;
      theControlInterface->onAuralSensorData();
      break;
    case SENSOR_ACTUATORS:
      // Should not be posible!
      RS_WARNING(currentTime << "SENSOR_ACTUATORS should not be possible to get!");
      break;
    case SENSOR_INIT:
      theControlInterface->onInit();
      break;
    case SENSOR_ERROR:
      theControlInterface->onSensorError();
      break;
    default:
      RS_ERROR(currentTime << ": Strange sensor input!!!");
    }
  }

#ifdef RS_ENABLE_COACH_DEBUGGING
  // If we are using the coach then we should also check the coach socket
  // for data.
  if (theCoachInterface) {
    while ( theCoachInterface->receiveInput(sensorData) ) {
      theSensors->analyzeCoachMessage(sensorData);
    }
  }
#endif
  
  // Check if the currentTime differs from the lastKnownServerTime.
  if ( false && currentTime != lastKnownServerTime )
    RS_WARNING("Diff: " << lastKnownServerTime 
	       << " != " << currentTime 
	       << " (" << lastSenseTime << ", " << lastVisualTime
	       << ", " << lastAuralTime << ") (" 
	       << estimatedTime << ", " << ticks << ")");

  // Check if input from more than one cycle was received at once.
  if ( currentTime.getT() - time_of_first_data > 1 )
    RS_WARNING(currentTime
	       << ": Received several steps at once! "
	       << time_of_first_data << " to " << currentTime
	       << " lastActTime = " << lastActTime);
  
  sigalarmsSinceSigio = 0;

  //std::cout << "Current time: " << currentTime << std::endl;
  
  // Notify the decision maker that a new sensors data was processed.
  theControlInterface->onAfterSensorData();

  //  sigprocmask(SIG_UNBLOCK, &allSigsMask, NULL);
  sigprocmask(SIG_UNBLOCK, &sigalarmask, NULL);

  RS_LOG( LA_CONTROLLER, "Leaving SigIO..." );
}


bool Controller::updateTime(const UInt16 newTime)
  throw()
{
  Time lastTime = currentTime;
  //UInt16 last_estimatedTime = estimatedTime;

  // Update the current time with the time of the last sensor data.
  // Since that time is assumed to be correct if the game isn't stopped
  // reset the estimated time if the game is running.
  if ( !currentTime.isClockStopped() ) {
    if ( lastKnownServerTime != newTime ) {
      currentTime.setS(0);
      stoppedCycles = 0;
    }
    estimatedTime = 0;
    // If the clock is stopped then the time shouldn't be different...
  } else if ( lastKnownServerTime != newTime ) {
    RS_WARNING("Clock is stopped but the current time is not equal to the new time.");
    // But since it is start the clock again...
    currentTime.startClock();
  }

  // Update the last known server time and the current time.
  lastKnownServerTime = newTime;
  currentTime.setT(lastKnownServerTime);

  switch ( theSensors->getLastSensorType() ) {
  case SENSOR_SENSE:
    // This is used to start the timer after the first sense_body
    if ( startClock ) {
      commandTimer.start(tickLength / 2, tickLength);
      senseTicks = 1;
      lastSenseTime = currentTime;
      startClock = false;
    } else {
      // If the server and the internal clock are running on time the
      // ticks should be zero, since the estimated time should be updated
      // before a new time arrive.
      if ( false && abs(ticks - ticksBetweenCycles) > 1 )
	RS_WARNING(currentTime << ": ticks = " << ticks);
      // If the game is stopped then the
      if ( currentTime.isClockStopped() &&
	   currentTime.getT() == lastSenseTime.getT() ) {
	if ( false )
	  RS_WARNING(currentTime << ": stoppedCycles = " << stoppedCycles
		     << " estimatedTime = " << estimatedTime
		     << " ticks " << ticks);
	if ( estimatedTime > 0 ) {
	  stoppedCycles+=estimatedTime;
	  estimatedTime = 0;
	  if ( ticks == ticksBetweenCycles ) {
	    stoppedCycles++;
	    ticks = 0;
	  }
	} else {
	  stoppedCycles++;
	  ticks = 0;
	}
	currentTime.setS(stoppedCycles);
	if ( false )
	  RS_WARNING(currentTime << ": stoppedCycles = " << stoppedCycles
		     << " estimatedTime = " << estimatedTime
		     << " ticks " << ticks);
      } else {
	// Reset the ticks since last cycle.
	estimatedTime = 0;
	ticks = 0;
      }
	
      // If senseTicks isn't equal to ticksBetweenSenses either the
      // server or the internal clock is not running correctly.
      if ( false && abs(senseTicks - ticksBetweenSenses) > 1 )
	RS_WARNING(currentTime << ": senseTicks = " << senseTicks);
      // Reset the number of ticks since last sense.
      senseTicks = 0;
      
    }
    break;
  case SENSOR_VISUAL:
    // If visualTicks isn't equal to ticksBetweenVisuals either the
    // server or the internal clock is not running correctly.
    if ( false && abs(visualTicks - ticksBetweenVisuals) > 1 )
      RS_WARNING(currentTime << ": visualTicks = " << visualTicks);
    // Reset the number of ticks since last visual.

    visualTicks = 0;
    break;
  case SENSOR_AURAL:
    break;
  case SENSOR_ACTUATORS:
    // Should not be posible!
    RS_WARNING(currentTime << "SENSOR_ACTUATORS should not be possible to get!");
    break;
  case SENSOR_INIT:
    break;
  default:
    RS_ERROR(currentTime << ": Strange sensor input!!!");
  }

  if ( currentTime.getT() - lastCycleTime.getT() > 1 ||
       ( currentTime.isClockStopped() &&
	 currentTime.getS() - lastCycleTime.getS() > 1) )
   RS_WARNING(currentTime << ": Missed cycles! lastCycleTime = "
	      << lastCycleTime);

  // RS_LOG ( LA_CONTROLLER, "currentTime = " << currentTime << " lastCycleTime = " << lastCycleTime );
  bool new_cycle = ( currentTime > lastCycleTime );
  
  if ( new_cycle ) {
    lastCycleTime = currentTime;
    theControlInterface->onNewCycle(false);
  }

  if ( lastTime > currentTime ) {
    RS_WARNING(currentTime << ": Incorrectly estimated server time! "
	       << "lastTime = " << lastTime);
  }

  return new_cycle;
  
  //return ( currentTime > lastTime ||
  //   last_estimatedTime > estimatedTime );
}


bool Controller::sendCommand(ServerCommand* cmd)
  throw(RSException)
{

  RS_LOG( LA_CONTROLLER, theSensors->getCurrentTime() << " " << cmd );

  bool res = false;
  if ( cmd != NULL ) {
    res = theServerInterface->sendCommand(cmd->serverString());
    theSensors->analyze(cmd);
    theControlInterface->onActuatorSensorData();
  }
  return res;  
}


bool Controller::sendCommand(ControlCommand* cmd)
  throw()
{
  switch ( cmd->getType() ) {
  case CMD_SetCyclesSinceIOBeforeShutdown:
    maxCyclesSinceIOBeforeShutdown = cmd->getSetCyclesSinceIOBeforeShutdown()->getCycles();
    break;
  case CMD_SetMinTicksBeforeSendingCommand:
    minTicksBeforeSendingCommand = cmd->getSetMinTicksBeforeSendingCommand()->getTicks();
    break;
  case CMD_SetMaxTicksBeforeSendingCommand:
    maxTicksBeforeSendingCommand = cmd->getSetMaxTicksBeforeSendingCommand()->getTicks();
    break;
  case CMD_SetTicksBeforeCommandWarning:
    ticksBeforeCommandWarning = cmd->getSetTicksBeforeCommandWarning()->getTicks();
    break;
  case CMD_SetTicksBetweenCycles:
    ticksBetweenCycles = cmd->getSetTicksBetweenCycles()->getTicks();
    break;
  case CMD_SetTicksBetweenSenses:
    ticksBetweenSenses = cmd->getSetTicksBetweenSenses()->getTicks();
    break;
  case CMD_SetTicksBetweenVisuals:
    ticksBetweenVisuals = cmd->getSetTicksBetweenVisuals()->getTicks();
    break;
  default:
    RS_ERROR("Unknown controller command sent to the controller!");
  }

  return true;
}

bool Controller::receiveInput(char* const input)
  throw(RSException)
{
  return theServerInterface->receiveInput(input);
}

void Controller::delayedActions(Int16 dashes, Int16 kicks, Int16 says,
				Int16 turns, Int16 turn_necks)
  throw()
{
  RS_WARNING("The agent executed more actions than sent previous cycle! dashes = " +
	   toString(dashes) + " kicks = " + toString(kicks) +
	   " says = " + toString(says) + " turn = " + toString(turns) +
	   " turn_necks = " + toString(turn_necks));
  theControlInterface->onDelayedActions(dashes, kicks, says, turns, turn_necks);
}

void Controller::missingActions(Int16 dashes, Int16 kicks, Int16 says,
				 Int16 turns, Int16 turn_necks)
  throw()
{
  RS_WARNING("The agent executed less actions than sent previous cycle! dashes = " +
	   toString(dashes) + " kicks = " + toString(kicks) +
	   " says = " + toString(says) + " turn = " + toString(turns) +
	   " turn_necks = " + toString(turn_necks));
  theControlInterface->onMissingActions(dashes, kicks, says, turns, turn_necks);
}


// Ugly fix to avoid anachronism when initializing the signal handlers.
extern "C" RETSIGTYPE externSigAlarmHandler(int sig)
{
  Controller::currentControllerObject->sigAlarmHandler(sig);
}

// Ugly fix to avoid anachronism when initializing the signal handlers.
extern "C" RETSIGTYPE externSigIOHandler(int sig)
{
  Controller::currentControllerObject->sigIOHandler(sig);
}


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "Controller.icc"
#endif
