/* -*- Mode: C++ -*-
 * $Id: SensorInterface.cc,v 1.3 2000/09/13 00:01:51 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: SensorInterface.cc
 * Unitname: BasicSystem
 * $Revision: 1.3 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/13 00:01:51 $
 *
 *
 * HISTORY
 *
 * $Log: SensorInterface.cc,v $
 * Revision 1.3  2000/09/13 00:01:51  frehe
 * Made minor changes to the forEach methods.
 *
 * Revision 1.2  2000/09/05 21:15:53  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/03 20:41:13  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file SensorInterface.cc
 * Implementation file for the class SensorInterface.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.3 $
 * $Author: frehe $
 * $Date: 2000/09/13 00:01:51 $
 */


#include "SensorInterface.h"
#include "rs_error.h"
#include "rs_log.h"


RS_BEGIN_NAMESPACE


SensorInterface::SensorInterface(const std::string& my_team_name)
  : forcedNewCycle(false),
    forcedNewCycleTime(Time(0,0)),
    currentTime(Time::currentTime),
    clockStopped(Time::clockStopped),
    lastCycleTime(Time(0,0)),
    markers(),
    lines(),
    didSeeBall(false),
    ballDistance(),
    ballDirection(),
    ballDistanceChange(),
    ballDirectionChange(),
    players(21),
    viewQuality(VQ_HIGH),
    viewWidth(VW_NORMAL),
    stamina(),
    effort(),
    speed(0),
    headDirection(0),
    kicks(0),
    dashes(0),
    turns(0),
    says(0),
    turnNecks(0),
    ourTeamName(my_team_name),
    theirTeamName(""),
    ourSide(S_UNKNOWN),
    theirSide(S_UNKNOWN),
    uniformNumber(),
    playMode(PM_UNKNOWN),
    ourScore(0),
    theirScore(0),
    coachMessage(),
    playerMessage(),
    lastPeriodicCommand(NULL),
    lastImmediateCommands(),
    coachMessageTime(),
    ballPosition(PointUnknown()),
    ballSpeed(VectorUnknown()),
    agentPosition(PointUnknown()),
    agentSpeed(VectorUnknown()),
    agentDirection(AngleDegUnknown()), 
    agentNeckDirection(AngleDegUnknown()),
    opponents(),
    teammates()
{
}


SensorInterface::~SensorInterface()
{ 
} 


void SensorInterface::newCycle()
{
  if ( lastCycleTime < currentTime &&
       ( !forcedNewCycle ||
	 ( forcedNewCycle &&
	   forcedNewCycleTime < currentTime ) ) ) {
    //std::cout << currentTime << ": NEW CYCLE!" << endl;
    updateBeforeNewCycle();

    if ( currentTime.getT() - lastCycleTime.getT() > 1 ||
	   ( currentTime.isClockStopped() &&
	     currentTime.getS() - lastCycleTime.getS() > 1) )
      RS_WARNING(currentTime << ": [SensorInterface] Missed cycle(s)!"
		 << " lastCycleTime = " << lastCycleTime);
    
    lastCycleTime = currentTime;
    
    forcedNewCycle = false;
  }
  else {
    RS_WARNING(currentTime << ": SKIPPED NEW CYCLE SINCE FORCED NEW CYCLE ALREADY!");
  }
}


void SensorInterface::forceNewCycle()
{
  if (  lastCycleTime < currentTime &&
	( !forcedNewCycle ||
	  ( forcedNewCycle &&
	    forcedNewCycleTime < currentTime ) ) ) {
    forcedNewCycle = true;
    forcedNewCycleTime = currentTime;
    lastCycleTime = currentTime;
    RS_WARNING(currentTime << ": FORCED NEW CYCLE!");
    updateBeforeNewCycle();
  }
}


void SensorInterface::updateBeforeNewCycle()
{
  while ( !markers.empty() )
    markers.pop();
  while ( !lines.empty() )
    lines.pop();
  didSeeBall = false;
  ballDistance.makeUnknown();
  ballDirection.makeUnknown();
  ballDistanceChange.makeUnknown();
  ballDirectionChange.makeUnknown();
  players.clear();
  coachMessage = "";
  playerMessage = "";

  ballPosition.makeUnknown();
  ballSpeed.makeUnknown();
  agentPosition.makeUnknown();
  agentSpeed.makeUnknown();
  agentDirection.makeUnknown();
  agentNeckDirection.makeUnknown();
  opponents.clear();
  teammates.clear();
}

/*
UInt16 SensorInterface::estimatedTimeToNextSenseBody() const
{
  return 100 - (getCurrentTimeMSec()-lastSenseBodyTime);
}

UInt16 SensorInterface::estimatedTimeToNextSee() const
{
  return 150 - (getCurrentTimeMSec()-lastSeeTime);
}

UInt16 SensorInterface::estimatedTimeToNextCycle() const
{
  return 100 - (getCurrentTimeMSec()-lastSenseBodyTime);
}

UInt16 SensorInterface::estimatedTimeToNextAction() const
{
  return 70 - (getCurrentTimeMSec()-lastSenseBodyTime);
}
*/

/*
void SensorInterface::setTime(const UInt16 time)
{
  // This method should convert UInt16 into Time.
  if ( currentTime.getT() < time ) {
    // If the clock is stopped then the time shouldn't be different...
    if ( clockStopped ) {
      RS_WARNING("Clock is stopped but the current time is not equal to the new time.");
      // But since it is start the clock again...
      clockStopped = false;
    }
    currentTime = Time(time, 0);
  }
  
}
*/


void SensorInterface::forEachMarker(void (*fn)(const ServerMarkerObject&)) const
{
  std::priority_queue<ServerMarkerObject> tmp_markers(markers);
  while ( !tmp_markers.empty() ) {
    fn(tmp_markers.top());
    tmp_markers.pop();
  }
}


void SensorInterface::forEachLine(void (*fn)(const ServerLineObject&)) const
{
  std::priority_queue<ServerLineObject> tmp_lines(lines);
  while ( !tmp_lines.empty() ) {
    fn(tmp_lines.top());
    tmp_lines.pop();
  }
}


void SensorInterface::forEachPlayer(void (*fn)(const ServerPlayerObject&)) const
{
  for ( std::vector<ServerPlayerObject>::const_iterator i = players.begin();
	i != players.end(); ++i )
    fn(*i);
}


void SensorInterface::forEachOpponent(void (*fn)(const CoachPlayerObject&)) const
{
  for ( std::vector<CoachPlayerObject>::const_iterator i = opponents.begin();
	i != opponents.end(); ++i )
    fn(*i);
}


void SensorInterface::forEachTeammate(void (*fn)(const CoachPlayerObject&)) const
{
  for ( std::vector<CoachPlayerObject>::const_iterator i = teammates.begin();
	i != teammates.end(); ++i )
    fn(*i);
}


void SensorInterface::setLastSensorType(const SensorType sensor_type)
{
  //std::cerr << "In SetLastSensorType...\n";
  
  // This method is called *after* the whole server message is parsed.
  switch ( sensor_type ) {
  case SENSOR_AURAL:
    break;
  case SENSOR_VISUAL:
    break;
  case SENSOR_SENSE:
    break;
  case SENSOR_ACTUATORS:
    break;
  case SENSOR_INIT:
    break;
  case SENSOR_COACH:
    break;
  default:
    RS_ERROR("Unknown sensor type in [SensorInterface]:SetLastSensorType.");
  }
}


void SensorInterface::seeMarker(const MarkerName marker,
				const Float32U dist,
				const Float32U ang,
				const bool behind)
{
  static ServerMarkerObject new_marker;
  new_marker.markerId = marker;
  new_marker.distance = getFloatUnknown(dist);
  if ( ang.isUnknown() )
    new_marker.direction = AngleDegUnknown();
  else
    new_marker.direction = AngleDegUnknown(ang.getValue());
  new_marker.behind = behind;

  markers.push(new_marker);
}


void SensorInterface::seeLine(const SideLineName line,
			      const Float32U dist,
			      const Float32U ang)
{
  //std::cerr << "Entering SeeLine...\n";
  //std::cerr << "SeeLine...\n";

  static ServerLineObject new_line;
  new_line.lineId = line;
  new_line.distance = getFloatUnknown(dist);
  if ( ang.isUnknown() )
    new_line.direction = AngleDegUnknown();
  else
    new_line.direction = AngleDegUnknown(ang.getValue());
  
  lines.push(new_line);
  
  //std::cerr << "Leaving SeeLine...\n";
}


void SensorInterface::seeBall(const Float32U dist,
			      const Float32U ang,
			      const Float32U dist_chng,
			      const Float32U dir_chng)
{
  RS_LOG( LA_SENSOR_INTERFACE,
	  "Saw ball: dist = " << dist << " dir = " << ang );
	  
  didSeeBall = true;
  ballDistance = getFloatUnknown(dist);
  if ( ang.isUnknown() )
    ballDirection = AngleDegUnknown();
  else
    ballDirection = AngleDegUnknown(ang.getValue());
  ballDistanceChange = getFloatUnknown(dist_chng);
  ballDirectionChange = getFloatUnknown(dir_chng);
}


void SensorInterface::seePlayer(const Side player_side,
				const Team player_team,
				const UInt16U player_number,
				const Float32U dist,
				const Float32U ang,
				const Float32U dist_chng,
				const Float32U dir_chng,
				const Float32U body_dir,
				const Float32U face_dir)
{
  ServerPlayerObject new_player;
  new_player.side = player_side;
  new_player.team = player_team;
  new_player.uniformNumber = player_number;
  new_player.distance = getFloatUnknown(dist);
  if ( ang.isUnknown() )
    new_player.direction = AngleDegUnknown();
  else
    new_player.direction = AngleDegUnknown(ang.getValue());
  new_player.distanceChange = getFloatUnknown(dist_chng);
  new_player.directionChange = getFloatUnknown(dir_chng);
  if ( body_dir.isUnknown() )
    new_player.bodyDirection = AngleDegUnknown();
  else
    new_player.bodyDirection = AngleDegUnknown(body_dir.getValue());
  if ( face_dir.isUnknown() )
    new_player.faceDirection = AngleDegUnknown();
  else
    new_player.faceDirection = AngleDegUnknown(face_dir.getValue());

  players.push_back(new_player);
}


void SensorInterface::senseBody(const ViewQuality new_view_quality,
				const ViewWidth new_view_width,
				const Float32 new_stamina,
				const Float32 new_effort,
				const Float32 new_speed,
				const Float32 new_head_direction,
				const UInt16 new_kicks,
				const UInt16 new_dashes,
				const UInt16 new_turns,
				const UInt16 new_says,
				const UInt16 new_turn_necks)
{
  viewQuality = new_view_quality;
  viewWidth = new_view_width;
  stamina = new_stamina;
  effort = new_effort;
  speed = new_speed;
  headDirection = new_head_direction;
  kicks = new_kicks;
  dashes = new_dashes;
  turns = new_turns;
  says = new_says;
  turnNecks = new_turn_necks;
}


void SensorInterface::hearInitMessage(const char my_side,
				      const UInt16 my_number,
				      const PlayMode play_mode,
				      const bool)
{
  if ( my_side == 'l' )
    ourSide = S_LEFT;
  else if ( my_side == 'r' )
    ourSide = S_RIGHT;
  else
    ourSide = S_UNKNOWN;
  
  if ( ourSide == S_LEFT )
    theirSide = S_RIGHT;
  else if ( ourSide == S_RIGHT )
    theirSide = S_LEFT;
  uniformNumber = my_number;
  playMode = play_mode;
  // Currently init only checks if it's before kick off
  // otherwise it's play_on.
  if ( playMode == PM_BEFORE_KICK_OFF ) {
    clockStopped = true;
  } else
    clockStopped = false;
}


void SensorInterface::hearReferee(const PlayMode play_mode,
				  const Side side)
{
  switch ( play_mode ) {
  case PM_OFFSIDE_KICK:
    if ( side == ourSide )
      playMode = PM_OUR_OFFSIDE_KICK;
    else
      playMode = PM_THEIR_OFFSIDE_KICK;
    clockStopped = false;
    break;
  case PM_KICK_OFF:
  if ( side == ourSide )
    playMode = PM_OUR_KICK_OFF;
  else
    playMode = PM_THEIR_KICK_OFF;
  clockStopped = false;
  break;
  case PM_KICK_IN:
    if ( side == ourSide )
      playMode = PM_OUR_KICK_IN;
    else
      playMode = PM_THEIR_KICK_IN;
    clockStopped = false;
    break;
  case PM_FREE_KICK:
    if ( side == ourSide )
      playMode = PM_OUR_FREE_KICK;
    else
      playMode = PM_THEIR_FREE_KICK;
    clockStopped = false;
    break;
  case PM_CORNER_KICK:
    if ( side == ourSide )
      playMode = PM_OUR_CORNER_KICK;
    else
      playMode = PM_THEIR_CORNER_KICK;
    clockStopped = false;
    break;
  case PM_GOAL_KICK:
    if ( side == ourSide )
      playMode = PM_OUR_GOAL_KICK;
    else
      playMode = PM_THEIR_GOAL_KICK;
    clockStopped = false;
    break;
  case PM_GOAL:
    if ( side == ourSide )
      playMode = PM_OUR_GOAL;
    else
      playMode = PM_THEIR_GOAL;
    clockStopped = true;
    break;
  case PM_GOALIE_GOT_BALL:
    if ( side == ourSide )
      playMode = PM_OUR_GOALIE_GOT_BALL;
    else
      playMode = PM_THEIR_GOALIE_GOT_BALL;
    clockStopped = false;
    break;
  case PM_BEFORE_KICK_OFF:
    clockStopped = true;
    playMode = play_mode;
    break;
  case PM_TIME_OVER:
    clockStopped = true;
    playMode = play_mode;
    break;
  case PM_PLAY_ON:
    clockStopped = false;
    playMode = play_mode;
    break;
  case PM_DROP_BALL:
    clockStopped = false;
    playMode = play_mode;
    break;
  case PM_HALF_TIME:
    clockStopped = true;
    playMode = play_mode;
    break;
  case PM_TIME_UP:
    clockStopped = true;
    playMode = play_mode;
    break;
  case PM_EXTENDED_TIME:
    clockStopped = true;
    playMode = play_mode;
    break;
  default:
    playMode = PM_UNKNOWN;
    RS_ERROR("Unknown play mode in HearReferee!");
  }
}


void SensorInterface::hearScore(const Side side,
				const UInt16 new_score)
{
  if ( ourSide == side )
    ourScore = new_score;
  else
    theirScore = new_score;
}


void SensorInterface::hearCoach(const char* const message)
{
  coachMessage = message;
}


void SensorInterface::hearPlayer(const char* const message)
{
  playerMessage = message;
}


void SensorInterface::setLastCommand(ServerCommand* command)
{
  //std::cout << "SetLastCommand = " << command->ServerString() << std::endl;
  if ( lastPeriodicCommand != NULL )
    delete lastPeriodicCommand;
  lastPeriodicCommand = command;
}


void SensorInterface::setCoachTime(Int time)
{
  coachMessageTime = UInt(time);
}


void SensorInterface::coachSeeMarker(MarkerName,
				     Float32U, Float32U)
{
  // std::cerr << "Saw marker " << marker << " at position Point("
  //    << posX << ", " << posY << ")" << std::endl;
}


void SensorInterface::coachSeeBall(Float32U posX, Float32U posY,
				   Float32U speedX, Float32U speedY)
{
  ballPosition = PointUnknown(getFloatUnknown(posX), getFloatUnknown(posY));
  ballSpeed = VectorUnknown(PointUnknown(getFloatUnknown(speedX),
					 getFloatUnknown(speedY)));
  
  //std::cerr << "Saw ball at position Point(" << posX << ", " << posY << ")"
  //    << " with speed Point(" << speedX << ", " << speedY << ")"
  //    << " at time " << coachMessagetime << std::endl;
}


void SensorInterface::coachSeePlayer(Side player_side, Team player_team,
				     UInt16U player_number,
				     Float32U posX, Float32U posY,
				     Float32U speedX, Float32U speedY,
				     Float32U bodyDir, Float32U neckDir)
{
  CoachPlayerObject cPlr;
  cPlr.side = player_side;
  cPlr.team = player_team;
  cPlr.uniformNumber = player_number;
  cPlr.position = PointUnknown(getFloatUnknown(posX), getFloatUnknown(posY));
  cPlr.speed = VectorUnknown(PointUnknown(getFloatUnknown(speedX), getFloatUnknown(speedY)));
  cPlr.bodyDirection = AngleDegUnknown(getFloatUnknown(bodyDir));
  cPlr.neckDirection = AngleDegUnknown(getFloatUnknown(neckDir));
  
  if ( player_team == T_OUR_TEAM ) {
    if ( player_number == uniformNumber ) {
      agentPosition = cPlr.position;
      agentSpeed = cPlr.speed; 
      agentDirection = cPlr.bodyDirection;
      agentNeckDirection = cPlr.neckDirection;
    } else {
      teammates.push_back(cPlr);
    }
    //std::cerr << "Saw our player number " << player_number
    //     << " at position Point(" << posX << ", " << posY << ")"
    //     << " with speed Point(" << speedX << ", " << speedY << ")"
    //     << std::endl;
  } else if ( player_team == T_THEIR_TEAM ) {
    opponents.push_back(cPlr);
    // std::cerr << "Saw their player number " << player_number
    //	      << " at position Point(" << posX << ", " << posY << ")"
    //      << " with speed Point(" << speedX << ", " << speedY << ")"
    //      << std::endl;
  } else {
    std::cerr << "Coach saw unknown player number " << player_number
    	      << " at position Point(" << posX << ", " << posY << ")"
	      << " with speed Point(" << speedX << ", " << speedY << ")"
	      << std::endl;
  }

}


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "SensorInterface.icc"
#endif
