/* -*- Mode: C++ -*-
 * $Id: Sensors.cc,v 1.4 2000/10/24 15:15:17 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Sensors.cc
 * Unitname: BasicSystem
 * $Revision: 1.4 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/10/24 15:15:17 $
 *
 *
 * HISTORY
 *
 * $Log: Sensors.cc,v $
 * Revision 1.4  2000/10/24 15:15:17  frehe
 * Fixed the generation of the onNewCycle event.
 *
 * Revision 1.3  2000/09/05 21:15:53  frehe
 * Made minor changes.
 *
 * Revision 1.2  2000/09/04 08:46:45  frehe
 * Fixed a naming error in the sensors.
 *
 * Revision 1.1  2000/09/03 20:41:14  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file Sensors.cc
 * Implementation file for the class Sensors.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2000/10/24 15:15:17 $
 */


#include <cstdio>
#include <cstring>
#include <cctype>

#include "Sensors.h"
#include "Controller.h"
#include "SensorInterface.h"
#include "string_utils.h"
#include "rs_error.h"
#include "rs_log.h"


RS_BEGIN_NAMESPACE


Sensors::Sensors(SensorInterface* sensor_interface)
  : theController(NULL),
    theSensorInterface(sensor_interface),
    lastSensorType(SENSOR_UNKNOWN),
    currentTime(0),
    commandQueue()
{
}


void Sensors::init(Controller* controller)
{
  theController = controller;
}


void Sensors::analyze(ServerCommand* command)
{
  lastSensorType = SENSOR_ACTUATORS;
  commandQueue.push(command->getType());
  theSensorInterface->setLastCommand(command);
  theSensorInterface->setLastSensorType(lastSensorType);  
}


void Sensors::analyze(char* sensor_data)
{
  bool initmsg = false;

  //std::cerr << "Entering analyze...\n";

  lastSensorType = SENSOR_UNKNOWN;
  switch ( sensor_data[3] ) {            
  case 'e': lastSensorType = SENSOR_VISUAL; break; // see
  case 'n': lastSensorType = SENSOR_SENSE;  break; // sense
  case 'a': lastSensorType = SENSOR_AURAL;  break; // hear
  case 'i': initmsg = true;
            lastSensorType = SENSOR_INIT;   break; // init message
  default:  RS_ERROR("Sent an illegal message");   // error message
            lastSensorType = SENSOR_ERROR;  return;
  }

  if ( initmsg ) {
    analyzeInitMessage(sensor_data);
    RS_LOG( LA_SENSORS, toString(currentTime) + ": Got Init Message!");
  } else {
    currentTime = getInt(&sensor_data);
    if ( theController->updateTime(currentTime) ) {
      theSensorInterface->newCycle();
      RS_LOG( LA_SENSORS, toString(currentTime) + "--- NEW CYCLE ---");
    }
    
    switch ( lastSensorType ) {
    case SENSOR_VISUAL:
      analyzeVisualSensor(sensor_data);
      RS_LOG( LA_SENSORS, toString(currentTime) + ": Saw!");
      break;
    case SENSOR_SENSE:
      analyzeSenseSensor(sensor_data);
      RS_LOG( LA_SENSORS, toString(currentTime) + ": Sensed!");
      break;
    case SENSOR_AURAL:
      analyzeAuralSensor(sensor_data);
      RS_LOG( LA_SENSORS, toString(currentTime) + ": Heard!");
      break;
    default:
      RS_ERROR(toString(currentTime) + ": Unknown sensor!");
      return;
    }
  }

  theSensorInterface->setLastSensorType(lastSensorType);
  
  //std::cerr << "Leaving analyze...\n";
}


// Parse initialize message 
// Code adopted from CMUnited'98 
void Sensors::analyzeInitMessage(char* iniInfo)
{
  char side;
  int number;
  PlayMode play_mode;
  bool reconnect = false;
  
  char mode[100];

  if ( !(strncmp(iniInfo,"(init",4)) ) {
    // It's an init msg
    sscanf(iniInfo,"(init %c %d %[^)]",&side, &number, mode);
  }
  else if ( !(strncmp(iniInfo,"(reconnect",4)) ) {
    // It's a reconnect msg
    sscanf(iniInfo,"(reconnect %c %[^)]",&side, mode);
    reconnect = true;
    //printf("reconnecting to %d on side %c!\n",number,side);
  }

  if ( mode[0] == 'b' ){ // before_kick_off
    play_mode = PM_BEFORE_KICK_OFF;
    /*
    if ( my_side == S_LEFT ) {
      play_mode = PM_OUR_KICK_OFF;
    } else {
      play_mode = PM_THEIR_KICK_OFF;
    }
    */
  }
  // Maybe add a "real" parser for the play modes...
  else  // Act as if the game is in progress
    play_mode = PM_PLAY_ON;

  theSensorInterface->hearInitMessage(side, number, play_mode, reconnect);
  
  printf( "------------------------------------------------------------------------\n");
  printf( "Connected player %d on team %s\n", number, theSensorInterface->getOurTeamName().c_str());
  printf( "He's playing on the %c side of the field.\n", side);
  printf( "------------------------------------------------------------------------\n");
}


void Sensors::analyzeSenseSensor(char* sense_info)
{
  // A temporary hack...
  static UInt16 kicks      = 0;
  static UInt16 dashes     = 0;
  static UInt16 turns      = 0;
  static UInt16 says       = 0;
  static UInt16 turn_necks = 0;

  getWord(&sense_info);  
  sense_info += 10;       // "view_mode "

  ViewQuality view_quality = VQ_UNKNOWN;
  switch ( sense_info[0] ){
  case 'h': view_quality = VQ_HIGH; break;  // high 
  case 'l': view_quality = VQ_LOW;  break;  // low  
  default:  RS_ERROR("Unknown view quality");
  }

  getNextWord(&sense_info);  

  ViewWidth view_width = VW_UNKNOWN;
  switch ( sense_info[1] ){
  case 'o': view_width = VW_NORMAL; break;  // normal 
  case 'a': view_width = VW_NARROW; break;  // narrow 
  case 'i': view_width = VW_WIDE;   break;  // wide   
  default:  RS_ERROR("Unknown view width");
  }
   
  Float32 stamina  = getFloat(&sense_info);
  Float32 effort   = getFloat(&sense_info);
  Float32 speed    = getFloat(&sense_info);
  Float32 head_ang = getFloat(&sense_info);

  Int16 old_dashes = dashes;
  Int16 old_kicks = kicks;
  Int16 old_says = says;
  Int16 old_turns = turns;
  Int16 old_turn_necks = turn_necks;
  
  kicks      = getInt(&sense_info);
  dashes     = getInt(&sense_info);
  turns      = getInt(&sense_info);
  says       = getInt(&sense_info);
  turn_necks = getInt(&sense_info);
  
  old_dashes     = dashes - old_dashes;
  old_kicks      = kicks - old_kicks;
  old_says       = says - old_says;
  old_turns      = turns - old_turns;
  old_turn_necks = turn_necks - old_turn_necks;

  while (!commandQueue.empty()) {
    switch (commandQueue.front()) {
    case CMD_Dash:
      old_dashes--;
      break;
    case CMD_Kick:
      old_kicks--;
      break;
    case CMD_Say:
      old_says--;
      break;
    case CMD_Turn:
      old_turns--;
      break;
    case CMD_TurnNeck:
      old_turn_necks--;
      break;
    default:
      //CMD_Unknown, CMD_Bye, CMD_Catch, CMD_ChangeView, CMD_Init, CMD_Move, CMD_SenseBody
      break;
    }
    commandQueue.pop();
  }

  if ( old_dashes < 0 ||
       old_kicks < 0 ||
       old_says < 0 ||
       old_turns < 0 ||
       old_turn_necks < 0 )
    theController->delayedActions(old_dashes, old_kicks, old_says,
				  old_turns, old_turn_necks);

  if ( old_dashes > 0 ||
       old_kicks > 0 ||
       old_says > 0 ||
       old_turns > 0 ||
       old_turn_necks > 0 )
    theController->missingActions(old_dashes, old_kicks, old_says,
				  old_turns, old_turn_necks);

  theSensorInterface->senseBody(view_quality, view_width, stamina,
				effort, speed, head_ang, kicks,
				dashes, turns, says, turn_necks);

  //std::cerr << "sense_info = " << sense_info << endl;
  
}

void Sensors::analyzeVisualSensor(char* sightInfo)
{

  RS_LOG( LA_SENSORS, "SAW: " << sightInfo );
  
  static ObjectType object_type;  
  static ViewQuality view_qual;

  static Float32U dist;
  static Float32U ang;
  static Float32U dir_chng;
  static Float32U dist_chng;
  static Float32U body_dir;
  static Float32U face_dir;
  static UInt16U player_number;
  static char player_team_name[64];
  static Side player_side;
  static Team player_team;
  static MarkerName marker;
  static SideLineName line;
  static bool behind;

  while (*sightInfo != ')') {

    dist.makeUnknown();
    dir_chng.makeUnknown();
    dist_chng.makeUnknown();
    body_dir.makeUnknown();
    face_dir.makeUnknown();
    player_number.makeUnknown();
    player_team_name[0] = '\0';
    player_side = S_UNKNOWN;
    player_team = T_UNKNOWN;
    marker = UNKNOWN_MARKER;
    line = SL_UNKNOWN;
    behind = false;
 
    //std::cerr << sightInfo << std::endl; //"Before getWord...\n";
    sightInfo += 3;          // " ((" 
    //getWord(&sightInfo);            // " ((" 
    //std::cerr << sightInfo << std::endl;
    //std::cerr << "After getWord...\n";
    //std::cerr << ".";
    
    if ( *sightInfo=='g' ) {
      object_type = OBJ_MARKER;
      sightInfo+=5;          // "goal " 
      if ( *sightInfo=='r' )
	marker = GOAL_R;
      else if ( *sightInfo=='l' )
	marker = GOAL_L;
      else
	RS_ERROR("goal ?");
    } else if ( *sightInfo=='G' ){
      object_type = OBJ_MARKER_BEHIND;
      behind = true;
      marker = UNKNOWN_GOAL;
    } else if ( *sightInfo=='f' ){
      object_type = OBJ_MARKER;
      sightInfo+=5;          // "flag " 
      if ( *sightInfo=='r' ){
	sightInfo+=2;
	if ( *sightInfo=='0' )
	  marker = FLAG_R0;
	else if ( *sightInfo=='b' ){
	  sightInfo+=1;
	  if ( *sightInfo==')' )
	    marker = FLAG_RB; 
	  else {
	    sightInfo+=1;
	    if ( *sightInfo=='1' )   marker = FLAG_RB10; else
	      if ( *sightInfo=='2' )   marker = FLAG_RB20; else
		if ( *sightInfo=='3' )   marker = FLAG_RB30; else
		  RS_ERROR("flag r b ?");
	  }  
	} else
	  if ( *sightInfo=='t' ){
	    sightInfo+=1;
	    if ( *sightInfo==')' )     marker = FLAG_RT; 
	    else{
	      sightInfo+=1;
	      if ( *sightInfo=='1' )   marker = FLAG_RT10; else
		if ( *sightInfo=='2' )   marker = FLAG_RT20; else
		  if ( *sightInfo=='3' )   marker = FLAG_RT30; else
		    RS_ERROR("flag r t ?");
	    }  
	  } else
	    RS_ERROR("flag r ?");
      } else
	if ( *sightInfo=='l' ){
	  sightInfo+=2;
	  if ( *sightInfo=='0' )       marker = FLAG_L0;   else
	    if ( *sightInfo=='b' ){
	      sightInfo+=1;
	      if ( *sightInfo==')' )     marker = FLAG_LB; 
	      else{
		sightInfo+=1;
		if ( *sightInfo=='1' )   marker = FLAG_LB10; else
		  if ( *sightInfo=='2' )   marker = FLAG_LB20; else
		    if ( *sightInfo=='3' )   marker = FLAG_LB30; else
		      RS_ERROR("flag l b ?");
	      }  
	    } else
	      if ( *sightInfo=='t' ){
		sightInfo+=1;
		if ( *sightInfo==')' )     marker = FLAG_LT; 
		else{
		  sightInfo+=1;
		  if ( *sightInfo=='1' )   marker = FLAG_LT10; else
		    if ( *sightInfo=='2' )   marker = FLAG_LT20; else
		      if ( *sightInfo=='3' )   marker = FLAG_LT30; else
			RS_ERROR("flag l t ?");
		}  
	      } else
		RS_ERROR("flag l ?");
	} else
	  if ( *sightInfo=='t' ){
	    sightInfo+=2;
	    if ( *sightInfo=='0' )       marker = FLAG_T0;   else
	      if ( *sightInfo=='l' ){
		sightInfo+=2;
		if ( *sightInfo=='1' )     marker = FLAG_TL10; else
		  if ( *sightInfo=='2' )     marker = FLAG_TL20; else
		    if ( *sightInfo=='3' )     marker = FLAG_TL30; else
		      if ( *sightInfo=='4' )     marker = FLAG_TL40; else
			if ( *sightInfo=='5' )     marker = FLAG_TL50; else
			  RS_ERROR("flag t l ?");
	      } else
		if ( *sightInfo=='r' ){
		  sightInfo+=2;
		  if ( *sightInfo=='1' )     marker = FLAG_TR10; else
		    if ( *sightInfo=='2' )     marker = FLAG_TR20; else
		      if ( *sightInfo=='3' )     marker = FLAG_TR30; else
			if ( *sightInfo=='4' )     marker = FLAG_TR40; else
			  if ( *sightInfo=='5' )     marker = FLAG_TR50; else
			    RS_ERROR("flag t r ?");
		} else
		  RS_ERROR("flag t ?");
	  } else
	    if ( *sightInfo=='b' ){
	      sightInfo+=2;
	      if ( *sightInfo=='0' )       marker = FLAG_B0;   else
		if ( *sightInfo=='l' ){
		  sightInfo+=2;
		  if ( *sightInfo=='1' )     marker = FLAG_BL10; else
		    if ( *sightInfo=='2' )     marker = FLAG_BL20; else
		      if ( *sightInfo=='3' )     marker = FLAG_BL30; else
			if ( *sightInfo=='4' )     marker = FLAG_BL40; else
			  if ( *sightInfo=='5' )     marker = FLAG_BL50; else
			    RS_ERROR("flag b l ?");
		} else
		  if ( *sightInfo=='r' ){
		    sightInfo+=2;
		    if ( *sightInfo=='1' )     marker = FLAG_BR10; else
		      if ( *sightInfo=='2' )     marker = FLAG_BR20; else
			if ( *sightInfo=='3' )     marker = FLAG_BR30; else
			  if ( *sightInfo=='4' )     marker = FLAG_BR40; else
			    if ( *sightInfo=='5' )     marker = FLAG_BR50; else
			      RS_ERROR("flag b r ?");
		  } else
		    RS_ERROR("flag b ?");
	    } else
	      if ( *sightInfo=='c' ){
		sightInfo+=1;
		if ( *sightInfo==')' )       marker = FLAG_C;
		else{
		  sightInfo+=1;
		  if ( *sightInfo=='b' )     marker = FLAG_CB; else
		    if ( *sightInfo=='t' )     marker = FLAG_CT; else
		      RS_ERROR("flag c ?");
		}
	      } else
		if ( *sightInfo=='p' ){
		  sightInfo+=2;
		  if ( *sightInfo=='r' ){
		    sightInfo+=2;
		    if ( *sightInfo=='t')      marker = FLAG_PRT; else
		      if ( *sightInfo=='c')      marker = FLAG_PRC; else
			if ( *sightInfo=='b')      marker = FLAG_PRB; else	    
			  RS_ERROR("flag p r ?");
		  } else
		    if ( *sightInfo=='l' ){
		      sightInfo+=2;
		      if ( *sightInfo=='t')      marker = FLAG_PLT; else
			if ( *sightInfo=='c')      marker = FLAG_PLC; else
			  if ( *sightInfo=='b')      marker = FLAG_PLB; else	    
			    RS_ERROR("flag p l ?");
		    } else
		      RS_ERROR("flag p ?");
		} else
		  if ( *sightInfo=='g' ){
		    sightInfo+=2;
		    if ( *sightInfo=='l' ){
		      sightInfo+=2;
		      if ( *sightInfo=='t' )     marker = FLAG_GLT; else
			if ( *sightInfo=='b' )     marker = FLAG_GLB; else
			  RS_ERROR("flag g l ?");
		    } else
		      if ( *sightInfo=='r' ){
			sightInfo+=2;
			if ( *sightInfo=='t' )     marker = FLAG_GRT; else
			  if ( *sightInfo=='b' )     marker = FLAG_GRB; else
			    RS_ERROR("flag g r ?");
		      } else
			RS_ERROR("flag g ?");
		  } else
		    RS_ERROR("flag ?");
    } else
      if ( *sightInfo=='F' ){
	object_type = OBJ_MARKER_BEHIND;
	behind = true;
	marker = UNKNOWN_FLAG;
      } else
	if ( *sightInfo=='l' ){
	  object_type = OBJ_LINE;
	  sightInfo+=5;          // "line " 
	  if ( *sightInfo=='r' )         line   = SL_RIGHT;  else
	    if ( *sightInfo=='l' )         line   = SL_LEFT;   else
	      if ( *sightInfo=='t' )         line   = SL_TOP;    else
		if ( *sightInfo=='b' )         line   = SL_BOTTOM; else
		  RS_ERROR("line ?");
	} else
	  if ( *sightInfo=='p' || *sightInfo=='P' ){
	    object_type = OBJ_PLAYER;
	    sightInfo+=6;                          // "player" 
	    if ( *sightInfo == ' ' ){              // there's a team  
	      sightInfo++;
	      // get the team name
	      int a = 0;
	      while ( *sightInfo != ' ' && *sightInfo != ')' && a < 63 )
		player_team_name[a++]=*sightInfo++;
	      if ( a == 64 )
		RS_WARNING("team name WAY to long > 63 characters!");
	      player_team_name[a] = '\0';

	      if ( !strcmp(player_team_name, theSensorInterface->getOurTeamName().c_str()) ) {
		player_side = theSensorInterface->getOurSide();
		player_team = T_OUR_TEAM;
	      } else if ( theSensorInterface->getTheirTeamName() == "" ) {
		player_side = theSensorInterface->getTheirSide();
		player_team = T_THEIR_TEAM;
		theSensorInterface->setTheirTeamName(player_team_name);
	      } else if ( !strcmp(player_team_name,
				  theSensorInterface->getTheirTeamName().c_str()) ) {
		player_side = theSensorInterface->getTheirSide();
		player_team = T_THEIR_TEAM;
	      }
	
	      if ( *sightInfo== ' ' ){             // there's a number 
		player_number = getInt(&sightInfo);
	      }
	    }
	  } else
	    if ( *sightInfo=='b' || *sightInfo=='B' ) {
	      object_type = OBJ_BALL;
	    }
	    else {
	      RS_ERROR("unknown object in analyzeVisualSensors: " + std::string(sightInfo));
	      return;
	    }
  
    advanceTo(')',&sightInfo);              // advance to end of object 

    /************************************/

    ang = getFloat(&sightInfo);

    if ( *sightInfo != ')' ) {                  // 'high' quality     
      view_qual = VQ_HIGH;
      dist = ang;
      ang = getFloat(&sightInfo);
    }
    else {
      view_qual = VQ_LOW;
    }

    if ( *sightInfo != ')' ){
      dist_chng = getFloat(&sightInfo);
      dir_chng  = getFloat(&sightInfo);
    }

    if ( *sightInfo != ')' ){
      if (object_type != OBJ_PLAYER)
	RS_WARNING("Only players should have body_dir");
      body_dir = getFloat(&sightInfo);
    }

    if ( *sightInfo != ')' ){
      if (object_type != OBJ_PLAYER)
	RS_WARNING("Only players should have face_dir");
      face_dir = getFloat(&sightInfo);
    }

    if ( *sightInfo != ')' ) {
      RS_ERROR("Should be done with object info here");
      return;
    }

    sightInfo++;                                // ")" 

    /************************************/

    //std::cerr << "Before switch...\n";
    
    switch (object_type){
    case OBJ_MARKER_BEHIND:
    case OBJ_MARKER:
      theSensorInterface->seeMarker(marker, dist, ang, behind);
      //theSensorInterface->seeMarker(marker, dist, ang, dist_chng, dir_chng, behind);
      break;
    case OBJ_LINE:
      theSensorInterface->seeLine(line, dist, ang);
      break;
    case OBJ_BALL:
      theSensorInterface->seeBall(dist, ang, dist_chng, dir_chng);
      break;
    case OBJ_PLAYER:
      /*
	if ( !player_side ){                      // Too far for team or num 
	if ( view_qual == VQ_LOW )                // low quality   
	seePlayer(ang, time);
	else if ( dir_chng.isUnknown() )                   // high quality  
	seePlayer(dist, ang, time);
	else {                                        // know direction
	RS_ERROR("Shouldn't know dir_chng when the player's far");
	return;
	}
	}  

	else{
	if ( !player_number ){                  // Too far for number     
	if ( view_qual == VQ_LOW )                // low quality   
	seePlayer(player_side, ang, time);
	else if ( dir_chng.isUnknown() )                   // high quality  
	seePlayer(player_side, dist, ang, time);
	else {                                           // know direction
	RS_ERROR("Shouldn't know dir_chng when the team member's far");
	return;
	}
	}

	else{                                   // Know side AND number   
	if ( view_qual == VQ_LOW )                // low quality   
	seePlayer(player_side, player_number, ang, time);
	else if ( dir_chng.IsUnknown() ){                  // high quality  
	printf("%s\n",sightInfo-30);
	RS_ERROR("Should know dir_chng when know number");
	seePlayer(player_side, player_number, dist, ang, time);
	}
	else */                                       // know direction

      //cout << "Saw player " << player_side << ":" << player_number << endl;
      
      theSensorInterface->seePlayer(player_side, player_team, player_number,
				    dist, ang, dist_chng, dir_chng,
				    body_dir, face_dir);
      /*
	}
	}
	*/
    }
    //std::cerr << "sightInfo = " << sightInfo << std::endl;
  }
  //std::cerr << "sightInfo = " << sightInfo << std::endl;
}


/****************************************************************************************/

void Sensors::analyzeAuralSensor(char* sound_info)
{
  if ( sound_info[1] == 'r' ){   // Referee or Coach message 
    sound_info += 9;             // " referee "
    // I'm not sure this is correct after the last changes!!!
    if ( islower(sound_info[0]) )
      analyzeRefereeSound(sound_info);  // Coach messages start with capital letters or are numbers
    else
      // Return the rest of the string, even the last ')'
      theSensorInterface->hearCoach(sound_info);
    return;
  }
  
  // Return the rest of the string, even the last ')'
  theSensorInterface->hearPlayer(sound_info);
  
}

/****************************************************************************************/

void Sensors::analyzeRefereeSound(char* msg)
{

  static const char left_side = 'l';
  static const char right_side = 'r';
  
  PlayMode play_mode = PM_UNKNOWN;
  Side side = S_UNKNOWN;
  
  switch ( msg[0] ) {
  case 'p': play_mode = PM_PLAY_ON; break;           // play_on
  case 'k': 
    if ( msg[5] == 'i' ){                                  // kick_in
      play_mode = PM_KICK_IN;
      if ( msg[8] == left_side )
	side = S_LEFT;
      else if ( msg[8] == right_side )
	side = S_RIGHT;
      else 
	RS_ERROR("kick_in_?");
    }
    else if ( msg[5] == 'o' ){                            // kick_off
      play_mode = PM_KICK_OFF;
      if ( msg[9] == left_side )
	side = S_LEFT;
      else if ( msg[9] == right_side )
	side = S_RIGHT;
      else 
	RS_ERROR("kick_off_?");
    }
    else
      RS_ERROR("referee k..?");
    break;
  case 'g': 
    if ( msg[5] == 'k' ) {                                 // goal_kick 
      play_mode = PM_GOAL_KICK;
      if ( msg[10] == left_side )
	side = S_LEFT;
      else if ( msg[10] == right_side )
	side = S_RIGHT;
      else 
	RS_ERROR("goal_kick_?");
    }
    else if ( msg[5] == 'e' ){                           // goalie_catch_ball
      play_mode = PM_GOALIE_GOT_BALL;
      if ( msg[18] == left_side )
	side = S_LEFT;
      else if ( msg[18] == right_side )
	side = S_RIGHT;
      else
	RS_ERROR("goalie_catch_ball_?");
    }
    else if ( msg[5] == left_side ){                    // goal
      play_mode = PM_GOAL;
      side = S_LEFT;
      theSensorInterface->hearScore(side, getInt(&msg[7]));
    }
    else if ( msg[5] == right_side ){
      play_mode = PM_GOAL;
      side = S_RIGHT;
      theSensorInterface->hearScore(side, getInt(&msg[7]));
    }
    else 
      RS_ERROR("referee g..?");
    break;
  case 'c':                                               // corner_kick 
    play_mode = PM_CORNER_KICK;
    if ( msg[12] == left_side )
      side = S_LEFT;
    else if ( msg[12] == right_side )
      side = S_RIGHT;
    else 
      RS_ERROR("corner_kick_?");
    break;
  case 'd': play_mode = PM_DROP_BALL; break;        // drop_ball     
  case 'o':                                               // offside     
    play_mode = PM_OFFSIDE_KICK;
    if ( msg[8] == left_side )
      side = S_LEFT;
    else if ( msg[8] == right_side )
      side = S_RIGHT;
    else 
      RS_ERROR("offside_?");
    break;
  case 'f':
    if ( msg[5] == 'k' ){                                 // free_kick
      play_mode = PM_FREE_KICK;
      if ( msg[10] == left_side )
	side = S_LEFT;
      else if ( msg[10] == right_side )
	side = S_RIGHT;
      else 
	RS_ERROR("free_kick_?");
    }
    // ???? Why don't I have to do anything?
    else if ( msg[5] == left_side )                     // foul 
      ;
    else if ( msg[5] == right_side )
      ;
    else 
      RS_ERROR("referee f..?");
    break;
  case 'h':                                               // half_time 
    play_mode = PM_HALF_TIME;
    break;             
  case 'b':
    play_mode = PM_BEFORE_KICK_OFF;
    break;       // before_kick_off 
  case 't': 
    if ( msg[5] == 'u' ) {                             // time_up 
      play_mode = PM_TIME_UP; 
    }
    else if ( msg[5] == 'o' ) {                         // time_over 
      play_mode = PM_TIME_OVER;
      break;
    }
    else if ( msg[5] == 'e' ){                            // time_extended 
      play_mode = PM_EXTENDED_TIME;
    }
    else 
      RS_ERROR("referee t..?");
    break;
  default:
    RS_ERROR("Referee msg ????");
  }

  theSensorInterface->hearReferee(play_mode, side);

}

void Sensors::analyzeCoachMessage(char* sightInfo)
{
  static ObjectType object_type;  

  static Float32U posX;
  static Float32U posY;
  static Float32U speedX;
  static Float32U speedY;
  static Float32U bodyDir;
  static Float32U neckDir;
  static UInt16U player_number;
  static char player_team_name[64];
  static Side player_side;
  static Team player_team;
  static MarkerName marker;

  if ( sightInfo[2] != 'e' )
    return;
  
  theSensorInterface->setCoachTime(getInt(&sightInfo));

  while (*sightInfo != ')') {

    posX.makeUnknown();
    posY.makeUnknown();
    speedX.makeUnknown();
    speedY.makeUnknown();
    bodyDir.makeUnknown();
    neckDir.makeUnknown();
    player_number.makeUnknown();
    player_team_name[0] = '\0';
    player_side = S_UNKNOWN;
    player_team = T_UNKNOWN;
    marker = UNKNOWN_MARKER;
 
    //std::cerr << sightInfo << std::endl; //"Before getWord...\n";
    sightInfo += 3;          // " ((" 
    //getWord(&sightInfo);            // " ((" 
    //std::cerr << sightInfo << std::endl;
    //std::cerr << "After getWord...\n";
    //std::cerr << ".";
    
    if ( *sightInfo=='g' ) {
      object_type = OBJ_MARKER;
      sightInfo+=5;          // "goal " 
      if ( *sightInfo=='r' )
	marker = GOAL_R;
      else if ( *sightInfo=='l' )
	marker = GOAL_L;
      else
	RS_ERROR("goal ?");
    } else if ( *sightInfo=='f' ){
      object_type = OBJ_MARKER;
      sightInfo+=5;          // "flag " 
      if ( *sightInfo=='r' ){
	sightInfo+=2;
	if ( *sightInfo=='0' )
	  marker = FLAG_R0;
	else if ( *sightInfo=='b' ){
	  sightInfo+=1;
	  if ( *sightInfo==')' )
	    marker = FLAG_RB; 
	  else {
	    sightInfo+=1;
	    if ( *sightInfo=='1' )   marker = FLAG_RB10; else
	      if ( *sightInfo=='2' )   marker = FLAG_RB20; else
		if ( *sightInfo=='3' )   marker = FLAG_RB30; else
		  RS_ERROR("flag r b ?");
	  }  
	} else
	  if ( *sightInfo=='t' ){
	    sightInfo+=1;
	    if ( *sightInfo==')' )     marker = FLAG_RT; 
	    else{
	      sightInfo+=1;
	      if ( *sightInfo=='1' )   marker = FLAG_RT10; else
		if ( *sightInfo=='2' )   marker = FLAG_RT20; else
		  if ( *sightInfo=='3' )   marker = FLAG_RT30; else
		    RS_ERROR("flag r t ?");
	    }  
	  } else
	    RS_ERROR("flag r ?");
      } else
	if ( *sightInfo=='l' ){
	  sightInfo+=2;
	  if ( *sightInfo=='0' )       marker = FLAG_L0;   else
	    if ( *sightInfo=='b' ){
	      sightInfo+=1;
	      if ( *sightInfo==')' )     marker = FLAG_LB; 
	      else{
		sightInfo+=1;
		if ( *sightInfo=='1' )   marker = FLAG_LB10; else
		  if ( *sightInfo=='2' )   marker = FLAG_LB20; else
		    if ( *sightInfo=='3' )   marker = FLAG_LB30; else
		      RS_ERROR("flag l b ?");
	      }  
	    } else
	      if ( *sightInfo=='t' ){
		sightInfo+=1;
		if ( *sightInfo==')' )     marker = FLAG_LT; 
		else{
		  sightInfo+=1;
		  if ( *sightInfo=='1' )   marker = FLAG_LT10; else
		    if ( *sightInfo=='2' )   marker = FLAG_LT20; else
		      if ( *sightInfo=='3' )   marker = FLAG_LT30; else
			RS_ERROR("flag l t ?");
		}  
	      } else
		RS_ERROR("flag l ?");
	} else
	  if ( *sightInfo=='t' ){
	    sightInfo+=2;
	    if ( *sightInfo=='0' )       marker = FLAG_T0;   else
	      if ( *sightInfo=='l' ){
		sightInfo+=2;
		if ( *sightInfo=='1' )     marker = FLAG_TL10; else
		  if ( *sightInfo=='2' )     marker = FLAG_TL20; else
		    if ( *sightInfo=='3' )     marker = FLAG_TL30; else
		      if ( *sightInfo=='4' )     marker = FLAG_TL40; else
			if ( *sightInfo=='5' )     marker = FLAG_TL50; else
			  RS_ERROR("flag t l ?");
	      } else
		if ( *sightInfo=='r' ){
		  sightInfo+=2;
		  if ( *sightInfo=='1' )     marker = FLAG_TR10; else
		    if ( *sightInfo=='2' )     marker = FLAG_TR20; else
		      if ( *sightInfo=='3' )     marker = FLAG_TR30; else
			if ( *sightInfo=='4' )     marker = FLAG_TR40; else
			  if ( *sightInfo=='5' )     marker = FLAG_TR50; else
			    RS_ERROR("flag t r ?");
		} else
		  RS_ERROR("flag t ?");
	  } else
	    if ( *sightInfo=='b' ){
	      sightInfo+=2;
	      if ( *sightInfo=='0' )       marker = FLAG_B0;   else
		if ( *sightInfo=='l' ){
		  sightInfo+=2;
		  if ( *sightInfo=='1' )     marker = FLAG_BL10; else
		    if ( *sightInfo=='2' )     marker = FLAG_BL20; else
		      if ( *sightInfo=='3' )     marker = FLAG_BL30; else
			if ( *sightInfo=='4' )     marker = FLAG_BL40; else
			  if ( *sightInfo=='5' )     marker = FLAG_BL50; else
			    RS_ERROR("flag b l ?");
		} else
		  if ( *sightInfo=='r' ){
		    sightInfo+=2;
		    if ( *sightInfo=='1' )     marker = FLAG_BR10; else
		      if ( *sightInfo=='2' )     marker = FLAG_BR20; else
			if ( *sightInfo=='3' )     marker = FLAG_BR30; else
			  if ( *sightInfo=='4' )     marker = FLAG_BR40; else
			    if ( *sightInfo=='5' )     marker = FLAG_BR50; else
			      RS_ERROR("flag b r ?");
		  } else
		    RS_ERROR("flag b ?");
	    } else
	      if ( *sightInfo=='c' ){
		sightInfo+=1;
		if ( *sightInfo==')' )       marker = FLAG_C;
		else{
		  sightInfo+=1;
		  if ( *sightInfo=='b' )     marker = FLAG_CB; else
		    if ( *sightInfo=='t' )     marker = FLAG_CT; else
		      RS_ERROR("flag c ?");
		}
	      } else
		if ( *sightInfo=='p' ){
		  sightInfo+=2;
		  if ( *sightInfo=='r' ){
		    sightInfo+=2;
		    if ( *sightInfo=='t')      marker = FLAG_PRT; else
		      if ( *sightInfo=='c')      marker = FLAG_PRC; else
			if ( *sightInfo=='b')      marker = FLAG_PRB; else	    
			  RS_ERROR("flag p r ?");
		  } else
		    if ( *sightInfo=='l' ){
		      sightInfo+=2;
		      if ( *sightInfo=='t')      marker = FLAG_PLT; else
			if ( *sightInfo=='c')      marker = FLAG_PLC; else
			  if ( *sightInfo=='b')      marker = FLAG_PLB; else	    
			    RS_ERROR("flag p l ?");
		    } else
		      RS_ERROR("flag p ?");
		} else
		  if ( *sightInfo=='g' ){
		    sightInfo+=2;
		    if ( *sightInfo=='l' ){
		      sightInfo+=2;
		      if ( *sightInfo=='t' )     marker = FLAG_GLT; else
			if ( *sightInfo=='b' )     marker = FLAG_GLB; else
			  RS_ERROR("flag g l ?");
		    } else
		      if ( *sightInfo=='r' ){
			sightInfo+=2;
			if ( *sightInfo=='t' )     marker = FLAG_GRT; else
			  if ( *sightInfo=='b' )     marker = FLAG_GRB; else
			    RS_ERROR("flag g r ?");
		      } else
			RS_ERROR("flag g ?");
		  } else
		    RS_ERROR("flag ?");
    } else
      if ( *sightInfo=='p' ){
	object_type = OBJ_PLAYER;
	sightInfo+=6;                          // "player" 
	if ( *sightInfo == ' ' ){              // there's a team  
	  sightInfo++;
	  // get the team name
	  int a = 0;
	  while ( *sightInfo != ' ' && *sightInfo != ')' && a < 63 )
	    player_team_name[a++]=*sightInfo++;
	  if ( a == 64 )
	    RS_WARNING("team name WAY to long > 63 characters!");
	  player_team_name[a] = '\0';
	    
	  if ( !strcmp(player_team_name, theSensorInterface->getOurTeamName().c_str()) ) {
	    player_side = theSensorInterface->getOurSide();
	    player_team = T_OUR_TEAM;
	  } else if ( theSensorInterface->getTheirTeamName() == "" ) {
	    player_side = theSensorInterface->getTheirSide();
	    player_team = T_THEIR_TEAM;
	  } else if ( !strcmp(player_team_name,
			      theSensorInterface->getTheirTeamName().c_str()) ) {
	    player_side = theSensorInterface->getTheirSide();
	    player_team = T_THEIR_TEAM;
	  }
	    
	  if ( *sightInfo== ' ' ){             // there's a number 
	    player_number = getInt(&sightInfo);
	  }
	}
      } else
	  if ( *sightInfo=='b' ) {
	    object_type = OBJ_BALL;
	  }
	  else {
	    RS_ERROR("unknown object in analyzeCoachMessage " + std::string(sightInfo));
	    return;
	  }
    
    advanceTo(')',&sightInfo);              // advance to end of object 

    /************************************/

    // All objects should have an absolute position
    posX = getFloat(&sightInfo);
    posY = getFloat(&sightInfo);

    // Players have absolute body direction and and absolute neck direction too
    if ( object_type == OBJ_PLAYER ) {
      bodyDir = getFloat(&sightInfo);
      neckDir = getFloat(&sightInfo);
    }

    // All moving objects have an absolute speed
    if ( object_type == OBJ_PLAYER ||
	 object_type == OBJ_BALL ) {
      speedX = getFloat(&sightInfo);
      speedY = getFloat(&sightInfo);
    }

    sightInfo++;                                // ")" 

    /************************************/

    //std::cerr << "Before switch...\n";
    
    switch (object_type){
    case OBJ_MARKER:
      theSensorInterface->coachSeeMarker(marker, posX, posY);
      break;
    case OBJ_BALL:
      theSensorInterface->coachSeeBall(posX, posY, speedX, speedY);
      break;
    case OBJ_PLAYER:
      theSensorInterface->coachSeePlayer(player_side, player_team, player_number,
					 posX, posY, speedX, speedY, bodyDir, neckDir);
      break;
    default:
      RS_ERROR("Unknown object in analyzeCoachMessage!");
    }
    //std::cerr << "sightInfo = " << sightInfo << std::endl;
  }
  //std::cerr << "sightInfo = " << sightInfo << std::endl;

  theSensorInterface->setLastSensorType(SENSOR_COACH);
}

/****************************************************************************/
/* These routines are to save time instead of using sscanf or atof, etc.    */
/* When passing **str_ptr, the pointer is advanced past the number          */
/* When passing  *str    , the pointer remains where it was before          */
/****************************************************************************/

// They are taken from CMUnited'98 created by Peter Stone et al 

Float32 Sensors::getFloat(char** str_ptr){

  Float32 d_frac;
  Float32 result;
  int m_flag;
  int d_flag;

  d_frac = 1.0;
  result = 0.0;
  m_flag = d_flag = 0;

  // Advance to the beginning of the number 
  while( !isdigit(**str_ptr) && **str_ptr!='-' && **str_ptr!='.')
    (*str_ptr)++;

  // Process the number bit by bit 
  while((**str_ptr!=')') && (**str_ptr!=' ') && (**str_ptr!='\n') && (**str_ptr!=',')){
    if (**str_ptr=='.')
      d_flag=1;
    else if (**str_ptr=='-')
      m_flag=1;
    else if (d_flag){
      d_frac *= 10.0;
      result+=(Float32)(**str_ptr-'0')/d_frac;
    }
    else
      result=result*10.0+(Float32)(**str_ptr-'0');
    (*str_ptr)++;
  }
  if (m_flag)
    result=-result;

  //printf("%.1f\n",result);

  return result;
}


// get the number, but don't advance pointer
Float32 Sensors::getFloat(char* str){
  char **str_ptr = &str;
  return getFloat(str_ptr);
}


Int32 Sensors::getInt(char** str_ptr)
{
  Int32 result;
  Int32 m_flag;

  result = 0;
  m_flag = 0;

  // Advance to the beginning of the number
  while( !isdigit(**str_ptr) && **str_ptr!='-')
    (*str_ptr)++;

  // Process the number bit by bit
  while((**str_ptr!=')') && (**str_ptr!=' ') &&
	(**str_ptr!='\n') && (**str_ptr!=',')) {
    if (**str_ptr=='-')
      m_flag=1;
    else
      result=result*10+(Int32)(**str_ptr-'0');
    (*str_ptr)++;
  }
  if (m_flag)
    result=-result;

  return result;
}


int Sensors::getInt(char* str)
{
  char **str_ptr = &str;
  return getInt(str_ptr);
}


void Sensors::getWord(char** str_ptr)
{
  while ( !isalpha(**str_ptr) ) (*str_ptr)++;
}


void Sensors::getNextWord(char** str_ptr)
{
  while ( isalpha(**str_ptr) ) (*str_ptr)++;
  getWord(str_ptr);
}


void Sensors::advanceTo(char c, char** str_ptr)
{
  while ( **str_ptr != c ) (*str_ptr)++;
}


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "Sensors.icc"
#endif
