/* -*- Mode: C++ -*-
 * $Id: ServerInterface.cc,v 1.1 2000/09/03 20:41:14 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: ServerInterface.cc
 * Unitname: BasicSystem
 * $Revision: 1.1 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/03 20:41:14 $
 *
 *
 * HISTORY
 *
 * $Log: ServerInterface.cc,v $
 * Revision 1.1  2000/09/03 20:41:14  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file ServerInterface.cc
 * Implementation file for the class ServerInterface.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2000/09/03 20:41:14 $
 */


#include "ServerInterface.h"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_SYS_FILIO_H
#  include <sys/filio.h>
#elif defined(HAVE_SYS_IOCTL_H)
#  include <sys/ioctl.h>
#endif

#include "rs_error.h"


RS_BEGIN_NAMESPACE


ServerInterface::ServerInterface(const std::string& host,
				 const UInt16 port,
				 const UInt32 server_buffer_size)
  throw()
  : theSocket(host, port, server_buffer_size, SOCKET_TYPE_UDP),
    connected(false),
    theErrorMessage(SERVER_NO_ERROR)
{
}


bool ServerInterface::connect()
  throw(RSException)
{
  if ( connected ) {
    theErrorMessage = SERVER_ALREADY_CONNECTED;
    return false;
  }
    
  // Open the UDP connection
  if ( theSocket.open() ) {
    try {
      theSocket.setOwner(getpid());

      int flags = theSocket.getFlags();
      const int on = 1;

#ifdef HAVE_SYS_FILIO_H
      theSocket.setFlags(flags | O_NONBLOCK | FIOASYNC);
      ioctl(theSocket.getFD(), FIOASYNC, &on);
      ioctl(theSocket.getFD(), FIONBIO, &on);
#elif defined(HAVE_SYS_IOCTL_H)
      theSocket.setFlags(flags | O_NONBLOCK | O_ASYNC);
      ioctl(theSocket.getFD(), O_ASYNC, &on);
      ioctl(theSocket.getFD(), O_NONBLOCK, &on);
#endif

    }
    catch (RSException& e) {
      RS_ERROR("Caught exception:" + e.What());
      throw;
    }
    connected = true;
    return true;
  } else {
    theErrorMessage = SERVER_SOCKET_ERROR;
    return false;
  }
}


bool ServerInterface::disconnect()
  throw(RSException)
{
  if ( connected ) {
    if ( theSocket.close() ) {
      connected = false;
      return true;
    } else {
      theErrorMessage = SERVER_SOCKET_ERROR;
      return false;
    }    
  } else {
    theErrorMessage = SERVER_NOT_CONNECTED;
    return false;
  }
}


bool ServerInterface::sendCommand(const std::string& cmd)
  throw(RSException)
{
  if ( connected ) {
    if ( theSocket.writeData(cmd) ) {
      return true;
    } else {
      theErrorMessage = SERVER_SOCKET_ERROR;
      return false;
    }
  } else {
    theErrorMessage = SERVER_NOT_CONNECTED;
    return false;
  }
}


bool ServerInterface::receiveInput(char* const input)
  throw(RSException)
{
  static Int32 tmp_size;
  if ( connected ) {
    if ( theSocket.readData(input, &tmp_size) ) {
      return true;
    } else {
      if ( theSocket.getErrorMessage() == SOCKET_NO_DATA )
	theErrorMessage = SERVER_NO_DATA;
      else
	theErrorMessage = SERVER_SOCKET_ERROR;
      return false;
    }
  } else {
    theErrorMessage = SERVER_NOT_CONNECTED;
    return false;
  }
}


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "ServerInterface.icc"
#endif
