/* -*- Mode: C++ -*-
 * $Id: AgentView.h,v 1.4 2001/01/11 22:00:17 olalu526 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: AgentView.h
 * Unitname: Framework/view
 * $Revision: 1.4 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: olalu526 $ $Date: 2001/01/11 22:00:17 $
 *
 *
 * HISTORY
 *
 * $Log: AgentView.h,v $
 * Revision 1.4  2001/01/11 22:00:17  olalu526
 * Added the real world model with support for new simulation steps.
 *
 * Revision 1.3  2001/01/10 15:12:57  olalu526
 * Added support for WorldModel to some basic classes.
 *
 * Revision 1.2  2000/10/18 23:05:53  frehe
 * Added lastCoachMessage and lastPlayerMessage.
 *
 * Revision 1.1  2000/09/21 13:46:39  frehe
 * Move the RoboSoc views from rsviews to the rsframework library.
 *
 * Revision 1.2  2000/09/13 00:08:05  frehe
 * Updated the views to inherit from the ViewWithHistory instead and also changed those views that used a static pointer to itself to do process the data in the sensor interface to use the new forEach methods implemented in the FrameworkSensorInterface.
 *
 * Revision 1.1  2000/09/04 15:27:19  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file AgentView.h
 * Include file for the class AgentView.
 *
 * @todo Add more algorithms that determines the position
 *       of the agent.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.4 $
 * $Author: olalu526 $
 * $Date: 2001/01/11 22:00:17 $
 */


#ifndef AGENTVIEW_H
#define AGENTVIEW_H


#include "ViewWithSimulation.h"
#include "AgentObject.h"


RS_BEGIN_NAMESPACE


class ParameterView;
class CommandView;
class MarkerView;
class GameView;


/** Used to represent collections of agent objects. */
//typedef ObjectCollection<AgentObject> AgentCollection;

#define theAgent (*(worldModel->agentData()))

/**
 * @class AgentView AgentView.h
 * This view is used to keep track of the agent,
 * its position, speed and other properties.
 *
 * <!-- @invariant The class invariant is... -->
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.4 $
 * $Author: olalu526 $
 * $Date: 2001/01/11 22:00:17 $
 */
class AgentView : public ViewWithSimulation {
public:
  /** Create an agent view with ViewId @a id and the
    history length @a hist_len. It will be initialized
    with the parameters from the @a parameter_view. */
  AgentView(const ViewId id, 
	    ParameterView* const parameter_view,
	    const UInt hist_len = DEFAULT_HISTORY_LENGTH);

  /** Destroys the agent view. */
  virtual ~AgentView();


  /** Get a constant object representing the agent @a hist cycles ago. */
  const AgentObject& getAgentObject(const Int hist) const;


  /** Returns @c true iff the agent is outside the field. */
  bool isAgentOutsideField() const;


  /** Get the time of the last aural (hear) message. */
  Time getLastHearTime() const;

  /** Get the time of the visual (see) message. */
  Time getLastSeeTime() const;

  /** Get the time of the last sense body message. */
  Time getLastSenseTime() const;


  /** Get the position of the agent @a hist cycles ago. */
  PointUCO getAgentPosition(const Int hist) const;

  /** Get the face direction of the agent @a hist cycles ago. */
  AngleDegUCO getFaceDirection(const Int hist) const;

  /** Get the body direction of the agent @a hist cycles ago. */
  AngleDegUCO getBodyDirection(const Int hist) const;

  /** Get the neck direction of the agent @a hist cycles ago. */
  AngleDegUCO getNeckDirection(const Int hist) const;

  /** Get the effort of the agent @a hist cycles ago. */
  FloatUCO getEffort(const Int hist) const;

  /* Get the recovery of the agent @a hist cycles ago. */
  //FloatUCO getRecover(const Int hist) const;

  /** Get the stamina of the agent @a hist cycles ago. */
  FloatUCO getStamina(const Int hist) const;

  /** Get the amount of speed of the agent @a hist cycles ago. */
  FloatUCO getSpeed(const Int hist) const;

  /** Get the speed vector of the agent @a hist cycles ago. */
  VectorUCO getSpeedVector(const Int hist) const;
  
  /** Get the view width of the agent @a hist cycles ago. */
  ViewWidth getViewWidth(const Int hist) const;

  /** Get the view quality of the agent @a hist cycles ago. */
  ViewQuality getViewQuality(const Int hist) const;

  /** Get the view angle of the agent @a hist cycles ago. */
  AngleDegUnknown getViewAngle(const Int hist) const;

  
  /** Returns @c true iff the agent is the goal keeper. */
  bool isGoalie() const;


  /** Returns the uniform number of the agent. */
  UIntUnknown getUniformNumber() const;


  //Int getHearCapacityFromTeammate() const;
  //Int getHearCapacityFromOpponent() const;
  //Int getCyclesSinceCatch() const;
  //Int getCyclesSinceSay() const;


  /** Returns the number of dashes made upto @a hist cycles ago. */
  UInt getDashes(const Int hist) const;

  /** Returns the number of kicks made upto @a hist cycles ago. */
  UInt getKicks(const Int hist) const;

  /** Returns the number of says made upto @a hist cycles ago. */
  UInt getSays(const Int hist) const;

  /** Returns the number of turns made upto @a hist cycles ago. */
  UInt getTurns(const Int hist) const;

  /** Returns the number of turn necks made upto @a hist cycles ago. */
  UInt getTurnNecks(const Int hist) const;


  /** Returns the last message sent by the coach. */
  const std::string& getLastCoachMessage() const;

  /** Returns the last message sent by a player. */
  const std::string& getLastPlayerMessage() const;

  /** Return the actual index to the time step
    @a i steps from the current time step.
    If @a i is -1 then it means the previous time step
    and if @a i is 1 then it means the next time step.
    OBS! Since the history is implemented as a circular
    array getIndex will return a value that between 0 and
    getHistoryLength-1. It also means that getIndex(1) is the last
    element in the array and getIndex(0) = getIndex(getHistoryLength). */
  UInt getIndex(const Int i) const;

  /** Increase the index to point to the next time step. */
  void increaseIndex();

  
  /** Returns the length of the history, which is the maximum
    number of steps in the past that will be saved. */
  UInt getHistoryLength() const;
  
protected:
  /** A pointer to the parameter view. */
  ParameterView* parameterView;

  /** A pointer to the command view. */
  CommandView* commandView;

  /** A pointer to the game view. */
  GameView* gameView;

  /** A pointer to the marker view. */
  MarkerView* markerView;


  /** The history of the agent. */
  //AgentCollection theAgent;


  /** The time of the last hear message. */
  Time lastHearTime;

  /** The time of the see message. */
  Time lastSeeTime;

  /** The time of the last sense body message. */
  Time lastSenseTime;


  virtual void myNewCycle();
  virtual void myUpdateAfterCoachMessage();
  virtual void myUpdateAfterInit();
  virtual void myUpdateAfterSee();
  virtual void myUpdateAfterHear();
  virtual void myUpdateAfterSense();
  virtual void myUpdateAfterCommand();


  /** Calculates the face direction of the agent based
    on the seen lines. The algorithm is taken from AT Humboldt 97. */
  AngleDegUnknown computeFaceDirection() const;

  /** Calculates the position of the agent based on the
   closest marker and the face direction. */
  PointUnknown computeAgentPosition();
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "AgentView.icc"
#endif


#endif // !AGENTVIEW_H
