/* -*- Mode: C++ -*-
 * $Id: BallView.h,v 1.3 2001/01/11 22:00:17 olalu526 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: BallView.h
 * Unitname: Framework/view
 * $Revision: 1.3 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: olalu526 $ $Date: 2001/01/11 22:00:17 $
 *
 *
 * HISTORY
 *
 * $Log: BallView.h,v $
 * Revision 1.3  2001/01/11 22:00:17  olalu526
 * Added the real world model with support for new simulation steps.
 *
 * Revision 1.2  2001/01/10 15:12:57  olalu526
 * Added support for WorldModel to some basic classes.
 *
 * Revision 1.1  2000/09/21 13:46:40  frehe
 * Move the RoboSoc views from rsviews to the rsframework library.
 *
 * Revision 1.2  2000/09/13 00:08:06  frehe
 * Updated the views to inherit from the ViewWithHistory instead and also changed those views that used a static pointer to itself to do process the data in the sensor interface to use the new forEach methods implemented in the FrameworkSensorInterface.
 *
 * Revision 1.1  2000/09/04 15:27:20  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file BallView.h
 * Include file for the class BallView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.3 $
 * $Author: olalu526 $
 * $Date: 2001/01/11 22:00:17 $
 */


#ifndef BALLVIEW_H
#define BALLVIEW_H


#include "ViewWithSimulation.h"
#include "ParameterView.h"
#include "AgentView.h"
#include "BallObject.h"


RS_BEGIN_NAMESPACE


class GameView;
class CommandView;


/** Used to represent a collection of ball objects. */
//typedef ObjectCollection<BallObject> BallCollection;

#define theBall (*(worldModel->ballData()))

/**
 * @class BallView BallView.h
 * This view is used to keep track of the ball.
 *
 * <!-- @invariant The class invariant is... -->
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.3 $
 * $Author: olalu526 $
 * $Date: 2001/01/11 22:00:17 $
 */
class BallView : public ViewWithSimulation {
public:
  /** Create a ball view with ViewId @a id and the
    history length @a hist_len. */
  BallView(const ViewId id,
	   const UInt hist_len = DEFAULT_HISTORY_LENGTH);

  /** Destroys the ball view. */
  virtual ~BallView();


  /** Get a constant object representing the ball @a hist cycles ago. */
  const BallObject& getBallObject(const Int hist) const;


  /** Get the time of the last observation of the ball. */
  TimeUnknown getLastBallObservationTime() const;


  /** Get the relative vector to the ball @a hist cycles ago. */
  VectorUCO getRelativeBallVector(const Int hist) const;

  /** Get the relative direction to the ball @a hist cycles ago. */  
  AngleDegUCO getRelativeBallDirection(const Int hist) const;

  /** Get the absolute vector to the ball @a hist cycles ago. */ 
  VectorUCO getAbsoluteBallVector(const Int hist) const;

  /** Get the absolute direction to the ball @a hist cycles ago. */
  AngleDegUCO getAbsoluteBallDirection(const Int hist) const;

  /** Get the distance to the ball @a hist cycles ago. */
  FloatUCO getBallDistance(const Int hist) const;

  /** Get the position of the ball @a hist cycles ago. */
  PointUCO getBallPosition(const Int hist) const;

  /** Get the distance change of the ball @a hist cycles ago. */
  FloatUCO getBallDistChanged(const Int hist) const;

  /** Get the direction change of the ball @a hist cycles ago. */
  AngleDegUCO getBallDirChanged(const Int hist) const;

  /** Get the absolute speed vector of the ball @a hist cycles ago. 
    I.e. absolute amount of speed and speed direction. */
  VectorUCO getBallAbsoluteSpeedVector(const Int hist) const;

  /** Get the relative speed vector of the ball @a hist cycles ago.
    I.e. absolute amount of speed and relative speed direction. */
  VectorUCO getBallRelativeSpeedVector(const Int hist) const;

  /** Get the amount of speed of the ball @a hist cycles ago. */
  FloatUCO getBallSpeed(const Int hist) const;

  /** Get the absolute speed direction of the ball @a hist cycles ago. */
  AngleDegUCO getBallAbsoluteSpeedDirection(const Int hist) const;

  /** Get the relative speed direction of the ball @a hist cycles ago. */
  AngleDegUCO getBallRelativeSpeedDirection(const Int hist) const;


  /** Returns @c true iff the ball is kickable.
   Obsolete: it should be made into a predicate! */
  bool isBallKickable(const Int hist) const;

  /** Returns @c true iff the ball is catchable.
    Obsolete: it should be made into a predicate! */
  bool isBallCatchable(const Int hist) const;

  /** Return the actual index to the time step
    @a i steps from the current time step.
    If @a i is -1 then it means the previous time step
    and if @a i is 1 then it means the next time step.
    OBS! Since the history is implemented as a circular
    array getIndex will return a value that between 0 and
    getHistoryLength-1. It also means that getIndex(1) is the last
    element in the array and getIndex(0) = getIndex(getHistoryLength). */
  UInt getIndex(const Int i) const;

  /** Increase the index to point to the next time step. */
  void increaseIndex();

  
  /** Returns the length of the history, which is the maximum
    number of steps in the past that will be saved. */
  UInt getHistoryLength() const;
  
protected:
  /** A pointer to the parameter view. */
  ParameterView* parameterView;

  /** A pointer to the game view. */
  GameView* gameView;

  /** A pointer to the command view. */
  CommandView* commandView;

  /** A pointer to the agent view. */
  AgentView* agentView;


  /** The history of the ball. */
  //BallCollection theBall;

  
  virtual void myNewCycle();
  virtual void myUpdateAfterCoachMessage();
  virtual void myUpdateAfterInit();
  virtual void myUpdateAfterSee();
  virtual void myUpdateAfterHear();
  virtual void myUpdateAfterSense();
  virtual void myUpdateAfterCommand();


  /** Method for estimating the speed of the ball. */
  void estimateBallSpeed();


  /** The last observation time of the ball.
    Temporary hack!!! TO BE REMOVED!!! */
  TimeUnknown lastObservationTime;


private:
  // Hide copy-constructor and assignment.
  BallView(const BallView&);
  BallView& operator= (const BallView&);  
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "BallView.icc"
#endif


#endif // !BALLVIEW_H
