/* -*- Mode: C++ -*-
 * $Id: CoachView.h,v 1.1 2000/09/21 13:46:41 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: CoachView.h
 * Unitname: Framework/view
 * $Revision: 1.1 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/21 13:46:41 $
 *
 *
 * HISTORY
 *
 * $Log: CoachView.h,v $
 * Revision 1.1  2000/09/21 13:46:41  frehe
 * Move the RoboSoc views from rsviews to the rsframework library.
 *
 * Revision 1.3  2000/09/13 00:08:07  frehe
 * Updated the views to inherit from the ViewWithHistory instead and also changed those views that used a static pointer to itself to do process the data in the sensor interface to use the new forEach methods implemented in the FrameworkSensorInterface.
 *
 * Revision 1.2  2000/09/05 21:15:58  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/04 15:27:21  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file CoachView.h
 * Include file for the class CoachView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2000/09/21 13:46:41 $
 */


#ifndef COACHVIEW_H
#define COACHVIEW_H


#include "ViewWithHistory.h"
#include "PlayerObject.h"


RS_BEGIN_NAMESPACE


class ParameterView;
class GameView;
class AgentView;


/** Used to represent a collection of "teams" i.e. vectors of player objects. */
typedef ObjectCollection< std::vector<PlayerObject> > PlayersCollection;


/**
 * @class CoachView CoachView.h
 * This view is used to keep track of the coach
 * which is a separate agent although it is inside
 * the agent. It is used to debug the agent.
 *
 * <!-- @invariant The class invariant is... -->
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2000/09/21 13:46:41 $
 */
class CoachView : public ViewWithHistory {
public:
  /** Create a coach view with ViewId @a id and the
    history length @a hist_len. */
  CoachView(const ViewId id,
	    const UInt hist_len = DEFAULT_HISTORY_LENGTH);

  /** Destroys the coach view. */
  virtual ~CoachView();


  /** Get the time of the last coach message received @a hist cycles ago. */
  UIntUnknown getCoachMessageTime(const Int hist = NOW) const;


  /** Get the absolute ball position @a hist cycles ago. */
  PointUnknown getBallAbsolutePosition(const Int hist = NOW) const;

  /** Get the absolute ball speed vector @a hist cycles ago. */
  VectorUnknown getBallAbsoluteSpeed(const Int hist = NOW) const;

  /** Get the relative vector to the ball @a hist cycles ago. */
  VectorUnknown getRelativeBallVector(const Int hist = NOW) const;


  /** Get the absolute position of the agent @a hist cycles ago. */
  PointUnknown getAgentAbsolutePosition(const Int hist = NOW) const;

  /** Get the absolute speed vector of the agent @a hist cycles ago. */
  VectorUnknown getAgentAbsoluteSpeed(const Int hist = NOW) const;

  /** Get the absolute (body?) direction of the agent @a hist cycles ago. */
  AngleDegUnknown getAgentAbsoluteDirection(const Int hist = NOW) const;

  /** Get the absolute face direction of the agent @a hist cycles ago. */
  AngleDegUnknown getAgentAbsoluteFaceDirection(const Int hist = NOW) const;

  /** Get the absolute neck direction of the agent @a hist cycles ago. */
  AngleDegUnknown getAgentAbsoluteNeckDirection(const Int hist = NOW) const;


  /** Get a vector of player objects representing the opponents
    @a hist cycles ago. */
  std::vector<PlayerObject> getOpponents(const Int hist) const;

  /** Get a vector of player objects representing the teammates
    @a hist cycles ago. */
  std::vector<PlayerObject> getTeammates(const Int hist) const;

  
protected:
  /** A pointer to the parameter view. */
  ParameterView* parameterView;

  /** A pointer to the game view. */
  GameView* gameView;

  /** A pointer to the agent view. */
  AgentView* agentView;


  /** The history of the coach message time. */
  UIntUnknownCollection coachMessageTime;


  /** The history of the position of the ball. */
  PointUnknownCollection ballPosition;

  /** The history of the speed of the ball. */
  VectorUnknownCollection ballSpeed;


  /** The history of the position of the agent. */
  PointUnknownCollection agentPosition;

  /** The history of the speed of the agent. */
  VectorUnknownCollection agentSpeed;

  /** The history of the direction of the agent. */
  AngleDegUnknownCollection agentDirection;

  /** The history of the neck direction of the agent. */
  AngleDegUnknownCollection agentNeckDirection;


  /** The history of the opponents. */
  PlayersCollection opponents;

  /** The history of the teammates. */
  PlayersCollection teammates;

  
  virtual void myNewCycle();
  virtual void myUpdateAfterCoachMessage();
  virtual void myUpdateAfterInit();
  virtual void myUpdateAfterSee();
  virtual void myUpdateAfterHear();
  virtual void myUpdateAfterSense();
  virtual void myUpdateAfterCommand();


private:
  // Hide the copy-constructor and the assignment operator.
  CoachView(const CoachView&);
  CoachView& operator= (const CoachView&);
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "CoachView.icc"
#endif


#endif // !COACHVIEW_H
