/* -*- Mode: C++ -*-
 * $Id: CommandView.h,v 1.1 2000/09/21 13:46:41 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: CommandView.h
 * Unitname: Framework/view
 * $Revision: 1.1 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/21 13:46:41 $
 *
 *
 * HISTORY
 *
 * $Log: CommandView.h,v $
 * Revision 1.1  2000/09/21 13:46:41  frehe
 * Move the RoboSoc views from rsviews to the rsframework library.
 *
 * Revision 1.2  2000/09/13 00:08:07  frehe
 * Updated the views to inherit from the ViewWithHistory instead and also changed those views that used a static pointer to itself to do process the data in the sensor interface to use the new forEach methods implemented in the FrameworkSensorInterface.
 *
 * Revision 1.1  2000/09/04 15:27:22  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file CommandView.h
 * Include file for the class CommandView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2000/09/21 13:46:41 $
 */


#ifndef COMMANDVIEW_H
#define COMMANDVIEW_H


#include "library_types_with_unknown.h"
#include "ViewWithHistory.h"


RS_BEGIN_NAMESPACE


class GameView;
class BallView;
class AgentView;
class ParameterView;


/**
 * @class CommandView CommandView.h
 * This view is used to calculate the totaly effect on the agent
 * after a command has been executed.
 *
 * <!-- @invariant The class invariant is... -->
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2000/09/21 13:46:41 $
 */
class CommandView : public ViewWithHistory {
public:
  /** Create a command view with ViewId @a id and
    the history length @a hist_len. */
  CommandView(const ViewId id,
	      const UInt hist_len = DEFAULT_HISTORY_LENGTH);

  /** Destroy the command view. */
  virtual ~CommandView();


  /** Initalize the command view. This method must be called before
    the command view can be used. It is necessary since there
    are circular dependencies between the command view and the
    game, ball, and agent views. */
  virtual bool init();


  /** Get the total movement vector of the agent @a hist
    cycles ago. If the value is unknown then the zero-vector
    will be returned. */
  Vector getAgentMovement(const Int hist = PREV) const;

  /** Get the agent position as calculated @a hist cycles ago. */
  Point getAgentPosition(const Int hist = NOW) const;

  /** Get the speed of the agent as calculated @a hist cycles ago. */
  Vector getAgentSpeed(const Int hist = NOW) const;

  /** Get the total agent acceleration @a hist cycles ago. */
  Vector getAgentAcceleration(const Int hist = PREV) const;


  /** Get the total ball movement vector @a hist cycles ago. */
  Vector getBallMovement(const Int hist = PREV) const;

  /** Get the ball speed as calculated @a hist cycles ago. */
  Vector getBallSpeed(const Int hist = NOW) const;

  /** Get the total ball acceleration @a hist cycles ago. */
  Vector getBallAcceleration(const Int hist = PREV) const;


  /** Get the actual angle that agent turned @a hist cycles ago. */
  AngleDeg getActualTurn(const Int hist = PREV) const;

  /** Get the actual angle that the agent turned the neck @a hist cycles ago.
    This one is special since turn necks take effect immediately!
    All the other effects take place at the end of the cycle. */
  AngleDeg getActualTurnNeck(const Int hist = NOW) const;

  /** Get the actual dash power used to dash @a hist cycles ago. */
  Float getActualDashPower(const Int hist = PREV) const;

  /** Get the position after the move made @a hist cycles ago. */
  PointUnknown getMovePosition(const Int hist = NOW) const;


  /** Get the type of the last command executed. */
  CommandType getLastCommandType() const;

  /** Get the server string of the last executed command. */
  std::string getLastServerString() const;

  
protected:
  /** A pointer to the parameter view. */
  ParameterView* parameterView; 

  /** A pointer to the game view. */
  GameView* gameView;

  /** A pointer to the ball view. */
  BallView* ballView;

  /** A pointer to the agent view. */
  AgentView* agentView;


  /** The history of the agent movement vector. */
  VectorCollection agentMovement;

  /** The history of the ball movement vector. */
  VectorCollection ballMovement;

  /** The history of the actual turns made by the agent. */
  AngleDegCollection actualTurn;

  /** The history of the actual turn necks made by the agent. */
  AngleDegCollection actualTurnNeck;

  /** The history of the agent acceleration. */
  PointCollection agentAcceleration;

  /** The history of the ball acceleration. */
  PointCollection ballAcceleration;

  /** The history of the actual dash power used by the agent. */
  FloatCollection actualDashPower;

  /** The history of the positions move to by the agent using
    the move command. */
  PointUnknownCollection movePosition;
  

  /** The history of the speed of the agent. */
  PointCollection agentSpeed;

  /** The history of the position of the agent. */
  PointCollection agentPosition;

  /** The history of the speed of the ball. */
  PointCollection ballSpeed;


  /** The history of the stamina of the agent. */
  FloatCollection agentStamina;

  /** The history of the effort of the agent. */
  FloatCollection agentEffort;

  /** The history of the recovery of the agent. */
  FloatCollection agentRecovery;

  /** The history of the body direction of the agent. */
  AngleDegCollection agentBodyDirection;

  /** The history of the neck direction of the agent. */
  AngleDegCollection agentNeckDirection;


  /** The type of the last command. */
  CommandType lastCommandType;

  /** The server string of the last command. */
  std::string lastServerString;

  
  virtual void myNewCycle();
  virtual void myUpdateAfterCoachMessage();
  virtual void myUpdateAfterInit();
  virtual void myUpdateAfterSee();
  virtual void myUpdateAfterHear();
  virtual void myUpdateAfterSense();
  virtual void myUpdateAfterCommand();


private:
  // Hide copy-constructor and assignment.
  CommandView(const CommandView&);
  CommandView& operator= (const CommandView&);  
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "CommandView.icc"
#endif


#endif
