/* -*- Mode: C++ -*-
 * $Id: GameView.h,v 1.3 2001/01/11 22:00:18 olalu526 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: GameView.h
 * Unitname: Framework/view
 * $Revision: 1.3 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: olalu526 $ $Date: 2001/01/11 22:00:18 $
 *
 *
 * HISTORY
 *
 * $Log: GameView.h,v $
 * Revision 1.3  2001/01/11 22:00:18  olalu526
 * Added the real world model with support for new simulation steps.
 *
 * Revision 1.2  2001/01/10 15:12:58  olalu526
 * Added support for WorldModel to some basic classes.
 *
 * Revision 1.1  2000/09/21 13:46:42  frehe
 * Move the RoboSoc views from rsviews to the rsframework library.
 *
 * Revision 1.2  2000/09/13 00:08:08  frehe
 * Updated the views to inherit from the ViewWithHistory instead and also changed those views that used a static pointer to itself to do process the data in the sensor interface to use the new forEach methods implemented in the FrameworkSensorInterface.
 *
 * Revision 1.1  2000/09/04 15:27:23  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file GameView.h
 * Include file for the class GameView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.3 $
 * $Author: olalu526 $
 * $Date: 2001/01/11 22:00:18 $
 */


#ifndef GAMEVIEW_H
#define GAMEVIEW_H


#include "ViewWithSimulation.h"
#include "FrameworkSensorInterface.h"


RS_BEGIN_NAMESPACE


/** Represent a collection of play modes. */
//typedef ObjectCollection<PlayMode> PlayModeCollection;
#define playModes (*(worldModel->playModeData()))

/**
 * @class GameView GameView.h
 * This view is used to keep track of the current game.
 * It includes things like the current game time,
 * the current play mode and so on.
 *
 * <!-- @invariant The class invariant is... -->
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.3 $
 * $Author: olalu526 $
 * $Date: 2001/01/11 22:00:18 $
 */
class GameView : public ViewWithSimulation {
public:
  /** Construct a game view with ViewId @a id and
    the history length @a hist_len. */
  GameView(const ViewId id,
	   UInt16 hist_len = DEFAULT_HISTORY_LENGTH);

  /** Destroys the game view. */
  virtual ~GameView();

  /** Get the current time according to the agent. */
  Time getCurrentTime() const;

  
  /** Get the last play mode for the cycles that was @a hist cycles ago.
    @todo Get the last few play modes instead since they sometimes change
    very fast, several times in one cycle. */
  PlayMode getPlayMode(const Int hist = NOW) const;


  /** Get the side of the field that the agent is playing on. */
  Side getOurSide() const;

  /** Get the side of the field that the opponents are playing on. */
  Side getTheirSide() const;


  /** Get the score of the agent's team. */
  UInt16 getOurScore() const;

  /** Get the score for the opponents. */
  UInt16 getTheirScore() const;


  /** Get the teamname of the agent's team. */
  const std::string& getOurTeamName() const;

  /** Get the teamname of the opponents. */
  const std::string& getTheirTeamName() const;

  /** Return the actual index to the time step
    @a i steps from the current time step.
    If @a i is -1 then it means the previous time step
    and if @a i is 1 then it means the next time step.
    OBS! Since the history is implemented as a circular
    array getIndex will return a value that between 0 and
    getHistoryLength-1. It also means that getIndex(1) is the last
    element in the array and getIndex(0) = getIndex(getHistoryLength). */
  UInt getIndex(const Int i) const;

  /** Increase the index to point to the next time step. */
  void increaseIndex();

  
  /** Returns the length of the history, which is the maximum
    number of steps in the past that will be saved. */
  UInt getHistoryLength() const;
  
protected:
  /** The history of the play modes. */
  //PlayModeCollection playModes;

  virtual void myNewCycle();
  virtual void myUpdateAfterCoachMessage();
  virtual void myUpdateAfterInit();
  virtual void myUpdateAfterSee();
  virtual void myUpdateAfterHear();
  virtual void myUpdateAfterSense();
  virtual void myUpdateAfterCommand();


private:
  // Hide copy-constructor and assignment.
  GameView(const GameView&);
  GameView& operator= (const GameView&);  
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "GameView.icc"
#endif


#endif // !GAMEVIEW_H
