/* -*- Mode: C++ -*-
 * $Id: MarkerView.h,v 1.1 2000/09/21 13:46:42 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: MarkerView.h
 * Unitname: Framework/view
 * $Revision: 1.1 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/21 13:46:42 $
 *
 *
 * HISTORY
 *
 * $Log: MarkerView.h,v $
 * Revision 1.1  2000/09/21 13:46:42  frehe
 * Move the RoboSoc views from rsviews to the rsframework library.
 *
 * Revision 1.2  2000/09/13 00:08:09  frehe
 * Updated the views to inherit from the ViewWithHistory instead and also changed those views that used a static pointer to itself to do process the data in the sensor interface to use the new forEach methods implemented in the FrameworkSensorInterface.
 *
 * Revision 1.1  2000/09/04 15:27:24  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file MarkerView.h
 * Include file for the class MarkerView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2000/09/21 13:46:42 $
 */


#ifndef MARKERVIEW_H
#define MARKERVIEW_H


#include "framework_types.h"
#include "ViewWithHistory.h"
#include "MarkerObject.h"
#include "LineObject.h"


RS_BEGIN_NAMESPACE


class ParameterView;
class CommandView;
class GameView;
class AgentView;


/** Represent a collection of marker objects. */
typedef ObjectCollection<MarkerObject> MarkerCollection;

/** Represent a collection of line objects. */
typedef ObjectCollection<LineObject> LineCollection;


// This class might throw out_of_range exceptions, although I don't
// think there should ever be such a problem since everything goes through
// getIndex which never goes out of bounds.

/**
 * @class MarkerView MarkerView.h
 * This view is used to keep track of the lines and
 * markers in the current game.
 *
 * <!-- @invariant The class invariant is... -->
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2000/09/21 13:46:42 $
 */
class MarkerView : public ViewWithHistory {
public:
  /** Create a marker view with the ViewId @a id and
    the history length @a hist_len. Initialize it with
    the parameters found in the @a paramter_view. */
  MarkerView(const ViewId id,
	     ParameterView* const parameter_view,
	     const UInt hist_len = DEFAULT_HISTORY_LENGTH);

  /** Destroys the marker view. */
  virtual ~MarkerView();

  
  /** Initialize the marker view. Must be called before
   it is used! It is necessary since we have a circular
   dependency between the marker view and the agent view. */
  virtual bool init();

  
  /** Returns the number of side lines seen this cycle. */
  UInt getNumberOfSeenLines() const;

  /** Get the closest side line seen this cycle. */
  const LineObject& getClosestSeenLine() const;

  /** Get a line object representing the side line @a line
    @a hist cycles ago. */
  LineObject& getLine(const SideLineName line, Int hist);

  /** Get a constant line object representing the side line @a line
    @a hist cycles ago. */
  const LineObject& getLine(const SideLineName line, Int hist) const;

  /** Get a constant line object representing the side line @a line
    @a hist cycles ago. */
  const LineObject& getLine(const Int line, Int hist) const;

  
  /** Get an object representing the left side line @a hist cycles ago. */
  const LineObject& getLineL(const Int hist) const;

  /** Get an object representing the right side line @a hist cycles ago. */
  const LineObject& getLineR(const Int hist) const;
  
  /** Get an object representing the top side line @a hist cycles ago. */
  const LineObject& getLineT(const Int hist) const;

  /** Get an object representing the bottom side line @a hist cycles ago. */
  const LineObject& getLineB(const Int hist) const;

  
  /** Get a constant reference to the closest seen marker this cycle. */
  const MarkerObject& getClosestSeenMarker() const;

  /** Get a marker object with the marker name @a marker
    @a hist cycles ago. */
  MarkerObject& getMarker(const MarkerName marker, Int hist);

  /** Get a constant marker object with the marker name @a marker
    @a hist cycles ago. */
  const MarkerObject& getMarker(const MarkerName marker, Int hist) const;

  /** Get a constant marker object representing the marker with number
    @a marker @a hist cycles ago. */
  const MarkerObject& getMarker(const Int marker, Int hist) const;

  /** Get a reference to the marker object that is closest to the
    point @a pt @a hist cycles ago. */
  MarkerObject& getClosestMarkerTo(const PointUCO& pt, Int hist);

  /** Get the constant marker object that is closest to the point @a pt
    @a hist cycles ago. */
  const MarkerObject& getClosestMarkerTo(const PointUCO& pt, Int hist) const;

  /** Get the constant marker object that is the goal closest to the
    point @a pt @a hist cycles ago. */
  const MarkerObject& getClosestGoalTo(const PointUCO& pt, Int hist) const;


  /** Get a constant marker object representing the opponents' goal. */
  const MarkerObject& getTheirGoal(const Int hist) const;
  /** Get a constant marker object representing your own goal. */
  const MarkerObject& getOurGoal(const Int hist) const;


  //@{
  /** @name Get Marker Methods
    Methods for accessing the markers directly.
    Returns the constant marker object representing the state
    of the marker @a hist cycles ago. */
  const MarkerObject& getGoalL(const Int hist) const;
  const MarkerObject& getGoalR(const Int hist) const;
  const MarkerObject& getFlagC(const Int hist) const;
  const MarkerObject& getFlagCT(const Int hist) const;
  const MarkerObject& getFlagCB(const Int hist) const;
  const MarkerObject& getFlagLT(const Int hist) const;
  const MarkerObject& getFlagLB(const Int hist) const;
  const MarkerObject& getFlagRT(const Int hist) const;
  const MarkerObject& getFlagRB(const Int hist) const;
  const MarkerObject& getFlagPLT(const Int hist) const;
  const MarkerObject& getFlagPLC(const Int hist) const;
  const MarkerObject& getFlagPLB(const Int hist) const;
  const MarkerObject& getFlagPRT(const Int hist) const;
  const MarkerObject& getFlagPRC(const Int hist) const;
  const MarkerObject& getFlagPRB(const Int hist) const;
  const MarkerObject& getFlagGLT(const Int hist) const;
  const MarkerObject& getFlagGLB(const Int hist) const;
  const MarkerObject& getFlagGRT(const Int hist) const;
  const MarkerObject& getFlagGRB(const Int hist) const;
  const MarkerObject& getFlagTL50(const Int hist) const;
  const MarkerObject& getFlagTL40(const Int hist) const;
  const MarkerObject& getFlagTL30(const Int hist) const;
  const MarkerObject& getFlagTL20(const Int hist) const;
  const MarkerObject& getFlagTL10(const Int hist) const;
  const MarkerObject& getFlagT0(const Int hist) const;
  const MarkerObject& getFlagTR10(const Int hist) const;
  const MarkerObject& getFlagTR20(const Int hist) const;
  const MarkerObject& getFlagTR30(const Int hist) const;
  const MarkerObject& getFlagTR40(const Int hist) const;
  const MarkerObject& getFlagTR50(const Int hist) const;
  const MarkerObject& getFlagBL50(const Int hist) const;
  const MarkerObject& getFlagBL40(const Int hist) const;
  const MarkerObject& getFlagBL30(const Int hist) const;
  const MarkerObject& getFlagBL20(const Int hist) const;
  const MarkerObject& getFlagBL10(const Int hist) const;
  const MarkerObject& getFlagB0(const Int hist) const;
  const MarkerObject& getFlagBR10(const Int hist) const;
  const MarkerObject& getFlagBR20(const Int hist) const;
  const MarkerObject& getFlagBR30(const Int hist) const;
  const MarkerObject& getFlagBR40(const Int hist) const;
  const MarkerObject& getFlagBR50(const Int hist) const;
  const MarkerObject& getFlagLT30(const Int hist) const;
  const MarkerObject& getFlagLT20(const Int hist) const;
  const MarkerObject& getFlagLT10(const Int hist) const;
  const MarkerObject& getFlagL0(const Int hist) const;
  const MarkerObject& getFlagLB10(const Int hist) const;
  const MarkerObject& getFlagLB20(const Int hist) const;
  const MarkerObject& getFlagLB30(const Int hist) const;
  const MarkerObject& getFlagRT30(const Int hist) const;
  const MarkerObject& getFlagRT20(const Int hist) const;
  const MarkerObject& getFlagRT10(const Int hist) const;
  const MarkerObject& getFlagR0(const Int hist) const;
  const MarkerObject& getFlagRB10(const Int hist) const;
  const MarkerObject& getFlagRB20(const Int hist) const;
  const MarkerObject& getFlagRB30(const Int hist) const;
  //@}
  
  
protected:
  /** A pointer to the parameter view. */
  ParameterView* const parameterView;

  /** A pointer to the command view. */
  CommandView* commandView;

  /** A pointer to the game view. */
  GameView* gameView;

  /** A pointer to the agent view. */
  AgentView* agentView;


  /** Updates the view with the information stored in the
    sensor interface regarding the markers. */
  virtual void processMarker(const ServerMarkerObject& marker);

  /** Updates the view with the information stored in the
    sensor interface regarding the lines. */
  virtual void processLine(const ServerLineObject& line);

  
  /** Reset the history of all lines and markers. */
  void resetAll();

  /** Reset the @a i:th step in the history. */
  void reset(const Int i);


  /** Rotate the position of all markers. Used when the
    team plays on the right hand side of the field since
    the server doesn't rotate the markers itself. */
  void rotateAllMarkers();

  
  /** The number of seen lines this cycle. */
  UInt numberOfSeenLines;

  /** The closest line seen this cycle. */
  LineObject closestSeenLine;

  
  /** The history of the left side line. */
  LineCollection lineL;

  /** The history of the right side line. */
  LineCollection lineR;

  /** The history of the top side line. */
  LineCollection lineT;
  
  /** The history of the bottom side line. */
  LineCollection lineB;
  
  /** A line object representing an unknown line. */
  LineObject unknownLine;


  /** The number of seen markers this cycle. */
  UInt numberOfSeenMarkers;

  /** The closest marker seen this cycle. */
  MarkerObject closestSeenMarker;

  
  //@{
  /** @name Marker History Objects
    The history of the marker objects. */
  MarkerCollection goalL;
  MarkerCollection goalR;
  MarkerCollection flagC;
  MarkerCollection flagCT;
  MarkerCollection flagCB;
  MarkerCollection flagLT;
  MarkerCollection flagLB;
  MarkerCollection flagRT;
  MarkerCollection flagRB;
  MarkerCollection flagPLT;
  MarkerCollection flagPLC;
  MarkerCollection flagPLB;
  MarkerCollection flagPRT;
  MarkerCollection flagPRC;
  MarkerCollection flagPRB;
  MarkerCollection flagGLT;
  MarkerCollection flagGLB;
  MarkerCollection flagGRT;
  MarkerCollection flagGRB;
  MarkerCollection flagTL50;
  MarkerCollection flagTL40;
  MarkerCollection flagTL30;
  MarkerCollection flagTL20;
  MarkerCollection flagTL10;
  MarkerCollection flagT0;
  MarkerCollection flagTR10;
  MarkerCollection flagTR20;
  MarkerCollection flagTR30;
  MarkerCollection flagTR40;
  MarkerCollection flagTR50;
  MarkerCollection flagBL50;
  MarkerCollection flagBL40;
  MarkerCollection flagBL30;
  MarkerCollection flagBL20;
  MarkerCollection flagBL10;
  MarkerCollection flagB0;
  MarkerCollection flagBR10;
  MarkerCollection flagBR20;
  MarkerCollection flagBR30;
  MarkerCollection flagBR40;
  MarkerCollection flagBR50;
  MarkerCollection flagLT30;
  MarkerCollection flagLT20;
  MarkerCollection flagLT10;
  MarkerCollection flagL0;
  MarkerCollection flagLB10;
  MarkerCollection flagLB20;
  MarkerCollection flagLB30;
  MarkerCollection flagRT30;
  MarkerCollection flagRT20;
  MarkerCollection flagRT10;
  MarkerCollection flagR0;
  MarkerCollection flagRB10;
  MarkerCollection flagRB20;
  MarkerCollection flagRB30;
  //@}

  /** A marker object representing an unknown marker. */
  MarkerObject unknownMarker;
  
  virtual void myNewCycle();
  virtual void myUpdateAfterCoachMessage();
  virtual void myUpdateAfterInit();
  virtual void myUpdateAfterSee();
  virtual void myUpdateAfterHear();
  virtual void myUpdateAfterSense();
  virtual void myUpdateAfterCommand();

  
private:
  // Hide copy-constructor and assignment.
  MarkerView(const MarkerView&);
  MarkerView& operator= (const MarkerView&);  
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "MarkerView.icc"
#endif


#endif // !MARKERVIEW_H
