/* -*- Mode: C++ -*-
 * $Id: ParameterView.h,v 1.1 2000/09/21 13:46:43 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: ParameterView.h
 * Unitname: Framework/view
 * $Revision: 1.1 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/21 13:46:43 $
 *
 *
 * HISTORY
 *
 * $Log: ParameterView.h,v $
 * Revision 1.1  2000/09/21 13:46:43  frehe
 * Move the RoboSoc views from rsviews to the rsframework library.
 *
 * Revision 1.1  2000/09/04 15:27:25  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file ParameterView.h
 * Include file for the class ParameterView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2000/09/21 13:46:43 $
 */


#ifndef PARAMETERVIEW_H
#define PARAMETERVIEW_H


#include "View.h"


RS_BEGIN_NAMESPACE


/** Used to represent the different types of input value
  that the paramter view can read from the config files. */
enum InputType {
  V_UINT16,
  V_INT32,
  V_FLOAT32,
  V_BOOL,
  V_STRING,
  V_ONOFF,
  V_NONE
};


/**
 * @class ParameterView ParameterView.h
 * This view is used to store all the soccer server parameters
 * and also user defined client parameters.
 *
 * This class is based on the Parameter class in RoboLog
 *
 * @todo Add an associative array with client parameters and
 *       values. In this way any parameter can be added by the
 *       user without having to modify the parameter view.
 *
 * <!-- @invariant The class invariant is... -->
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2000/09/21 13:46:43 $
 */
class ParameterView : public View
{
public:
  /** Create a parameter view with all the parameters initialized
    to their default values. */
  ParameterView(const ViewId id);
  /** Create a parameter view and initialize its parameters
    with the command line options supplied by @a argc and @a argv. */
  ParameterView(const ViewId id, int argc, char** argv,
		int verbose = 0);
  /** Parameter view destructor. */
  virtual ~ParameterView();

  // I make these virtual so that the user can define its own
  // subclass and still be able to send pointers to it to the
  // classes that use ParameterView and still call the "right"
  // functions even though it's a pointer to a super class.
  /** Read and parse the command line options supported by the parameter view.
    They include @c -f @i file to read @i file as if it was a configuration
    file. You can also set individual parameters on the command line by
    @c -parameter @i value to set the parameter @c parameter to @i value.
    The parameter name is the same as used by the parameter view except for
    the prefix CP_ or SP_. */
  virtual void readCommandLine(int argc, char** argv, int verbose = 0);
  /** Read and parse a configuration file @a filename assigning parameters
    values. */ 
  virtual void readOptions(const std::string& filename, int verbose = 0);
  /** Read and parse a line buffer @a linebuf containing parameters and their
    values. */
  virtual bool readOption(const char* linebuf, int verbose = 0);
  /** Read and assign a single parameter @a option the value @a value. */
  virtual bool readOption(const char* option, const char* value);

  /*
  // These are included so that one can have pointers to this
  // class and get values from subclasses.
  // But it requires that you can connect a string to a value.
  virtual get(const std::string&, Float32*);
  virtual get(const std::string&, Int32*);
  virtual get(const std::string&, std::string*);
  */


  //@{
  /** @name Client parameters
    These are predefined client parameters that can be read and accessed
    by the parameter view. They are usually read from the client.conf file. */
  std::string CP_team_name;
  bool        CP_goalie;
  UInt16      CP_cycles_since_io_before_shutdown;
  UInt16      CP_min_ticks_before_sending_command;
  UInt16      CP_ticks_before_command_warning;
  UInt16      CP_max_ticks_before_sending_command;
  //@}
  

  //@{
  /** @name Server parameters
    These are predefined server parameters that can be read and accessed
    by the parameter view. They are usually read from the server.conf file,
    they have the same default value as the standard server.conf file. */    
  Float32     SP_pitch_length;
  Float32     SP_pitch_width;
  Float32     SP_pitch_margin;
  Float32     SP_penalty_area_length;
  Float32     SP_penalty_area_width;
  Float32     SP_goal_area_length;
  Float32     SP_goal_area_width;
  Float32     SP_penalty_spot_dist;
  Float32     SP_corner_arc_r;
  Float32     SP_free_kick_buffer;
  Int32       SP_after_goal_wait;
  Float32     SP_feel_distance;
  Int32       SP_num_lines;
  Int32       SP_num_markers;
  Float32     SP_unum_far_length;
  Float32     SP_unum_too_far_length;
  Float32     SP_team_far_length;
  Float32     SP_team_too_far_length;
  
  Float32     SP_version;
  Int32       SP_team_size;
  Int32       SP_half;
  std::string SP_host;
  Int32       SP_buffer_size; 
  
  // server.conf starts here
  Float32     SP_goal_width;
  
  Float32     SP_player_size;	
  Float32     SP_player_decay;    	
  Float32     SP_player_rand;	
  Float32     SP_player_weight;	
  Float32     SP_player_speed_max;

  Float32     SP_stamina_min;
  Float32     SP_stamina_max;
  Float32     SP_stamina_inc_max;
  Float32     SP_recover_dec_thr;
  Float32     SP_recover_dec;
  Float32     SP_recover_min;
  Float32     SP_recover_max;
        
  Float32     SP_effort_dec_thr;
  Float32     SP_effort_dec;
  Float32     SP_effort_inc_thr;
  Float32     SP_effort_inc;
  Float32     SP_effort_min;
  Float32     SP_effort_max;

  Int32       SP_hear_max;
  Int32       SP_hear_inc;
  Int32       SP_hear_decay;        

  Float32     SP_inertia_moment;
  
  Int32       SP_sense_body_step;
  
  Float32     SP_catchable_area_l;
  Float32     SP_catchable_area_w;
  Float32     SP_catch_probability;
  Int32       SP_catch_ban_cycle;
        
  Float32     SP_ball_size;		
  Float32     SP_ball_decay;	
  Float32     SP_ball_rand;		
  Float32     SP_ball_weight;
  Float32     SP_ball_speed_max;

  Float32     SP_wind_force;
  Float32     SP_wind_dir;        
  Float32     SP_wind_rand;
  bool        SP_wind_none;
  bool        SP_wind_random;

  Float32     SP_kickable_margin;
  Float32     SP_control_radius;
  Float32     SP_ckick_margin;
  Float32     SP_dash_power_rate;
  Float32     SP_kick_power_rate;	

  Float32     SP_visible_angle;
  Float32     SP_visible_distance;
  Float32     SP_audio_cut_dist;

  Float32     SP_quantize_step;
  Float32     SP_quantize_step_l;        

  Float32     SP_maxpower;		
  Float32     SP_minpower;
  Float32     SP_maxmoment;		
  Float32     SP_minmoment;
  Float32     SP_maxneckang;
  Float32     SP_minneckang;
  Float32     SP_maxneckmoment;
  Float32     SP_minneckmoment;

  Int32       SP_port;
  Int32       SP_coach_port;
  Int32       SP_olcoach_port;
        
  Int32       SP_say_coach_cnt_max;
  Int32       SP_say_coach_msg_size;
  Int32       SP_send_vi_step;
  
  Int32       SP_simulator_step;
  Int32       SP_send_step;
  Int32       SP_recv_step;
  Int32       SP_half_time;        
  Int32       SP_say_msg_size;
  
  bool	      SP_use_offside;
  Float32     SP_offside_active_area_size;
  bool	      SP_forbid_kick_off_offside;
  
  bool	      SP_verbose;
  Int32       SP_record_version;
  bool        SP_record_log;
  bool        SP_send_log;
  std::string SP_log_file;
  std::string SP_record;
  std::string SP_replay;
  
  bool	      SP_coach;
  bool	      SP_coach_w_referee;
  bool        SP_old_coach_hear;
  // server.conf ends here
  
  Float32    SP_kickable_area;
  //@}
  

protected:
  virtual void myNewCycle();
  virtual void myUpdateAfterCoachMessage();
  virtual void myUpdateAfterInit();
  virtual void myUpdateAfterSee();
  virtual void myUpdateAfterHear();
  virtual void myUpdateAfterSense();
  virtual void myUpdateAfterCommand();

private:
  /** Internal type used to parse the parameter options. */
  struct OptionType
  {
    char name[35];
    void *vptr;
    InputType vsize;
  };

  bool readOption(const char* linebuf, int verbose, const OptionType*);
  bool readOption(const char* option, const char* value, const OptionType*);

  /** Set the total number of parameters that can be stored in the
    parameter view. */
  static const int optsize = 110; // The current number is 108
  OptionType options[optsize];
        
  void parameterInit();
  void initOptions();  

  // Hide copy-constructor and assignment.
  ParameterView(const ParameterView&);
  ParameterView& operator= (const ParameterView&);  
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "ParameterView.icc"
#endif


#endif // !PARAMETERVIEW_H
