/* -*- Mode: C++ -*-
 * $Id: PlayerView.h,v 1.3 2001/01/11 22:00:18 olalu526 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: PlayerView.h
 * Unitname: Framework/view
 * $Revision: 1.3 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: olalu526 $ $Date: 2001/01/11 22:00:18 $
 *
 *
 * HISTORY
 *
 * $Log: PlayerView.h,v $
 * Revision 1.3  2001/01/11 22:00:18  olalu526
 * Added the real world model with support for new simulation steps.
 *
 * Revision 1.2  2001/01/10 15:12:58  olalu526
 * Added support for WorldModel to some basic classes.
 *
 * Revision 1.1  2000/09/21 13:46:43  frehe
 * Move the RoboSoc views from rsviews to the rsframework library.
 *
 * Revision 1.2  2000/09/13 00:08:10  frehe
 * Updated the views to inherit from the ViewWithHistory instead and also changed those views that used a static pointer to itself to do process the data in the sensor interface to use the new forEach methods implemented in the FrameworkSensorInterface.
 *
 * Revision 1.1  2000/09/04 15:27:26  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file PlayerView.h
 * Include file for the class PlayerView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.3 $
 * $Author: olalu526 $
 * $Date: 2001/01/11 22:00:18 $
 */


#ifndef PLAYERVIEW_H
#define PLAYERVIEW_H


#include <vector>

#include "framework_types.h"
#include "ViewWithSimulation.h"
#include "PlayerObject.h"


RS_BEGIN_NAMESPACE


class ParameterView;
class CommandView;
class AgentView;


/** Used to represent a collection of "teams" i.e. vectors of player objects. */
//typedef ObjectCollection< std::vector<PlayerObject> > PlayersCollection;

/**
 * @class PlayerView PlayerView.h
 * This view is used to keep track of the players,
 * their position, speed and other properties.
 *
 * <!-- @invariant The class invariant is... -->
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.3 $
 * $Author: olalu526 $
 * $Date: 2001/01/11 22:00:18 $
 */
class PlayerView : public ViewWithSimulation {
public:
#define theOpponents (*(worldModel->opponentsData()))
#define theTeammates (*(worldModel->teammatesData()))
#define theUnknownPlayers (*(worldModel->unknownPlayersData()))
#define theClosestOpponentIndex (*(worldModel->closestOpponentIndexData()))
#define theClosestTeammateIndex (*(worldModel->closestTeammateIndexData()))
#define theClosestUnknownPlayerIndex (*(worldModel->closestUnknownPlayerIndexData()))

  /** Create a player view with ViewId @a id and
    the history length @a hist_len. */
  PlayerView(const ViewId id,
	     const UInt hist_len = DEFAULT_HISTORY_LENGTH);

  /** Destroys the player view. */
  ~PlayerView();


  /** Get a pointer to the closest player @a hist cycles ago. */
  PlayerObject* getClosestPlayer(const Int hist);

  /** Get a pointer to the closest opponent @a hist cycles ago. */
  PlayerObject* getClosestOpponent(const Int hist);

  /** Get a pointer to the closest teammate @a hist cycles ago. */
  PlayerObject* getClosestTeammate(const Int hist);

  /** Get a pointer to the closest unknown player @a hist cycles ago. */
  PlayerObject* getClosestUnknownPlayer(const Int hist);

  
  // Warning temporary hack!
  /** Get a vector with the opponents (a copy of the actual data). */
  std::vector<PlayerObject> getOpponents(const Int hist) const;

  /** Get a vector with the teammates (a copy of the actual data). */
  std::vector<PlayerObject> getTeammates(const Int hist) const;

  /** Get a vector with the unknown players (a copy of the actual data). */
  std::vector<PlayerObject> getUnknownPlayers(const Int hist) const;

  
protected:
  /** A pointer to the parameter view. */
  ParameterView* parameterView;

  /** A pointer to the command view. */
  CommandView* commandView;

  /** A pointer to the agent view. */
  AgentView* agentView;

  
  /** The index in @a unknownPlayers that is the closest
    of those players. */
  //UIntUnknownCollection closestUnknownPlayerIndex;

  /** The index in @a opponents that is the closest
    of those players. */
  //UIntUnknownCollection closestOpponentIndex;
  
  /** The index in @a teammates that is the closest
    of those players. */
  //UIntUnknownCollection closestTeammateIndex;


  /** Stores all opponents seen a cycle.
   No matching is currently done between cycles. */
  //PlayersCollection opponents;

  /** Stores all teammates seen a cycle.
   No matching is currently done between cycles. */  
  //PlayersCollection teammates;

  /** Stores all unknown players seen a cycle.
   No matching is currently done between cycles. */  
  //PlayersCollection unknownPlayers;


  /** Update the player vectors with the information
    stored in the sensor interface. */
  virtual void processPlayer(const ServerPlayerObject& plr);

  
  /** Update the closest unknown player. */
  void setClosestUnknownPlayer(const PlayerObject&);

  /** Update the closest opponent. */
  void setClosestOpponent(const PlayerObject&);

  /** Update the closest teammate. */
  void setClosestTeammate(const PlayerObject&);

  
  virtual void myNewCycle();
  virtual void myUpdateAfterCoachMessage();
  virtual void myUpdateAfterInit();
  virtual void myUpdateAfterSee();
  virtual void myUpdateAfterHear();
  virtual void myUpdateAfterSense();
  virtual void myUpdateAfterCommand();


private:
  // Hide copy-constructor and assignment.
  PlayerView(const PlayerView&);
  PlayerView& operator= (const PlayerView&);  
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "PlayerView.icc"
#endif


#endif // !PLAYERVIEW_H
