/* -*- Mode: C++ -*-
 * $Id: SkillWithArguments.h,v 1.1 2000/09/13 03:10:38 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: SkillWithArguments.h
 * Unitname: Framework
 * $Revision: 1.1 $
 * Created by: Fredrik Heintz 2000-09-13
 * Last modified by $Author: frehe $ $Date: 2000/09/13 03:10:38 $
 *
 *
 * HISTORY
 *
 * $Log: SkillWithArguments.h,v $
 * Revision 1.1  2000/09/13 03:10:38  frehe
 * Added a new class SkillWithArguments to be used by all skills using arguments.
 *
 */

/**
 * @file SkillWithArguments.h
 * Include file for the class SkillWithArguments.
 *
 * @par Unit:
 *      Framework
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2000/09/13 03:10:38 $
 */


#ifndef SKILLWITHARGUMENTS_H
#define SKILLWITHARGUMENTS_H


#include "Skill.h"
#include "TooFewArgumentsException.h"


RS_BEGIN_NAMESPACE


/**
 * @class SkillWithArguments SkillWithArguments.h
 * This abstract class represent skill that uses arguments.
 *
 * @par Unit:
 *      Framework
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2000/09/13 03:10:38 $
 */
class SkillWithArguments : public Skill {
public:
  /** Default constructor, it is ok since the class is abstract
    all subclasses should make sure to get default values for the
    arguments. */
  SkillWithArguments();

  /** Destructor cleaning up after the skill. */
  virtual ~SkillWithArguments();


  /** Used to set the skill's arguments in a generic way.
   This method must be overridden by subclasses.
   It should transform the generic vector of floats to
   the actual arguments.
   It will throw TooFewArgumentsException if not enough arguments
   are given. */
  virtual void setArguments(const std::vector<Float>& arguments)
    throw(TooFewArgumentsException) = 0;

  
  /** This method is used to generate a plan for how the goal of
    the skill should be obtained by the agent. Abstract since
    no concrete instances of this class should be created. */
  virtual Plan generatePlan() = 0;

  /** This method is used to first set the arguments of the skill
    and then generate a plan for how the goal of the skill should
    be obtained by the agent. It uses the methods setArguments and
    generatePlan(), so it should not be overriden unless it should
    made unavailable. */
  virtual Plan generatePlan(const std::vector<Float>& arguments);

  
  /** This method should return a FuzzyBool telling the caller
    if the skill wants to be called again. Abstract since
    no concrete instances of this class should be created. */
  virtual FuzzyBool persistent() = 0;

  /** This method should implement the precondition specified in
   the description of the skill. Abstract since no concrete instances
   of this class should be created. */
  virtual FuzzyBool applicable() = 0;

  
private:
  /** Hide copy-constructor. */
  SkillWithArguments(const SkillWithArguments&);
  /** Hide assignment operator. */
  SkillWithArguments& operator= (const SkillWithArguments&);
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "SkillWithArguments.icc"
#endif


#endif // !SKILLWITHARGUMENTS_H
