/* -*- Mode: C++ -*-
 * $Id: SkillWithPosition.h,v 1.2 2000/09/13 03:10:39 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: SkillWithPosition.h
 * Unitname: Framework
 * $Revision: 1.2 $
 * Created by: Fredrik Heintz 2000-09-09
 * Last modified by $Author: frehe $ $Date: 2000/09/13 03:10:39 $
 *
 *
 * HISTORY
 *
 * $Log: SkillWithPosition.h,v $
 * Revision 1.2  2000/09/13 03:10:39  frehe
 * Added a new class SkillWithArguments to be used by all skills using arguments.
 *
 * Revision 1.1  2000/09/09 11:53:02  frehe
 * Added four new abstract skill classes, SkillWithDirection, Vector, Position and Speed.
 *
 */

/**
 * @file SkillWithPosition.h
 * Include file for the class SkillWithPosition.
 *
 * @par Unit:
 *      Framework
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2000/09/13 03:10:39 $
 */


#ifndef SKILLWITHPOSITION_H
#define SKILLWITHPOSITION_H


#include "SkillWithArguments.h"
#include "library_types.h"


RS_BEGIN_NAMESPACE


/**
 * @class SkillWithPosition SkillWithPosition.h
 * This abstract class represent skill that uses
 * a position as an argument.
 *
 * @par Unit:
 *      Framework
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2000/09/13 03:10:39 $
 */
class SkillWithPosition : public SkillWithArguments {
public:
  /** Constructor taking the default argument for the position
    used as an argument in the skill. */
  SkillWithPosition(const Point& pos);

  /** Destructor cleaning up after the skill. */
  virtual ~SkillWithPosition();


  /** Used to set the skills's arguments in a generic way.
    The first value in the vector will be interpreted as the
    X-coordinate and the second value as the Y-coordinate. */
  virtual void setArguments(const std::vector<Float>& arguments)
     throw(TooFewArgumentsException);


  /** Used to set the skill's position argument to @a pos. */
  virtual void setPosition(const Point& pos);

  /** Return the skill's position argument. */
  virtual Point getPosition() const;

  
  /** This method is used to generate a plan for how the goal of
    the skill should be obtained by the agent. Abstract since
    no concrete instances of this class should be created. */
  virtual Plan generatePlan() = 0;

  /** This method is used to first set the argument of the skill
    and then generate a plan for how the goal of the skill should
    be obtained by the agent. It uses the methods setPosition and
    generatePlan(), so it should not be overriden unless it should
    made unavailable. */
  virtual Plan generatePlan(const Point& pos);

  /** This method should return a FuzzyBool telling the caller
    if the skill wants to be called again. Abstract since
    no concrete instances of this class should be created. */
  virtual FuzzyBool persistent() = 0;

  /** This method should implement the precondition specified in
   the description of the skill. Abstract since no concrete instances
   of this class should be created. */
  virtual FuzzyBool applicable() = 0;

  
protected:  
  /** The position argument. */
  Point position;

  
private:
  /** Hide copy-constructor. */
  SkillWithPosition(const SkillWithPosition&);
  /** Hide assignment operator. */
  SkillWithPosition& operator= (const SkillWithPosition&);
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "SkillWithPosition.icc"
#endif


#endif // !SKILLWITHPOSITION_H
