/* -*- Mode: C++ -*-
 * $Id: SkillWithSpeed.h,v 1.2 2000/09/13 03:10:40 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: SkillWithSpeed.h
 * Unitname: Framework
 * $Revision: 1.2 $
 * Created by: Fredrik Heintz 2000-09-09
 * Last modified by $Author: frehe $ $Date: 2000/09/13 03:10:40 $
 *
 *
 * HISTORY
 *
 * $Log: SkillWithSpeed.h,v $
 * Revision 1.2  2000/09/13 03:10:40  frehe
 * Added a new class SkillWithArguments to be used by all skills using arguments.
 *
 * Revision 1.1  2000/09/09 11:53:02  frehe
 * Added four new abstract skill classes, SkillWithDirection, Vector, Position and Speed.
 *
 */

/**
 * @file SkillWithSpeed.h
 * Include file for the class SkillWithSpeed.
 *
 * @par Unit:
 *      Framework
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2000/09/13 03:10:40 $
 */


#ifndef SKILLWITHSPEED_H
#define SKILLWITHSPEED_H


#include "SkillWithArguments.h"


RS_BEGIN_NAMESPACE


/**
 * @class SkillWithSpeed SkillWithSpeed.h
 * This abstract class represent skill that uses
 * a speed as an argument.
 *
 * @par Unit:
 *      Framework
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2000/09/13 03:10:40 $
 */
class SkillWithSpeed : public SkillWithArguments {
public:
  /** Constructor taking the default argument for the speed
    used as an argument in the skill. */
  SkillWithSpeed(const Float spd);

  /** Destructor cleaning up after the skill. */
  virtual ~SkillWithSpeed();


  /** Used to set the skills's arguments in a generic way.
    The first value in the vector will be interpreted as the speed. */
  virtual void setArguments(const std::vector<Float>& arguments)
     throw(TooFewArgumentsException);


  /** Used to set the skill's speed argument to @a spd. */
  virtual void setSpeed(const Float spd);

  /** Return the skill's speed argument. */
  virtual Float getSpeed() const;

  
  /** This method is used to generate a plan for how the goal of
    the skill should be obtained by the agent. Abstract since
    no concrete instances of this class should be created. */
  virtual Plan generatePlan() = 0;

  /** This method is used to first set the argument of the skill
    and then generate a plan for how the goal of the skill should
    be obtained by the agent. It uses the methods setSpeed and
    generatePlan(), so it should not be overriden unless it should
    made unavailable. */
  virtual Plan generatePlan(const Float spd);

  /** This method should return a FuzzyBool telling the caller
    if the skill wants to be called again. Abstract since
    no concrete instances of this class should be created. */
  virtual FuzzyBool persistent() = 0;

  /** This method should implement the precondition specified in
   the description of the skill. Abstract since no concrete instances
   of this class should be created. */
  virtual FuzzyBool applicable() = 0;

  
protected:  
  /** The speed argument. */
  Float speed;

  
private:
  /** Hide copy-constructor. */
  SkillWithSpeed(const SkillWithSpeed&);
  /** Hide assignment operator. */
  SkillWithSpeed& operator= (const SkillWithSpeed&);
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "SkillWithSpeed.icc"
#endif


#endif // !SKILLWITHSPEED_H
