/* -*- Mode: C++ -*-
 * $Id: View.h,v 1.4 2000/09/13 00:04:53 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: View.h
 * Unitname: Framework
 * $Revision: 1.4 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/13 00:04:53 $
 *
 *
 * HISTORY
 *
 * $Log: View.h,v $
 * Revision 1.4  2000/09/13 00:04:53  frehe
 * Added process-methods that corresponds to the forEach methods in the FrameworkSensorInterface.
 *
 * Revision 1.3  2000/09/05 21:15:55  frehe
 * Made minor changes.
 *
 * Revision 1.2  2000/09/04 15:25:04  frehe
 * Changed the accessability of currentUpdateNumber.
 *
 * Revision 1.1  2000/09/03 22:08:44  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file View.h
 * Include file for the class View.
 *
 * @par Unit:
 *      Framework
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2000/09/13 00:04:53 $
 */


#ifndef VIEW_H
#define VIEW_H


#include "basic_system_types.h"


RS_BEGIN_NAMESPACE


class ViewManager;
class ParameterView;
class FrameworkSensorInterface;


/** Represent the identity of a view, must be unique. */
typedef UInt16 ViewId;


/** Represent the result of updateing a view. */
enum ViewUpdateResult {
  UPDATE_OK,
  UPDATE_ALREADY_MADE,
  UPDATE_MISSING_UPDATES,
  UPDATE_ERROR
};


/** Use getView to get a pointer to the view with ViewId @a view_id.
  The reason for not using view* getView(ViewId) is that
  C++ doesn't allow overloading on the return type. */
template<class view>
void getView(view*& tmp, ViewId view_id)
{
  tmp = dynamic_cast<view*>(ViewManager::instance()->getView(view_id));
  if ( tmp == NULL ) {
    tmp = new view(view_id);
    ViewManager::instance()->addView(tmp, true);
  }
}


/** Use getView to get a pointer to the view with ViewId @a view_id
  if the view needs a history length to be created.
  The reason for not using view* getView(ViewId) is that
  C++ doesn't allow overloading on the return type. */
template<class view>
void getView(view*& tmp, ViewId view_id, UInt hist_len)
{
  tmp = dynamic_cast<view*>(ViewManager::instance()->getView(view_id));
  if ( tmp == NULL ) {
    tmp = new view(view_id, hist_len);
    ViewManager::instance()->addView(tmp, true);
  }
}


/** Use this getView to get a pointer to the view with ViewId @a view_id
  if the view needs a pointer to a ParameterView to be created.
  The reason for not using view* getView(ViewId, ParameterView*) is that
  C++ doesn't allow overloading on the return type. */
template<class view>
void getView(view*& tmp, ViewId view_id, ParameterView* paramView)
{
  tmp = dynamic_cast<view*>(ViewManager::instance()->getView(view_id));
  if ( tmp == NULL ) {
    tmp = new view(view_id, paramView);
    ViewManager::instance()->addView(tmp, true);
  }
}


/** Use this getView to get a pointer to the view with ViewId @a view_id
  if the view needs both a pointer to a ParameterView and a history
  length to be created.
  The reason for not using view* getView(ViewId, ParameterView*) is that
  C++ doesn't allow overloading on the return type. */
template<class view>
void getView(view*& tmp, ViewId view_id,
	     ParameterView* paramView, UInt hist_len)
{
  tmp = dynamic_cast<view*>(ViewManager::instance()->getView(view_id));
  if ( tmp == NULL ) {
    tmp = new view(view_id, paramView, hist_len);
    ViewManager::instance()->addView(tmp, true);
  }
}


/**
 * @class View View.h
 * This class is the abstract base class for all views
 * in the RoboSoc system.
 *
 * @par Unit:
 *      Framework
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2000/09/13 00:04:53 $
 */
class View {
public:
  /** Constructs a view with the ViewId @a id. */
  View(const ViewId id);

  /** Destroys the view. */
  virtual ~View();

  
  /** Get the ViewId of this view. */
  ViewId getId() const;


  /** A pointer to the sensor interface, use it to get the
    information stored in the interface. */
  static const FrameworkSensorInterface* sensorInterface;


  /** Will check and see if the view has already been updated
    if it has not then call the myNewCycle method. */
  ViewUpdateResult newCycle(const UInt32 update_no);
  
  /** Will check and see if the view has already been updated
    if it has not then call the myUpdateAfterCoachMessage method. */
  ViewUpdateResult updateAfterCoachMessage(const UInt32 update_no);

  /** Will check and see if the view has already been updated
    if it has not then call the myUpdateAfterInit method. */
  ViewUpdateResult updateAfterInit(const UInt32 update_no);

  /** Will check and see if the view has already been updated
    if it has not then call the myUpdateAfterSee method. */
  ViewUpdateResult updateAfterSee(const UInt32 update_no);

  /** Will check and see if the view has already been updated
    if it has not then call the myUpdateAfterHear method. */
  ViewUpdateResult updateAfterHear(const UInt32 update_no);

  /** Will check and see if the view has already been updated
    if it has not then call the myUpdateAfterSense method. */
  ViewUpdateResult updateAfterSense(const UInt32 update_no);

  /** Will check and see if the view has already been updated
    if it has not then call the myUpdateAfterCommand method. */
  ViewUpdateResult updateAfterCommand(const UInt32 update_no);


  /** This method could be used to initialize a view. */
  virtual bool init();

  
protected:
  /** A pointer to the view manager. */
  ViewManager* const viewManager;


  /** Called when the sensor interface detects a new cycle.
    It is called *before* the values in SensorInterface
    are erased, so this is the last chance to use them. */
  virtual void myNewCycle() = 0;

  /** Called after a new message from the coach has been recieved. */
  virtual void myUpdateAfterCoachMessage() = 0;

  /** Called after an init message has been received. */
  virtual void myUpdateAfterInit() = 0;

  /** Called after an visual (see) message has been received. */
  virtual void myUpdateAfterSee() = 0;

  /** Called after an aural (hear) message has been received. */
  virtual void myUpdateAfterHear() = 0;

  /** Called after a sense body message has been received. */
  virtual void myUpdateAfterSense() = 0;

  /** Called after a command has been executed. */
  virtual void myUpdateAfterCommand() = 0;

  
  /** The current update number used when explicitly
    updating views that the view depends upon. */
  UInt32 currentUpdateNumber;


  /** True iff the view is initialize with the init method. */
  bool initialized;

  
  /** A dummy method included here so that they can be used
    by the framework sensor interface to process markers. */
  virtual void processMarker(const ServerMarkerObject& marker);

  /** A dummy method included here so that they can be used
    by the framework sensor interface to process lines. */
  virtual void processLine(const ServerLineObject& line);

  /** A dummy method included here so that they can be used
    by the framework sensor interface to process players. */
  virtual void processPlayer(const ServerPlayerObject& player);
  
  /** A dummy method included here so that they can be used
    by the framework sensor interface to process opponents
    as seen by the coach. */
  virtual void processOpponent(const CoachPlayerObject& player);
  
  /** A dummy method included here so that they can be used
    by the framework sensor interface to process teammates
    as seen by the coach. */
  virtual void processTeammate(const CoachPlayerObject& player);
  
  
private:
  ViewId id;
  UInt32 lastUpdateNumber;
  bool updateBegun;

  // Hide copy-constructor and assignment.
  View(const View&);
  View& operator= (const View&);
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "View.icc"
#endif


#endif // !VIEW_H
