/* -*- Mode: C++ -*-
 * $Id: ClearBall.cc,v 1.10 2001/06/19 14:01:42 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *0
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: ClearBall.cc
 * Unitname: Framework/skill
 * $Revision: 1.10 $
 * Created by:Chris & Govert
 * Last modified by $Author: frehe $ $Date: 2001/06/19 14:01:42 $
 *
 *
 * HISTORY
 */

/**
 * @file ClearBall.cc
 * Implementation file for the skill class ClearBall.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.10 $
 * $Author: frehe $
 * $Date: 2001/06/19 14:01:42 $
 */


#include "ClearBall.h"
#include "view_constants.h"
#include "KickBallInDirection.h"
#include "ParameterView.h"
#include "MarkerView.h"
#include "AgentView.h"
#include "PlayerView.h"
#include "BallView.h"
#include "rs_log.h"


RS_BEGIN_NAMESPACE

const float rangeX = 52.5;
const float rangeY = 34.0;

/**
 * This class describes the clearball skill. It's called in an emergency (probably 
 * in front of the own goal) and it results in kicking the ball to an open area on 
 * the field (preferably towards the enemy side of the field)
 *
 * Preconditions
 *  The player needs to be in kicking distance to the ball
 *  It is impossible to give a good pass to a friendly player
 *  It is to dangerous to try to dribble (e.g. if the enemy players get the ball they will score)
 * Postconditions
 *  The ball is kicked to an empty area on the field if an appropriate area exists.
 *  if no good area exists within the view of the player he tries to kick it in the direction 
 *  of the enemy goal
 *  the ball is NEVER kicked in the direction of the own side of the field (because this could 
 *  cause a more dangerous situation)
 *  The ball is only kicked outside of the field if there is no better solution
 */
ClearBall::ClearBall()
  : Skill(),
    parameterView(NULL),
    markerView(NULL),
    agentView(NULL),
    ballView(NULL),
    playerView(NULL)
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(markerView, MARKER_VIEW_ID, parameterView);
  getView(agentView, AGENT_VIEW_ID, parameterView);
  getView(ballView, BALL_VIEW_ID);
  getView(playerView, PLAYER_VIEW_ID);
} 


ClearBall::~ClearBall()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(MARKER_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
}

//-------------------------------------------------------------



float calculateDistance(PointUCO Point1, PointUCO Point2) {
  return  sqrt(  sqr(Point1.getX().getValue().getValue().getValue() - 
		     Point2.getX().getValue().getValue().getValue()) + 
		 sqr(Point1.getY().getValue().getValue().getValue()  -
		     Point2.getY().getValue().getValue().getValue()));
}
PointUCO checkUpperAndBottomLine(float y, float m, float b) {
    float x = (y - b) / m;
    PointUCO p; //= new PointUCO();
    p.setX(x);
    p.setY(y);
    if ((x < (-rangeX)) || (x > rangeX))
      p.makeUnknown();
    return p; 
  }

PointUCO checkLeftAndRightLine(float x, float m, float b) {
    float y = m * x + b;
    PointUCO p;//= new PointUCO(x,y);
    p.setX(x);
    p.setY(y);
    if ((y<(-rangeY))||(y>rangeY)) 
       p.makeUnknown();
    return p;
  }

float calculateDistanceToFieldBorder(PointUCO player, float absAngle) {
  float playerX = player.getX().getValue().getValue().getValue();
  float playerY = player.getY().getValue().getValue().getValue();
  PointUCO intersection;
  float distance;
  
  //RS_LOG( LA_SCORE, "  calculate distance: x=" << playerX << " y=" << playerY << " absAngle:" << absAngle);
  if ((absAngle > 89) && (absAngle < 91))
    distance = abs(rangeY - playerY);
  else if ((absAngle > -91) && (absAngle < -89))
    distance = abs(playerY - rangeY);
  else {  //y=mx+b
    float m = tan(absAngle);
    float b = playerY - m * playerX;  //b=y-mx
    if (absAngle > 0)
      intersection = checkUpperAndBottomLine(rangeY, m, b); 
    else
      intersection = checkUpperAndBottomLine(-rangeY, m, b);   
   if (intersection.isUnknown())
     {
     if ((abs(absAngle) < 90))
        intersection = checkLeftAndRightLine(rangeX,m ,b); 
      else 
        intersection = checkLeftAndRightLine(-rangeX,m ,b);
     }
    //if (intersecton == null) then fatal error
    distance = calculateDistance(player, intersection); 
  }
  //RS_LOG( LA_SCORE, "  calculate distance returns " << distance);
  return distance;
}


/**
 *This method calculates a value for an angle. The value describes how good
 * it is to Clear the ball in this direction
 * Arguments
 * width: the width between two opponents associated with this angle
 * angle: the angle between two opponents (relative to the player)
 * distanceA: the distance to the first enemy
 * distanceB: the distance to the second enemy
 * goaldir: the direction towards the enemy goal
 * Preconditions
 *  The bodydirection of the player must be known
 * Postconditions
 *  A value describing the quality of the angle is returned ass a float
 */
float ClearBall::calculateValue(float width, float angle,
                   float distanceA, float distanceB,float goaldir)
  //width is the angle between two opponents
  //angle is the relative kickdirection
{ float modifier=1.0;

  //distance
  if(distanceA<3) modifier=modifier*0.1;
  if(distanceB<3)modifier=modifier*0.1;
  
 
  if (!agentView->getBodyDirection(NOW).isUnknown()) { 
    float absAngle = angle - agentView->getBodyDirection(NOW).getDeg(); 
    
    //distance to outside the field
    /*
    PointUCO p = agentView->getAgentPosition(NOW);
    if (!p.isUnknown())
    {
     float distanceC = calculateDistanceToFieldBorder(p, absAngle);
     if(!(distanceC>25))
       {
	 if(distanceC<2)modifier=modifier*0.01;
	 else if (distanceC<5) modifier=modifier*0.1;
	 else if (distanceC<10) modifier=modifier*0.3;
	 else if (distanceC<15) modifier=modifier*0.5;
	 else modifier=modifier*0.8;
       }
    }
    */

    //direction
    //if the angle is towards the own side of the field:modifier=0
    if ((absAngle >  90) ||
        (absAngle < -90)) 
	modifier=0.01;
    else
      //best case is 1, worst case is 0.5 (absAngle-goaldir==180 degrees)
      modifier=modifier* (1 - ((abs(absAngle - goaldir) / 180)) / 2);
  }
  
  return (width*modifier);
}


/*
 * returns the width between two angles as a positive float
 */
float getWidth(float firstAngle, float secondAngle)
{
  return abs(secondAngle - firstAngle);
}


/*
 * returns the middle of two angles
 */
float getAngleBetween(float firstAngle, float secondAngle)
  // the two angles are relativ to avoid problems
{
  return (firstAngle + secondAngle) / 2;
}


/**
 * This method returns the relative angle(?) of the direction to the 
 * opponents' goal
 * If the goal can't be seen the angle towards the enemy side of the field is 
 * returned (in this case 0)
 */
float ClearBall::theirGoal()
{ 
  PointUCO pt = agentView->getAgentPosition(NOW);
  VectorUCO vec = VectorUCO(pt, markerView->getTheirGoal(NOW).getPosition());  
  if (vec.isUnknown()) {
    if  (agentView->getBodyDirection(NOW).isUnknown())
      return  0.0;
    else
      return agentView->getBodyDirection(NOW).getDeg();
  }
  else
    return vec.getAngle().getDeg();
}


/*
 * a method for sorting the angles from smallest to largest
 */
/*
void  bubbleSort(vector<float>* A, vector<float>* B,int size)
{
     float temp;
     for (int pass = 0; pass < size - 1; pass++)
         {
	   for (int i = size - 1; i > pass; i--)
                 if (*A[i - 1] > *A[i])
                 {
                     temp = *A[i - 1];
		     *A[i - 1] = *A[i];
                     *A[i] = temp;
                 
		     temp = *B[i - 1];
		     *B[i - 1] = *B[i];
                     *B[i] = temp;
		 }
	 }
 }
 */
                                

/**
 * A method for calculating the best angle (relative to the player) to clear the ball
 * a value for each angle between two opponents is calculated and 
 * the best one is returned
 * when there are no opponents in view the player tries to kick to the direction most 
 * towards the enemy goal
 * when the opponent can't kick in any direction in his view he kicks to the enemy goal
 */
float ClearBall::calculateBestAngle()
{
  /*struct angleAndDistance {
    float angle;
    float distance;
  };*/

  AngleDegUCO neckAngle = agentView->getNeckDirection(NOW);
  AngleDegUnknown angleSize = agentView->getViewAngle(NOW);
  float bestAngle = 0;
  if (neckAngle.isUnknown() || angleSize.isUnknown())
    bestAngle = theirGoal();   //improve: direction to their goal instead 0.0
  else {
      //get all opponents
      std::vector<PlayerObject> opponentVector = playerView->getOpponents(NOW);
      std::vector<float> opponentAngles;       
      std::vector<float> opponentDistances;
  
      //get all the angles of the opponents and include the two borders of the view
      if (opponentVector.size() > 0) {
        RS_LOG( LA_SCORE, "\n " << opponentVector.size() << " opponents in view:");
        opponentAngles.   resize(opponentVector.size() + 2);
        opponentDistances.resize(opponentVector.size() + 2);     
	opponentAngles[0] = -angleSize.getAngle().getDeg() /2; //left border
        opponentDistances[0] = 100.0;
        for (unsigned int i=0; i < opponentVector.size(); i++) {
            VectorUCO angleVector = VectorUCO(
	      agentView->getAgentPosition(NOW),    // own position
              opponentVector[i].getPosition());    // position of enemy
            opponentAngles   [i+1] = angleVector.getAngle().getDeg();  //rel. angle?
            opponentDistances[i+1] = 1.0; // for test purpose only
            //RS_LOG( LA_SCORE, "\n - opponentAngles: " << opponentAngles[i+1]);
            
            /* normiliation is not necessary, fredrick told
            if (opponentAngles[i+1] > 180)
              opponentAngles[i+1] = opponentAngles[i+1] - 360;*/ 
            
            if (!agentView->getBodyDirection(NOW).isUnknown())  
              opponentAngles[i+1] = opponentAngles[i+1]
		- agentView->getBodyDirection(NOW).getDeg(); //rel. angle	    
        }
        opponentAngles[opponentVector.size()+1] = -opponentAngles[0];   //right border
	opponentDistances[opponentVector.size()+1] = 100.0;

	//bubbleSort(&opponentAngles, &opponentDistances, opponentAngles.size());
      float temp;
      for (unsigned int pass = 0; pass < opponentAngles.size() - 1; pass++)
         {
	   for (unsigned int i = opponentAngles.size() - 1; i > pass; i--)
                 if (opponentAngles[i - 1] > opponentAngles[i])
                 {
                     temp = opponentAngles[i - 1];
		     opponentAngles[i - 1] = opponentAngles[i];
                     opponentAngles[i] = temp;
                 
		     temp = opponentDistances[i - 1];
		     opponentDistances[i - 1] = opponentDistances[i];
                     opponentDistances[i] = temp;
		 }
	 }

      } 
      else {	        
	//bestAngle = theirGoal();
        RS_LOG( LA_SCORE, "\n no opponents in view"); 
        opponentAngles.   resize(5);
        opponentDistances.resize(5);
        opponentAngles[0] = -angleSize.getAngle().getDeg() /2;
        opponentAngles[1] = -angleSize.getAngle().getDeg() /4;
        opponentAngles[2] = 0.0;
        opponentAngles[3] = +angleSize.getAngle().getDeg() /4;
        opponentAngles[4] = +angleSize.getAngle().getDeg() /2; 
        for (unsigned int i=0; i < opponentDistances.size(); i++) 
          opponentDistances[i] = 100.0;    
      }
      
        //calculate the best kickdirection
        float newValue;
        float newAngle;
        float bestValue = 0;
        for (unsigned int i=0; i < opponentAngles.size()-1; i++) {
          //RS_LOG( LA_SCORE, "\n opponentAngles: " << opponentAngles[i] << " and " << opponentAngles[i+1]);
          //calculate possible kick direction
          newAngle = getAngleBetween(opponentAngles[i], opponentAngles[i+1]);
          //calculate if it makes sense or not 
	  newValue = calculateValue(getWidth(opponentAngles[i],opponentAngles[i+1]), 
                  newAngle, opponentDistances[i], opponentDistances[i+1], theirGoal());
          //RS_LOG( LA_SCORE, "\n rel.angle:" << newAngle << ", value:" << newValue);    
	  if (newValue > bestValue) {
	    bestValue = newValue;
            bestAngle = newAngle;
       	  }
        }
	if (bestValue<1) {
          RS_LOG( LA_SCORE, "\n no good kickdirection was found"); 
	  if (!agentView->getBodyDirection(NOW).isUnknown())  
            bestAngle= theirGoal(); //+ agentView->getBodyDirection(NOW).getDeg();
	  else 
            bestAngle=0.0;
        }
  }
  
  /*
  int extremelyimportantvariable;
  int Frederikisourfriend;
  int RoboSocFormattingInProgress;
  int tigredInReadingWarnings;
  int X;
  int good_;
  int for_;
  int Nothing;
  int ParadiseLostFanPage;
  int LOOK_FOR_SEX_on_www_beuzel_de;
  int or_www_whocare_com;
  int _WE_LOVE_YOU_;
  int _JoTakesDrugs_;
  int I_will_marry_manson;
  int x_______________________________a;
  int x_______________________________b;
  int x_______________________________c;
  int ____RUFUS_AND_MARTIN_ARE_GAY____a;
  int ____RUFUS_AND_MARTIN_ARE_GAY____b;
  int ____RUFUS_AND_MARTIN_ARE_GAY____c;
  int ____RUFUS_AND_MARTIN_ARE_GAY____d;
  int x_______________________________d;
  int x_______________________________e;
  int x_______________________________f;
  int this__;
  int is_;
  int stupid;
  int and_;
  int childish;
  int compiling_sucks;
  int _TERRITORY_CONTROL_VIEW_ID;
  int funnyPicturesOn_;
  int www_dot_i_dot_am_slash_schweden;
  int Thanks;
  int to_;
  int the_;
  int audience;
  int the_entertaining_company;
  int supergroovy_and_;
  int the_artist_formarly_known_as_Chris;
  int wanna_buy_marihuana_;
  int contact_RUFUS_AND_MARTIN;
  int ten_kronar_or_the_natural_way;
  int we_are_clean;
  */

  return bestAngle;  //important: is this abs or rel? its better to do rel  
}

//------------------------------------------------------------

Plan ClearBall::generatePlan()
{
  //cout << "\ngeneratePlan: DRINK DEVIL BEER!";

  //static Command* cmd;
  KickBallInDirection* pKick;
  static VectorUCO vec;
  static Plan plan;

  //PointUCO pt = agentView->getAgentPosition(NOW);
  //vec = VectorUCO(pt, markerView->getTheirGoal(NOW).getPosition());
 
  //RS_LOG( LA_SCORE, "\n****************************");
  RS_LOG( LA_SCORE, "\n execute calculateAngle:");
  float ourangle = calculateBestAngle();
  RS_LOG( LA_SCORE, "\n return angle " << ourangle);
  //RS_LOG( LA_SCORE, "\n****************************");

  
  if ( agentView->getBodyDirection(NOW).isUnknown() ) {
    pKick = new KickBallInDirection();

    std::vector<Float> vect(2);
    vect[0] = 150;
    vect[1] = ourangle;

    pKick->setArguments( vect );
  }
  else {
    pKick = new KickBallInDirection;

    std::vector<Float> vect(2);
    vect[0] = 150;
    vect[1] = ourangle - 
      agentView->getBodyDirection(NOW).getDeg();

    pKick->setArguments( vect );
  }

  plan = pKick->generatePlan();

  delete pKick;
  pKick = NULL;
  
  return plan;
}


FuzzyBool ClearBall::persistent()
{
  return FALSE;
}


FuzzyBool ClearBall::applicable()
{
  if ( ballView->getRelativeBallDirection(NOW).isKnown() &&
       ballView->getBallDistance(NOW) < 1 )
    return TRUE;
  else
    return FALSE;
}


RS_END_NAMESPACE
