/* -*- Mode: C++ -*-
 * $Id: CoordRectangle.h,v 1.3 2001/06/19 14:01:43 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: CoordRectangle.h
 * Unitname: Library
 * $Revision: 1.3 $
 * Created by: Ola Lundqvist 1999-09-22
 * Last modified by $Author: frehe $ $Date: 2001/06/19 14:01:43 $
 *
 *
 * HISTORY
 *
 * $Log: CoordRectangle.h,v $
 * Revision 1.3  2001/06/19 14:01:43  frehe
 * Made minor changes to compile RoboSoc with gcc-3.0
 *
 * Revision 1.2  2000/11/09 07:53:54  olalu526
 * Fixed a nice little bug.
 *
 * Revision 1.1  2000/11/07 17:23:59  olalu526
 * Generalized the zone concept to handle various type of objects. It now handles a set of rectangles and circles to form an area. This affected quite a lot of skills.
 *
 *
 */

/**
 * @file CoordRectangle.h
 * Include file for the template CoordRectangle.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.3 $
 * $Author: frehe $
 * $Date: 2001/06/19 14:01:43 $
 */


#ifndef COORDRECTANGLE_H
#define COORDRECTANGLE_H


#include <iostream>

#include "library_types.h"
#include "math_utils.h"
#include "AngleDegrees.h"
#include "CoordVector.h"
#include "CoordPoint.h"
#include "CoordArea.h"


/** String to use as the header when outputing a CoordRectangle. */
#define RECTANGLE_HEAD std::string("#Rectangle")


RS_BEGIN_NAMESPACE


template <class T> class CoordVector;
template <class T> class CoordPoint;
template <class T> class AngleDegrees;


/**
 * @class CoordRectangle CoordRectangle.h
 * Template class for representing a cartesian two dimensional
 * coordinate with a radiues, forming a rectangle.
 * The X, Y and R values are of type T.
 *
 * <!-- @todo -->
 *
 * @par Unit:
 *      Library
 * $Revision: 1.3 $
 * $Author: frehe $
 * $Date: 2001/06/19 14:01:43 $
 */
template<class T> class CoordRectangle : public CoordArea<T> {
public:
  CoordRectangle();
  CoordRectangle(const T&, const T&, const T&, const T&);
  CoordRectangle(const CoordRectangle<T>&);
  CoordRectangle(const CoordPoint<T> &, const CoordPoint<T>&);

  CoordRectangle<T> operator - () const;
  CoordRectangle<T> operator - (const CoordRectangle<T>&) const;
  CoordRectangle<T> operator - (const CoordVector<T>&) const;
  CoordRectangle<T> operator - (const CoordPoint<T>&) const;
  CoordRectangle<T> operator + (const CoordRectangle<T>&) const;
  CoordRectangle<T> operator + (const CoordVector<T>&) const;
  CoordRectangle<T> operator + (const CoordPoint<T>&) const;
  CoordRectangle<T> operator * (const T&) const;
  CoordRectangle<T> operator * (const CoordRectangle<T>&) const;
  CoordRectangle<T> operator / (const T&) const;
  CoordRectangle<T> operator / (const CoordRectangle<T>&) const;

  CoordRectangle<T>& operator =(const CoordRectangle<T>&);

  CoordRectangle<T>& operator -=(const CoordRectangle<T>&);
  CoordRectangle<T>& operator -=(const CoordVector<T>&);
  CoordRectangle<T>& operator -=(const CoordPoint<T>&);
  CoordRectangle<T>& operator +=(const CoordRectangle<T>&);
  CoordRectangle<T>& operator +=(const CoordVector<T>&);
  CoordRectangle<T>& operator +=(const CoordPoint<T>&);
  CoordRectangle<T>& operator *=(const T&);
  CoordRectangle<T>& operator *=(const CoordRectangle<T>&);
  CoordRectangle<T>& operator /=(const T&);
  CoordRectangle<T>& operator /=(const CoordRectangle<T>&);
  
  bool operator ! () const;
  bool operator ==(const CoordRectangle<T>&) const;
  bool operator !=(const CoordRectangle<T>&) const;

  T getX1() const;
  T getY1() const;
  T getX2() const;
  T getY2() const;
  CoordPoint<T>& getP1() const;
  CoordPoint<T>& getP2() const;
  CoordRectangle<T>& setX1(const T&);
  CoordRectangle<T>& setX2(const T&);
  CoordRectangle<T>& setY1(const T&);
  CoordRectangle<T>& setY2(const T&);
  CoordRectangle<T>& setP1(const CoordPoint<T>&);
  CoordRectangle<T>& setP2(const CoordPoint<T>&);
  CoordRectangle<T>& set (const CoordPoint<T>& pt_x, const CoordPoint<T>& pt_y);
  
  CoordVector<T> getVectorTo(const CoordRectangle<T>& pt) const;
  CoordVector<T> getVectorTo(const CoordPoint<T>& pt) const;
  CoordPoint<T> getPointOf() const;
  AngleDegrees<T> getDirectionTo(const CoordRectangle<T>& pt) const;
  AngleDegrees<T> getDirectionTo(const CoordPoint<T>& pt) const;
  T getDistanceTo(const CoordRectangle<T>& pt) const;
  T getDistanceTo(const CoordPoint<T>& pt) const;

  FuzzyBool inArea(const CoordPoint<T>& pt);
  //FuzzyBool throughArea(const CoordPoint<T>& pt);
  
protected:
  CoordPoint<T> p1;
  CoordPoint<T> p2;
};

typedef CoordRectangle<Float> Rectangle;

template<class T>
inline FuzzyBool CoordRectangle<T>::inArea(const CoordPoint<T> & pt) {
  //cerr << "Rectangle::inArea" << endl;
  T minx = min(p1.getX(),p2.getX());
  T miny = min(p1.getY(),p2.getY());
  T maxx = max(p1.getX(),p2.getX());
  T maxy = max(p1.getY(),p2.getY());
  //cerr << "inRArea: (" << p1 << ", " << p2 << ") " << pt;
  if (pt.getX() < maxx &&
      pt.getX() > minx &&
      pt.getY() < maxy &&
      pt.getY() > miny) {
    //    cerr << " TRUE!" << endl << flush;
    return TRUE;
  }
  //cerr << " FALSE!" << endl << flush;
  return FALSE;
}

/*
 * X, Y is inverted (negated) but the radius remains unchanged.
 */
template<class T>
inline CoordRectangle<T> CoordRectangle<T>::operator -() const
{
  return CoordRectangle<T>(-p1,-p2);
}

template<class T>
inline CoordRectangle<T> CoordRectangle<T>::operator -(const CoordRectangle<T>& pt) const
{
  return CoordRectangle<T>(p1 - pt.p1, p2 - pt.p2);
}

template<class T>
inline CoordRectangle<T> CoordRectangle<T>::operator -(const CoordVector<T>& pt) const
{
  return CoordRectangle<T>(p1 - pt, p2 - pt);
}

template<class T>
inline CoordRectangle<T> CoordRectangle<T>::operator -(const CoordPoint<T>& pt) const
{
  return CoordRectangle<T>(p1 - pt, p2 - pt);
}

template<class T>
inline CoordRectangle<T> CoordRectangle<T>::operator *(const T& a) const
{
  return CoordRectangle<T>(p1 * a, p2 * a);
}

template<class T>
inline CoordRectangle<T> CoordRectangle<T>::operator *(const CoordRectangle<T>& pt) const
{
  return CoordRectangle<T>(p1 * pt.p1, p2 * pt.p2);
}

template<class T>
inline CoordRectangle<T> operator *(const T& a, const CoordRectangle<T>& pt)
{
  return pt * a;
}

template<class T>
inline CoordRectangle<T> CoordRectangle<T>::operator /(const T& a) const
{
#if CHECK_DIVISION_BY_ZERO == 1
  if (a == T(0))
    return COORDRECTANGLE_DIVISION_BY_ZERO_RESULT;
  else
#endif
    return CoordRectangle<T>(p1 / a, p2 / a);
}

template<class T>
inline CoordRectangle<T> CoordRectangle<T>::operator /(const CoordRectangle<T>& pt) const
{
#if CHECK_DIVISION_BY_ZERO == 1
  if ( pt.p1 == T(0) ) {
    if ( pt.p2 == T(0) ) {
      return CoordRectangle_DIVISION_BY_ZERO_RESULT;
    } else {
      return CoordRectangle<T>(T_DIVISION_BY_ZERO_RESULT, p2 / pt.p2);
    }
  } else {
    if ( pt.p2 == T(0) ) {
      return CoordRectangle<T>(p1 / pt.p1, T_DIVISION_BY_ZERO_RESULT);
    } else {
#endif
      return CoordRectangle<T>(p1 / pt.p1, p2 / pt.p2);
#if CHECK_DIVISION_BY_ZERO == 1      
    }
  }
#endif
}

template<class T>
inline CoordRectangle<T> CoordRectangle<T>::operator +(const CoordRectangle<T>& pt) const
{
  return CoordRectangle<T>(p1 + pt.p1, p2 + pt.p2);
}

template<class T>
inline CoordRectangle<T> CoordRectangle<T>::operator +(const CoordVector<T>& pt) const
{
  return CoordRectangle<T>(p1 + pt, p2 + pt);
}

template<class T>
inline CoordRectangle<T> CoordRectangle<T>::operator +(const CoordPoint<T>& pt) const
{
  return CoordRectangle<T>(p1 + pt, p2 + pt);
}

template<class T>
inline CoordRectangle<T>& CoordRectangle<T>::operator-=(const CoordRectangle<T>& pt)
{
  p1 -= pt.p1;
  p2 -= pt.p2;
  return *this;
}

template<class T>
inline CoordRectangle<T>& CoordRectangle<T>::operator-=(const CoordPoint<T>& pt)
{
  p1 -= pt;
  p2 -= pt;
  return *this;
}

template<class T>
inline CoordRectangle<T>& CoordRectangle<T>::operator-=(const CoordVector<T>& vec)
{
  *this = *this - vec;
  return *this;
}

template<class T>
inline CoordRectangle<T>& CoordRectangle<T>::operator+=(const CoordRectangle<T>& pt)
{
  p1 += pt.p1;
  p2 += pt.p2;
  return *this;
}

template<class T>
inline CoordRectangle<T>& CoordRectangle<T>::operator+=(const CoordPoint<T>& pt)
{
  p1 += pt;
  p2 += pt;
  return *this;
}

template<class T>
inline CoordRectangle<T>& CoordRectangle<T>::operator+=(const CoordVector<T>& vec)
{
  *this = *this + vec;
  return *this;
}

template<class T>
inline CoordRectangle<T>& CoordRectangle<T>::operator*=(const T& a)
{
  p1 *= a;
  p2 *= a;
  return *this;
}

template<class T>
inline CoordRectangle<T>& CoordRectangle<T>::operator*=(const CoordRectangle<T>& pt)
{
  p1 *= pt.p1;
  p2 *= pt.p2;
  return *this;
}

template<class T>
inline CoordRectangle<T>& CoordRectangle<T>::operator/=(const T& a)
{
#if CHECK_DIVISION_BY_ZERO == 1
  if ( a == T(0)) {
    p1 = T_DIVISION_BY_ZERO_RESULT;
    p2 = T_DIVISION_BY_ZERO_RESULT;
  } else {
#endif
    p1 /= a;
    p2 /= a;
#if CHECK_DIVISION_BY_ZERO == 1
  }
#endif
  return *this;
}

template<class T>
inline CoordRectangle<T>& CoordRectangle<T>::operator/=(const CoordRectangle<T>& pt)
{
#if CHECK_DIVISION_BY_ZERO == 1
  if ( pt.p1 == T(0) )
    p1 = T_DIVISION_BY_ZERO_RESULT;
  else
#endif
    p1 /= pt.p1;
  
#if CHECK_DIVISION_BY_ZERO == 1
  if ( pt.p2 == T(0) )
    p2 = T_DIVISION_BY_ZERO_RESULT;
  else
#endif
    p2 /= pt.p2;

  return *this;
}

template<class T>
CoordRectangle<T>& CoordRectangle<T>::operator =(const CoordRectangle<T>& pt)
{
  if ( this != &pt ) {
    p2 = pt.p2;
    p1 = pt.p1;
  }
  return *this;
}

template<class T>
inline bool CoordRectangle<T>::operator !() const
{
  return (!p2) && (!p1);
}

template<class T>
inline bool CoordRectangle<T>::operator ==(const CoordRectangle<T>& pt) const
{
  return (p2 == pt.p2) && (p1 == pt.p1);
}

template<class T>
inline bool CoordRectangle<T>::operator !=(const CoordRectangle<T>& pt) const
{
  return (p2 != pt.p2) || (p1 != pt.p1) || (r != pt.r);
}

/*
inline bool CoordRectangle<T>::InFront(const CoordRectangle<T>& pt) const
{
  return (p1 > pt.p1);
}

inline bool CoordRectangle<T>::Behind(const CoordRectangle<T>& pt) const
{
  return (p1 < pt.p1);
}
*/

/*
inline bool CoordRectangle<T>::LeftOf(const CoordRectangle<T>& pt) const
{
  return (p1 > pt.p1);
}

inline bool CoordRectangle<T>::RightOf(const CoordRectangle<T>& pt) const
{
  return (p1 < pt.p1);
}

inline bool CoordRectangle<T>::Above(const CoordRectangle<T>& pt) const
{
  return (p2 < pt.p2);
}

inline bool CoordRectangle<T>::Below(const CoordRectangle<T>& pt) const
{
  return (p2 > pt.p2);
}
*/

template<class T>
inline T CoordRectangle<T>::getX1() const
{
  return p1.getX();
}

template<class T>
inline T CoordRectangle<T>::getY1() const
{
  return p1.getX();
}

template<class T>
inline T CoordRectangle<T>::getX2() const
{
  return p2.getX();
}

template<class T>
inline T CoordRectangle<T>::getY2() const
{
  return p2.getY();
}

template<class T>
inline CoordPoint<T>& CoordRectangle<T>::getP1() const
{
  return p1;
}

template<class T>
inline CoordPoint<T>& CoordRectangle<T>::getP2() const
{
  return p2;
}

template<class T>
inline CoordRectangle<T>& CoordRectangle<T>::setP1(const CoordPoint<T>& pt_x)
{
  p1 = pt_x;
  return *this;
}

template<class T>
inline CoordRectangle<T>& CoordRectangle<T>::setP2(const CoordPoint<T>& pt_y)
{
  p2 = pt_y;
  return *this;
}

template<class T>
inline CoordRectangle<T>& CoordRectangle<T>::setX1(const T& pt_x)
{
  p1.setX(pt_x);
  return *this;
}

template<class T>
inline CoordRectangle<T>& CoordRectangle<T>::setY1(const T& pt_y)
{
  p1.setY(pt_y);
  return *this;
}

template<class T>
inline CoordRectangle<T>& CoordRectangle<T>::setX2(const T& pt_x)
{
  p2.setX(pt_x);
  return *this;
}

template<class T>
inline CoordRectangle<T>& CoordRectangle<T>::setY2(const T& pt_y)
{
  p2.setY(pt_y);
  return *this;
}

template<class T>
inline CoordRectangle<T>& CoordRectangle<T>::set(const CoordPoint<T>& pt_x, const CoordPoint<T>& pt_p2)
{
  p1 = pt_x;
  p2 = pt_p2;
  return *this;
}

template<class T>
inline CoordVector<T> CoordRectangle<T>::getVectorTo(const CoordRectangle<T>& pt) const
{
  return Vector(pt.getPointOf() - getPointOf());
}

template<class T>
inline CoordVector<T> CoordRectangle<T>::getVectorTo(const CoordPoint<T>& pt) const
{
  return Vector(pt - getPointOf());
}

template<class T>
inline AngleDegrees<T> CoordRectangle<T>::getDirectionTo(const CoordRectangle<T>& pt) const
{
  return Vector(pt.getPointOf() - getPointOf()).getAngle();
}

template<class T>
inline AngleDegrees<T> CoordRectangle<T>::getDirectionTo(const CoordPoint<T>& pt) const
{
  return Vector(pt - getPointOf()).getAngle();
}

template<class T>
inline T CoordRectangle<T>::getDistanceTo(const CoordRectangle<T>& pt) const
{
  return Vector(pt.getPointOf() - getPointOf()).getLength();;
}

template<class T>
inline T CoordRectangle<T>::getDistanceTo(const CoordPoint<T>& pt) const
{
  return Vector(pt - getPointOf()).getLength();;
}

template<class T>
inline std::ostream& operator<<(std::ostream& os, const CoordRectangle<T>& pt)
{
#ifdef RECTANGLE_HEAD
  return os << RECTANGLE_HEAD << "(" << pt.getX() << ", " << pt.getY() << ", " << pt.getR() << ")";
#else
  return os << "(" << pt.getX() << ", " << pt.getY() << ", " << pt.getR() << ")";
#endif
}

template<class T>
inline const std::string ToString(const CoordRectangle<T>& pt)
{
#ifdef RECTANGLE_HEAD
  return RECTANGLE_HEAD + "(" + ToString(pt.getX()) + ", " + ToString(pt.getY()) + ", " + ToString(pt.getR()) + ")";
#else
  return "(" + ToString(pt.getX()) + ", " + ToString(pt.getY()) + ", " + ToString(pt.getR()) + ")";
#endif
}

template<class T>
CoordRectangle<T>::CoordRectangle()
  : CoordArea<T>('r'),
    p1(),
    p2()
{
}

template<class T>
CoordRectangle<T>::CoordRectangle(const T& pt_x1, const T& pt_y1, const T& pt_x2, const T& pt_y2)
  : CoordArea<T>('r'),
    p1(pt_x1, pt_y1),
    p2(pt_x2, pt_y2)
{
}

template<class T>
CoordRectangle<T>::CoordRectangle(const CoordRectangle<T>& pt)
  : CoordArea<T>('r'),
    p1(pt.p1),
    p2(pt.p2)
{
}

template<class T>
CoordRectangle<T>::CoordRectangle(const CoordPoint<T>& pt1, const CoordPoint<T>& pt2) : CoordArea<T>('r'),
  p1(pt1),
  p2(pt2)
{
}

template<class T>
CoordPoint<T> CoordRectangle<T>::getPointOf() const
{
  return (p1+p2)/2;
}

template<class T>
std::istream& operator>>(std::istream& is, CoordRectangle<T>& pt)
{
  /* Input formats: f is a T
	(f,f) = (f,f)
  */

  T pt_x1;
  T pt_x2;
  T pt_y1;
  T pt_y2;
  char c = 0;

  is >> std::ws >> c;
  if (c == '(') {
    is >> std::ws >> pt_x1 >> std::ws >> c;
    if (c == ',') is >> std::ws >> pt_y1 >> std::ws >> c;
    if (c != ',') is.clear(std::ios::badbit); //error...;
    if (c == ',') is >> std::ws >> pt_x2 >> std::ws >>  c;
    if (c != ',') is.clear(std::ios::badbit); //error...;
    if (c == ',') is >> std::ws >> pt_y2 >> std::ws >> c;
    if (c != ')') is.clear(std::ios::badbit); //error...;
  }
  else {
    is.clear(std::ios::badbit); //error...;
  }

  if (is) pt = CoordRectangle<T>(CoordPoint<T>(pt_x1,pt_y1),
				 CoordPoint<T>(pt_x2,pt_y2));

  return is;
}


RS_END_NAMESPACE


#endif // COORDRECTANGLE_H
