/* -*- Mode: C++ -*-
 * $Id: CutOffGoalAngleStamina.cc,v 1.1 2000/11/07 18:05:35 mikbr212 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: CutOffGoalAngleStamina.cc
 * Unitname: Framework/skill
 * $Revision: 1.1 $
 * Last modified by $Author: mikbr212 $ $Date: 2000/11/07 18:05:35 $
 *
 *
 * HISTORY
 * $Log: CutOffGoalAngleStamina.cc,v $
 * Revision 1.1  2000/11/07 18:05:35  mikbr212
 * Created
 *
 */

/**
 * @file CutOffGoalAngleStamina.cc
 * Implementation file for the skill class CutOffGoalAngleStamina.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.1 $
 * $Author: mikbr212 $
 * $Date: 2000/11/07 18:05:35 $
 */


#include "CutOffGoalAngleStamina.h"
#include "view_constants.h"
#include "TurnCommand.h"
#include "DashCommand.h"
#include "ParameterView.h"
#include "MarkerView.h"
#include "AgentView.h"
#include "BallView.h"
#include "rs_log.h"


RS_BEGIN_NAMESPACE


CutOffGoalAngleStamina::CutOffGoalAngleStamina()
  : Skill(),
    goalLine(Point(),Vector()),
    pos(Point(0,0)),
    parameterView(NULL),
    markerView(NULL),
    agentView(NULL),
    ballView(NULL),
    kickBallView(NULL)
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(markerView, MARKER_VIEW_ID, parameterView);
  getView(ballView, BALL_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID, parameterView);
  CommandView* commandView = NULL;
  getView(commandView, COMMAND_VIEW_ID);
  getView(kickBallView, KICKBALL_VIEW_ID, parameterView,
	  ballView, agentView, commandView);
  
  goalLine = Line(Point(-52.5,-parameterView->SP_goal_width/2),
		  Point(-52.5, parameterView->SP_goal_width/2));
  blockRadius = max(parameterView->SP_catchable_area_l,
		    parameterView->SP_player_size + parameterView->SP_ball_size) * 1.1;
  maxDistFromGoal = parameterView->SP_penalty_area_length;
} 


CutOffGoalAngleStamina::~CutOffGoalAngleStamina()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(COMMAND_VIEW_ID);
  viewManager->releaseView(KICKBALL_VIEW_ID);
  viewManager->releaseView(MARKER_VIEW_ID);
}


Point CutOffGoalAngleStamina::getOptimum(const Point& ballPos) {
  // calculate optimal position
  Vector leftVec = Vector(goalLine.getStartPoint() - ballPos);
  Vector rightVec = Vector(goalLine.getEndPoint() - ballPos);
  
  Vector midVec = leftVec + rightVec;
  float angleDiff = abs((leftVec.getAngle() - midVec.getAngle()).getDeg());
  midVec.setLength(blockRadius / tan(angleDiff));
  //  cout << "midVec = " << midVec.getLength() << " v " << midVec.getAngle().getDeg() << endl;

  Point optimalPos = ballPos + midVec;
  
  //  cout << "ballPos: " << ballPos.getX() << ", " << ballPos.getY() << endl;

  //  cout << "pos: " << optimalPos.getX() << ", " << optimalPos.getY() << endl;
  
  Vector goalToOptimum = Vector(optimalPos - Point(-52.5,0));
  if(goalToOptimum.getLength() > maxDistFromGoal) {
    goalToOptimum.setLength(maxDistFromGoal);
    optimalPos = Point(goalToOptimum) + Point(-52.5,0);
    //    cout << "crop\n";
    }
  
  //  cout << "pos: " << optimalPos.getX() << ", " << optimalPos.getY() << endl;
  return optimalPos;
}
  
PointUCO CutOffGoalAngleStamina::getAbsoluteBallPosition() {
  return agentView->getAgentPosition(NOW) + 
    PointUCO(ballView->getAbsoluteBallVector(NOW));
}


/**
 * Go to optimal position.
 */ 
Plan CutOffGoalAngleStamina::generatePlan()
{

  if(!getAbsoluteBallPosition().isKnown())
    return Plan();

  Point ballPos = getAbsoluteBallPosition().getPoint();

  if(timeStamp + 1 < Time::currentTime) {
    pos = getOptimum(ballPos);
    timeStamp = Time::currentTime;
  }

  return goSkill.generatePlan(pos);
}


FuzzyBool CutOffGoalAngleStamina::persistent()
{
  return !(timeStamp + 3 < Time::currentTime) && goSkill.persistent();
}


/**
 * True if goal is not blocked.
 */
FuzzyBool CutOffGoalAngleStamina::applicable()
{
  if(!getAbsoluteBallPosition().isKnown())
    return false;

  Point ballPos = getAbsoluteBallPosition().getPoint();

  // Project ball to goal-line
  Line right = Line(ballPos, agentView->getAgentPosition(NOW).getPoint(),
		    blockRadius, true);
  Line left = Line(ballPos, agentView->getAgentPosition(NOW).getPoint(),
		   blockRadius, false);
  
  double leftS, leftT, rightS, rightT;
  right.extendToLine(goalLine, rightS, rightT);
  left.extendToLine(goalLine, leftS, leftT);
  
  if(leftS <= 0 || rightS <= 0) 
    return false; // ball is between me and the goal

  if(leftT < 0 && rightT > 1)
    return false;

  Point bestPos = getOptimum(ballPos);
  if(Vector(agentView->getAgentPosition(NOW).getPoint() -
	    bestPos).getLength() < 0.3)
    return false;
    
  return true;
}


RS_END_NAMESPACE
