/* -*- Mode: C++ -*-
 * $Id: DribbleMM.cc,v 1.7 2001/06/19 14:01:43 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: DribbleMM.cc
 * Unitname: Framework/skill
 * $Revision: 1.7 $
 * Created by: Magnus Ingmarsson and Magnus Wahlstrm 2000-09-17
 * Last modified by $Author: frehe $ $Date: 2001/06/19 14:01:43 $
 *
 *
 * HISTORY
 *
 */

/**
 * @file DribbleMM.cc
 * Implementation file for the skill class DribbleMM.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.7 $
 * $Author: frehe $
 * $Date: 2001/06/19 14:01:43 $
 */


#include <vector>

#include "DribbleMM.h"
#include "BallView.h"
#include "PlayerView.h"
#include "AgentView.h"
#include "MarkerView.h"
#include "GameView.h"
#include "view_constants.h"

#include "KickCommand.h"


RS_BEGIN_NAMESPACE


// --- CONSTRUCTORS ---



DribbleMM::DribbleMM()
  : SkillWithDirection(AngleDeg(0)),
    ballView(NULL),
    parameterView(NULL),
    playerView(NULL),
    agentView(NULL),
    markerView(NULL),
    gameView(NULL)    
  
{
  std::cerr << "A DribbleMM initialized without arguments" << std::endl;
  getView(ballView, BALL_VIEW_ID);
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(playerView, PLAYER_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID, parameterView);
  getView(markerView, MARKER_VIEW_ID, parameterView);
  getView(gameView, GAME_VIEW_ID);
  kickpower = 0.45;
  d_safe=30.0; d_close=15.0; d_horror=10.0;
  for (int i=0; i<8; i++)
    dangers[i] = NULL;
}

DribbleMM::DribbleMM(AngleDeg dir)
  // This is an initializer list
  // which means that all the objects inside the
  // class can be initialized after the colon.
  // You initialize the variable myVar with
  // myVar(constructor arguments).
  // You should also initialize the base classes, but
  // only the classes that you directly inherit from.
  // If you do not initialize a variable its default
  // constructor will be called.
  : SkillWithDirection(dir),
    ballView(NULL),
    parameterView(NULL),
    playerView(NULL),
    agentView(NULL),
    markerView(NULL),
    gameView(NULL)    
{
  std::cerr << "A DribbleMM initialized" << std::endl;
  getView(ballView, BALL_VIEW_ID);
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(playerView, PLAYER_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID, parameterView);
  getView(markerView, MARKER_VIEW_ID, parameterView);
  getView(gameView, GAME_VIEW_ID);
  kickpower = 0.45;
  d_safe=30.0; d_close=15.0; d_horror=10.0;
  for (int i=0; i<8; i++)
    dangers[i] = NULL;
}


// --- DESTRUCTOR ---

DribbleMM::~DribbleMM()
{
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(GAME_VIEW_ID);
}

// --- METHODS ---

 // This might still change

void DribbleMM::setPrecision(int prec)
{
  if ((prec > 0) && (prec < 10)) {
    d_horror = 10.0 + Float(prec-5);
    d_close = 15.0 + Float((prec-5)*2);
    d_safe = 30.0 + Float((prec-5)*4);
  }
  else
    std::cerr << "setPrecision called with illegal value: " << prec << std::endl;
}

void DribbleMM::setDistances(Float vclose, Float close, Float far) {
  d_horror = vclose;
  d_close = close;
  d_safe = far;
}

Plan DribbleMM::generatePlan(AngleDeg dir)
{
  direction = dir;
  return generatePlan();
}

int DribbleMM::sector(Float dir)
{
  int sec = int(floor((dir + 22.5 - direction.getDeg())/8)) % 8;
  if (sec < 0)
    sec += 8;
  return sec;
}

VectorUCO *DribbleMM::edgeVector(double dir, PointUCO pos)
{
  // This needs to consider the enemie's goal as safe
  double x,y;
  VectorUCO *vec = new VectorUCO(AngleDeg(dir), 1.0);

  x=cos(dir); y = sin(dir);

  if (x==0.0)
    x = 111.0;
  else if (x < 0.0)
    x = (52.0 + pos.getPoint().getX())/abs(x);
  else
    x = (52.0 - pos.getPoint().getX())/x;
  
  if (y==0.0)
    y = 111.0;
  else if (y < 0.0)
    y = (34.0 + pos.getPoint().getY())/abs(y);
  else
    y = (34.0 - pos.getPoint().getY())/y;

  if (x<y && x < d_horror)
    vec->setLength(x);
  else if (y < d_horror)
    vec->setLength(y);
  else { // Far away, don't bother
    delete vec;
    return NULL;
  }
  
  return vec;
}


/** The worst threat of two vectors */
VectorUCO *DribbleMM::worst(VectorUCO *a, VectorUCO *b)
{
  if (!a)
    return b;
  else if(!b)
    return a;
  else if(a->getLength() > b->getLength())
    return b;
  else
    return a;
}

/** The best choice of two sectors */
int DribbleMM::bestSector(int a, int b)
{
  if (!dangers[a])
    return a;
  else if (!dangers[b])
    return b;
  else if (dangers[a]->getLength() > dangers[b]->getLength())
    return a;
  else
    return b;
}

/** Is the sector almost empty? */
bool DribbleMM::safeSector(int sec)
{
  return (!dangers[sec] || (dangers[sec]->getLength() > d_safe));
}

/** Internal. Iterates through the sectors until it finds a good one. 
    Iterative order: How much it has to steer off course*/
int DribbleMM::pickSector() {
  return pickSector(0);
}

int DribbleMM::pickSector(int left)
{
  int best = bestSector(left, (8-left)%8); // Best of +/-left

  if (left == 4) {
    //    cerr << "Sector bottomed out." << endl;
    return best;
  }
  else if (safeSector(best)) {
    //    cerr << "Approved #" << best << endl;
    return best; // left==4: This is the last sector. 
                 // safe: This is the first OK sector.
  }
  else {
    //    cerr << "Dangers #" << left << ": " << dangers[left]->getVector() <<
    //  " and #" << (8-left)%8 << ": " << dangers[(8-left)%8]->getVector() << endl;
    return bestSector(pickSector(left+1), best); 
    // In case we're surrounded, keep track of the least bad
  }
}

Plan DribbleMM::generatePlan()
{
  static PointUCO pos;
  static AngleDeg wanteddir;
  static VectorUCO agentvec, oppvec;
  static Vector ballvec;
  //  static KickBallToSpeedVector kickit(ballvec);
  Float kickfactor=1.0, dir;
  Plan plan;
  int sec,tmp;

  if ( !ballView->getRelativeBallDirection(NOW).isKnown() ||
       ballView->getBallDistance(NOW) > parameterView->SP_kickable_area )
    return Plan();

  pos = agentView->getAgentPosition(NOW);
  if (!pos.isKnown())
  {
    // This might not be needed with some rewriting...
    return Plan();
  }

  // Initialize the "dangers" array with the edges of the field
  for(sec = 0; sec < 8; sec++) {
    if (dangers[sec])
      delete dangers[sec];
    dangers[sec] = edgeVector(Float(sec*45) + direction.getDeg(), pos);
  }
  
  // Vectors for later use
  agentvec = agentView->getSpeedVector(NOW);


  // This code is still slightly bad
  // It divides the directions into 8 arcs, and finds the best
  // arc to go in. Confusing details further down.

  std::vector<PlayerObject> opps = playerView->getOpponents(NOW);

  if ( !opps.empty() )
    {
      std::vector<PlayerObject>::iterator p;
      // cerr << "Opponents: " << end;
      for(p = opps.begin(); p != opps.end(); ++p) 
	{
	  oppvec = p->getVector();
	  //    cerr << "Vector: " << oppvec;
	  
	  if  (oppvec.isUnknown() || oppvec.getAngle().isUnknown() || 
	       oppvec.getLength().isUnknown()) 
	    continue;
	  sec = sector(oppvec.getAngle().getDeg());
	
	  //    cerr << " (sec " << sec << ")" << endl;
	  dangers[sec] = worst(dangers[sec], &oppvec);
	  
	  if (oppvec.getLength() < d_close) { // Big threat, avoid alot
	    tmp = sector(oppvec.getAngle().getDeg() - 45.0);
	    dangers[tmp] = worst(dangers[tmp], &oppvec);
	    
	    tmp = sector(oppvec.getAngle().getDeg() + 45.0);
	    dangers[tmp] = worst(dangers[tmp], &oppvec);
	    if (oppvec.getLength() < d_horror) { // OHMYGODHESCOMINGRIGHTFORUS!
	      tmp = sector(oppvec.getAngle().getDeg() - 90.0);
	      dangers[tmp] = worst(dangers[tmp], &oppvec);
	      
	      tmp = sector(oppvec.getAngle().getDeg() + 90.0);
	      dangers[tmp] = worst(dangers[tmp], &oppvec);
	    }
	  }
	}
    }
  else

  //  cerr << "Stnenoppo." << endl;

  // Find a decent direction. It tries, in order, sectors 0, 1 or 7, 2 or 6, 3 or 5, and 4.
  // The reason for this order, is that 0 is in the wanted direction, while 4 is away from it.
  // A sector is acceptable if it is empty, or if the closest player in the sector is 
  // at least d_safe steps away. If a sector is not found acceptable, it checks the next
  // sector or pair of sectors and sees if it is acceptable.
  // If no sector is found acceptable, it goes for the least bad one.

  // It probably should have a prettier implementation than this... 
  // It probably should look at the opponents' speed and direction and do something clever.
  // But it does care about the edge of the field!
  sec = pickSector(); 
  
  //  cerr << "Decision: sector " << sec;
  
  // We know our sector. We just go for the middle of it. Well... Improve!
  dir = Float(sec*45) + direction.getDeg() - agentvec.getVector().getAngle().getDeg();
  if (dir > 180.0)
    dir -= 360.0;
  wanteddir = AngleDeg(dir);
  
  // Here we might need to do a more specific check with how the ball is 
  // located in relation to our body.
  if (wanteddir > 90) 
    wanteddir = 90;
  else if (wanteddir < -90) 
    wanteddir = -90;

  //  cerr << "Decision: #" << sec << ", " << wanteddir.getDeg() << 
  //    " ideal " << direction.getDeg() << 
  //    " us " << agentvec.getVector().getAngle().getDeg() << endl;


  //  Point(0,0)+agentvec.getVector().rotate(-wanteddir)
  //  X value of this is how much "help" we have from the agent's speed
  // if (dangers[sec] && dangers[sec]->getLength < d_safe)
  // This tests if you need to be careful
  // But it seems we're alredy using too much CPU power!?!
  ballvec.setLength(kickpower);
  ballvec.setAngle(wanteddir);
//  if (kickit.applicable(ballvec))
//    plan = kickit.generatePlan(ballvec);
//  else {
    // If the reason that KickBallToSpeedVector is unapplicable is that the
    // kick would be too hard (since the ball has too high speed in the wrong
    // direction), this kick strength is certainly wrong.
    plan.addStep(Step(new KickCommand(7.0, wanteddir)));
    //  }
  if (plan.empty())
    plan.addStep(Step(new KickCommand(7.0, wanteddir)));
  return plan;
}


FuzzyBool DribbleMM::persistent()
{
  // Since we have no clearly defined end goal we just dribble until
  // we are not called anymore.
  return FALSE;
}


FuzzyBool DribbleMM::applicable()
{
  // If ball is within kicking distance return true
  if ( ballView->getRelativeBallDirection(NOW).isKnown() &&
       ballView->getBallDistance(NOW) <= parameterView->SP_kickable_area &&
       agentView->getAgentPosition(NOW).isKnown() &&
       agentView->getSpeedVector(NOW).isKnown() )
    return TRUE;
  else
    return FALSE;

}


RS_END_NAMESPACE


