/* -*- Mode: C++ -*-
 * $Id: FaceDirection.cc,v 1.4 2001/06/19 14:01:44 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: FaceDirection.cc
 * Unitname: Framework/skill
 * $Revision: 1.4 $
 * Created by: Ola Lundqvist 2000-09-27
 * Last modified by $Author: frehe $ $Date: 2001/06/19 14:01:44 $
 *
 *
 * HISTORY
 *
 * $Log: FaceDirection.cc,v $
 * Revision 1.4  2001/06/19 14:01:44  frehe
 * Made minor changes to compile RoboSoc with gcc-3.0
 *
 * Revision 1.3  2000/10/01 13:05:54  olalu526
 * Updated FacingDirection, fixing bugs and stuff.
 *
 * Revision 1.2  2000/09/30 15:35:15  olalu526
 * Bug fixed MoveToPositionFacingDirection, now works really fine.
 *
 * Revision 1.1  2000/09/29 17:55:01  olalu526
 * Added FaceDirection, fixed some in FindBall (now works as expected), fixed segmentation fault in TurnDirection and more.
 *
 * Revision 1.3  2000/09/28 20:39:05  olalu526
 * Updated for testing, segmentation fault...
 *
 * Revision 1.2  2000/09/28 20:17:37  olalu526
 * Update FaceDirection.
 *
 * Revision 1.1  2000/09/28 20:12:27  olalu526
 * Added FaceDirection.
 *
 *
 */

/**
 * @file FaceDirection.cc
 * Implementation file for the skill class FaceDirection.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2001/06/19 14:01:44 $
 */


#include "FaceDirection.h"


RS_BEGIN_NAMESPACE


// --- CONSTRUCTORS ---


FaceDirection::FaceDirection()
  : SkillWithArguments(),
    angle(),
    bodyturn(),
    agentView(),
    parameterView()
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID, parameterView);
}


FaceDirection::FaceDirection(const AngleDeg &a)
  : SkillWithArguments(),
    angle(a),
    bodyturn(),
    agentView(),
    parameterView()
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID, parameterView);
}


// --- DESTRUCTOR ---

FaceDirection::~FaceDirection()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
}


// --- METHODS ---

void FaceDirection::setAngle(const AngleDeg &a) {
  angle = a;
}

void FaceDirection::setArguments(const std::vector<Float>& arguments)
  throw(TooFewArgumentsException)
{
  if (arguments.size() < 1) {
    RS_THROW(TooFewArgumentsException,
	     "FaceDirection got less than three arguments.");
  }
  else {
    setAngle(AngleDeg(arguments[0]));
  }
}


Plan FaceDirection::generatePlan(const AngleDeg& a) {
  setAngle(a);
  noBodyTurn();
  return generatePlan();
}

Plan FaceDirection::generatePlan(const AngleDeg& a, const AngleDeg& bt) {
  setAngle(a);
  setBodyTurn(bt);
  return generatePlan();
}

void FaceDirection::noBodyTurn() {
  bodyturn = AngleDeg(Float(0));
}

void FaceDirection::setBodyTurn(const AngleDeg& bt) {
  bodyturn = bt;
}

AngleDeg FaceDirection::getBodyTurn() {
  return bodyturn;
}

AngleDeg FaceDirection::getDiffAngle() const {
    AngleDeg agd = (agentView->getFaceDirection(NOW)).getDeg();
    /*    cout << "FaceDire: " << (angle-agd-bodyturn).getDeg()
	 << " angle: " << angle.getDeg()
	 << " agfac: " << agd.getDeg()
	 << " bodyt: " << bodyturn.getDeg()
	 << endl << flush;*/
    return angle - agd - bodyturn;
}

Plan FaceDirection::generatePlan()
{
  // Return an empty plan.
  static Command* cmd;
  static Plan plan;
  plan.makeEmpty();
  
  if ((agentView->getFaceDirection(NOW)).isKnown() ) {    
    AngleDeg ang = getDiffAngle();
    noBodyTurn();
    cmd = new TurnNeckCommand(ang.getDeg());
    plan.addStep(Step(cmd));
    delete cmd;
    cmd = NULL;
    return plan;
  }
  std::cout << "FaceDirection: This should NEVER happen!" << std::flush;
  noBodyTurn();
  return Plan();
}


FuzzyBool FaceDirection::persistent()
{
  // Never persistent.
  return FALSE;
}


FuzzyBool FaceDirection::applicable(const AngleDeg& a) {
  setAngle(a);
  return applicable();
}

FuzzyBool FaceDirection::applicable(const AngleDeg& a, const AngleDeg& bt) {
  setBodyTurn(bt);
  return applicable(a);
}


FuzzyBool FaceDirection::applicable()
{
  // Almost always applicable.
  if ((agentView->getFaceDirection(NOW)).isKnown()) {
    if (abs(getDiffAngle()) > 5) {
      return TRUE;
    }
  }
  return FALSE;
}


RS_END_NAMESPACE


// Remove from here
/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "FaceDirection.icc"
#endif
// To here if you do not have any inline definitions!
