/* -*- Mode: C++ -*-
 * $Id: InterceptBallStamina.h,v 1.7 2001/01/11 12:28:01 olalu526 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: InterceptBallStamina.h
 * Unitname: Framework/skill
 * $Revision: 1.7 $
 * Created by: Mattias Karlsson and Per Olof Pettersson 2000/09/19
 * Last modified by $Author: olalu526 $ $Date: 2001/01/11 12:28:01 $
 *
 *
 * HISTORY
 *
 * $Log: InterceptBallStamina.h,v $
 * Revision 1.7  2001/01/11 12:28:01  olalu526
 * Updated with some improvements from the first competition.
 *
 * Revision 1.6  2000/11/08 20:58:19  mikbr212
 * Changed persistent so that this skill won't be persistent
 * if the agent is losing the ball (ball moving further away)
 * or when no progress (in chasing the ball) is made.
 * These changes can be disabled by setting neverGiveUp to true (default).
 *
 * Added setNeverGiveUp
 * Added some attributes
 *
 * Changed Persistent
 * Changed generatePlan
 * Changed setArguments
 * Changed constructor
 *
 * Revision 1.5  2000/10/11 07:12:47  matka134
 * Fixed some typos in the documentation
 *
 * Revision 1.4  2000/10/10 15:12:33  perpe427
 * Removed alot of functions that are not used any more or have been transfered to FutureBallView and MoveToPositionFacingDirection. Documented code.
 *
 * Revision 1.3  2000/10/09 22:29:46  matka134
 * Major rewrite of the code in InterceptBallStamina. Moste of the code has been moved from this class to some MoveToPosition classes. It now uses MoveToPositinFacingBall to do most of the calulations.
 * It will hopefully work better now that we take the direction of the neck into account and such stuff. I havn't been testing it much, but it seem to work quite ok.
 *
 * Revision 1.2  2000/09/28 13:32:42  olalu526
 * A lot of changes.
 *
 * Revision 1.1  2000/09/27 12:17:02  perpe427
 * Renamed InterceptBallPoM as InterceptBallStamina
 *
 */

/**
 * @file InterceptBallStamina.h
 * Include file for the skill class InterceptBallStamina.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.7 $
 * $Author: olalu526 $
 * $Date: 2001/01/11 12:28:01 $
 */


#ifndef INTERCEPTBALLSTAMINA_H
#define INTERCEPTBALLSTAMINA_H


#include "SkillWithArguments.h"
#include "FuzzyBool.h"
#include "MoveToPositionFacingBall.h"


RS_BEGIN_NAMESPACE

class BallView;
class ParameterView;
class AgentView;
class FutureBallView;

/**
 * @class InterceptBallStamina InterceptBallStamina.h
 * Intercepts a moving ball not going below a certain level of stamina.  
 * This is accomplished by calculating the interception point closest in
 * time using the FutureBallView and the skill MoveToPositionFacingBall.
 * The currently used MoveToPosition-skill will not avoid obstacles in
 * between the agent and the ball, but this will change as soon as we
 * have checked that MoveToPositionAvoidingObstacles works well.  For
 * now the skill assumes that the ball is freely moving (speed
 * decaying) and ignores the possibility of the ball being kicked by
 * someone before interception but reacts immediatly if this happens.
 * In the future we plan to improve its ability to intercept a ball
 * controlled by an opponent.  This will probably be done, at least in
 * part, by improving the FutureBallView.
 *
 * @pre The preconditions for the skill are BallPositionKnown, BallNotKickable
 * @post The intended postconditions for the skill are BallKickable
 *
 * @todo 
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.7 $
 * $Author: olalu526 $
 * $Date: 2001/01/11 12:28:01 $ */
class InterceptBallStamina : public SkillWithArguments {
public:
  /** Default constructor. 
      @param minStamina The minimum level of stamina when reaching
      ball.  If minStamina is not specified the default
      value for the MoveToPositionFacingBall-skill will be used.
  */
  InterceptBallStamina(Float minStamina = -1,
		       bool _neverGiveUp = true);

  /** Destructor. */
  virtual ~InterceptBallStamina();

  /** This method is used to generate a plan for how the goal of
    the skill should be obtained by the agent. 
    @return Plan for intercepting ball.
  */
  virtual Plan generatePlan();

  /** This method tells you if the skill wants to be called again.
      @return True if the skill wants to be called again.
  */
  virtual FuzzyBool persistent();

  /** Call this function to see if the skill is applicable.
      @return True if the ball is not kickable and the position of
      the ball is known.
   */
  virtual FuzzyBool applicable();


  /** Set the amount of stamina that must be left when the skill is
      done.
      @param minStamina The minimum level of stamina when reaching
      ball. 
  */
  void setMinStamina(Float minStamina);
  /** Set the amount of stamina that must be left when the skill is
      done.  
      @param minStamina The minimum level of stamina when
      reaching ball.  
  */
  void setArguments(const std::vector<Float>& arguments) throw(TooFewArgumentsException);

  virtual void setEmergency(const FuzzyBool&);
  // *** ADDED BY MICKE ***
  void setNeverGiveUp(bool flag = true,
		      float _giveUpDist = 3,
		      int _giveUpTime = 10);
  // *** END ***

protected:
  /** A view that store information about the ball. */
  BallView* ballView;
  /** The parameter view. */
  ParameterView* parameterView;
  /** A view that store informatin about the agent. */
  AgentView* agentView;
  /** A view for estimate where the ball will be in the fututure. */
  FutureBallView* futureBallView;
  /** A pointer to a MoveToPosition object.
      (MoveToPositionFacingBall even, can change in the future...) */
  MoveToPositionFacingBall* moveToPos;


  // *** ADDED BY MICKE ***
  float closestBallDistance;
  Time closestTime;
  Time timeStamp;
  bool neverGiveUp;
  float giveUpDistance;
  int giveUpTime;
  // *** END ***
  
private:
  /** Hide copy-constructor. */
  InterceptBallStamina(const InterceptBallStamina&);
  /** Hide assignment operator. */
  InterceptBallStamina& operator= (const InterceptBallStamina&);
  FuzzyBool em;
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "InterceptBallStamina.icc"
#endif

#endif // INTERCEPTBALLSTAMINA_H
