#include "KickBallInDirection.h"
#include "BallView.h"
#include "PlayerView.h"
#include "AgentView.h"
#include "KickView.h"
#include "view_constants.h"
#include "KickCommand.h"

RS_BEGIN_NAMESPACE


KickBallInDirection::KickBallInDirection(Float eps, bool UseMaxPower)
  : SkillWithArguments(),
    ballView(NULL)
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(ballView, BALL_VIEW_ID);
  getView(playerView, PLAYER_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID, parameterView);
  getView(kickView, KICK_VIEW_ID);

  ballSpeedVector = ballView->getBallAbsoluteSpeedVector(NOW); // not sure: relative or absolute ??

  Epsilon = eps;
  bUseMaxPower = UseMaxPower;
} 


KickBallInDirection::~KickBallInDirection()
{
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(KICK_VIEW_ID);
}

void KickBallInDirection::setArguments(const std::vector<Float> & args)
  throw(RS::TooFewArgumentsException)
{
  power = args[0];
  angle = args[1];

  // angle adaptation if we know agent's body direction, since the shoot will be done according to the agent
  if(agentView->getBodyDirection(NOW).isKnown())
    {
      angle -= agentView->getBodyDirection(NOW).getAngle().getDeg();
    }
  /*else RS_LOG(LA_SCORE, "AGENT BODY DIRECTION UNKNOWN!!!!!");*/
}


VectorUnknown KickBallInDirection::computeKickSpeedVector(Float desiredPower)
{
  // 1- builds up a vector for the desired power and angle values
  VectorUnknown DesiredSpeedVector = VectorUnknown(angle, desiredPower * kickView->kickPowerRate());
  
  // 2- let's find the real kick vector needed for kicking according to the desired values
  VectorUnknown KickSpeedVector = DesiredSpeedVector - ballSpeedVector;

  return KickSpeedVector;
}


VectorUnknown KickBallInDirection::findBestKickSpeedVector()
{
  Float tmpPower, currentPower = power, maxPower = kickView->howHardAgentCanKick();

  //std::cerr << "mapower: " << maxPower;

  VectorUnknown KickSpeedVector;

  bool found = false;
  while(!found)
    {
      // update vars with new values and loop again
      KickSpeedVector = computeKickSpeedVector(currentPower);
      
      if(KickSpeedVector.isUnknown())
	break;

      tmpPower = KickSpeedVector.getLength().getValue();

      //std::cerr << "tmpPower: " << tmpPower << "maxpower: " << maxPower << std::endl;

      if(currentPower <= Epsilon)
	{
	  //RS_ERROR("humm, strange, power == zero");
	  break;
	}


      if( abs(tmpPower - maxPower) <= Epsilon ) // ok, power is correct
	{
	  // we got it!
	  // the actual KickSpeedVector is the one we want to shoot with
	  found = true;
	}
      else
	{
	  if(tmpPower > maxPower)
	    {
	      // power is too high, so reduce it by a half
	      currentPower *= 0.5;
	    }
	  else
	    {
	      // power is too small, so increase it by a half
	      currentPower *= 1.5;
	    }
	}
    }

  return KickSpeedVector;
}


Plan KickBallInDirection::generatePlan()
{
  Command* cmd;
  Plan plan;

  VectorUnknown KickSpeedVector = computeKickSpeedVector(power);
  Float tmpPower = 0, maxPower = kickView->howHardAgentCanKick();

  if(KickSpeedVector.isKnown())
    {
      tmpPower = KickSpeedVector.getLength().getValue();
  
      // 1- check how hard we should kick

      if(tmpPower < maxPower)
	{
	  // we can kick stronger
	  
	  // we are also ask to kick as strong as possible
	  if(bUseMaxPower)
	    {
	      // so adapt kick vector to use max power 
	      KickSpeedVector = findBestKickSpeedVector();
	    }
	}

      if(tmpPower > maxPower)
	{
	  // we cannot kick that strong
	  
	  // so adapt kick vector to keep direction (kick length might be reduced)
	  KickSpeedVector = findBestKickSpeedVector();
	}
    }

      
  // 2- generate the kick command
  if(KickSpeedVector.isKnown())
    {
      cmd = new KickCommand(KickSpeedVector.getLength().getValue() / kickView->kickPowerRate(), KickSpeedVector.getAngle().getDeg());

    }
  else // in the case we are dealing with unknown values
    {
      cmd = new KickCommand(power, angle);
    }

  plan = Plan(Step(cmd));
  delete cmd;
  cmd = NULL;

  return plan;
}


FuzzyBool KickBallInDirection::persistent()
{
  return FALSE;
}


FuzzyBool KickBallInDirection::applicable()
{
  if(ballView->getBallDistance(NOW).isKnown())
    {
      // if the ball is close enough
      return (ballView->getBallDistance(NOW).getValue().getValue().getValue() <= parameterView->SP_kickable_area);
    }
  else
    {
      return FALSE;
    }
}


RS_END_NAMESPACE

