/* -*- Mode: C++ -*-
 * $Id: KickBallToSpeedVectorMulti.cc,v 1.7 2000/11/07 18:10:44 mikbr212 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: KickBallToSpeedVectorMulti.cc
 * Unitname: Framework/skill
 * $Revision: 1.7 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: mikbr212 $ $Date: 2000/11/07 18:10:44 $
 *
 *
 * HISTORY
 *
 * $Log: KickBallToSpeedVectorMulti.cc,v $
 * Revision 1.7  2000/11/07 18:10:44  mikbr212
 * changed persistent
 *
 * Revision 1.6  2000/11/01 16:38:58  mikbr212
 * fixed constructor
 *
 * Revision 1.5  2000/11/01 16:01:39  mikbr212
 * changed constructor
 *
 * Revision 1.4  2000/11/01 12:42:59  mikbr212
 * Added possibility to do a fast kick
 *
 * Revision 1.3  2000/10/30 18:32:45  mikbr212
 * added comments
 *
 */

/**
 * @file KickBallToSpeedVectorMulti.cc
 * Implementation file for the skill class KickBallToSpeedVectorMulti.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.7 $
 * $Author: mikbr212 $
 * $Date: 2000/11/07 18:10:44 $
 */


#include "KickBallToSpeedVectorMulti.h"


RS_BEGIN_NAMESPACE

/**
 * Constructor.
 * When fast kick is forced (fast = true, adaptive = false), this skill won't do
 * multiple kicks.
 * @param fast True if a fast kick should be applied, default = true.
 * @param adaptive True if this skill should check if a fast kick should be applied,
 * default = false. This argument overries the fast argument if true.
 */

KickBallToSpeedVectorMulti::KickBallToSpeedVectorMulti(const Vector& vec, bool fast,
						       bool adaptive)
  : SkillWithVector(vec),
    phase(0),
    timeStamp(0,0),
    doFast(fast),
    doAdaptive(adaptive),
    emergency(3.0), // 3.0 is temporary, should be calculated!
    canKick(vec),
    parameterView(NULL),
    ballView(NULL),
    agentView(NULL),
    kickBallView(NULL),
    kickSkill(Vector(0,0)),
    turnSkill(Vector(0,0))
{
  init();
}

void KickBallToSpeedVectorMulti::init() {
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(ballView, BALL_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID, parameterView);
  CommandView* commandView = NULL;
  getView(commandView, COMMAND_VIEW_ID);
  getView(kickBallView, KICKBALL_VIEW_ID, parameterView,
	  ballView, agentView, commandView);

  bufMin = parameterView->SP_player_size + parameterView->SP_ball_size; 
  kickMax = parameterView->SP_kickable_area;
  
  bufR = bufMin;
  kickR = kickMax;

  setPlayerMargin();
  setKickableMargin();
}

/**
 * Destructor.
 */
KickBallToSpeedVectorMulti::~KickBallToSpeedVectorMulti()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(COMMAND_VIEW_ID);
  viewManager->releaseView(KICKBALL_VIEW_ID);
}

/**
 * Recalculates the 'back kick' ball positions (there are two: left and right).
 * Called after setPlayerMargin and setKickableMargin
 */
void KickBallToSpeedVectorMulti::recalcDesPos() {
  // calculate back-kick position
  const double avgKick = 0.75;

  double deltaX = 100 / 2 * parameterView->SP_kick_power_rate * avgKick;
  double deltaY = bufR;
  if(deltaX*deltaX + deltaY*deltaY > kickR*kickR)
    deltaX = sqrt(kickR*kickR - deltaY*deltaY);

  desBallPosR = Vector(Point(-deltaX,deltaY));
  desBallPosL = Vector(Point(-deltaX,-deltaY));

}

/**
 * Sets the wanted speedvector.
 */
void KickBallToSpeedVectorMulti::setVector(const Vector& vec) {
  vector = vec;
}

/**
 * Retrieves the wanted speedvector.
 */
Vector KickBallToSpeedVectorMulti::getVector() const {
  return vector;
}

/**
 * Sets the margin in percent of kickable_margin between the player and the ball.
 */
bool KickBallToSpeedVectorMulti::setPlayerMargin(double x) {
  if(kickR < (bufMin + x / 100.0 * parameterView->SP_kickable_margin) || 
     x < 0.0 || x >= 100.0)
    return FALSE;
  bufR = bufMin + x / 100.0 * parameterView->SP_kickable_margin;
  recalcDesPos();
  return TRUE;
}

/**
 * Sets the margin in percent of kickable_margin between the kickable_area-radius and the 
 * allowed ball_area.
 */
bool KickBallToSpeedVectorMulti::setKickableMargin(double x) {
  if((kickMax - x / 100.0 * parameterView->SP_kickable_margin) <= bufR || 
     x < 0.0 || x >= 100.0)
    return FALSE;
  kickR = kickMax - x / 100.0 * parameterView->SP_kickable_margin;
  recalcDesPos();
  return TRUE;
}

/**
 * Set fast kick parameters.
 * A fast kick is a simple kick when possible, and othervise
 * a turnball followed by a simple kick. Note: the fast kick can't
 * get the ball up to the same speed as the full double kick.
 * @param fast If true a fast kick will be applied.
 * @param adaptive If true, the skill will choose when to do a fast or
 * double kick. If an opponent is able to intercept the ball before it
 * has been kicked away using a double kick, a simple kick will be choosen.
 * Note: The simple kick will always be used if that is enough.
 */
void KickBallToSpeedVectorMulti::setFastKick(bool fast, bool adaptive) {
  doFast = fast;
  doAdaptive = adaptive;
}

/**
 * Generates a plan.
 * The plan will consist of only one step. To get more steps
 * simply call this skill next cycle. 
 * This method runs setVector(vec) and then generatePlan().
 */
Plan KickBallToSpeedVectorMulti::generatePlan(const Vector& vec) {
  setVector(vec);
  return generatePlan();
}

/**
 * Generates a plan.
 * The plan will consist of only one step. To get more steps
 * simply call this skill next cycle. 
 * If the desired ball speed can be reached with a simple
 * kick, that will be returned. Otherwise the ball will
 * be turned to a 'back kick' position. This may take
 * zero (if the ball is already there) or more steps
 * depending on the parameters set to the turnball skill and
 * how far away the ball is from its final position. 
 * After the turnball two kicks will be returned.
 */
Plan KickBallToSpeedVectorMulti::generatePlan() {
  if(!ballView->isBallKickable(NOW) || 
     ballView->getRelativeBallVector(NOW).isUnknown())
    return Plan(); 

  if(timeStamp + 1 != Time::currentTime) { // new kick?
    phase = turnPhase;
    if(doAdaptive) // only change to fast or multi kick when new kick
      doFast = emergency();
  }

  // do a simple kick if that is enough 
  if(kickBallView->canShootToSpeedVector(vector) ||
     (doFast && canKick(vector)))
    phase = simpleKickPhase; 
  
  //  cout << "phase = " << phase;

  Vector desBallPos(0,0); 
  if(phase == turnPhase && !doFast) {
    if(vector.getAngle().getDeg() > 0) 
      desBallPos = desBallPosL;
    else 
      desBallPos = desBallPosR;
    
    desBallPos.rotate(vector.getAngle());
    
    if((ballView->getRelativeBallVector(NOW).getVector() - desBallPos).getLength()
       < 0.2)
      phase = backKickPhase; // skip to next phase
  } else if(phase == turnPhase && doFast) {
    // set desBallPos between player and kickable_area (25% from player)
    desBallPos = Vector(vector.getAngle(),parameterView->SP_kickable_area -
			parameterView->SP_kickable_margin*0.75);
  }

  if(phase == turnPhase) { // still turnPhase?
    //    cout << "Multi: turn\n";
    timeStamp = Time::currentTime;
    return turnSkill.generatePlan(desBallPos);
  } 

  if(phase == backKickPhase) {
    Line kickLine = Line(Point(ballView->getRelativeBallVector(NOW).getVector()),
			 vector);

    kickLine.extendToCircle(Point(0,0),kickR);

    //    cout << "Multi: kick from back\n";
    phase = simpleKickPhase; // set next phase
    timeStamp = Time::currentTime;
    return kickSkill.generatePlan(kickLine.getVector());
  }

  //  cout << "Multi: simple kick\n";
  /* Note: no timeStamp = no furter steps */
  return kickSkill.generatePlan(vector);
}

/**
 * Checks if this skill should be called again (this cycle).
 */
FuzzyBool KickBallToSpeedVectorMulti::persistent()
{
  return (timeStamp + 1 == Time::currentTime);
}

/**
 * Checks if preconds are true.
 */
FuzzyBool KickBallToSpeedVectorMulti::applicable(const Vector& vec) {
  setVector(vec);
  return applicable();
}

/**
 * Checks if preconds are true.
 */
FuzzyBool KickBallToSpeedVectorMulti::applicable() {
  return ballView->isBallKickable(NOW);
}



RS_END_NAMESPACE


