/* -*- Mode: C++ -*-
 * $Id: MoveToPositionAvoidingObstacles.cc,v 1.2 2001/01/09 15:37:33 olalu526 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Jan Axelson and
 * Christer Rosenquist.
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: MoveToPositionAvoidingObstacles.cc
 * Unitname: Framework/skill
 * $Revision: 1.2 $
 * Created by: Jan Axelson and Christer Rosenquist 2000-09-24
 * Last modified by $Author: olalu526 $ $Date: 2001/01/09 15:37:33 $
 *
 *
 * HISTORY
 *
 */

/**
 * @file MoveToPositionAvoidingObstacles.cc
 *  
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.2 $
 * $Author: olalu526 $
 * $Date: 2001/01/09 15:37:33 $
 */


#include "MoveToPositionAvoidingObstacles.h"
#include "view_constants.h"
#include "Command.h"
#include "TurnCommand.h"
#include "DashCommand.h"
#include "Step.h"
#include "PlayerObject.h"
#include "ObjectCollection.h"

RS_BEGIN_NAMESPACE

const float MARGINAL = 0.2;
const float AT_POSTITION = 0.2;

// --- CONSTRUCTORS ---

MoveToPositionAvoidingObstacles::MoveToPositionAvoidingObstacles()
  : MoveToPosition(),
    playerView(NULL)
{
  getView(playerView, PLAYER_VIEW_ID);
}

MoveToPositionAvoidingObstacles::MoveToPositionAvoidingObstacles(const Point &position)
  : MoveToPosition(position),
    playerView(NULL)
{
  getView(playerView, PLAYER_VIEW_ID);
}

MoveToPositionAvoidingObstacles::MoveToPositionAvoidingObstacles(const int dash)
  : MoveToPosition(dash),
    playerView(NULL)
{
  getView(playerView, PLAYER_VIEW_ID);
}

MoveToPositionAvoidingObstacles::MoveToPositionAvoidingObstacles(const Point &position, const int dash)
  : MoveToPosition(position, dash),
    playerView(NULL)
{
  getView(playerView, PLAYER_VIEW_ID);
}

// --- DESTRUCTOR ---

MoveToPositionAvoidingObstacles::~MoveToPositionAvoidingObstacles()
{
  viewManager->releaseView(PLAYER_VIEW_ID);
}


// --- METHODS ---

Plan MoveToPositionAvoidingObstacles::generatePlan()
{
  static Plan plan;
  static PlayerObject* closestObstacle = NULL;
  static float distToPos;

 
  // Get hold of the closest obstacle.
  closestObstacle = playerView->getClosestPlayer(NOW);

  // Slow down if approaching the position.
  if( (agentView->getAgentPosition(NOW)).isKnown() ){
    distToPos = getPosition().getDistanceTo((static_cast<PointUnknown>
					     (agentView->getAgentPosition(NOW))).getPoint());    

    if( distToPos < 1 ) 
      setDash(25);
  }

  // If obstacle between Agent and obstacle.
  if( closestObstacle != NULL  &&
      closestObstacle->getPosition().isKnown() &&
      (inTheCorridor(closestObstacle)) ) 
    plan = goToXY(createPoint(closestObstacle));// the position to go.
  else
    plan =  goToXY(getPosition());// Move straight ahead to position.
  
  return plan;
}

// Set new position and return a 'Plan' to obtain that goal. 
Plan MoveToPositionAvoidingObstacles::generatePlan(const Point& position){
  setPosition(position);
  return generatePlan();
}

// Set new dash and return a 'Plan' to obtain that goal.
Plan MoveToPositionAvoidingObstacles::generatePlan(const int dash){
  
  setDash(dash);
  return generatePlan();
}

// Set new position and new dash. Return a 'Plan'.
Plan MoveToPositionAvoidingObstacles::generatePlan(const Point &position, const int dash){
  setPosition(position);
  setDash(dash);
  return generatePlan();
}

// Returns a 'Plan' to get closer to the toPosition.
Plan MoveToPositionAvoidingObstacles::goToXY(Point toPosition){
  static Command* cmd = NULL;
  static Plan plan;
  static PointUnknown agentPosition;
  static AngleDegUnknown toDirection;
  static AngleDegUnknown bodyDirection;
  static AngleDeg angle;
  static const AngleDeg stuckAngle(90);
  static bool lastActionTurn = false;

  if( !plan.empty() )
    plan.makeEmpty();

  if(applicable().isTrue() &&
     agentView->getAgentPosition(NOW).isKnown() &&
     agentView->getBodyDirection(NOW).isKnown() ){
    agentPosition = agentView->getAgentPosition(NOW);
    toDirection = agentPosition.getDirectionTo(toPosition);
    bodyDirection = agentView->getBodyDirection(NOW);
    
    // Never do more than one TurnCommand in row.
    angle = toDirection.getAngle() - bodyDirection.getAngle();
    if( !lastActionTurn &&
	    bodyDirection.isKnown() &&
	    abs(angle) > 5 ){
      cmd = new TurnCommand(angle);
      lastActionTurn = true;
    }
    else{
      cmd = new DashCommand(getDash());
      lastActionTurn = false;
    }
  }
  else if(applicable().isFalse())
    cmd = new DashCommand(0);

  if(cmd == NULL) cmd = new DashCommand(13); // Can't stand still.
  plan = Plan(Step(cmd));
  delete cmd;
  cmd = NULL;
  return plan;
}

// The created point is located 2 times the radius of the agent + MARGINAL beside
// the obstacle and at the same distance from the agent that the obstacle is. 
// Wich side of the obstacle the created point is situated at, is determined by
// the closest way to the wanna go to position.
Point MoveToPositionAvoidingObstacles::createPoint(PlayerObject* obstacle){
  static AgentObject agent;
  static float alfa, beta; 
  static PointUnknown  obstaclePosition;

  agent = agentView->getAgentObject(NOW);
  obstaclePosition = obstacle->getPosition();

  // alfa is the angle between the agent direction to the position and
  // the agent direction to the obstacle.
  alfa = (agent.getRelativeDirectionTo(getPosition()) -
    agent.getRelativeDirectionTo(obstaclePosition)).getDeg();

  // The vector from agent to obstacle.
  VectorUnknown vec = (agent.getPosition()).getVectorTo(obstacle->getPosition());
 
  beta = (2*0.3 + MARGINAL) * 360 / (2 * PI * (vec.getLength()).getValue());

 // This is done to select the closest way to the position or the subgoal.
  beta = toLeft(obstacle) ? beta : -beta;

  // The vector to the subgoal.
  vec = vec.rotate(beta);
  
  // The subgoal point.
  PointUnknown toPoint(agent.getPosition(), vec); 
  Point p((toPoint.getX()).getValue(), (toPoint.getY()).getValue());
  
  return p;
}

// Determine if the obstacle is to left of a straight line between
// the agent and the goal position.
bool MoveToPositionAvoidingObstacles::toLeft(PlayerObject* obstacle){
  static AgentObject agent;
  bool toLeft;
  
  agent = agentView->getAgentObject(NOW);
  toLeft = ((agent.getRelativeDirectionTo(getPosition())).getDeg() -
	  (agent.getRelativeDirectionTo(obstacle->getPosition())).getDeg()) > 0;
  
  return toLeft;
}

// Determinates if the obstacle is in a 'corridor'.
// The corridor is 2 * (the agent radie) wide and the distance between the
//  agent and either the distance to the goal position long,
//  or at max 3 meters long.
//  The corridor is situated between the agent and the goal position.
bool MoveToPositionAvoidingObstacles::inTheCorridor(PlayerObject* obstacle){
  static AgentObject agent;
  static bool inCorridor;
  
  agent = agentView->getAgentObject(NOW);

  if((agent.getDistanceTo(getPosition()) - obstacle->getDistanceTo(getPosition()) > 0) &&
     (agent.getDistanceTo(getPosition()) - obstacle->getDistanceTo(getPosition()) < 3))

    inCorridor = (obstacle->getDistanceTo(agentView->getAgentPosition(NOW)) * 
		  sin(abs(agent.getRelativeDirectionTo(getPosition()) - 
	agent.getRelativeDirectionTo(obstacle->getPosition())))) < (2*0.3);
  else
    inCorridor = false;

  return inCorridor;
}

RS_END_NAMESPACE



