/**
 * @file Pass.cc
 * Implementation of the Pass skill.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.8 $
 * $Author: frehe $
 * $Date: 2001/06/19 14:01:45 $
 */

#include "Pass.h"
#include "KickCommand.h"
#include "KickBallToSpeedVector.h"

RS_BEGIN_NAMESPACE

// --- CONSTRUCTORS ---

/**
 * Constructor.
 */

Pass::Pass()
  : SkillWithArguments(),
    parameterView(NULL), //,
    //    markerView(NULL),
    agentView(NULL),
    ballView(NULL),
    playerView(NULL)
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(playerView, PLAYER_VIEW_ID);
  //  getView(markerView, MARKER_VIEW_ID, parameterView);
  getView(agentView, AGENT_VIEW_ID, parameterView);
  getView(ballView, BALL_VIEW_ID);
  getView(territoryControlView, TERRITORY_CONTROL_VIEW_ID,
	  playerView, agentView, 64, 64, 64, NULL, NULL, NULL);

  victims.resize(0);
}

// --- DESTRUCTOR ---

Pass::~Pass()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  //  viewManager->releaseView(MARKER_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
}


// --- METHODS ---


/**
 * Sets the arguments to the skill.
 * The arguments is a vector containing possible pass victims.
 * If no arguments are supplied the pass will be sent to the
 * best player according to the TerritoryControlView.
 */

void Pass::setArguments(const std::vector<Float>& arguments)
  throw(RS::TooFewArgumentsException)
{
  victims.resize(arguments.size());
  for(unsigned int i = 0; i < arguments.size(); i++) {
    victims[i] = (int) arguments[i];
  }
}

PlayerObject Pass::findClosestTeammate() {
  //  if(tm.size() == 0)
    // ARG! PANIC
  
  std::vector<PlayerObject> tm = playerView->getTeammates(NOW);
  unsigned int i;
  // Find the closest player.

  // Double looping now to check if the recipiant is
  // a valid passvictim.

  float nearest = 1000000, curDist;
  int closeNum = 0, uNum;

  for(i = 0; i < tm.size(); i++) {
    // Check if we can find the distance to
    // the current player beeing considerd
    // in a pass.

    if(tm[i].getDistance().isKnown()) {
      curDist = (float) tm[i].getDistance().getValue().getValue().getValue();

      // Check if we have some passvictims defined.
      // If we don't just pass the nearest player.

      if(curDist < nearest) {

	if(victims.size() > 0) {
	  if(tm[i].getUniformNumber().isKnown()) {
	    uNum = tm[i].getUniformNumber().getValue();

	    for(unsigned int j = 0; j < victims.size(); j++) {
	      if(uNum == victims[j]) {
		closeNum = i;
		nearest = curDist;
	      }
	    }
	  }
	} else {
	  // No passvictims specified.
	  // Pass the nearest player.
	  closeNum = i;
	  nearest = curDist;
	}
      }
    }
  }

  //  std::cout << "Find closest.end res: "<< closeNum << std::endl;

  return tm[closeNum];
  
}

/**
 * Same as
 * setArguments(foo);
 * generatePlan();
 */

Plan Pass::generatePlan(const std::vector<Float>& arguments) {
  setArguments(arguments);
  return generatePlan();
}

/**
 * Genrates a plan for achieving a pass.
 * Chooses the closest of the passvictims if the argument is
 * supplied, otherwise pass the best player according to 
 * TerritoryControlView
 */

Plan Pass::generatePlan()
{
  // Get the possible pass victims.
  std::vector<PlayerObject> tm = playerView->getTeammates(NOW);

  // Debug printing.
  //  std::cout << "--------------------------------------" << std::endl;
  //  std::cout << "Pass.generatePlan()" << std::endl;
  //  std::cout << "Num players close: " << tm.size() << std::endl;

  Vector passVector = calculatePassVector();
  //  std::cout << "Desired pass speed: " << passVector.getLength() << std::endl;

  std::cout << "Pass Vector: " << passVector.getLength() << " v "
       << passVector.getAngle().getDeg() << std::endl;

  // Create the plan.
  KickBallToSpeedVector kbt(passVector);
  if(kbt.applicable())
    return kbt.generatePlan(); 
  
  return Plan();
}

std::vector<PlayerObject> Pass::seenVictims() {
    std::vector<PlayerObject> tm = playerView->getTeammates(NOW); 
    std::vector<PlayerObject> seen;
    seen.resize(0);
    
    // No teammates or victims.
    if(tm.size() == 0 || victims.size() == 0)
      return seen;

    for(unsigned int i = 0; i < tm.size(); i++) {

      if(tm[i].getUniformNumber().isKnown()) {
	int curNum = tm[i].getUniformNumber().getValue();
	for(unsigned int j = 0; j < victims.size(); j++) {
	  if(curNum == victims[j]) {
	    seen.insert(seen.begin(), tm[i]);
	    break;
	  }
	}
      }
    }

    //    std::cout << "Seen victims: " << seen.size() << " seen players: " << tm.size() << std::endl;
    return seen;
}

Vector Pass::calculatePassVector() {
  // Calculates the speedvector to a desired pass.
  // There are actually two modes.
  // 1) Unspecified mode. Passes to some player, according to
  // the weightmap in TerrContView
  // 2) Specified mode. Tries to pass to one of the players
  // with the uniform number specified in the argument vector.
  
  Vector passVector(0,0);

  if(victims.size() == 0) {
    // Unspecified pass.

    // Weights. 
    // 1) Fristd::endly wieght. 
    // 2) Unfristd::endly weight
    // 3) Advance, positive towords enemy goal.
    // 4) Precision, longer passes are less important.
    passVector = territoryControlView->getMax(1.0, -1.0, 0.2, 0.8);
  } else {
    // Specified pass.
    
    //    std::vector<PlayerObject> tm = playerView->getTeammates(NOW); 
    std::vector<PlayerObject> seen = seenVictims();
    
    // If no players are seen we can't pass.
    if(seen.size() == 0)
      return Vector(0,0);

    // Search for the closest one.
    PlayerObject closest;
    Float64 dist = 10000;
    for(unsigned int i = 0; i < seen.size(); i++) {
      PlayerObject curPl = seen[i];
      if(curPl.getDistance().isKnown()) {
	if(curPl.getDistance().getValue() < dist) {
	  closest = curPl;
	  dist = curPl.getDistance().getValue().getValue().getValue();
	}
      }
    }

    // If we can't tell the distance to any players.
    if (dist == 10000)
      return Vector(0,0);

    // Calculate the actual passvector.
    if(closest.getVector().isKnown()) {
      passVector += closest.getVector().getVector();
    } else {
      if(closest.getDirection().isKnown())
	// Assumption here, if the distance to the
	// recieving player is unkonwn he's at 20 meters
	passVector = Vector(closest.getDirection().getDeg(), 20);
    }

    // Try to get the pass to a spot in front of the recieving player.
    // the bodydirection of the player always seem to be unkown thou!
    // So it is decommented to avoid inappropriate behaviour if
    // We happen to actually recieve it.

    /*
      if(po.getBodyDirection().isKnown()) {
      Vector bodyVector(po.getBodyDirection().getDeg(), 2); 
      passVector += bodyVector;
      } */
  }

  // KickBall is relative to the player.
  if(agentView->getBodyDirection(NOW).isKnown()) {
    passVector.setAngle(passVector.getAngle().getDeg()
			- agentView->getBodyDirection(NOW).getDeg());
  }
  
  // Debug printing
  /*std::cout << "Pass Vector: " << passVector.getLength() << " v "
       << passVector.getAngle().getDeg() << std::endl;
  */
  // Normalize the speedvector.
  Vector speedVector = passVector;
  speedVector.setLength(1);

  // This is supposed to create a kick with a limitid velocity
  // at the destination.
  // The speed at the destination should be 0.3 m/s
  speedVector =
    speedVector * 
    (0.3 + passVector.getLength() * (1 - parameterView->SP_ball_decay));

  return speedVector;
}


/**
 * Persistent.
 * FALSE
 */
FuzzyBool Pass::persistent()
{
  // Never persistent.
  return FALSE;
}

/**
 * Same as:
 * setArguments(foo);
 * applicable();
 */

FuzzyBool Pass::applicable(const std::vector<Float>& arguments) {
  setArguments(arguments);
  return applicable();
}

/**
 * Returns the applicable value
 */

FuzzyBool Pass::applicable()
{


  // Check if we can kick the ball before we check the territory control view.
  if(!ballView->getRelativeBallVector(NOW).isKnown())
    return FALSE;

  if(!ballView->getRelativeBallVector(NOW).getVector().getLength() > parameterView->SP_kickable_area)
    return FALSE;

  std::vector<PlayerObject> tm = playerView->getTeammates(NOW);
  // We need to see a player to be able to pass.
  if(tm.size() != 0) {
    Vector passVector = calculatePassVector();
    if(passVector.getLength() == 0)
      return FALSE;

    KickBallToSpeedVector kbt(passVector);
    return kbt.applicable();
  }
  return FALSE;
}

RS_END_NAMESPACE
