
/* -*- Mode: C++ -*-
 * $Id: Score.cc,v 1.10 2000/11/26 15:46:49 danbe628 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Score.cc
 * Unitname: Framework/skill
 * $Revision: 1.10 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: danbe628 $ $Date: 2000/11/26 15:46:49 $
 *
 *
 * HISTORY
 */

/**
 * @file Score.cc
 * Implementation file for the skill class Score.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.10 $
 * $Author: danbe628 $
 * $Date: 2000/11/26 15:46:49 $
 */


#include "Score.h"
#include "view_constants.h"
#include <algo.h>

#include "KickBallInDirection.h"

#include "AbleToKickBallToPoint.h"

#include "ParameterView.h"
#include "MarkerView.h"
#include "AgentView.h"
#include "BallView.h"
#include "PlayerView.h"

#include "rs_log.h"

#include "Triangle.h" // for math functions on triangle topics

RS_BEGIN_NAMESPACE


Score::Score()
  : Skill(),
    parameterView(NULL),
    markerView(NULL),
    agentView(NULL),
    ballView(NULL),
    playerView(NULL)
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(markerView, MARKER_VIEW_ID, parameterView);
  getView(agentView, AGENT_VIEW_ID, parameterView);
  getView(ballView, BALL_VIEW_ID);
  getView(playerView, PLAYER_VIEW_ID);
} 


Score::~Score()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(MARKER_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
}


void Score::findScoringDirections()
{
  Point goalBot, goalTop, ballPos;

  int hist = NOW; 
  
  // handles unknown values by looking back in the history
  while(ballView->getBallPosition(hist).isUnknown() && (hist<20))
    hist++;
  if(hist == 20)
    {
      hist = NOW;
      while(agentView->getAgentPosition(hist).isUnknown() && (hist<20))
	hist++;
      
      if(hist == 20)
	ballPos = Point(0,0);
      else
	ballPos = agentView->getAgentPosition(hist).getPoint();
    }
  else
    ballPos = ballView->getBallPosition(hist).getPoint();
 
  goalBot = goalTop = markerView->getTheirGoal(NOW).getPosition().getPoint();
  goalTop.setY(goalTop.getY() - parameterView->SP_goal_width/2);
  goalBot.setY(goalBot.getY() + parameterView->SP_goal_width/2);

  // 1- get the delimiters of the shoot-triangle
  Triangle tri(ballPos, goalTop, goalBot );

  //RS_LOG(LA_SCORE, "Score skill: shoot triangle, ballPos= " << ballPos << goalTop << goalBot );

  // 2- find the players in this triangle field
  std::vector<PlayerObject> oppList = playerView->getOpponents(NOW);
  
  playersInField.clear(); // empty vector, no player

  register int pos;
  PlayerObject oppPlayer;

  for(pos = 0; pos < (int)oppList.size(); pos++) // for each opponent player...
    {
      oppPlayer = oppList[pos];

      // check if position is known and if player is in the scoring field
      if( oppPlayer.getPosition().isKnown() && tri.isPointInTriangle(oppPlayer.getPosition().getPoint(), Epsilon) )
	{
	  // ok, so add the player
	  playersInField.push_back(oppPlayer);
	}
    }
  // RS_LOG(LA_SCORE, "Score skill: number of players in field: " << playersInField.size());

  // 3- find the directions towards the goal from the ball to all these players 
  EqFormulaList dirList; // create list of formulas
  Point pt;

  for(pos = 0; pos < (int)playersInField.size(); pos++) // for each player in field found in preceding step...
    {
      if(playersInField[pos].getPosition().isKnown())
	{
	  EqFormula form;
	  pt = playersInField[pos].getPosition().getPoint(); // get his position
	  
	  // check to not divide by zero
	  if( abs(pt.getX() - ballPos.getX()) > Epsilon )
	    {
	      // fill in the formula
	      form.a = (pt.getY() - ballPos.getY()) / (pt.getX() - ballPos.getX()); // the slope
	      
	      form.b = ballPos.getY() - form.a*ballPos.getX();
	      
	      // add it to list
	      dirList.push_back(form);
	    }
	  // else: the X difference equals zero, the opponent player is on the same X position as the ball
	  // this means the player is on the vertical line to the ball
	  // if the player is not "on" the ball, shooting is possible
	  // if the player is "on" the ball, we also consider that the shoot is possible (check this better maybe later on)
	}
    }

  // 4- find the intersection points in the goal (intersections with goal line and equations found in preceding step)
  std::vector<Float> ptList; // list of intersection points

  for(pos = 0; pos < (int)dirList.size(); pos++) // for each formula...
    {
      // store only the Y coordinate, the X is known (X coordinate of their goal position)
      ptList.push_back( dirList[pos].a * goalTop.getX() + dirList[pos].b );
    }

  // 5- determine the mid-points: middle of each segment formed by the intersection points
  // and give them a weight, depending on the size of each kickable zone in the goal (mid point beeing the middle of each zone)
  
  std::sort(ptList.begin(), ptList.end()); // sort the list of intersection points in goal
                                      // Y coord. increase when you go from the top to the bot. of the goal

  std::vector<WeightenMidPoint> midPoints; // list of mid points

  if( ptList.size() == 0 )
    {
      // list is empty (no intersection in the goal with player-ball lines, goal is empty!), so mid point is the middle of the goal
      WeightenMidPoint mp; // new mid point

      mp.midPoint.setY( (goalTop.getY() + goalBot.getY())/2 ); // compute Y coord.
      mp.midPoint.setX( goalBot.getX() );

      mp.weight = 1; // weight is 1, goal is empty

      midPoints.push_back(mp); // add this weighten mid point
      
      // RS_LOG(LA_SCORE, "Score skill: new mid point: " << mp.midPoint);
    }
  else
    {
      // first mid point is between the top limit of the goal and the first intersection point
      
      WeightenMidPoint mp; // new mid point
      mp.midPoint.setY( (goalTop.getY() + ptList[0]) / 2 ); // compute Y coord.
      mp.weight = (ptList[0] - goalTop.getY()) / parameterView->SP_goal_width; // compute weight

      midPoints.push_back(mp);

      // RS_LOG(LA_SCORE, "Score skill: new mid point, Y= " << mp.midPoint.getY());
      
      // compute the intermediate mid points
      for(pos = 0; pos <(int) ptList.size() - 1; pos++)
	{
	  mp.midPoint.setY( (ptList[pos] + ptList[pos + 1]) / 2 ); // compute Y coord.
	  mp.weight = (ptList[pos + 1] - ptList[pos]) / parameterView->SP_goal_width; // compute weight

	  midPoints.push_back(mp);

	  // RS_LOG(LA_SCORE, "Score skill: new mid point, Y= " << mp.midPoint.getY());

	}

      // for the last mid point, the one between the last intersection point and the bottom of the goal
      mp.midPoint.setY( (goalBot.getY() + ptList[ptList.size() - 1]) / 2 ); // compute Y coord.
      mp.weight = (ptList[ptList.size() - 1] - goalBot.getY()) / parameterView->SP_goal_width; // compute weight

      midPoints.push_back(mp);

      // RS_LOG(LA_SCORE, "Score skill: new mid point, Y= " << mp.midPoint.getY());
    }

  // 6- gets the values of the angles made up with lines to the mid-points

  // get orthogonal distance from ball to the goal (simply X difference)
  // this value has to be positive; and it should not be zero, if zero that means the ball is in the goal
  Float distToGoal = abs(ballPos.getX() - goalTop.getX());

  shootAngles.clear(); // empty vector

  WeightenShootAngle ang;

  for(pos = 0; pos < (int)midPoints.size(); pos++) // for each midPoint...
    {
      // get dist. form mid-point to crossing point in goal (same Y coord. value as ball's one)
      // -->(warning: dist. here can be negative)
      // the sign of the calculated dist. is the sign of the desired angle
      // (if the mid-point is upper than the crossing point, the sign is negative
      // if the mid-point is lower than the crossing point, the sign is positive)
      
      // and divide this dist. value by the distance to the goal

      // and get the inv tan of this --> that's our angle in radian (with the appropriate sign)
      if(distToGoal != 0)
	{
	  // little conversion from rad to deg (180/pi = 57.297)
	  ang.angle = 57.297 * (atan( (midPoints[pos].midPoint.getY() - ballPos.getY()) / distToGoal ));

	  //RS_LOG(LA_SCORE, "opp= " << (midPoints[pos].midPoint.getY() - ballPos.getY()) << "adj= " << distToGoal );
	  //RS_LOG(LA_SCORE, "opp/adj= " << (midPoints[pos].midPoint.getY() - ballPos.getY()) / distToGoal );
	  //RS_LOG(LA_SCORE, "atan= " << atan( (midPoints[pos].midPoint.getY() - ballPos.getY()) / distToGoal ));
	}
      else
	ang.angle = 0;

      //RS_LOG(LA_SCORE, "Score skill: possible angle = " << ang.angle);
      ang.weight = midPoints[pos].weight; // associated angle to mid point, so simply transfer weight from mid point

      shootAngles.push_back(ang);
    }
}

bool Score::findBestScoringDirection(AngleDeg& ang)
{
  if(shootAngles.size() > 0)
    {
      // start with this value
      Float wg = shootAngles[0].weight;
      ang = shootAngles[0].angle.getDeg();

      for(register int pos = 1; pos < (int)shootAngles.size(); pos++) // for each shoot angle...
	{
	  // look which has the best weight
	  if(wg < shootAngles[pos].weight)
	    {
	      wg = shootAngles[pos].weight; // update weight
	      ang = shootAngles[pos].angle; //getDeg(); // update angle
	    }
	}
      // RS_LOG(LA_SCORE, "Score skill: best angle found: " << ang);
      return true; // ok, could find an angle
    }
  else
    {
      //RS_LOG(LA_SCORE, "ERROR: Score skill; trying to find an angle from an empty shoot Angle list");
      return false; // error, list is empty, no correct angle
    }
}

Plan Score::generatePlan()
{
  KickBallInDirection* pKick = NULL;

  // 1- get all the possible scoring directions (updates also the list of players in the triangle field)
  findScoringDirections();

  // 2- find the best scoring dir.
  AngleDeg ang;
  
  if( findBestScoringDirection(ang) )
    {
      // ok, found it

      // generate a new kick
      std::vector<Float> args(2);
      args[0] = 100; // power rate
      args[1] = ang.getDeg();

      pKick = new KickBallInDirection; // new kick
      pKick->setUseMaxPower();
      pKick->setArguments(args); // update args
      //RS_LOG(LA_SCORE, "Score skill: will kick with angle: " << ang.getDeg());
    }
  else
    {
      // problem, no best angle
      RS_LOG(LA_SCORE, "Score skill: will do nothing, did not find any angle to shoot");
      // do nothing
      // pKick should be NULL
    }

  // 3- shoot!
  static Plan plan;

  // do that only if pKick exists
  if(pKick)
    {
      // ok
      if(pKick->applicable())
	{
	  // Plan
	  plan = pKick->generatePlan();
	  
	  //RS_LOG(LA_SCORE, "Score skill: plan generated");
	} 

      delete pKick;
      pKick = NULL;
    }
  else
    RS_LOG(LA_SCORE, "Score skill: pKick is null, no plan will be generated");

  return plan;
}


FuzzyBool Score::persistent()
{
  return FALSE;
}


FuzzyBool Score::applicable()
{
  if(ballView->getBallDistance(NOW).isKnown() && ballView->getBallPosition(NOW).isKnown())
    if( (ballView->getBallDistance(NOW) < parameterView->SP_kickable_margin) && // are we close enough to ball ?
	markerView->getTheirGoal(NOW).getPosition().isKnown() ) // do we know the goal position ?
      {
	if ( ballView->getRelativeBallDirection(NOW).isKnown() )                // do we know ball direction ?
	  {
	    AbleToKickBallToPoint predATKBTP; // predicate
	    std::vector<Float> args(2);
	    args[0] = markerView->getTheirGoal(NOW).getPosition().getPoint().getX();
	    args[1] = markerView->getTheirGoal(NOW).getPosition().getPoint().getY();
	    
	    predATKBTP.setArguments(args); // update position in predicate
	    
	    if( predATKBTP.evaluate() > FuzzyBool(0.9) )
	      {
		return TRUE;
	      }
	    
	    if( predATKBTP.evaluate() > FuzzyBool(0.4) )// == UNKNOWN )
	      {
		return UNKNOWN;
	      }
	  }
      }
 
  return FALSE;
}


RS_END_NAMESPACE
