/* -*- Mode: C++ -*-
 * $Id: TurnBall.cc,v 1.10 2000/11/07 18:10:43 mikbr212 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: TurnBall.cc
 * Unitname: Framework/skill
 * $Revision: 1.10 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: mikbr212 $ $Date: 2000/11/07 18:10:43 $
 *
 *
 * HISTORY
 * 
 * $Log: TurnBall.cc,v $
 * Revision 1.10  2000/11/07 18:10:43  mikbr212
 * changed persistent
 *
 * Revision 1.9  2000/10/30 18:32:45  mikbr212
 * added comments
 *
 */

/**
 * @file TurnBall.cc
 * Implementation file for the skill class TurnBall.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.10 $
 * $Author: mikbr212 $
 * $Date: 2000/11/07 18:10:43 $
 */

#include "TurnBall.h"


RS_BEGIN_NAMESPACE

/**
 * Constructor.
 */
TurnBall::TurnBall(const Vector& vec)
  : SkillWithVector(vec),
    timeStamp(Time(0,0)),
    //    newBallPos(Vector(0,0)),
    parameterView(NULL),
    ballView(NULL),
    kickBallView(NULL)
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(ballView, BALL_VIEW_ID);
  AgentView* agentView = NULL;
  getView(agentView, AGENT_VIEW_ID, parameterView);
  CommandView* commandView = NULL;
  getView(commandView, COMMAND_VIEW_ID);
  getView(kickBallView, KICKBALL_VIEW_ID, parameterView,
	  ballView, agentView, commandView);

  bufMin = parameterView->SP_player_size + parameterView->SP_ball_size; 
  kickMax = parameterView->SP_kickable_area;
  
  bufR = bufMin;
  kickR = kickMax;
  
  setAccuracy();
  setPlayerMargin();
  setKickableMargin();
  setMaxAcceleration();
  
  /*
    cout << "bufR = " << bufR << ", kickR = " << kickR << ", accR = " 
    << accR << ", maxAcc = " << maxAccelerate << ", kickMax = " << kickMax
    << "bufMin = " << bufMin << ", kickable_margin = " 
    << parameterView->SP_kickable_margin << endl;
  */
} 

/**
 * Destructor.
 */
TurnBall::~TurnBall()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(KICKBALL_VIEW_ID);
  viewManager->releaseView(COMMAND_VIEW_ID);
}

/**
 * Sets vector.
 */
void TurnBall::setVector(const Vector& vec) {
  vector = vec;
}

/**
 * Retrieves vector.
 */
Vector TurnBall::getVector() const {
  return vector;
}

/**
 * Sets the accuracy.
 * The if the ball is within the accuracy-radius then the skill
 * is not applicable anymore.
 */
bool TurnBall::setAccuracy(double x) {
  if(x < 0.0)
    return FALSE;
  accR = x;
  return TRUE;
}

/**
 * Sets the margin in percent of kickable_margin between the player and the ball.
 */
bool TurnBall::setPlayerMargin(double x) {
  if(kickR < (bufMin + x / 100.0 * parameterView->SP_kickable_margin) || 
     x < 0.0 || x >= 100.0)
    return FALSE;
  bufR = bufMin + x / 100.0 * parameterView->SP_kickable_margin;
  return TRUE;
  //  cout << "ply true\n";
}

/**
 * Sets the margin in percent of kickable_margin between the kickable_area-radius and the 
 * allowed ball_area.
 */
bool TurnBall::setKickableMargin(double x) {
  if((kickMax - x / 100.0 * parameterView->SP_kickable_margin) <= bufR || 
     x < 0.0 || x >= 100.0)
    return FALSE;
  kickR = kickMax - x / 100.0 * parameterView->SP_kickable_margin;
  return TRUE;
  //  cout << "kick true\n";
}

/**
 * Sets the maximum acceleration given to ball.
 */
bool TurnBall::setMaxAcceleration(double x) {
  if(x < 0.0)
    return false;
  maxAccelerate = x;
  return true;
}

/**
 * Generates a plan.
 * The plan will consist of only one step. To get more steps
 * simply call this skill next cycle. 
 * This method runs setVector(vec) and then generatePlan().
 */
Plan TurnBall::generatePlan(const Vector& vec) {
  setVector(vec);
  return generatePlan();
}

/**
 * Generates a plan.
 * The plan will consist of only one step. To get more steps
 * simply call this skill next cycle. 
 */
Plan TurnBall::generatePlan() {
  if(!ballView->isBallKickable(NOW) || 
     ballView->getRelativeBallVector(NOW).isUnknown()) 
    return Plan(); // empty plan

  Point ballPos = (Point) ballView->getRelativeBallVector(NOW).getVector();
  Vector kickVec(0,0);
  
  // Can I shoot straight to desired ball position?
  Line straightLine = Line(ballPos, (Point)vector);
  if(!straightLine.isThroughCircle(Point(0,0),min(bufR,Vector(ballPos).getLength()))) {
    kickVec = straightLine.getVector();
  } else { // not straight line
    bool kickCW = (vector.getAngle().getDeg()
		   - Vector(ballPos).getAngle().getDeg()) > 0;

    // get lines from pos to tangent of circle
    Line nowBallLine = Line(ballPos, Point(0,0), bufR, kickCW);
    Line desBallLine = Line(Point(vector), Point(0,0), bufR, !kickCW);

    Vector crossPoint = (Vector) nowBallLine.getIntersection(desBallLine);
 
    if(crossPoint.getLength() <= kickR) {
      kickVec = crossPoint - Vector(ballPos);
    } else { // crosspoint outside kickable area (minus margin)
      // Just put ball on circle between bufR & kickR
      nowBallLine.extendToCircle(Point(0,0),(bufR+kickR)/2);
      kickVec = nowBallLine.getVector();
    }
  }
  
  return getKickPlan(kickVec);
}

Plan TurnBall::getKickPlan(Vector vec) {
  Vector ballPos(0,0);
  Vector ballSpeed(0,0);
  kickBallView->initBall(ballSpeed,ballPos,true);
  
  Vector kickVec = kickBallView->getKickVector(ballSpeed,ballPos,vec,
    					       maxAccelerate);

  timeStamp = Time::currentTime;
  //  newBallPos = ballPos + ballSpeed + kickVec;

  Command *cmd = new KickCommand(kickVec.getLength() /
				 kickBallView->kickPowerRate(),
				 kickVec.getAngle().getDeg());
  return Plan(Step(cmd));
}

/**
 * Checks if this skill should be called again (this cycle), ie if
 * (1) generatePlan was executed last cycle,
 * (2) the ball position is kickable, and
 * (3) the ball is not at desired position
 * @see setAccuracy(double)
 */
FuzzyBool TurnBall::persistent()
{
  return (timeStamp + 1 == Time::currentTime &&
	  applicable());

  //newBallPos.getLength() < parameterView->SP_kickable_area &&
  //	  (newBallPos - vector).getLength() > accR);
}

/**
 * Checks if preconds are true.
 */
FuzzyBool TurnBall::applicable(const Vector& vec) {
  setVector(vec);
  return applicable();
}

/**
 * Checks if preconds are true.
 */
FuzzyBool TurnBall::applicable()
{
  if(ballView->isBallKickable(NOW) &&
     ballView->getRelativeBallVector(NOW).isKnown()) {
    Vector ballPos = ballView->getRelativeBallVector(NOW).getVector();

    if((ballPos - vector).getLength() > accR) {
      return true;
    }
  }
  return false;
}


RS_END_NAMESPACE






