/* -*- Mode: C++ -*-
 * $Id: AgentView.cc,v 1.5 2001/01/11 22:00:20 olalu526 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: AgentView.cc
 * Unitname: Framework/view
 * $Revision: 1.5 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: olalu526 $ $Date: 2001/01/11 22:00:20 $
 *
 *
 * HISTORY
 *
 * $Log: AgentView.cc,v $
 * Revision 1.5  2001/01/11 22:00:20  olalu526
 * Added the real world model with support for new simulation steps.
 *
 * Revision 1.4  2001/01/10 15:12:58  olalu526
 * Added support for WorldModel to some basic classes.
 *
 * Revision 1.3  2000/10/18 23:06:35  frehe
 * Added lastCoachMessage and lastPlayerMessage.
 *
 * Revision 1.2  2000/09/25 13:34:18  frehe
 * Made minor modifications.
 *
 * Revision 1.1  2000/09/21 13:46:45  frehe
 * Move the RoboSoc views from rsviews to the rsframework library.
 *
 * Revision 1.3  2000/09/13 00:08:05  frehe
 * Updated the views to inherit from the ViewWithHistory instead and also changed those views that used a static pointer to itself to do process the data in the sensor interface to use the new forEach methods implemented in the FrameworkSensorInterface.
 *
 * Revision 1.2  2000/09/05 21:15:57  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/04 15:27:18  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file AgentView.cc
 * Implementation file for the class AgentView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.5 $
 * $Author: olalu526 $
 * $Date: 2001/01/11 22:00:20 $
 */


#include "AgentView.h"
#include "ViewManager.h"
#include "ParameterView.h"
#include "CommandView.h"
#include "MarkerView.h"
#include "GameView.h"
#include "view_constants.h"
#include "rs_error.h"
#include "rs_log.h"


RS_BEGIN_NAMESPACE


AgentView::AgentView(const ViewId id,
		     ParameterView* const parameter_view,
		     const UInt hist_len)
  : ViewWithSimulation(id),
    // parameterView is now depricated because of the worldModelInterface but
    // is here for old things to work.
    parameterView(parameter_view),
    commandView(NULL),
    gameView(NULL),
    markerView(NULL),
    lastHearTime(Time()),
    lastSeeTime(Time()),
    lastSenseTime(Time())
{
  worldModel->setAgentHistoryLength(hist_len);
  // This is just to increase the reference count for the parameter view.
  viewManager->getView(PARAMETER_VIEW_ID);

  //getView(parameterView, PARAMETER_VIEW_ID);
  getView(commandView, COMMAND_VIEW_ID, hist_len);
  getView(gameView, GAME_VIEW_ID, hist_len);
  getView(markerView, MARKER_VIEW_ID, parameterView, hist_len);
}


AgentView::~AgentView()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(COMMAND_VIEW_ID);
  viewManager->releaseView(GAME_VIEW_ID);
  viewManager->releaseView(MARKER_VIEW_ID);
}


bool AgentView::isAgentOutsideField() const
{
  static const Float offset_x = 0.75;
  static const Float offset_y = 0.75;
  
  PointUnknown apos = theAgent[getIndex(NOW)].getPosition();
  if ( (apos.getX().isKnown() &&
	abs(apos.getX()) > (parameterView->SP_pitch_length/2 + offset_x )) ||
       (apos.getY().isKnown() &&
	abs(apos.getY()) > (parameterView->SP_pitch_width/2 + offset_y )) ) {
    return true;
  }

  return false;
}


// NewCycle is called *before* the values in CSensorInterface
// are erased, so this is the last chance to use them...
void AgentView::myNewCycle()
{
  commandView->newCycle(currentUpdateNumber);

  //RS_LOG( LA_AGENT_VIEW, "Agent pos = " << theAgent[getIndex(NOW)].getPosition() );
  
  // Start with updating after the immediate actions that have already
  // been executed *last* cycle.
  // If the agent made a turn_neck update the neck direction
  theAgent[getIndex(NOW)].setNeckDirection(getAgentObject(NOW).getNeckDirection() +
					   commandView->getActualTurnNeck(PREV));
  theAgent[getIndex(NOW)].setFaceDirection(getAgentObject(NOW).getFaceDirection() +
					   commandView->getActualTurnNeck(PREV));

  // If the agent made a move update the speed and position of the agent
  if ( commandView->getMovePosition(PREV).isKnown() ) {
    theAgent[getIndex(NOW)].setPosition(commandView->getMovePosition(PREV));
    theAgent[getIndex(NOW)].setAbsoluteSpeedVector(VectorUCO(0,0));
  }

  theAgent.copy(getIndex(NOW), getIndex(NEXT));

  // Update the time step in the history.
  increaseIndex();

  // Update the agent position based on the estimation done in the command view
  theAgent[getIndex(NOW)].setPosition(theAgent[getIndex(NOW)].getPosition() +
				      VectorUCO(VectorUnknown(commandView->getAgentMovement(PREV))));

  // Bad attempt to calculate the real position of the agent
  // by estimating after dash, the problem is the initial position.
  //theAgent[getIndex(NOW)].setPosition(commandView->getAgentPosition(NOW));
  
  // Update the speed of the agent
  theAgent[getIndex(NOW)].setAbsoluteSpeedVector(VectorUCO(VectorUnknown(commandView->getAgentSpeed(NOW))));

  // Update the body direction of the agent
  if ( theAgent[getIndex(NOW)].getFaceDirection().isKnown() ) {
    theAgent[getIndex(NOW)].setFaceDirection(theAgent[getIndex(NOW)].getFaceDirection() +
					     commandView->getActualTurn(PREV));
  }
}


void AgentView::myUpdateAfterInit()
{
  theAgent[getIndex(NOW)].setGoalie(parameterView->CP_goalie);
  theAgent[getIndex(NOW)].setUniformNumber(sensorInterface->getUniformNumber());
  theAgent[getIndex(NOW)].setSide(sensorInterface->getOurSide());
}


void AgentView::myUpdateAfterSee()
{
  gameView->updateAfterSee(currentUpdateNumber);
  markerView->updateAfterSee(currentUpdateNumber);

  lastSeeTime = gameView->getCurrentTime();
  

  theAgent[getIndex(NOW)].setFaceDirection(computeFaceDirection());

  /* This should be uncommented if you rotate all markers
  // Compensate if we play on the right side of the field.
  if ( gameView->getOurSide() == S_Right ) {
    agentFaceDirection = agentFaceDirection - 180;
  }
  */

  theAgent[getIndex(NOW)].setPosition(computeAgentPosition());

  // If the agent is outside the field or playing on the
  // right hand side of the field then the face direction
  // and position should be adjusted.
  // The face direction computation assumes that the agent is inside the field
  // and is playing on the left side of the field.
//   if ( (IsAgentOutsideField() &&
// 	gameView->getOurSide() == S_Left) ||
//        (!IsAgentOutsideField() &&
// 	gameView->getOurSide() == S_Right) )

  // If the agent is playing on the right hand side of the field we
  // must adjust the face direction and the position.
  if ( gameView->getOurSide() == S_RIGHT ) {
    theAgent[getIndex(NOW)].setPosition(-theAgent[getIndex(NOW)].getPosition());
    theAgent[getIndex(NOW)].setFaceDirection(theAgent[getIndex(NOW)].getFaceDirection() -
					     180);
  }
}


void AgentView::myUpdateAfterHear()
{
  //gameView->updateAfterHear(currentUpdateNumber);
  //markerView->UpdateAfterHear(currentUpdateNumber);

  lastHearTime = gameView->getCurrentTime();
  theAgent[getIndex(NOW)].setLastCoachMessage(sensorInterface->getCoachMessage());
  theAgent[getIndex(NOW)].setLastPlayerMessage(sensorInterface->getPlayerMessage());
}


void AgentView::myUpdateAfterSense()
{
  gameView->updateAfterSense(currentUpdateNumber);
  //markerView->UpdateAfterSense(currentUpdateNumber);

  lastSenseTime = gameView->getCurrentTime();
  
  theAgent[getIndex(NOW)].setNeckDirection(sensorInterface->getHeadDirection());
  theAgent[getIndex(NOW)].setSpeed(sensorInterface->getSpeed());
  theAgent[getIndex(NOW)].setAbsoluteSpeedDirection(getBodyDirection(NOW));
  theAgent[getIndex(NOW)].setEffort(sensorInterface->getEffort());
  theAgent[getIndex(NOW)].setStamina(sensorInterface->getStamina());
  theAgent[getIndex(NOW)].setViewWidth(sensorInterface->getViewWidth());
  theAgent[getIndex(NOW)].setViewQuality(sensorInterface->getViewQuality());
  theAgent[getIndex(NOW)].setDashes(sensorInterface->getDashes());
  theAgent[getIndex(NOW)].setKicks(sensorInterface->getKicks());
  theAgent[getIndex(NOW)].setSays(sensorInterface->getSays());
  theAgent[getIndex(NOW)].setTurns(sensorInterface->getTurns());
  theAgent[getIndex(NOW)].setTurnNecks(sensorInterface->getTurnNecks());
}


void AgentView::myUpdateAfterCommand()
{
  /* Since we currently send all commands last in the cycle, even
     immediate actions, we can do all the updating at the begining
     of the new cycle.
  commandView->UpdateAfterCommand(currentUpdateNumber);

  // If the agent made a turn_neck update the neck direction
  if ( commandView->getLastCommandType() == CMD_TurnNeck ) {
    //if ( getAgentObject(NOW).getNeckDirection().IsKnown() )
    theAgent[getIndex(NOW)].setNeckDirection(getAgentObject(NOW).getNeckDirection() +
					     commandView->getActualTurnNeck(NOW));
    theAgent[getIndex(NOW)].setFaceDirection(getAgentObject(NOW).getFaceDirection() +
					     commandView->getActualTurnNeck(NOW));
  }

  // If the agent made a move update the speed and position of the agent
  if ( commandView->getLastCommandType() == CMD_Move ) {
    theAgent[getIndex(NOW)].setPosition(commandView->getMovePosition(NOW));
    theAgent[getIndex(NOW)].setAbsoluteSpeedVector(VectorUCO(0,0));
  }
  */
}


AngleDegUnknown AgentView::computeFaceDirection() const
  // This is taken from ATHumboldt'97
{
  if ( markerView->getNumberOfSeenLines() == 0 ) {
    RS_WARNING("Didn't see any lines! Face direction is unknown!");
    return AngleDegUnknown();
  } else {
    LineObject line = markerView->getClosestSeenLine();
    AngleDegUnknown angle = line.getDirection();

    if ( angle.isUnknown() )
      return angle;
    else {

      switch ( line.getLineId() ) {
      case SL_TOP:
	if ( angle > 0 ) {
	  return -angle;
	}
	else {
	  return AngleDeg(180) - angle;
	} 
	break;
      case SL_BOTTOM:
	if ( angle > 0 ) {
	  return AngleDeg(180) - angle;
	}
	else {
	  return -angle;
	}
	break;
      case SL_LEFT:
	if ( angle > 0 ) {
	  return AngleDeg(270) - angle;
	}
	else if ( angle < 0 ) {
	  return AngleDeg(90) - angle;
	} 
	else {
	  RS_WARNING("Couldn't compute face dir. Special case 1 is not treated.\n");
	  return AngleDegUnknown();
	}
	break;
      case SL_RIGHT:
	if ( angle > 0 ) {
	  return AngleDeg(90) - angle;
	}
	else if ( angle < 0 ) {
	  return AngleDeg(270) - angle;
	} 
	else {
	  RS_WARNING("Couldn't compute face dir. Special case 2 is not treated.\n");
	  return AngleDegUnknown();
	}
	break;
      default:
	RS_WARNING("Unknown line id: " + toString(line) + " in computeFaceDir.\n");
	return AngleDegUnknown();
      }
    }
  }

  return AngleDegUnknown();

}


PointUnknown AgentView::computeAgentPosition()
{
  MarkerObject flag = markerView->getClosestSeenMarker();

  if ( flag.getVector().isUnknown() )
    return PointUnknown();
  else {
    VectorUnknown rel_pos(flag.getDirection() +
			   AngleDegUCO(theAgent[getIndex(NOW)].getFaceDirection()),
			   flag.getDistance());
    PointUnknown abs_pos = flag.getPosition();
    /*
    log << "face dir = " << agentFaceDirection
	<< " flag_id = " << flag.getMarkerId()
	<< " flag_pos = " << abs_pos
	<< "\nrel_pos = " << rel_pos
	<< " my_pos = " << abs_pos - rel_pos
	<< endl;
	*/
    
    abs_pos = abs_pos - rel_pos;

    return abs_pos;
  }
}


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "AgentView.icc"
#endif
