/* -*- Mode: C++ -*-
 * $Id: CoachView.cc,v 1.1 2000/09/21 13:46:46 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: CoachView.cc
 * Unitname: Framework/view
 * $Revision: 1.1 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/21 13:46:46 $
 *
 *
 * HISTORY
 *
 * $Log: CoachView.cc,v $
 * Revision 1.1  2000/09/21 13:46:46  frehe
 * Move the RoboSoc views from rsviews to the rsframework library.
 *
 * Revision 1.3  2000/09/13 00:08:06  frehe
 * Updated the views to inherit from the ViewWithHistory instead and also changed those views that used a static pointer to itself to do process the data in the sensor interface to use the new forEach methods implemented in the FrameworkSensorInterface.
 *
 * Revision 1.2  2000/09/05 21:15:57  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/04 15:27:21  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file CoachView.cc
 * Implementation file for the class CoachView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2000/09/21 13:46:46 $
 */


#include "CoachView.h"
#include "ViewManager.h"
#include "ParameterView.h"
#include "GameView.h"
#include "AgentView.h"
#include "view_constants.h"
#include "rs_error.h"


RS_BEGIN_NAMESPACE


CoachView::CoachView(const ViewId id,
		     const UInt hist_len)
  : ViewWithHistory(id, hist_len),
    parameterView(NULL),
    gameView(NULL),
    agentView(NULL),
    coachMessageTime(hist_len, UIntUnknown()),
    ballPosition(hist_len, PointUnknown()),
    ballSpeed(hist_len, VectorUnknown()),
    agentPosition(hist_len, PointUnknown()),
    agentSpeed(hist_len, VectorUnknown()),
    agentDirection(hist_len, AngleDegUnknown()),
    agentNeckDirection(hist_len, AngleDegUnknown()),
    opponents(hist_len, std::vector<PlayerObject>(11, PlayerObject())),
    teammates(hist_len, std::vector<PlayerObject>(10, PlayerObject()))
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(gameView, GAME_VIEW_ID, hist_len);
  getView(agentView, AGENT_VIEW_ID, parameterView, hist_len);
}


CoachView::~CoachView()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(GAME_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
}


void CoachView::myNewCycle()
{
  // Update the current time step in the history.
  increaseIndex();

  coachMessageTime[getIndex(NOW)] = UIntUnknown();
  ballPosition[getIndex(NOW)] = PointUnknown();
  ballSpeed[getIndex(NOW)] = VectorUnknown();
  agentPosition[getIndex(NOW)] = PointUnknown();
  agentSpeed[getIndex(NOW)] = VectorUnknown();
  agentDirection[getIndex(NOW)] = AngleDegUnknown();
  agentNeckDirection[getIndex(NOW)] = AngleDegUnknown();

  for ( std::vector<PlayerObject>::iterator plr = opponents[getIndex(NOW)].begin();
	plr != opponents[getIndex(NOW)].end(); ++plr ) {
    plr->setPosition(PointUCO());
    plr->setVector(VectorUCO());
  }

  for ( std::vector<PlayerObject>::iterator plr = teammates[getIndex(NOW)].begin();
	plr != teammates[getIndex(NOW)].end(); ++plr ) {
    plr->setPosition(PointUCO());
    plr->setVector(VectorUCO());
  }
}


void CoachView::myUpdateAfterCoachMessage()
{
  coachMessageTime[getIndex(NOW)] = sensorInterface->getCoachMessageTime();
  ballPosition[getIndex(NOW)] = sensorInterface->getBallPosition();
  ballSpeed[getIndex(NOW)] = sensorInterface->getBallSpeed();
  agentPosition[getIndex(NOW)] = sensorInterface->getAgentPosition();
  agentSpeed[getIndex(NOW)] = sensorInterface->getAgentSpeed();
  agentDirection[getIndex(NOW)] = sensorInterface->getAgentDirection();
  agentNeckDirection[getIndex(NOW)] = sensorInterface->getAgentNeckDirection();

  std::vector<CoachPlayerObject> playerVector;

  playerVector = sensorInterface->getOpponents();
  for ( std::vector<CoachPlayerObject>::const_iterator plr = playerVector.begin();
	plr != playerVector.end(); ++plr ) {
    PlayerObject player(plr->team,
			plr->side,
			plr->uniformNumber,
			plr->neckDirection,
			plr->bodyDirection + plr->neckDirection,
			plr->position,
			VectorUCO(),
			FloatUCO(),
			AngleDegUCO(),
			plr->speed);
    if ( plr->uniformNumber.isKnown() )
      opponents[getIndex(NOW)][plr->uniformNumber.getValue()] = player;
    else
      RS_WARNING("Coach saw an opponent player with unknown uniform number!");
  }
  
  playerVector = sensorInterface->getTeammates();
  for ( std::vector<CoachPlayerObject>::const_iterator plr = playerVector.begin();
	plr != playerVector.end(); ++plr ) {
    PlayerObject player(plr->team,
			plr->side,
			plr->uniformNumber,
			plr->neckDirection,
			plr->bodyDirection + plr->neckDirection,
			plr->position,
			VectorUCO(agentPosition[getIndex(NOW)], plr->position),
			FloatUCO(),
			AngleDegUCO(),
			plr->speed);
    if ( plr->uniformNumber.isKnown() )
      teammates[getIndex(NOW)][plr->uniformNumber.getValue()] = player;
    else
      RS_WARNING("Coach saw a teammate player with unknown uniform number!");
  }

}


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "CoachView.icc"
#endif
