/* -*- Mode: C++ -*-
 * $Id: CommandView.cc,v 1.2 2000/10/18 23:06:30 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: CommandView.cc
 * Unitname: Framework/view
 * $Revision: 1.2 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/10/18 23:06:30 $
 *
 *
 * HISTORY
 *
 * $Log: CommandView.cc,v $
 * Revision 1.2  2000/10/18 23:06:30  frehe
 * Fixed a bug that an object collection was not initialized.
 *
 * Revision 1.1  2000/09/21 13:46:46  frehe
 * Move the RoboSoc views from rsviews to the rsframework library.
 *
 * Revision 1.3  2000/09/13 00:08:07  frehe
 * Updated the views to inherit from the ViewWithHistory instead and also changed those views that used a static pointer to itself to do process the data in the sensor interface to use the new forEach methods implemented in the FrameworkSensorInterface.
 *
 * Revision 1.2  2000/09/05 21:15:58  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/04 15:27:22  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file CommandView.cc
 * Implementation file for the class CommandView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2000/10/18 23:06:30 $
 */


#include "CommandView.h"
#include "CatchCommand.h"
#include "DashCommand.h"
#include "KickCommand.h"
#include "MoveCommand.h"
#include "TurnCommand.h"
#include "TurnNeckCommand.h"
#include "ViewManager.h"
#include "ParameterView.h"
#include "GameView.h"
#include "BallView.h"
#include "AgentView.h"
#include "view_constants.h"
#include "rs_error.h"
#include "rs_assert.h"
#include "rs_log.h"


RS_BEGIN_NAMESPACE


CommandView::CommandView(const ViewId id,
			 const UInt16 hist_len)
  : ViewWithHistory(id, hist_len),
    parameterView(NULL),
    gameView(NULL),
    ballView(NULL),
    agentView(NULL),
    agentMovement(hist_len, Vector(0,0)),
    ballMovement(hist_len, Vector(0,0)),
    actualTurn(hist_len, AngleDeg(0)),
    actualTurnNeck(hist_len, AngleDeg(0)),
    agentAcceleration(hist_len, Point(0,0)),
    ballAcceleration(hist_len, Point(0,0)),
    actualDashPower(hist_len, Float(0)),
    movePosition(hist_len, PointUnknown()),
    agentSpeed(hist_len, Point(0,0)),
    agentPosition(hist_len, Point(0,0)),
    ballSpeed(hist_len, Point(0,0)),
    agentStamina(hist_len, Float(0)),
    agentEffort(hist_len, Float(1)),
    agentRecovery(hist_len, Float(1)),
    agentBodyDirection(hist_len, AngleDeg(0)),
    agentNeckDirection(hist_len, AngleDeg(0))
{
  getView(parameterView, PARAMETER_VIEW_ID);

  // Set the default values for some of the attributes
  agentStamina[getIndex(NOW)] = parameterView->SP_stamina_max;
  agentEffort[getIndex(NOW)] = parameterView->SP_effort_max;
  agentRecovery[getIndex(NOW)] = parameterView->SP_recover_max;
}


CommandView::~CommandView()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  if ( initialized ) {
    viewManager->releaseView(GAME_VIEW_ID);
    viewManager->releaseView(BALL_VIEW_ID);
    viewManager->releaseView(AGENT_VIEW_ID);
  }
}


bool CommandView::init()
{
  getView(gameView, GAME_VIEW_ID, getHistoryLength());
  getView(ballView, BALL_VIEW_ID, getHistoryLength());
  getView(agentView, AGENT_VIEW_ID, parameterView, getHistoryLength());
  initialized = true;
  return true;  
}


void CommandView::myNewCycle()
{

  try {

    // Save the neck direction
    agentNeckDirection[getIndex(NEXT)] = agentNeckDirection[getIndex(NOW)];

  
    // Update the body direction of the agent
    agentBodyDirection[getIndex(NEXT)] = agentBodyDirection[getIndex(NOW)] + actualTurn[getIndex(NOW)];

  
    // Calculate the movement of the agent the last cycle.
    Point agentU = agentAcceleration[getIndex(NOW)];
    if ( agentView->getSpeed(NOW).isUnknown() ) {
      // agentU += agentSpeed[getIndex(NOW)];
    } else if ( agentView->getSpeedVector(NOW).getAngle().isUnknown() ) {
      if ( agentView->getBodyDirection(NOW).isUnknown() ) {
	//agentU += Point(Vector(agentBodyDirection[getIndex(NOW)],
	//		       FloatUnknown(agentView->getSpeed(NOW)).getValue()));
      } else {
	agentU += Point(Vector(agentView->getBodyDirection(NOW).getDeg(),
			       FloatUnknown(agentView->getSpeed(NOW)).getValue()));
      }
    } else {
      agentU += Point(agentView->getSpeedVector(NOW).getVector());
    }

    // Calculate the movement vector of the agent
    agentMovement[getIndex(NOW)] = Vector(agentU);

    // Update the calculated agent position
    agentPosition[getIndex(NEXT)] = agentPosition[getIndex(NOW)] + agentU;

    // Calculate the new agent speed
    agentSpeed[getIndex(NEXT)] = agentU * parameterView->SP_player_decay;

    // Reset the acceleration of the agent.
    agentAcceleration[getIndex(NEXT)] = Point(0,0);
  
  } catch (RSException &excpt) {
    RS_FATAL("Caught an exception in myNewCycle part 1: " << excpt.What());
  }

  try {
    // Calculate the movement of the ball the last cycle.
    Point ballU = ballAcceleration[getIndex(NOW)];
    if ( ballView->getBallSpeed(NOW).isUnknown() ||
	 ballView->getBallAbsoluteSpeedDirection(NOW).isUnknown() ) {
      // ballU += ballSpeed[getIndex(NOW)];
    } else
      ballU += Point(ballView->getBallAbsoluteSpeedVector(NOW).getVector());

    // Calculate the movement vector of the ball
    ballMovement[getIndex(NOW)] = Vector(ballU);

    // Calculate the new ball speed
    ballSpeed[getIndex(NEXT)] = ballU * parameterView->SP_ball_decay;
  
    // Reset the acceleration of the ball.
    ballAcceleration[getIndex(NEXT)] = Point(0,0);

  } catch (RSException &excpt) {
    RS_FATAL("Caught an exception in myNewCycle part 2: " << excpt.What());
  }

  try {
    
    // Decrease the stamina with the power used by dash
    agentStamina[getIndex(NEXT)] = max(Float(parameterView->SP_stamina_min),
				       agentStamina[getIndex(NOW)] - actualDashPower[getIndex(NOW)]);

    // Update the recovery
    agentRecovery[getIndex(NEXT)] = agentRecovery[getIndex(NOW)];
    if ( agentStamina[getIndex(NEXT)] <= parameterView->SP_recover_dec_thr * parameterView->SP_stamina_max )
      agentRecovery[getIndex(NEXT)] = max(Float(parameterView->SP_recover_min),
					  agentRecovery[getIndex(NEXT)] - parameterView->SP_recover_dec);

    // Update the effort
    agentEffort[getIndex(NEXT)] = agentEffort[getIndex(NOW)];
    if ( agentStamina[getIndex(NEXT)] <= parameterView->SP_effort_dec_thr * parameterView->SP_stamina_max )
      agentEffort[getIndex(NEXT)] = max(Float(parameterView->SP_effort_min),
					agentEffort[getIndex(NEXT)] - parameterView->SP_effort_dec);
    else if ( agentStamina[getIndex(NEXT)] >= parameterView->SP_effort_inc_thr * parameterView->SP_stamina_max )
      agentEffort[getIndex(NEXT)] = min(Float(parameterView->SP_effort_max),
					agentEffort[getIndex(NEXT)] + parameterView->SP_effort_inc);

    // Recover stamina
    agentStamina[getIndex(NEXT)] = min(Float(parameterView->SP_stamina_max),
				       agentStamina[getIndex(NEXT)] +
				       agentRecovery[getIndex(NEXT)]*parameterView->SP_stamina_inc_max);
  
  
    // Then update the state of the view
    increaseIndex();
  
    movePosition.reset(getIndex(NOW));

    agentMovement.reset(getIndex(NOW));
    ballMovement.reset(getIndex(NOW));
    actualTurn.reset(getIndex(NOW));
    actualTurnNeck.reset(getIndex(NOW));
    actualDashPower.reset(getIndex(NOW));

  } catch (RSException &excpt) {
    RS_FATAL("Caught an exception in myNewCycle part 3: " << excpt.What());
  }
}


void CommandView::myUpdateAfterCommand()
{
  if ( !initialized )
    init();

  RS_REQUIRE( sensorInterface->getLastCommand() != NULL );
  
  RS_LOG( LA_COMMAND_VIEW, "lastCommand = " << sensorInterface->getLastCommand() );
  const ServerCommand* const last_cmd = sensorInterface->getLastCommand();
  lastCommandType = last_cmd->getType();
  lastServerString = sensorInterface->getLastCommand()->serverString();

  //std::cerr << gameView->getCurrentTime()
  //          << "Updated after " << sensorInterface->getLastCommand()->ServerString()
  //	      << std::endl;
  
  try {
  
    if ( lastCommandType == CMD_Catch ) {
      const CatchCommand* const cmd = last_cmd->getConstCatchCommand();
    } else if ( lastCommandType == CMD_Dash ) {
      const DashCommand* const cmd = last_cmd->getConstDashCommand();
      if ( gameView->getPlayMode() == PM_BEFORE_KICK_OFF ) {
	// Dash shouldn't work since you're not allowed to dash in this play mode!
      } else {
	// First we need to normalize the power, since we can't be sure that the
	// power is actually within its bounds.
	Float power = cmd->getPower();
	Float real_power = minMax(Float(parameterView->SP_minpower),
				  power,
				  Float(parameterView->SP_maxpower));
	Float needed_power = real_power;
	// Calculate the maximum needed power the agent can use
	Float max_needed_power = -parameterView->SP_stamina_min;
	if ( agentView->getStamina(NOW).isUnknown() )
	  max_needed_power += agentStamina[getIndex(NOW)];
	else
	  max_needed_power += FloatUnknown(agentView->getStamina(NOW)).getValue();
      
	// If we do a negative dash the cost is doubled.
	if ( power < 0 ) {
	  needed_power *= -2;
	}

	// Check if the agent has enough stamina to do the dash,
	// otherwise decrease the power so that the stamina
	// will be SP_stamina_min afterwards.
	if ( needed_power > max_needed_power ) {
	  needed_power = max_needed_power;
	  // Since the cost of negative dash is doubled half the
	  // real power if negative dash.
	  real_power = ( power < 0 ) ? needed_power/-2 : needed_power;
	}

	AngleDeg bodyDir;
	if ( agentView->getBodyDirection(NOW).isUnknown() )
	  bodyDir = agentBodyDirection[getIndex(NOW)];
	else
	  bodyDir = agentView->getBodyDirection(NOW).getDeg();
      
	Float effort;
	if ( agentView->getEffort(NOW).isUnknown() )
	  effort = agentEffort[getIndex(NOW)];
	else
	  effort = FloatUnknown(agentView->getEffort(NOW)).getValue();

	// Now update the acceleration of the agent.
	real_power = real_power * effort * parameterView->SP_dash_power_rate;
	agentAcceleration[getIndex(NOW)] = Point(real_power * cos(bodyDir),
						 real_power * sin(bodyDir));
      
	// Then store the power used.
	actualDashPower[getIndex(NOW)] = needed_power;
      }
    } else if ( lastCommandType == CMD_Kick ) {
      const KickCommand* const cmd = last_cmd->getConstKickCommand();
      Float power = cmd->getPower();
      AngleDeg angle = cmd->getAngle();
      if ( gameView->getPlayMode() == PM_BEFORE_KICK_OFF ||
	   gameView->getPlayMode() == PM_OUR_GOAL ||
	   gameView->getPlayMode() == PM_THEIR_GOAL ||
	   gameView->getPlayMode() == PM_OUR_OFFSIDE_KICK ||
	   gameView->getPlayMode() == PM_THEIR_OFFSIDE_KICK ) {
	// Kick shouldn't work since you're not allowed to kick in these play modes!
      } else if ( ballView->getBallDistance(NOW).isUnknown() ||
		  ballView->getBallDistance(NOW) > parameterView->SP_kickable_area ) {
	// Kick shouldn't work since the ball is too far away!
	// If the distance is unknown this test will also fail.
      
	//    cout << "Didn't update after kick! distance: " << ballView->getDistance(NOW)
	//   << std::endl;
      } else {
	Float dir_diff = 0;
	if ( ballView->getRelativeBallDirection(NOW).isKnown() )
	  dir_diff = abs(ballView->getRelativeBallDirection(NOW).getDeg());
      
	Float dist_diff = FloatUnknown(ballView->getBallDistance(NOW)).getValue() - parameterView->SP_player_size - parameterView->SP_ball_size;
	Float kick_power = parameterView->SP_kick_power_rate * power * (1.0 - 0.25*dir_diff/180.0 - 0.25*dist_diff/parameterView->SP_kickable_margin);

	AngleDeg ball_dir = angle;
	if ( agentView->getBodyDirection(NOW).isUnknown() ) {
	  // ball_dir += agentBodyDirection[getIndex(NOW)];
	} else {
	  ball_dir += agentView->getBodyDirection(NOW).getDeg();
	}
      
	ballAcceleration[getIndex(NOW)] = Point(kick_power * cos(ball_dir),
						kick_power * sin(ball_dir));
      }
    } else if ( lastCommandType == CMD_Move ) {
      const MoveCommand* const cmd = last_cmd->getConstMoveCommand();
      // If it is successful then the player should be at the intended position.
      // If the position is legal that is. Otherwise the position will be totally random.
      // The speed of the agent should be zero, and the neck and body dir unchanged.
      Point pos = cmd->getDestination();
      // This is not true since it is the play mode at the execution of the command
      // that is intresting!!!
      if ( ( ( gameView->getPlayMode() == PM_BEFORE_KICK_OFF ||
	       gameView->getPlayMode() == PM_OUR_GOAL ||
	       gameView->getPlayMode() == PM_THEIR_GOAL ) &&
	     // All players can move when before kick off and after a goal
	     // If they move within their own half.
	     ( pos.getX() <= 0 &&
	       abs(pos.getY()) > parameterView->SP_pitch_width/2 ) ) ||
	   ( gameView->getPlayMode() == PM_OUR_GOALIE_GOT_BALL &&
	     agentView->isGoalie() &&
	     // The goalie is allowed to move if it has catched the ball
	     // and it moves within its penalty area.
	     pos.getX() <= (parameterView->SP_penalty_area_length - parameterView->SP_pitch_length/2) &&
	     abs(pos.getY()) > parameterView->SP_penalty_area_width/2 ) ) {
	movePosition[getIndex(NOW)] = pos;
	agentPosition[getIndex(NOW)] = pos;
	agentSpeed[getIndex(NOW)] = Point(0,0);
      }
    } else if ( lastCommandType == CMD_Turn ) {
      const TurnCommand* const cmd = last_cmd->getConstTurnCommand();
      if ( agentView->getSpeed(NOW).isUnknown() )
	actualTurn[getIndex(NOW)] = cmd->getAngle(); // /(1 + parameterView->SP_inertia_moment*Vector(agentSpeed[getIndex(NOW)]).getLength());
      else
	actualTurn[getIndex(NOW)] = cmd->getAngle()/(1 + parameterView->SP_inertia_moment*FloatUnknown(agentView->getSpeed(NOW)).getValue());
    } else if ( lastCommandType == CMD_TurnNeck ) {
      const TurnNeckCommand* const cmd = last_cmd->getConstTurnNeckCommand();
      actualTurnNeck[getIndex(NOW)] += cmd->getAngle();
      agentNeckDirection[getIndex(NOW)] += cmd->getAngle();
    }

  } catch (RSException &excpt) {
    RS_FATAL("Caught an exception myUpdateAfterCommand: " << excpt.What());
  }

}


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "CommandView.icc"
#endif
