/* -*- Mode: C++ -*-
 * $Id: FrameworkAgent.cc,v 1.5 2000/10/24 15:15:33 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: FrameworkAgent.cc
 * Unitname: Framework
 * $Revision: 1.5 $
 * Created by: Fredrik Heintz 2000-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/10/24 15:15:33 $
 *
 *
 * HISTORY
 *
 * $Log: FrameworkAgent.cc,v $
 * Revision 1.5  2000/10/24 15:15:33  frehe
 * Fixed startWithCoach method.
 *
 * Revision 1.4  2000/10/11 18:39:31  frehe
 * Rearrange the startWithCoach method.
 *
 * Revision 1.3  2000/09/05 21:15:56  frehe
 * Made minor changes.
 *
 * Revision 1.2  2000/09/04 15:25:40  frehe
 * Fixed some minor bugs.
 *
 * Revision 1.1  2000/09/03 22:08:46  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file FrameworkAgent.cc
 * Implementation file for the class FramworkAgent.
 *
 * @par Unit:
 *      Framework
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2000/10/24 15:15:33 $
 */


#include "FrameworkAgent.h"

#include "ServerInterface.h"
#include "FrameworkSensorInterface.h"
#include "Sensors.h"
#include "ActuatorInterface.h"
#include "Actuators.h"
#include "Decision.h"
#include "ControlInterface.h"
#include "Controller.h"

#include "ViewManager.h"
#include "view_constants.h"
#include "ParameterView.h"
#include "GameView.h"
#include "CommandView.h"
#include "MarkerView.h"
#include "BallView.h"
#include "AgentView.h"
#include "PlayerView.h"
#include "CoachView.h"

#include "rs_log.h"


RS_BEGIN_NAMESPACE


FrameworkAgent::FrameworkAgent(Decision* _decision,
			       int argc, char** argv)
  : viewManager(ViewManager::instance()),
    parameterView(new ParameterView(PARAMETER_VIEW_ID,
				     argc, argv)),
    serverInterface(new ServerInterface(parameterView->SP_host,
					 parameterView->SP_port,
					 UInt32(parameterView->SP_buffer_size))),
    sensorInterface(new FrameworkSensorInterface(parameterView->CP_team_name)),
    sensors(new Sensors(sensorInterface)),
    actuatorInterface(new ActuatorInterface()),
    actuators(new Actuators(actuatorInterface)),
    decision(_decision),
    controlInterface(new ControlInterface(decision)),
    controller(new Controller(serverInterface, sensors,
			      actuators, controlInterface,
			      parameterView->SP_simulator_step,
			      parameterView->SP_recv_step,
			      parameterView->SP_sense_body_step,
			      parameterView->SP_send_step,
			      parameterView->CP_cycles_since_io_before_shutdown,
			      parameterView->CP_min_ticks_before_sending_command,
			      parameterView->CP_ticks_before_command_warning,
			      parameterView->CP_max_ticks_before_sending_command))
{
  View::sensorInterface = sensorInterface;
  viewManager->addView(parameterView, true);
};


FrameworkAgent::FrameworkAgent(Decision* _decision,
			       ParameterView* parameter_view)
  : viewManager(ViewManager::instance()),
    parameterView(parameter_view),
    serverInterface(new ServerInterface(parameterView->SP_host,
					parameterView->SP_port,
					UInt32(parameterView->SP_buffer_size))),
    sensorInterface(new FrameworkSensorInterface(parameterView->CP_team_name)),
    sensors(new Sensors(sensorInterface)),
    actuatorInterface(new ActuatorInterface()),
    actuators(new Actuators(actuatorInterface)),
    decision(_decision),
    controlInterface(new ControlInterface(decision)),
    controller(new Controller(serverInterface, sensors,
			      actuators, controlInterface,
			      parameterView->SP_simulator_step,
			      parameterView->SP_recv_step,
			      parameterView->SP_sense_body_step,
			      parameterView->SP_send_step,
			      parameterView->CP_cycles_since_io_before_shutdown,
			      parameterView->CP_min_ticks_before_sending_command,
			      parameterView->CP_ticks_before_command_warning,
			      parameterView->CP_max_ticks_before_sending_command))
{
  View::sensorInterface = sensorInterface;
  viewManager->addView(parameterView, true);
};


#if defined(RS_USE_SERVER_CONSTANTS) && defined(RS_USE_CLIENT_CONSTANTS)
FrameworkAgent::FrameworkAgent(Decision* _decision)
  : viewManager(ViewManager::instance()),
    parameterView(new ParameterView(PARAMETER_VIEW_ID)),
    serverInterface(new ServerInterface(parameterView->SP_host,
					parameterView->SP_port,
					UInt32(parameterView->SP_buffer_size))),
    sensorInterface(new FrameworkSensorInterface(parameterView->CP_team_name,
						 viewManager)),
    sensors(new Sensors(sensorInterface)),
    actuatorInterface(new ActuatorInterface()),
    actuators(new Actuators(actuatorInterface)),
    decision(_decision),
    controlInterface(new ControlInterface(decision)),
    controller(new Controller(serverInterface, sensors,
			      actuators, controlInterface,
			      parameterView->SP_simulator_step,
			      parameterView->SP_recv_step,
			      parameterView->SP_sense_body_step,
			      parameterView->SP_send_step,
			      parameterView->CP_cycles_since_io_before_shutdown,
			      parameterView->CP_min_ticks_before_sending_command,
			      parameterView->CP_ticks_before_command_warning,
			      parameterView->CP_max_ticks_before_sending_command))
{
  View::sensorInterface = sensorInterface;
  viewManager->addView(parameterView, true);
};
#endif


FrameworkAgent::~FrameworkAgent()
{
  delete controller;
  delete controlInterface;
  delete actuators;
  delete actuatorInterface;
  delete sensors;
  delete sensorInterface;
  delete serverInterface;

  viewManager->releaseView(PARAMETER_VIEW_ID);
}


bool FrameworkAgent::start()
{
  RS_LOG( LA_FRAMEWORK_AGENT, "Starting game view..." );
  GameView* gameView;
  getView(gameView, GAME_VIEW_ID);
  
  RS_LOG( LA_FRAMEWORK_AGENT, "Starting command view..." );
  CommandView* commandView;
  getView(commandView, COMMAND_VIEW_ID);
  
  RS_LOG( LA_FRAMEWORK_AGENT, "Starting marker view..." );
  MarkerView* markerView;
  getView(markerView, MARKER_VIEW_ID, parameterView);
  
  RS_LOG( LA_FRAMEWORK_AGENT, "Starting agent view..." );
  AgentView* agentView;
  getView(agentView, AGENT_VIEW_ID, parameterView);
  
  RS_LOG( LA_FRAMEWORK_AGENT, "Starting ball view..." );
  BallView* ballView;
  getView(ballView, BALL_VIEW_ID);
  
  RS_LOG( LA_FRAMEWORK_AGENT, "Starting player view..." );
  PlayerView* playerView;
  getView(playerView, PLAYER_VIEW_ID);
  
  RS_LOG( LA_FRAMEWORK_AGENT, "Initializing the controller..." );
  if ( !controller->init() ) {
    RS_FATAL("The controller could not be initialized!");
    return false;
  }
  
  RS_LOG( LA_FRAMEWORK_AGENT, "Initializing the decision maker..." );
  if ( !decision->init(actuatorInterface) ) {
    RS_FATAL("The decision maker could not be initialized!");
    return false;
  }
  
  RS_LOG( LA_FRAMEWORK_AGENT, "Start the controller!" );
  controller->start();

  viewManager->releaseView(PLAYER_VIEW_ID);
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(MARKER_VIEW_ID);
  viewManager->releaseView(COMMAND_VIEW_ID);
  viewManager->releaseView(GAME_VIEW_ID);
  
  return true;
}


bool FrameworkAgent::startWithCoach()
{
  RS_LOG( LA_FRAMEWORK_AGENT, "Starting game view..." );
  GameView* gameView;
  getView(gameView, GAME_VIEW_ID);
  
  RS_LOG( LA_FRAMEWORK_AGENT, "Starting command view..." );
  CommandView* commandView;
  getView(commandView, COMMAND_VIEW_ID);
  
  RS_LOG( LA_FRAMEWORK_AGENT, "Starting marker view..." );
  MarkerView* markerView;
  getView(markerView, MARKER_VIEW_ID, parameterView);
  
  RS_LOG( LA_FRAMEWORK_AGENT, "Starting agent view..." );
  AgentView* agentView;
  getView(agentView, AGENT_VIEW_ID, parameterView);
  
  RS_LOG( LA_FRAMEWORK_AGENT, "Starting ball view..." );
  BallView* ballView;
  getView(ballView, BALL_VIEW_ID);
  
  RS_LOG( LA_FRAMEWORK_AGENT, "Starting player view..." );
  PlayerView* playerView;
  getView(playerView, PLAYER_VIEW_ID);
  
  RS_LOG( LA_FRAMEWORK_AGENT, "Starting coach view..." );
  CoachView* coachView;
  getView(coachView, COACH_VIEW_ID);

  RS_LOG( LA_FRAMEWORK_AGENT, "Initializing the controller..." );
  if ( !controller->init() ) {
    RS_FATAL("The controller could not be initialized!");
    return false;
  }

  
  RS_LOG( LA_FRAMEWORK_AGENT, "Start the coach..." );
  if ( !controller->startCoach(parameterView->SP_host,
			       parameterView->SP_coach_port,
			       UInt32(parameterView->SP_buffer_size)) ) {
    RS_ERROR("Coach could not be started!");
  }

  RS_LOG( LA_FRAMEWORK_AGENT, "Send (eye on) to coach agent." );
  controller->sendCoachMessage("(eye on)");

  
  RS_LOG( LA_FRAMEWORK_AGENT, "Initializing the decision maker..." );
  if ( !decision->init(actuatorInterface) ) {
    RS_FATAL("The decision maker could not be initialized!");
    return false;
  }
  
  RS_LOG( LA_FRAMEWORK_AGENT, "Start the controller!" );
  controller->start();

  viewManager->releaseView(COACH_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(MARKER_VIEW_ID);
  viewManager->releaseView(COMMAND_VIEW_ID);
  viewManager->releaseView(GAME_VIEW_ID);
  
  return true;
}


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "FrameworkAgent.icc"
#endif
