/* -*- Mode: C++ -*-
 * $Id: MarkerView.cc,v 1.1 2000/09/21 13:46:47 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: MarkerView.cc
 * Unitname: Framework/view
 * $Revision: 1.1 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/21 13:46:47 $
 *
 *
 * HISTORY
 *
 * $Log: MarkerView.cc,v $
 * Revision 1.1  2000/09/21 13:46:47  frehe
 * Move the RoboSoc views from rsviews to the rsframework library.
 *
 * Revision 1.2  2000/09/13 00:08:09  frehe
 * Updated the views to inherit from the ViewWithHistory instead and also changed those views that used a static pointer to itself to do process the data in the sensor interface to use the new forEach methods implemented in the FrameworkSensorInterface.
 *
 * Revision 1.1  2000/09/04 15:27:24  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file MarkerView.cc
 * Implementation file for the class MarkerView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2000/09/21 13:46:47 $
 */


#include "MarkerView.h"
#include "ViewManager.h"
#include "ParameterView.h"
#include "CommandView.h"
#include "GameView.h"
#include "AgentView.h"
#include "view_constants.h"


RS_BEGIN_NAMESPACE


MarkerView::MarkerView(const ViewId id, 
		       ParameterView* const parameter_view,
		       const UInt hist_len)
  : ViewWithHistory(id, hist_len),
    parameterView(parameter_view),
    commandView(NULL),
    gameView(NULL),
    agentView(NULL),

    numberOfSeenLines(0),
    closestSeenLine(SL_UNKNOWN),

    lineL(hist_len, LineObject(SL_LEFT)),
    lineR(hist_len, LineObject(SL_RIGHT)),
    lineT(hist_len, LineObject(SL_TOP)),
    lineB(hist_len, LineObject(SL_BOTTOM)),
    unknownLine(SL_UNKNOWN),

    numberOfSeenMarkers(0),
    closestSeenMarker(UNKNOWN_MARKER),

    goalL(hist_len,
	  MarkerObject(GOAL_L,
		       PointUCO(-parameterView->SP_pitch_length/2.0, 0.0))),
    goalR(hist_len,
	  MarkerObject(GOAL_R,
		       PointUCO(parameterView->SP_pitch_length/2.0, 0.0))),
    flagC(hist_len,
	  MarkerObject(FLAG_C,
		       PointUCO(0.0, 0.0))),
    
    flagCT(hist_len,
	   MarkerObject(FLAG_CT,
			PointUCO(0.0, -parameterView->SP_pitch_width/2.0))),
    flagCB(hist_len,
	   MarkerObject(FLAG_CB,
			PointUCO(0.0, parameterView->SP_pitch_width/2.0))),
    flagLT(hist_len,
	   MarkerObject(FLAG_LT,
			PointUCO(-parameterView->SP_pitch_length/2.0,
				 -parameterView->SP_pitch_width/2.0))),
    flagLB(hist_len,
	   MarkerObject(FLAG_LB,
			PointUCO(-parameterView->SP_pitch_length/2.0,
				 parameterView->SP_pitch_width/2.0))),
    flagRT(hist_len,
	   MarkerObject(FLAG_RT,
			PointUCO(parameterView->SP_pitch_length/2.0,
				 -parameterView->SP_pitch_width/2.0))),
    flagRB(hist_len,
	   MarkerObject(FLAG_RB,
			PointUCO(parameterView->SP_pitch_length/2.0,
				 parameterView->SP_pitch_width/2.0))),
    flagPLT(hist_len,
	    MarkerObject(FLAG_PLT,
			 PointUCO(-parameterView->SP_pitch_length/2.0
				  +parameterView->SP_penalty_area_length,
				  -parameterView->SP_penalty_area_width/2.0))),
    flagPLC(hist_len,
	    MarkerObject(FLAG_PLC,
			  PointUCO(-parameterView->SP_pitch_length/2.0
				   +parameterView->SP_penalty_area_length, 0))),
    flagPLB(hist_len,
	    MarkerObject(FLAG_PLB,
			 PointUCO(-parameterView->SP_pitch_length/2.0
				  +parameterView->SP_penalty_area_length,
				  parameterView->SP_penalty_area_width/2.0))),
    flagPRT(hist_len,
	    MarkerObject(FLAG_PRT,
			 PointUCO(parameterView->SP_pitch_length/2.0-parameterView->SP_penalty_area_length,
				  -parameterView->SP_penalty_area_width/2.0))),
    flagPRC(hist_len,
	    MarkerObject(FLAG_PRC,
			 PointUCO(parameterView->SP_pitch_length/2.0-parameterView->SP_penalty_area_length, 0))),
    flagPRB(hist_len,
	    MarkerObject(FLAG_PRB,
			 PointUCO(parameterView->SP_pitch_length/2.0-parameterView->SP_penalty_area_length,
				  parameterView->SP_penalty_area_width/2.0))),
    flagGLT(hist_len,
	    MarkerObject(FLAG_GLT,
			 PointUCO(-parameterView->SP_pitch_length/2.0,
				  -parameterView->SP_goal_width/2.0 ))),
    flagGLB(hist_len,
	    MarkerObject(FLAG_GLB,
			 PointUCO(-parameterView->SP_pitch_length/2.0,
				  parameterView->SP_goal_width/2.0 ))),
    flagGRT(hist_len,
	    MarkerObject(FLAG_GRT,
			 PointUCO(parameterView->SP_pitch_length/2.0,
				  -parameterView->SP_goal_width/2.0 ))),
    flagGRB(hist_len,
	    MarkerObject(FLAG_GRB,
			 PointUCO(parameterView->SP_pitch_length/2.0,
				  parameterView->SP_goal_width/2.0 ))),
    
    flagTL50(hist_len,
	     MarkerObject(FLAG_TL50,
			  PointUCO(-50.0,
				   -parameterView->SP_pitch_width/2.0-parameterView->SP_pitch_margin))),
    flagTL40(hist_len,
	     MarkerObject(FLAG_TL40, 
			  PointUCO(-40.0,
				     -parameterView->SP_pitch_width/2.0-parameterView->SP_pitch_margin))),
    flagTL30(hist_len,
	     MarkerObject(FLAG_TL30,
			   PointUCO(-30.0,
				    -parameterView->SP_pitch_width/2.0-parameterView->SP_pitch_margin))),
    flagTL20(hist_len,
	     MarkerObject(FLAG_TL20,
			  PointUCO(-20.0,
				   -parameterView->SP_pitch_width/2.0-parameterView->SP_pitch_margin))),
    flagTL10(hist_len,
	     MarkerObject(FLAG_TL10,
			  PointUCO(-10.0,
				   -parameterView->SP_pitch_width/2.0-parameterView->SP_pitch_margin))),
    flagT0(hist_len,
	   MarkerObject(FLAG_T0,
			PointUCO(0.0,
				 -parameterView->SP_pitch_width/2.0-parameterView->SP_pitch_margin))),
    flagTR10(hist_len,
	     MarkerObject(FLAG_TR10,
			  PointUCO(10.0,
				   -parameterView->SP_pitch_width/2.0-parameterView->SP_pitch_margin))),
    flagTR20(hist_len,
	     MarkerObject(FLAG_TR20,
			  PointUCO(20.0,
				   -parameterView->SP_pitch_width/2.0-parameterView->SP_pitch_margin))),
    flagTR30(hist_len,
	     MarkerObject(FLAG_TR30,
			  PointUCO(30.0,
				   -parameterView->SP_pitch_width/2.0-parameterView->SP_pitch_margin))),
    flagTR40(hist_len,
	     MarkerObject(FLAG_TR40,
			  PointUCO(40.0,
				   -parameterView->SP_pitch_width/2.0-parameterView->SP_pitch_margin))),
    flagTR50(hist_len,
	     MarkerObject(FLAG_TR50,
			  PointUCO(50.0,
				   -parameterView->SP_pitch_width/2.0-parameterView->SP_pitch_margin))),
    
    flagBL50(hist_len,
	     MarkerObject(FLAG_BL50,
			   PointUCO(-50.0,
				    parameterView->SP_pitch_width/2.0+parameterView->SP_pitch_margin))),
    flagBL40(hist_len,
	     MarkerObject(FLAG_BL40,
			  PointUCO(-40.0,
				   parameterView->SP_pitch_width/2.0+parameterView->SP_pitch_margin))),
    flagBL30(hist_len,
	     MarkerObject(FLAG_BL30,
			  PointUCO(-30.0,
				   parameterView->SP_pitch_width/2.0+parameterView->SP_pitch_margin))),
    flagBL20(hist_len,
	     MarkerObject(FLAG_BL20,
			  PointUCO(-20.0,
				   parameterView->SP_pitch_width/2.0+parameterView->SP_pitch_margin))),
    flagBL10(hist_len,
	     MarkerObject(FLAG_BL10,
			  PointUCO(-10.0,
				   parameterView->SP_pitch_width/2.0+parameterView->SP_pitch_margin))),
    flagB0(hist_len,
	     MarkerObject(FLAG_B0,
			  PointUCO(0.0,
				   parameterView->SP_pitch_width/2.0+parameterView->SP_pitch_margin))),
    flagBR10(hist_len,
	     MarkerObject(FLAG_BR10,
			  PointUCO(10.0,
				   parameterView->SP_pitch_width/2.0+parameterView->SP_pitch_margin))),
    flagBR20(hist_len,
	     MarkerObject(FLAG_BR20,
			  PointUCO(20.0,
				   parameterView->SP_pitch_width/2.0+parameterView->SP_pitch_margin))),
    flagBR30(hist_len,
	     MarkerObject(FLAG_BR30,
			  PointUCO(30.0,
				   parameterView->SP_pitch_width/2.0+parameterView->SP_pitch_margin))),
    flagBR40(hist_len,
	     MarkerObject(FLAG_BR40,
			  PointUCO(40.0,
				   parameterView->SP_pitch_width/2.0+parameterView->SP_pitch_margin))),
    flagBR50(hist_len,
	     MarkerObject(FLAG_BR50,
			  PointUCO(50.0,
				   parameterView->SP_pitch_width/2.0+parameterView->SP_pitch_margin))),
    
    flagLT30(hist_len,
	     MarkerObject(FLAG_LT30,
			  PointUCO(-parameterView->SP_pitch_length/2.0-parameterView->SP_pitch_margin,
				   -30))),
    flagLT20(hist_len,
	     MarkerObject(FLAG_LT20,
			  PointUCO(-parameterView->SP_pitch_length/2.0-parameterView->SP_pitch_margin,
				     -20))),
    flagLT10(hist_len,
	     MarkerObject(FLAG_LT10,
			  PointUCO(-parameterView->SP_pitch_length/2.0-parameterView->SP_pitch_margin,
				   -10))),
    flagL0(hist_len,
	   MarkerObject(FLAG_L0,
			PointUCO(-parameterView->SP_pitch_length/2.0-parameterView->SP_pitch_margin,
				 0))),
    flagLB10(hist_len,
	     MarkerObject(FLAG_LB10,
			  PointUCO(-parameterView->SP_pitch_length/2.0-parameterView->SP_pitch_margin,
				     10))),
    flagLB20(hist_len,
	     MarkerObject(FLAG_LB20,
			  PointUCO(-parameterView->SP_pitch_length/2.0-parameterView->SP_pitch_margin,
				     20))),
    flagLB30(hist_len,
	     MarkerObject(FLAG_LB30,
			  PointUCO(-parameterView->SP_pitch_length/2.0-parameterView->SP_pitch_margin,
				   30))),
    
    flagRT30(hist_len,
	     MarkerObject(FLAG_RT30,
			  PointUCO(parameterView->SP_pitch_length/2.0+parameterView->SP_pitch_margin,
				   -30))),
    flagRT20(hist_len,
	     MarkerObject(FLAG_RT20,
			  PointUCO(parameterView->SP_pitch_length/2.0+parameterView->SP_pitch_margin,
				   -20))),
    flagRT10(hist_len,
	     MarkerObject(FLAG_RT10,
			  PointUCO(parameterView->SP_pitch_length/2.0+parameterView->SP_pitch_margin,
				   -10))),
    flagR0(hist_len,
	   MarkerObject(FLAG_R0,
			PointUCO(parameterView->SP_pitch_length/2.0+parameterView->SP_pitch_margin,
				 0))),
    flagRB10(hist_len,
	     MarkerObject(FLAG_RB10,
			  PointUCO(parameterView->SP_pitch_length/2.0+parameterView->SP_pitch_margin,
				   10))),
    flagRB20(hist_len,
	     MarkerObject(FLAG_RB20,
			  PointUCO(parameterView->SP_pitch_length/2.0+parameterView->SP_pitch_margin,
				   20))),
    flagRB30(hist_len,
	     MarkerObject(FLAG_RB30,
			  PointUCO(parameterView->SP_pitch_length/2.0+parameterView->SP_pitch_margin,
				   30))),
    
    unknownMarker(UNKNOWN_MARKER, PointUCO(), VectorUCO())
{
  // This is just to increase the reference count for the parameter view.
  viewManager->getView(PARAMETER_VIEW_ID);

  getView(commandView, COMMAND_VIEW_ID, hist_len);
  getView(gameView, GAME_VIEW_ID, hist_len);
}


MarkerView::~MarkerView()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(COMMAND_VIEW_ID);
  viewManager->releaseView(GAME_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
}


bool MarkerView::init()
{
  getView(agentView, AGENT_VIEW_ID, parameterView, getHistoryLength());
  initialized = true;
  return true;
}


const LineObject& MarkerView::getLine(const SideLineName line,
				      const Int hist) const
{
  switch (line) {
    case SL_LEFT   : return lineL[getIndex(hist)]; break;
    case SL_RIGHT  : return lineR[getIndex(hist)]; break;
    case SL_TOP    : return lineT[getIndex(hist)]; break;
    case SL_BOTTOM : return lineB[getIndex(hist)]; break;
      
    default:
    return unknownLine;
  }
}


LineObject& MarkerView::getLine(const SideLineName line, const Int hist)
{
  switch (line) {
    case SL_LEFT   : return lineL[getIndex(hist)];
    case SL_RIGHT  : return lineR[getIndex(hist)];
    case SL_TOP    : return lineT[getIndex(hist)];
    case SL_BOTTOM : return lineB[getIndex(hist)];
      
    default:
    return unknownLine;
  }
}


const LineObject& MarkerView::getLine(const Int line,
				      const Int hist) const
{
  if ( line >= 0 && line < parameterView->SP_num_lines )
    return getLine(SideLineName(line), hist);
  else
    return unknownLine;
}


const MarkerObject& MarkerView::getMarker(const MarkerName marker,
					  const Int hist) const
{
  switch (marker) {
    case GOAL_L    : return goalL[getIndex(hist)];
    case GOAL_R    : return goalR[getIndex(hist)];

    case FLAG_C    : return flagC[getIndex(hist)];
    case FLAG_CT   : return flagCT[getIndex(hist)];
    case FLAG_CB   : return flagCB[getIndex(hist)];
    case FLAG_LT   : return flagLT[getIndex(hist)];
    case FLAG_LB   : return flagLB[getIndex(hist)];
    case FLAG_RT   : return flagRT[getIndex(hist)];
    case FLAG_RB   : return flagRB[getIndex(hist)];

    case FLAG_PLT  : return flagPLT[getIndex(hist)];
    case FLAG_PLC  : return flagPLC[getIndex(hist)];
    case FLAG_PLB  : return flagPLB[getIndex(hist)];
    case FLAG_PRT  : return flagPRT[getIndex(hist)];
    case FLAG_PRC  : return flagPRC[getIndex(hist)];
    case FLAG_PRB  : return flagPRB[getIndex(hist)];

    case FLAG_GLT  : return flagGLT[getIndex(hist)];
    case FLAG_GLB  : return flagGLB[getIndex(hist)];
    case FLAG_GRT  : return flagGRT[getIndex(hist)];
    case FLAG_GRB  : return flagGRB[getIndex(hist)];

    case FLAG_TL50 : return flagTL50[getIndex(hist)];
    case FLAG_TL40 : return flagTL40[getIndex(hist)];
    case FLAG_TL30 : return flagTL30[getIndex(hist)];
    case FLAG_TL20 : return flagTL20[getIndex(hist)];
    case FLAG_TL10 : return flagTL10[getIndex(hist)];
    case FLAG_T0   : return flagT0[getIndex(hist)];
    case FLAG_TR10 : return flagTR10[getIndex(hist)];
    case FLAG_TR20 : return flagTR20[getIndex(hist)];
    case FLAG_TR30 : return flagTR30[getIndex(hist)];
    case FLAG_TR40 : return flagTR40[getIndex(hist)];
    case FLAG_TR50 : return flagTR50[getIndex(hist)];

    case FLAG_BL50 : return flagBL50[getIndex(hist)];
    case FLAG_BL40 : return flagBL40[getIndex(hist)];
    case FLAG_BL30 : return flagBL30[getIndex(hist)];
    case FLAG_BL20 : return flagBL20[getIndex(hist)];
    case FLAG_BL10 : return flagBL10[getIndex(hist)];
    case FLAG_B0   : return flagB0[getIndex(hist)];
    case FLAG_BR10 : return flagBR10[getIndex(hist)];
    case FLAG_BR20 : return flagBR20[getIndex(hist)];
    case FLAG_BR30 : return flagBR30[getIndex(hist)];
    case FLAG_BR40 : return flagBR40[getIndex(hist)];
    case FLAG_BR50 : return flagBR50[getIndex(hist)];

    case FLAG_LT30 : return flagLT30[getIndex(hist)];
    case FLAG_LT20 : return flagLT20[getIndex(hist)];
    case FLAG_LT10 : return flagLT10[getIndex(hist)];
    case FLAG_L0   : return flagL0[getIndex(hist)];
    case FLAG_LB10 : return flagLB10[getIndex(hist)];
    case FLAG_LB20 : return flagLB20[getIndex(hist)];
    case FLAG_LB30 : return flagLB30[getIndex(hist)];

    case FLAG_RT30 : return flagRT30[getIndex(hist)];
    case FLAG_RT20 : return flagRT20[getIndex(hist)];
    case FLAG_RT10 : return flagRT10[getIndex(hist)];
    case FLAG_R0   : return flagR0[getIndex(hist)];
    case FLAG_RB10 : return flagRB10[getIndex(hist)];
    case FLAG_RB20 : return flagRB20[getIndex(hist)];
    case FLAG_RB30 : return flagRB30[getIndex(hist)];
      
    default        : return unknownMarker;
  }
}


MarkerObject& MarkerView::getMarker(const MarkerName marker, const Int hist)
{
  switch (marker) {
    case GOAL_L    : return goalL[getIndex(hist)];
    case GOAL_R    : return goalR[getIndex(hist)];

    case FLAG_C    : return flagC[getIndex(hist)];
    case FLAG_CT   : return flagCT[getIndex(hist)];
    case FLAG_CB   : return flagCB[getIndex(hist)];
    case FLAG_LT   : return flagLT[getIndex(hist)];
    case FLAG_LB   : return flagLB[getIndex(hist)];
    case FLAG_RT   : return flagRT[getIndex(hist)];
    case FLAG_RB   : return flagRB[getIndex(hist)];

    case FLAG_PLT  : return flagPLT[getIndex(hist)];
    case FLAG_PLC  : return flagPLC[getIndex(hist)];
    case FLAG_PLB  : return flagPLB[getIndex(hist)];
    case FLAG_PRT  : return flagPRT[getIndex(hist)];
    case FLAG_PRC  : return flagPRC[getIndex(hist)];
    case FLAG_PRB  : return flagPRB[getIndex(hist)];

    case FLAG_GLT  : return flagGLT[getIndex(hist)];
    case FLAG_GLB  : return flagGLB[getIndex(hist)];
    case FLAG_GRT  : return flagGRT[getIndex(hist)];
    case FLAG_GRB  : return flagGRB[getIndex(hist)];

    case FLAG_TL50 : return flagTL50[getIndex(hist)];
    case FLAG_TL40 : return flagTL40[getIndex(hist)];
    case FLAG_TL30 : return flagTL30[getIndex(hist)];
    case FLAG_TL20 : return flagTL20[getIndex(hist)];
    case FLAG_TL10 : return flagTL10[getIndex(hist)];
    case FLAG_T0   : return flagT0[getIndex(hist)];
    case FLAG_TR10 : return flagTR10[getIndex(hist)];
    case FLAG_TR20 : return flagTR20[getIndex(hist)];
    case FLAG_TR30 : return flagTR30[getIndex(hist)];
    case FLAG_TR40 : return flagTR40[getIndex(hist)];
    case FLAG_TR50 : return flagTR50[getIndex(hist)];

    case FLAG_BL50 : return flagBL50[getIndex(hist)];
    case FLAG_BL40 : return flagBL40[getIndex(hist)];
    case FLAG_BL30 : return flagBL30[getIndex(hist)];
    case FLAG_BL20 : return flagBL20[getIndex(hist)];
    case FLAG_BL10 : return flagBL10[getIndex(hist)];
    case FLAG_B0   : return flagB0[getIndex(hist)];
    case FLAG_BR10 : return flagBR10[getIndex(hist)];
    case FLAG_BR20 : return flagBR20[getIndex(hist)];
    case FLAG_BR30 : return flagBR30[getIndex(hist)];
    case FLAG_BR40 : return flagBR40[getIndex(hist)];
    case FLAG_BR50 : return flagBR50[getIndex(hist)];

    case FLAG_LT30 : return flagLT30[getIndex(hist)];
    case FLAG_LT20 : return flagLT20[getIndex(hist)];
    case FLAG_LT10 : return flagLT10[getIndex(hist)];
    case FLAG_L0   : return flagL0[getIndex(hist)];
    case FLAG_LB10 : return flagLB10[getIndex(hist)];
    case FLAG_LB20 : return flagLB20[getIndex(hist)];
    case FLAG_LB30 : return flagLB30[getIndex(hist)];

    case FLAG_RT30 : return flagRT30[getIndex(hist)];
    case FLAG_RT20 : return flagRT20[getIndex(hist)];
    case FLAG_RT10 : return flagRT10[getIndex(hist)];
    case FLAG_R0   : return flagR0[getIndex(hist)];
    case FLAG_RB10 : return flagRB10[getIndex(hist)];
    case FLAG_RB20 : return flagRB20[getIndex(hist)];
    case FLAG_RB30 : return flagRB30[getIndex(hist)];
      
    default        : return unknownMarker;
  }
}


const MarkerObject& MarkerView::getMarker(const Int marker,
					  const Int hist) const
{
  if ( marker >= 0 && marker < parameterView->SP_num_markers )
    return getMarker(MarkerName(int(marker)), hist);
  else
    return unknownMarker;
}


MarkerObject& MarkerView::getClosestMarkerTo(const PointUCO& pt, Int hist)
{
  FloatUCO dist;
  FloatUCO tmp;
  UInt marker = parameterView->SP_num_markers; // unknownMarker
  UInt i;
  dist = 9999;
  for (i = 0; i < parameterView->SP_num_markers; ++i) {
    if ( gameView->getOurSide() == S_RIGHT )
      tmp = pt.getDistanceTo(-getMarker(MarkerName(i), hist).getPosition());
    else
      tmp = pt.getDistanceTo(getMarker(MarkerName(i), hist).getPosition());
    if ( tmp < dist ) {
      marker = i;
      dist = tmp;
    }
  }
  return getMarker(MarkerName(marker), hist);
}


const MarkerObject& MarkerView::getClosestMarkerTo(const PointUCO& pt,
						   Int hist) const
{
  FloatUCO dist;
  FloatUCO tmp;
  UInt marker = parameterView->SP_num_markers; // unknownMarker
  UInt i;
  dist = 9999;
  for (i = 0; i < parameterView->SP_num_markers; ++i) {
    if ( gameView->getOurSide() == S_RIGHT )
      tmp = pt.getDistanceTo(-getMarker(MarkerName(i), hist).getPosition());
    else
      tmp = pt.getDistanceTo(getMarker(MarkerName(i), hist).getPosition());
    if ( tmp < dist ) {
      marker = i;
      dist = tmp;
    }
  }
  return getMarker(MarkerName(marker), hist);
}


const MarkerObject& MarkerView::getClosestGoalTo(const PointUCO& pt,
						 Int hist) const
{
  // Should be enough to check the x-coordinate.
  // But then I need data from the agent view.
  if ( pt.getDistanceTo(getTheirGoal(hist).getPosition()) >
       pt.getDistanceTo(getOurGoal(hist).getPosition()) )
    return getOurGoal(hist);
  else
    return getTheirGoal(hist);
}


const MarkerObject& MarkerView::getTheirGoal(const Int hist) const
{
  //if( gameView->getOurSide() == S_Left )
    return goalR[getIndex(hist)];
    /*
  else
  return goalL[getIndex(hist)];
  */
} 


const MarkerObject& MarkerView::getOurGoal(const Int hist) const
{
  /*
  if( gameView->getOurSide() == S_Right )
    return goalR[getIndex(hist)];
  else
  */
    return goalL[getIndex(hist)];
}


void MarkerView::rotateAllMarkers()
{
  UInt j;
  for (UInt i = 0; i < parameterView->SP_num_markers; ++i)
    for (j = 0; j < getHistoryLength(); ++j)
    getMarker(MarkerName(i), j).rotate();
}


void MarkerView::myNewCycle()
{
  //commandView->NewCycle(currentUpdateNumber);
  //agentView->NewCycle(currentUpdateNumber);
  
  //static t_side prev_side = S_Left;

  // Update the current observation time.
  increaseIndex();

  /*
  if ( prev_side != gameView->getOurSide() ) {
    //cout << gameView->getCurrentTime() << "Rotate all markers!\n";
    rotateAllMarkers();
    prev_side = gameView->getOurSide();
  }
  */
  
  // Set the number of seen lines to zero
  //numberOfLines = 0;

  // Set the number of seen markers to zero
  //numberOfSeenMarkers = 0;
  
  // get the movement vector and actual turn for the agent from
  // commandView and add it to all views' vectors.
  /*
  for (int i = 0; i < parameterView->SP_num_markers; ++i) {
    getMarker(MarkerName(i), NOW).SetVector(agentView->getAgentObject(NOW).getRelativeVectorTo(getMarker(MarkerName(i), NOW).getPosition()));
  }
  */
}


void MarkerView::processMarker(const ServerMarkerObject& marker)
{
  static FloatUnknown closest_dist;

  if ( !initialized )
    init();
  
  if ( numberOfSeenMarkers == 0 )
    closest_dist.makeUnknown();
  
  if ( marker.behind ) {
    // This does not work if we just made a move!!!
    MarkerObject& marker_ref = getClosestMarkerTo(agentView->getAgentPosition(NOW), NOW);

    //cout << "Used a marker behind " << marker_ref.getMarkerId() << endl;
    marker_ref.setDistance(marker.distance);
    marker_ref.setDirection(marker.direction);
    marker_ref.setBehind(true);

    if ( marker.distance.isKnown() &&
	 marker.distance < closest_dist &&
	 closest_dist.isUnknown() ) {
      closestSeenMarker = MarkerObject(marker_ref);
      closest_dist = marker.distance;
    }
    
    numberOfSeenMarkers++;

  } else {
    MarkerObject& marker_ref = getMarker(marker.markerId, NOW);
    marker_ref.setDistance(marker.distance);
    marker_ref.setDirection(marker.direction);
    marker_ref.setBehind(false);

    if ( (marker.distance.isKnown() &&
	  marker.distance < closest_dist) ||
	 closestSeenMarker.isBehind() ) {
      closestSeenMarker = MarkerObject(marker_ref);
      closest_dist = marker.distance;
    }
    
    numberOfSeenMarkers++;
  }
}


void MarkerView::processLine(const ServerLineObject& line)
{
  static FloatUCO dist;

  dist = line.distance;
  
  LineObject& line_ref = getLine(line.lineId, NOW);

  line_ref.setDistance(line.distance);
  line_ref.setDirection(line.direction);
  
  if ( ( numberOfSeenLines == 0 ) ||
       ( closestSeenLine.getDistance() < dist ) ) {
    closestSeenLine = LineObject(line_ref);
  }

  /*
  cout << "Line h0: " << getLine(line.line_id, 0)
       << "\nLine h1: " << getLine(line.line_id, 1)
       << "\nLine h2: " << getLine(line.line_id, 2)
       << endl;
  
  cout << "Saw line " << line.line_id << " matched it with "
       << line_ref.getLineId() << " closest line is "
       << closestSeenLine << endl;
       */
  
  numberOfSeenLines++;
}


void MarkerView::myUpdateAfterSee()
{
  // Go through the markers list and update all
  // markers that are seen,
  numberOfSeenMarkers = 0;
  sensorInterface->forEachMarker(&View::processMarker, this);

  // Go through all the lines
  numberOfSeenLines = 0;
  sensorInterface->forEachLine(&View::processLine, this);
}


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "MarkerView.icc"
#endif
