/* -*- Mode: C++ -*-
 * $Id: PlayerView.cc,v 1.4 2001/01/11 22:00:20 olalu526 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: PlayerView.cc
 * Unitname: Framework/view
 * $Revision: 1.4 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: olalu526 $ $Date: 2001/01/11 22:00:20 $
 *
 *
 * HISTORY
 *
 * $Log: PlayerView.cc,v $
 * Revision 1.4  2001/01/11 22:00:20  olalu526
 * Added the real world model with support for new simulation steps.
 *
 * Revision 1.3  2001/01/10 15:12:59  olalu526
 * Added support for WorldModel to some basic classes.
 *
 * Revision 1.2  2000/09/25 09:48:57  frehe
 * Fixed a bug in the CoachView.
 *
 * Revision 1.1  2000/09/21 13:46:48  frehe
 * Move the RoboSoc views from rsviews to the rsframework library.
 *
 * Revision 1.2  2000/09/13 00:08:09  frehe
 * Updated the views to inherit from the ViewWithHistory instead and also changed those views that used a static pointer to itself to do process the data in the sensor interface to use the new forEach methods implemented in the FrameworkSensorInterface.
 *
 * Revision 1.1  2000/09/04 15:27:25  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file PlayerView.cc
 * Implementation file for the class PlayerView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.4 $
 * $Author: olalu526 $
 * $Date: 2001/01/11 22:00:20 $
 */


#include "PlayerView.h"
#include "ViewManager.h"
#include "ParameterView.h"
#include "CommandView.h"
#include "AgentView.h"
#include "FrameworkSensorInterface.h"
#include "view_constants.h"


RS_BEGIN_NAMESPACE


PlayerView::PlayerView(const ViewId id,
		       const UInt hist_len)
  : ViewWithSimulation(id),
    parameterView(NULL),
    commandView(NULL),
    agentView(NULL)
{
  worldModel->setOpponentsHistoryLength(hist_len);
  worldModel->setTeammatesHistoryLength(hist_len);
  worldModel->setUnknownPlayersHistoryLength(hist_len);
  worldModel->setClosestOpponentIndexHistoryLength(hist_len);
  worldModel->setClosestTeammateIndexHistoryLength(hist_len);
  worldModel->setClosestUnknownPlayerIndexHistoryLength(hist_len);
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(commandView, COMMAND_VIEW_ID, hist_len);
  getView(agentView, AGENT_VIEW_ID, parameterView, hist_len);
}


PlayerView::~PlayerView()
{
  viewManager->releaseView(parameterView->getId());
  viewManager->releaseView(commandView->getId());
  viewManager->releaseView(agentView->getId());
}


void PlayerView::processPlayer(const ServerPlayerObject& plr)
{
  static FloatUnknown closest_uplr;
  static FloatUnknown closest_opp;
  static FloatUnknown closest_mate;
  
  if ( getClosestOpponent(NOW) == NULL )
    closest_opp = 9999;

  if ( getClosestTeammate(NOW) == NULL )
    closest_mate = 9999;

  if ( getClosestUnknownPlayer(NOW) == NULL )
    closest_uplr = 9999;

  // Need to adjust the body and neck directions, since they
  // are relative to my own body and neck directions!
  // Martin: hack?!?
  AngleDegUnknown myFaceDir = agentView->getFaceDirection(NOW);
  PlayerObject player(plr.team,
		      plr.side,
		      plr.uniformNumber,
		      AngleDegUnknown(),
		      AngleDegUnknown(),
		      PointUCO(agentView->getAgentPosition(NOW) +
			       VectorUCO(plr.direction+myFaceDir, plr.distance)),
		      VectorUCO(plr.direction+myFaceDir, plr.distance),
		      plr.distanceChange,
		      plr.directionChange,
		      VectorUCO());

  // TODO:
  // The control of the vector size below is because a bug in
  // the SensorInterface or the calling of it.  When the
  // time is stopped (I think) the methid
  // SensorInterface::updateBeforeNewCycle() is never called
  // which makes the vector CBasicSensorInterface::players grow.
  // Workaround by martin.
  switch ( plr.team ) {
  case T_OUR_TEAM:
    if (theTeammates[worldModel->getTeammatesIndex(NOW)].size() <= 9)
      theTeammates[worldModel->getTeammatesIndex(NOW)].push_back(player);
    break;
  case T_THEIR_TEAM:
    if (theOpponents[worldModel->getOpponentsIndex(NOW)].size() <= 10)
      theOpponents[worldModel->getOpponentsIndex(NOW)].push_back(player);
    break;
  case T_UNKNOWN:
    if (theUnknownPlayers[worldModel->getUnknownPlayersIndex(NOW)].size() <= 20)
      theUnknownPlayers[worldModel->getUnknownPlayersIndex(NOW)].push_back(player);
  }

  if ( plr.team == T_UNKNOWN &&
       plr.distance < closest_uplr ) {
    //cout << "Found new closest unknown player, no " << player.getUniformNumber() << endl;
    setClosestUnknownPlayer(player);
    closest_uplr = plr.distance;
  }

  if ( plr.team == T_THEIR_TEAM &&
       plr.distance < closest_opp ) {
    //cout << "Found new closest opponent, no " << player.getUniformNumber() << endl;
    setClosestOpponent(player);
    closest_opp = plr.distance;
  }

  if ( plr.team == T_OUR_TEAM &&
       plr.distance < closest_mate ) {
    //cout << "Found new closest teammate, no " << player.getUniformNumber() << endl;
    setClosestTeammate(player);
    closest_mate = plr.distance;
  }
}


PlayerObject* PlayerView::getClosestPlayer(const Int hist)
{
  FloatUnknown dist_opp;
  FloatUnknown dist_tmt;
  FloatUnknown dist_upl;

  if ( getClosestOpponent(hist) != NULL )
    dist_opp = getClosestOpponent(hist)->getDistance();
  if ( getClosestTeammate(hist) != NULL )
    dist_tmt = getClosestTeammate(hist)->getDistance();
  if ( getClosestUnknownPlayer(hist) != NULL )
    dist_upl = getClosestUnknownPlayer(hist)->getDistance();

  if ( dist_opp.isKnown() || dist_tmt.isKnown() || dist_upl.isKnown() ) {
    if ( dist_opp > dist_tmt )
      if ( dist_tmt > dist_upl )
	return getClosestUnknownPlayer(hist);
      else
	return getClosestTeammate(hist);
    else
      if ( dist_opp > dist_upl )
	return getClosestUnknownPlayer(hist);
      else
	return getClosestOpponent(hist);
  } else
    return NULL;
  
}


void PlayerView::setClosestUnknownPlayer(const PlayerObject&)
{
  theClosestUnknownPlayerIndex[worldModel->getClosestUnknownPlayerIndexIndex(NOW)] = theUnknownPlayers[worldModel->getUnknownPlayersIndex(NOW)].size() - 1;
}


void PlayerView::setClosestOpponent(const PlayerObject&)
{
  theClosestOpponentIndex[worldModel->getClosestOpponentIndexIndex(NOW)] = theOpponents[worldModel->getOpponentsIndex(NOW)].size() - 1;
}


void PlayerView::setClosestTeammate(const PlayerObject&)
{
  theClosestTeammateIndex[worldModel->getClosestTeammateIndexIndex(NOW)] = theTeammates[worldModel->getTeammatesIndex(NOW)].size() - 1;
}


void PlayerView::myNewCycle()
{
  
  commandView->newCycle(currentUpdateNumber);
  agentView->newCycle(currentUpdateNumber);

  /*
  // Update the direction to all players.
  // Start with the closest players.
  if ( theClosestOpponent != NULL )
    theClosestOpponent->AddDirection(commandView->getActualTurn(PREV));
  if ( theClosestTeammate != NULL )
    theClosestTeammate->AddDirection(commandView->getActualTurn(PREV));
  if ( theClosestPlayer != NULL )
    theClosestPlayer->AddDirection(commandView->getActualTurn(PREV));

  // Then all the stored players.
  std::vector<PlayerObject>::iterator p;
  for ( p = theOpponents.begin(); p != theOpponents.end(); ++p )
    p->AddDirection(commandView->getActualTurn(PREV));
  for ( p = theTeammates.begin(); p != theTeammates.end(); ++p )
    p->AddDirection(commandView->getActualTurn(PREV));
  for ( p = theUnknownPlayers.begin(); p != theUnknownPlayers.end(); ++p )
    p->AddDirection(commandView->getActualTurn(PREV));
    */

  // Copy last cycles players to this cycle
  theOpponents[worldModel->getOpponentsIndex(NEXT)] =
    theOpponents[worldModel->getOpponentsIndex(NOW)];
  theTeammates[worldModel->getTeammatesIndex(NEXT)] =
    theTeammates[worldModel->getTeammatesIndex(NOW)];
  theUnknownPlayers[worldModel->getUnknownPlayersIndex(NEXT)] =
    theUnknownPlayers[worldModel->getUnknownPlayersIndex(NOW)];

  theClosestOpponentIndex[worldModel->getClosestOpponentIndexIndex(NEXT)] =
    theClosestOpponentIndex[worldModel->getClosestOpponentIndexIndex(NOW)];
  theClosestTeammateIndex[worldModel->getClosestTeammateIndexIndex(NEXT)] =
    theClosestTeammateIndex[worldModel->getClosestTeammateIndexIndex(NOW)];
  theClosestUnknownPlayerIndex[worldModel->getClosestUnknownPlayerIndexIndex(NEXT)] =
    theClosestUnknownPlayerIndex[worldModel->getClosestUnknownPlayerIndexIndex(NOW)];
  
  // Update the vector to the players.
  /* These do not have to be done since they are only pointers to the reak
     objects and are therefore also updated below.
  // Start with the closest players.
  if ( getClosestOpponent(NOW) != NULL ) {
    getClosestOpponent(NEXT)->SetVector(getClosestOpponent(NEXT)->getVector() -
					commandView->getAgentMovement(PREV));
    //SetVector(agentView->getAgentObject(NOW).getRelativeVectorTo(getClosestOpponent(NEXT)->getPosition()));
    getClosestOpponent(NEXT)->SetDirection(getClosestOpponent(NEXT)->getDirection() -
					   commandView->getActualTurn(PREV));
  }
  if ( getClosestTeammate(NEXT) != NULL ) {
    getClosestTeammate(NEXT)->SetVector(getClosestTeammate(NEXT)->getVector() -
					commandView->getAgentMovement(PREV));
    //SetVector(agentView->getAgentObject(NOW).getRelativeVectorTo(getClosestTeammate(NEXT)->getPosition()));
    getClosestTeammate(NEXT)->SetDirection(getClosestTeammate(NEXT)->getDirection() -
					   commandView->getActualTurn(PREV));
  }
  if ( getClosestPlayer(NEXT) != NULL ) {
    getClosestPlayer(NEXT)->SetVector(getClosestPlayer(NEXT)->getVector() -
				      commandView->getAgentMovement(PREV));
    //SetVector(agentView->getAgentObject(NOW).getRelativeVectorTo(getClosestPlayer(NEXT)->getPosition()));
    getClosestPlayer(NEXT)->SetDirection(getClosestPlayer(NEXT)->getDirection() -
					 commandView->getActualTurn(PREV));
  }
  */

  // Then all the stored players.
  std::vector<PlayerObject>::iterator p;
  for ( p = theOpponents[worldModel->getOpponentsIndex(NEXT)].begin();
	p != theOpponents[worldModel->getOpponentsIndex(NEXT)].end(); ++p ) {
    p->setVector(p->getVector() - commandView->getAgentMovement(PREV));
    //setVector(agentView->getAgentObject(NOW).getRelativeVectorTo(p->getPosition()));
    p->setDirection(p->getDirection() - commandView->getActualTurn(PREV));
  }
  for ( p = theTeammates[worldModel->getTeammatesIndex(NEXT)].begin();
	p != theTeammates[worldModel->getTeammatesIndex(NEXT)].end(); ++p ) {
    p->setVector(p->getVector() - commandView->getAgentMovement(PREV));
    //setVector(agentView->getAgentObject(NOW).getRelativeVectorTo(p->getPosition()));
    p->setDirection(p->getDirection() - commandView->getActualTurn(PREV));
  }
  for ( p = theUnknownPlayers[worldModel->getUnknownPlayersIndex(NEXT)].begin();
	p != theUnknownPlayers[worldModel->getUnknownPlayersIndex(NEXT)].end(); ++p ) {
    p->setVector(p->getVector() - commandView->getAgentMovement(PREV));
    //setVector(agentView->getAgentObject(NOW).getRelativeVectorTo(p->getPosition()));
    p->setDirection(p->getDirection() - commandView->getActualTurn(PREV));
  }

  // Update the current observation.
  worldModel->increaseOpponentsIndex();
  worldModel->increaseTeammatesIndex();
  worldModel->increaseUnknownPlayersIndex();
  worldModel->increaseClosestOpponentIndexIndex();
  worldModel->increaseClosestTeammateIndexIndex();
  worldModel->increaseClosestUnknownPlayerIndexIndex();
}


void PlayerView::myUpdateAfterSee()
{
  // Reset the pointers to the closest seen players last time I saw.
  theClosestUnknownPlayerIndex[worldModel->getClosestUnknownPlayerIndexIndex(NOW)].makeUnknown();
  theClosestOpponentIndex[worldModel->getClosestOpponentIndexIndex(NOW)].makeUnknown();
  theClosestTeammateIndex[worldModel->getClosestTeammateIndexIndex(NOW)].makeUnknown();

  // Reset the queues of players seen last time I saw.
  theOpponents[worldModel->getOpponentsIndex(NOW)].clear();
  theTeammates[worldModel->getTeammatesIndex(NOW)].clear();
  theUnknownPlayers[worldModel->getUnknownPlayersIndex(NOW)].clear();
  
  sensorInterface->forEachPlayer(&View::processPlayer, this);

  /*
  std::cerr << "Saw " << sensorInterface->getNumberOfSeenPlayers() << " players." << std::endl;

  if ( getClosestOpponent(NOW) != NULL )
    std::cerr << "Stored " << theOpponents[getIndex(NOW)].size() << " theOpponents. Closest opponent is "
	      << getClosestOpponent(NOW)->getSide() << ":"
	      << getClosestOpponent(NOW)->getUniformNumber() << std::endl;
  if ( getClosestTeammate(NOW) != NULL )
    std::cerr << "Stored " << theTeammates[getIndex(NOW)].size() << " theTeammates. Closest teammate is "
	      << getClosestTeammate(NOW)->getSide() << ":"
	      << getClosestTeammate(NOW)->getUniformNumber() << std::endl;
  if ( getClosestPlayer(NOW) != NULL )
    std::cerr << "Stored " << theUnknownPlayers[getIndex(NOW)].size() << " unknown players. Closest player is "
	      << getClosestPlayer(NOW)->getSide() << ":"
	      << getClosestPlayer(NOW)->getUniformNumber() << std::endl;
  */
}


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "PlayerView.icc"
#endif
