/* -*- Mode: C++ -*-
 * $Id: WorldModel.cc,v 1.4 2001/01/23 15:33:05 olalu526 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *                     2001 Ola Lundqvist, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: WorldModel.cc
 * Unitname: Framework
 * $Revision: 1.4 $
 * Created by: Ola Lundqvist 2001-01-02
 * Last modified by $Author: olalu526 $ $Date: 2001/01/23 15:33:05 $
 *
 *
 * HISTORY
 *
 * $Log: WorldModel.cc,v $
 * Revision 1.4  2001/01/23 15:33:05  olalu526
 * Fixed major bug in World Model.
 *
 * Revision 1.3  2001/01/12 12:40:34  olalu526
 * Added some documentation.
 *
 *
 */

/**
 * @file WorldModel.cc
 * Implementation file for the class WorldModel
 *
 * @par Unit:
 *      Framework
 * $Revision: 1.4 $
 * $Author: olalu526 $
 * $Date: 2001/01/23 15:33:05 $
 */

#include "WorldModel.h"

RS_BEGIN_NAMESPACE

WorldModel::WorldModel(ParameterView* parameterView, const UInt hist_len) :
  playModes(hist_len, PM_UNKNOWN),
  currentPlayModeObservation(0),
  playModeHistoryLength(hist_len),

  ball(hist_len,
       BallObject(PointUCO(),
		  VectorUCO(),
		  FloatUCO(),
		  AngleDegUCO(),
		  VectorUCO(),
		  PointUnknown(),
		  VectorUnknown(),
		  TimeUnknown())),
  currentBallObservation(0),
  ballHistoryLength(hist_len),

  opponents(hist_len,
	    std::vector<PlayerObject>()),
  currentOpponentsObservation(0),
  opponentsHistoryLength(hist_len),

  teammates(hist_len,
	    std::vector<PlayerObject>()),
  currentTeammatesObservation(0),
  teammatesHistoryLength(hist_len),

  unknowns(hist_len,
	   std::vector<PlayerObject>()),
  currentUnknownPlayersObservation(0),
  unknownPlayersHistoryLength(hist_len),

  closestOpponentIndex(hist_len,
		       UIntUnknown()),
  currentClosestOpponentIndexObservation(0),
  closestOpponentIndexHistoryLength(hist_len),

  closestTeammateIndex(hist_len,
		       UIntUnknown()),
  currentClosestTeammateIndexObservation(0),
  closestTeammateIndexHistoryLength(hist_len),

  closestUnknownPlayerIndex(hist_len,
			    UIntUnknown()),
  currentClosestUnknownPlayerIndexObservation(0),
  closestUnknownPlayerIndexHistoryLength(hist_len),

  agent(hist_len,
	AgentObject(T_OUR_TEAM,
		    S_UNKNOWN,
		    UIntUnknown(),
		    AngleDegUCO(),
		    AngleDegUCO(),
		    PointUCO(),
		    VectorUCO(),
		    FloatUCO(),
		    AngleDegUCO(),
		    VectorUCO(),
		    parameterView->CP_goalie,
		    parameterView->SP_stamina_max,
		    parameterView->SP_effort_max,
		    parameterView->SP_recover_max,
		    parameterView->SP_hear_max,
		    VW_NORMAL,
		    VQ_HIGH)),
  currentAgentObservation(0),
  agentHistoryLength(hist_len)
{
}

WorldModel::~WorldModel() {
}

/* ====== */

PlayModeCollection* WorldModel::playModeData() {
  return &playModes;
}

UInt WorldModel::getPlayModeIndex(const int i) const {
  return playModes.getLength() + (currentPlayModeObservation + i) % playModes.getLength();
}

void WorldModel::setPlayModeHistoryLength(const int l) {
  if (playModes.getLength() < l) {
    playModes.setLength(l);
  }
  playModeHistoryLength = l;
}

UInt WorldModel::getPlayModeHistoryLength() {
  return playModeHistoryLength;
}

void WorldModel::increasePlayModeIndex() {
  currentPlayModeObservation = getPlayModeIndex(1);
}

/* ====== */

BallCollection* WorldModel::ballData() {
  return &ball;
}

UInt WorldModel::getBallIndex(const int i) const {
  return ball.getLength() + (currentBallObservation + i) % ball.getLength();
}

void WorldModel::setBallHistoryLength(const int l) {
  if (ball.getLength() < l) {
    ball.setLength(l);
  }
  ballHistoryLength = l;
}

UInt WorldModel::getBallHistoryLength() {
  return ballHistoryLength;
}

void WorldModel::increaseBallIndex() {
  currentBallObservation = getBallIndex(1);
}

/** ====== */

PlayersCollection* WorldModel::opponentsData() {
  return &opponents;
}

UInt WorldModel::getOpponentsIndex(const int i) const {
  return opponents.getLength() + (currentOpponentsObservation + i) % opponents.getLength();
}

void WorldModel::setOpponentsHistoryLength(const int l) {
  if (opponents.getLength() < l) {
    opponents.setLength(l);
  }
  opponentsHistoryLength = l;
}

UInt WorldModel::getOpponentsHistoryLength() {
  return opponentsHistoryLength;
}

void WorldModel::increaseOpponentsIndex() {
  currentOpponentsObservation = getOpponentsIndex(1);
}

/** ====== */

PlayersCollection* WorldModel::teammatesData() {
  return &teammates;
}

UInt WorldModel::getTeammatesIndex(const int i) const {
  return teammates.getLength() + (currentTeammatesObservation + i) % teammates.getLength();
}

void WorldModel::setTeammatesHistoryLength(const int l) {
  if (teammates.getLength() < l) {
    teammates.setLength(l);
  }
  teammatesHistoryLength = l;
}

UInt WorldModel::getTeammatesHistoryLength() {
  return teammatesHistoryLength;
}

void WorldModel::increaseTeammatesIndex() {
  currentTeammatesObservation = getTeammatesIndex(1);
}

/** ====== */

PlayersCollection* WorldModel::unknownPlayersData() {
  return &unknowns;
}

UInt WorldModel::getUnknownPlayersIndex(const int i) const {
  return unknowns.getLength() + (currentUnknownPlayersObservation + i) % unknowns.getLength();
}

void WorldModel::setUnknownPlayersHistoryLength(const int l) {
  if (unknowns.getLength() < l) {
    unknowns.setLength(l);
  }
  unknownPlayersHistoryLength = l;
}

UInt WorldModel::getUnknownPlayersHistoryLength() {
  return unknownPlayersHistoryLength;
}

void WorldModel::increaseUnknownPlayersIndex() {
  currentUnknownPlayersObservation = getUnknownPlayersIndex(1);
}

/** ====== */

UIntUnknownCollection* WorldModel::closestOpponentIndexData() {
  return &closestOpponentIndex;
}

UInt WorldModel::getClosestOpponentIndexIndex(const int i) const {
  return closestOpponentIndex.getLength() + (currentClosestOpponentIndexObservation + i) % closestOpponentIndex.getLength();
}

void WorldModel::setClosestOpponentIndexHistoryLength(const int l) {
  if (closestOpponentIndex.getLength() < l) {
    closestOpponentIndex.setLength(l);
  }
  closestOpponentIndexHistoryLength = l;
}

UInt WorldModel::getClosestOpponentIndexHistoryLength() {
  return closestOpponentIndexHistoryLength;
}

void WorldModel::increaseClosestOpponentIndexIndex() {
  currentClosestOpponentIndexObservation = getClosestOpponentIndexIndex(1);
}

/** ====== */

UIntUnknownCollection* WorldModel::closestTeammateIndexData() {
  return &closestTeammateIndex;
}

UInt WorldModel::getClosestTeammateIndexIndex(const int i) const {
  return closestTeammateIndex.getLength() + (currentClosestTeammateIndexObservation + i) % closestTeammateIndex.getLength();
}

void WorldModel::setClosestTeammateIndexHistoryLength(const int l) {
  if (closestTeammateIndex.getLength() < l) {
    closestTeammateIndex.setLength(l);
  }
  closestTeammateIndexHistoryLength = l;
}

UInt WorldModel::getClosestTeammateIndexHistoryLength() {
  return closestTeammateIndexHistoryLength;
}

void WorldModel::increaseClosestTeammateIndexIndex() {
  currentClosestTeammateIndexObservation = getClosestTeammateIndexIndex(1);
}

/* ====== */

UIntUnknownCollection* WorldModel::closestUnknownPlayerIndexData() {
  return &closestUnknownPlayerIndex;
}

UInt WorldModel::getClosestUnknownPlayerIndexIndex(const int i) const {
  return closestUnknownPlayerIndex.getLength() + (currentClosestUnknownPlayerIndexObservation + i) % closestUnknownPlayerIndex.getLength();
}

void WorldModel::setClosestUnknownPlayerIndexHistoryLength(const int l) {
  if (closestUnknownPlayerIndex.getLength() < l) {
    closestUnknownPlayerIndex.setLength(l);
  }
  closestUnknownPlayerIndexHistoryLength = l;
}

UInt WorldModel::getClosestUnknownPlayerIndexHistoryLength() {
  return closestUnknownPlayerIndexHistoryLength;
}

void WorldModel::increaseClosestUnknownPlayerIndexIndex() {
  currentClosestUnknownPlayerIndexObservation = getClosestUnknownPlayerIndexIndex(1);
}

/* ====== */

AgentCollection* WorldModel::agentData() {
  return &agent;
}

UInt WorldModel::getAgentIndex(const int i) const {
  return agent.getLength() + (currentAgentObservation + i) % agent.getLength();
}

void WorldModel::setAgentHistoryLength(const int l) {
  if (agent.getLength() < l) {
    agent.setLength(l);
  }
  agentHistoryLength = l;
}

UInt WorldModel::getAgentHistoryLength() {
  return agentHistoryLength;
}

void WorldModel::increaseAgentIndex() {
  currentAgentObservation = getAgentIndex(1);
}

RS_END_NAMESPACE
