/* -*- Mode: C++ -*-
 * $Id: KickBallView.cc,v 1.8 2000/10/28 20:22:02 mikbr212 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: KickBallView.cc
 * Unitname: Framework/view
 * $Revision: 1.8 $
 * Created by: Fredrik Heintz 2000-09-12
 * Last modified by $Author: mikbr212 $ $Date: 2000/10/28 20:22:02 $
 *
 *
 * HISTORY
 *
 * $Log: KickBallView.cc,v $
 * Revision 1.8  2000/10/28 20:22:02  mikbr212
 * changed initBall: now ballSpeed is retrieved from commandView if not known in ballView
 *
 * Revision 1.7  2000/10/26 16:00:45  mikbr212
 * Fixed bug in kickPowerRate: kickable_area to kickable_margin
 *
 * Revision 1.6  2000/10/09 18:46:15  mikbr212
 * Fixed serious wrong-calculations in getKickVector
 *
 * Revision 1.5  2000/10/09 10:17:44  mikbr212
 * added comments
 *
 * Revision 1.4  2000/10/08 20:05:01  mikbr212
 * Removed spam messages
 *
 * Revision 1.3  2000/10/08 12:57:28  mikbr212
 * removed comment
 *
 * Revision 1.2  2000/10/07 19:42:01  olalu526
 * Uncommented part using line.
 *
 * Revision 1.1  2000/10/07 14:29:25  mikbr212
 * created
 *
 */

/**
 * @file KickBallView.cc
 * Implementation file for the view class KickBallView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.8 $
 * $Author: mikbr212 $
 * $Date: 2000/10/28 20:22:02 $
 */


#include "KickBallView.h"

RS_BEGIN_NAMESPACE

/**
 * Constructor. Use getView instead.
 * @see getView
 */
KickBallView::KickBallView(const ViewId id,
			   ParameterView* _parameterView,
			   BallView* _ballView,
			   AgentView* _agentView,
			   CommandView* _commandView)
  : View(id),
    parameterView(_parameterView),
    ballView(_ballView),
    agentView(_agentView),
    commandView(_commandView) 
{
}

KickBallView::~KickBallView()
{
}

/**
 * Calculates the kickPowerRate for the current ballPosition.
 * Should only be used when ball is inside kickable_area.
 */
double KickBallView::kickPowerRate()  {
  if(ballView->getRelativeBallVector(NOW).isKnown()) {
    Vector ballPos = ballView->getRelativeBallVector(NOW).getVector();
    return kickPowerRate(ballPos);
  } else {
    return parameterView->SP_kick_power_rate * 0.75; // 0.75 is avg kick_rate
  }
}    

/**
 * Calculates the kickPowerRate for given ballPosition.
 * Should only be used when ball is inside kickable_area.
 * @param ballPos The relative ball position
 */
double KickBallView::kickPowerRate(const Vector& ballPos)  {
  return parameterView->SP_kick_power_rate 
    * (1.0 - 0.25 * abs(ballPos.getAngle().getDeg())
       / 180 - 0.25 * (ballPos.getLength() 
		       - parameterView->SP_player_size 
		       - parameterView->SP_ball_size)
       / parameterView->SP_kickable_margin);
}

/**
 * Retrieves the ball and speed position from ballView.
 * If assume is true the ball speed and position will be given a
 * default value if they are unknown.
 * Default is that the ball is in front of the player and is not moving.
 * @param ballSpeed Will be set to relative ball speed.
 * @param ballPos Will be sett to relative ball position.
 * @param assume True if default values may be used if the real values are unknown.
 * @returns True if both values were known (ie not set to default values).
 */
bool KickBallView::initBall(Vector& ballSpeed, Vector& ballPos, bool assume) {
  bool known = true;
  if(ballView->getRelativeBallVector(NOW).isKnown()) {
    ballPos = ballView->getRelativeBallVector(NOW).getVector();
  } else if(assume) { // assume the ball is in front of the agent
    ballPos = Vector(0,(parameterView->SP_kickable_area
			+ parameterView->SP_player_size
			+ parameterView->SP_ball_size) / 2);
    known=false;
  }

  if(ballView->getBallRelativeSpeedVector(NOW).isKnown()) {

    /* ball was seen this cycle -> get observated speed */
    ballSpeed = ballView->getBallRelativeSpeedVector(NOW).getVector();

  } else if(agentView->getBodyDirection(NOW).isKnown()) {

    /* ball was not seen this cycle -> get estimation */
    ballSpeed = commandView->getBallSpeed();

    // rotate to relative vector (absolute length and relative angle)
    ballSpeed.setAngle(ballSpeed.getAngle() 
		       - agentView->getBodyDirection(NOW).getAngle().getAngle().getAngle().getDeg());

  } else if(assume) { // assume it is not moving
    ballSpeed = Vector(0,0);
    known=false;
  }
  return known;
}


/**
 * Calculates the kick vector to get the desired ball speed.
 * @param desBallSpeed The desired ball speed
 * @param makKick Restricts how hard the kick can be.
 * If this value is negative, max (100*kickPowerRate) is used. Default is negative.
 * @returns the kick vector, length is the ball acceleration.
 */
Vector KickBallView::getKickVector(const Vector& desBallSpeed, double maxKick)  {
  Vector ballSpeed(0,0);
  Vector ballPos(0,0);
  initBall(ballSpeed,ballPos,true);
  return getKickVector(ballSpeed, ballPos, desBallSpeed, maxKick);
}

/**
 * Calculates the kick vector to get the desired ball speed.
 * In case desired ball speed can't be reached (because of the current
 * ball speed etc) then try to get the ball in the right direction, and if
 * that is not possible then try to put the ball in the right direction
 * (atleast more than it is now).
 * @param ballSpeed The ball speed.
 * @param ballPos The ball position.
 * @param desBallSpeed The desired ball speed.
 * @param maxKick Restricts how hard the kick can be.
 * If this value is negative, max (100*kickPowerRate) is used. Default is negative.
 * @returns the kick vector, length is the ball acceleration.
 */
Vector KickBallView::getKickVector(const Vector& ballSpeed,
				   const Vector& ballPos,
				   const Vector& desBallSpeed,
				   double maxKick)  {
  if(maxKick <= 0.0) // if negative, use default (max)
    maxKick = 100.0 * kickPowerRate(ballPos);

  Vector kickVec(0,0);
  if((desBallSpeed - ballSpeed).getLength() > maxKick) {
    // I can't shoot to desired ball speed
   Line wantedDir(Point(0,0), desBallSpeed);
   double t;
   Vector d = wantedDir.getDistance(Point(ballSpeed),t);

   if(abs((ballSpeed.getAngle() - desBallSpeed.getAngle()).getDeg()) >= 90) {
     // Ball moving in the wrong direction (diff > 90)

     if(ballSpeed.getLength() >= maxKick) {
       // ball moving to fast, try slow it down
       kickVec = - ballSpeed;
       kickVec.setLength(maxKick); 
     } else {
       //maxKick > ballSpeed & ballSpeed > d ==>> maxKick > d
       kickVec = -d; 
       kickVec += Vector(desBallSpeed.getAngle().getDeg(),
			 sqrt(maxKick*maxKick - d.getLength()*d.getLength()));
     }
   } else {
     // Ball moving in the right direction (diff < 90)
     
     if(d.getLength() >= maxKick) {
       // I ball moving to fast, try to put it in almost the right direction
       kickVec = - d;
       kickVec.setLength(maxKick);
     } else {
       // Just shoot in the right direction, but not with the right speed
       kickVec = -d;

       if(t > 1.0) // ball will move to fast (in the right dir)
	 kickVec -= Vector(desBallSpeed.getAngle().getDeg(),
			   sqrt(maxKick*maxKick - d.getLength()*d.getLength()));
       else // ball will move to slow (in the right dir)
	 kickVec += Vector(desBallSpeed.getAngle().getDeg(),
			   sqrt(maxKick*maxKick - d.getLength()*d.getLength()));
     }
   }
  } else {
    // Just kick!
    kickVec = desBallSpeed - ballSpeed;
  }
  
  return kickVec;
}

/**
 * Checks if we can shoot to desired ball speed.
 * Checks if current ball speed is too high and if we will
 * kick through ourself.
 * @param desBallSpeed The desired ball speed.
 * @param makKick Restricts how hard the kick should be.
 * @returns True if it is possible.
 */
bool KickBallView::canShootToSpeedVector(const Vector& desBallSpeed,
					    double maxKick)  {
  Vector ballSpeed(0,0);
  Vector ballPos(0,0);
  if(!initBall(ballSpeed,ballPos))
    return false;
  return canShootToSpeedVector(ballSpeed, ballPos, desBallSpeed, maxKick);
}

/**
 * Checks if we can shoot to desired ball speed.
 * Checks if current ball speed is too high and if we will
 * kick through ourself.
 * @param ballSpeed The ball speed.
 * @param ballPos The ball position.
 * @param desBallSpeed The desired ball speed.
 * @param makKick Restricts how hard the kick should be.
 * @returns True if it is possible.
 */
bool KickBallView::canShootToSpeedVector(const Vector& ballSpeed,
					const Vector& ballPos,
					const Vector& desBallSpeed,
					double maxKick)  {
  if(maxKick < 0.0) // if negative, use default (max)
    maxKick = 100.0 * kickPowerRate(ballPos);
  
  if(ballPos.getLength() > parameterView->SP_kickable_area)
    return false;

  // Check if new ballPos will be inside player
  // Note: it is possible to shoot through myself
  if((ballPos + desBallSpeed).getLength() < (parameterView->SP_player_size
					   + parameterView->SP_ball_size))
    return false; 
  
  return ((desBallSpeed - ballSpeed).getLength() <= maxKick);
}

void KickBallView::myNewCycle()
{
}

void KickBallView::myUpdateAfterCoachMessage(void)
{
}

void KickBallView::myUpdateAfterSense(void)
{
}

void KickBallView::myUpdateAfterSee(void)
{
}

void KickBallView::myUpdateAfterInit(void)
{
}

void KickBallView::myUpdateAfterHear(void)
{
}

void KickBallView::myUpdateAfterCommand(void)
{
}

/**
 * Retrieves the KickBallView from the viewmanager if it exists, otherwise it will be created.
 */
void getView(KickBallView*& tmp, ViewId view_id, 
	     ParameterView* paramView,
	     BallView* ballView,
	     AgentView* agentView,
	     CommandView* commandView)
{
  tmp = dynamic_cast<KickBallView*>(ViewManager::instance()->getView(view_id));
  if ( tmp == NULL ) {
    tmp = new KickBallView(view_id, paramView, ballView,
			   agentView, commandView);
    ViewManager::instance()->addView(tmp, true);
  }
}


RS_END_NAMESPACE



