/* -*- Mode: C++ -*-
 * $Id: KickView.cc,v 1.5 2000/11/26 15:47:18 danbe628 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <danbe628@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~danbe628/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: KickView.cc
 * Unitname: Framework/view
 * $Revision: 1.5 $
 * Created by: Fredrik Heintz 2000-09-12
 * Last modified by $Author: danbe628 $ $Date: 2000/11/26 15:47:18 $
 *
 *
 * HISTORY
 *
 * $Log: KickView.cc,v $
 * Revision 1.5  2000/11/26 15:47:18  danbe628
 * unknown values fixes
 *
 * Revision 1.4  2000/11/09 18:59:37  marsc818
 * some improvements
 *
 * Revision 1.3  2000/10/12 02:57:53  marsc818
 * fixed in some way
 *
 * Revision 1.2  2000/10/11 16:51:51  danbe628
 * KickBallInDirection now handles moving balls
 * Some fixes to KickView
 *
 * Revision 1.1  2000/09/27 13:52:37  marsc818
 * Added the kick view and abletokickballtopoint predicate.
 *
 * Revision 1.2  2000/09/17 14:17:33  danbe628
 * Made minor updates to the KickView.
 *
 * Revision 1.1  2000/09/13 00:06:21  danbe628
 * Added a view template called KickView.
 *
 */

/**
 * @file KickView.cc
 * Implementation file for the view class KickView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.5 $
 * $Author: danbe628 $
 * $Date: 2000/11/26 15:47:18 $
 */

#include "BallView.h"
#include "view_constants.h"
#include "ParameterView.h"
#include "AgentView.h"
#include "KickView.h"

#include "ViewManager.h"

//#include "view_constants.h"

RS_BEGIN_NAMESPACE

bool KickView::init()
{
  return true;
}

// If you do not want history then remove the second argument,
// and initialize View(id) instead of ViewWithHistory(id, hist_len).
// KickView::KickView(const ViewId id)
//   : View(id)
KickView::KickView(const ViewId id)//,
  /*const UInt hist_len*/
  : View/*WithHistory*/(id/*, hist_len*/),
    agentView(NULL),
    parameterView(NULL),
    ballView(NULL)
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(ballView, BALL_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID, parameterView);
}


KickView::~KickView()
{
  viewManager->releaseView(ballView->getId());
  viewManager->releaseView(agentView->getId());
  viewManager->releaseView(parameterView->getId());
} 


void KickView::myNewCycle()
{
  // This should be removed if not using history!
  // Update the current time step.
  //increaseIndex();
}

Float KickView::howHardAgentCanKick()
{
  PointUCO BallPos;
  PointUCO AgentPos;

  // 1. calculate the dist betw. agent and ball
  
  // handles unknown values by looking back in the history
  int hist = NOW;
  while(ballView->getBallPosition(hist).isUnknown() && (hist < 20))
    hist++;
  if(hist < 20)
    BallPos = ballView->getBallPosition(hist);
  else
    return 0;
  
  hist = NOW;
  while(agentView->getAgentPosition(hist).isUnknown() && (hist < 20))
    hist++;
  if(hist < 20)
    AgentPos = agentView->getAgentPosition(hist);
  else return 0;

  Float32 d = sqrt(pow(BallPos.getX().getValue().getValue().getValue() - AgentPos.getX().getValue().getValue().getValue(), 2) +
		   pow(BallPos.getY().getValue().getValue().getValue() - AgentPos.getY().getValue().getValue().getValue(), 2));

  // 2. check if ball is close enough and if dir is known
  if(d <= parameterView->SP_kickable_area)
    {
      return parameterView->SP_maxpower * kickPowerRate();
    }
  else
    return 0;
}

Float KickView::kickPowerRate()
{
  // handles unknown values by looking back in the history
  int hist1 = NOW;
  while(ballView->getRelativeBallDirection(hist1).isUnknown() && (hist1 < 20))
    hist1++;
  
  int hist2 = NOW;
  while(ballView->getBallDistance(hist2).isUnknown() && (hist2 < 20))
    hist2++;

  if((hist1 == 20) || (hist2 == 20))
    return 0;

  return parameterView->SP_kick_power_rate * (1.0 - .25 * ballView->getRelativeBallDirection(hist1).getAngle().getAngle().getAngle().getDeg() / 180 - .25 * (ballView->getBallDistance(hist2).getValue().getValue().getValue() - parameterView->SP_player_size - parameterView->SP_ball_size) / parameterView->SP_kickable_margin);
}



Float KickView::howFarAgentCanKick()
{
  // handles unknown values by looking back in the history
  int hist = NOW;
  while(ballView->getBallSpeed(hist).isUnknown() && (hist < 20))
    hist++;
  if(hist == 20)
    return 0;

  // ballSpeed after kick = possible kick power + initial ball speed
  Float ballSpeed = howHardAgentCanKick() + ballView->getBallSpeed(hist).getValue().getValue().getValue();
  
  // just a check if we divide by zero
  // in this case there must be a problem with SP_ball_decay because its value is 1
  if( (parameterView->SP_ball_decay - 1) == 0 )
    {
      RS_LOG(LA_VIEW, "KickView: ERROR! SP_ball_decay = 1 ?!? try to divide by zero");
      return 99999;
    }
  
  // compute how far we can kick ("formule de suite geometrique")
  Float howFar = ballSpeed * (1-pow(parameterView->SP_ball_decay,50)) / (1-parameterView->SP_ball_decay);
  
  return howFar;
}


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "KickView.icc"
#endif
