/*
 * $Id: Line.cc,v 1.7 2000/11/07 18:11:08 mikbr212 Exp $
 * 
 * HISTORY 
 *
 * $Log: Line.cc,v $
 * Revision 1.7  2000/11/07 18:11:08  mikbr212
 * added methods
 *
 * Revision 1.6  2000/10/30 18:32:22  mikbr212
 * added comments
 *
 *
 */

/**
 * @file Line.cc
 * Implementation file for the class Line.
 *
 * $Revision: 1.7 $
 * $Author: mikbr212 $
 * $Date: 2000/11/07 18:11:08 $
 */

#include "Line.h"

RS_BEGIN_NAMESPACE

/** 
 * Constructor.
 * Constructs line from point to point+vector.
 */
Line::Line(const Point& p1, const Vector& v1) {
  p = p1;
  v = (Point)v1;
}
/**
 * Constructor.
 * Constructs line from first point to second point.
 */
Line::Line(const Point& p1, const Point& p2) {
  p = p1;
  v = p2 - p1;
}
/**
 * Constructor.
 * Constructs line from first point to tangent of circle.
 * @param p1 Starting point
 * @param c Center of circle
 * @param radius Radius of circle
 * @param clockwise True if a clockwise tangent is wanted (default is true)
 */
Line::Line(const Point& p1, const Point& c, const double& radius, bool clockwise) {
  p = p1;
  Vector pc = (Vector)(c-p);
  double alpha;
  if(pc.getLength() >= radius) { // starting point is outside circle
    alpha = ASin(radius / pc.getLength()).getDeg(); // in degrees
    pc.setLength(sqrt(pc.getLength()*pc.getLength() - radius*radius)); //radius / tan(alpha));
  } else { // starting point is in circle
    alpha = 90;
    pc.setLength(sqrt(radius*radius - pc.getLength()*pc.getLength()));
  }
  if(clockwise)
    pc.setAngle(pc.getAngle() - alpha);
  else
    pc.setAngle(pc.getAngle() + alpha);
  
  v = (Point)pc;
}
  
/**
 * Descructor.
 */
Line::~Line() {
}

/**
 * Returns starting point of line.
 */
Point Line::getPoint() {
  return p;
}

/**
 * Returns vector of line.
 */
Vector Line::getVector() {
  return (Vector)v;
}

/**
 * Sets starting point.
 * @param p1 The starting point.
 */
void Line::setPoint(const Point& p1) {
  p = p1;
}

/**
 * Sets line-vector.
 * @param v1 The vector.
 */
void Line::setVector(const Vector& v1) {
  v = (Point)v1;
}

/**
 * Checks if the lines are parallell.
 * @param l The line to compare to.
 * @returns True if the lines are parallell.
 */
bool Line::isParallel(const Line& l) {
  return (Vector(v).getAngle() == Vector(l.v).getAngle()) ||
    ((-Vector(v)).getAngle() == Vector(l.v).getAngle());
}

/**
 * Checks if the line is through the circle (exclusive)
 * @param c Center of circle.
 * @param radius Radius of circle.
 */
bool Line::isThroughCircle(const Point& c, const double& radius) {
  double t;
  return (getDistance(c,t).getLength() < radius) &&
    (0.0 <= t && t <= 1.0);
}

/**
 * Calculates crosspoint of two lines.
 * @param l The other line.
 * @returns The crosspoint
 */
Point Line::getIntersection(const Line& l) {
  if(isParallel(l))
    return Point(0,0);

  Point a = l.p - p;
  double k = v.getY() * l.v.getX() - v.getX() * l.v.getY();
  double t = (a.getY() * l.v.getX() - a.getX() * l.v.getY()) / k;

  return p + t * v;
}

/**
 * Calculates crosspoint of two lines.
 * @param l The other line.
 * @param t Is set to the scaling-factor of vector where line l crosses this line.
 * ie if t is negative then the crosspoint is 'behind' the line, and if it is greater
 * than 1 it is in front of the line.
 * @param s Like t but for the other line.
 * @returns The crosspoint (starting point + t * vector = l.starting point + s * l.vector).
 */
Point Line::getIntersection(const Line& l, double& t, double& s) {
  if(isParallel(l))
    return Point(0,0);

  Point a;
  a = l.p - p;
  double k = v.getY() * l.v.getX() - v.getX() * l.v.getY();
  
  t = (a.getY() * l.v.getX() - a.getX() * l.v.getY()) / k;
  s = (a.getY() * v.getX() - a.getX() * v.getY()) / k;

  return p + t * v;
}
/**
 * Calculates the vector from a point to the nearest point on this line.
 * @param p2 From where the distance is to be calculated.
 * @returns The distance vector
 */
Vector Line::getDistance(const Point& p2) {
  double t;
  return getDistance(p2,t);
}
/**
 * Calculates the vector from a point to the nearest point on this line.
 * @param p2 From where the distance is to be calculated.
 * @param t is set to the scaling-factor of vector where the nearest point is.
 * if t is in [0,1] then the nearest point is on this line.
 * @returns The distance vector
 */
Vector Line::getDistance(const Point& p2, double& t) {
  Point a = p2 - p;

  t = (a.getX() * v.getX() + a.getY() * v.getY()) /
    (v.getX() * v.getX() + v.getY() * v.getY());

  return (Vector) (a - t * v);
}

/**
 * Extends (or crops) line to a circle.
 * If line (or it's extension) isn't through circle then nothing is done.
 * Note: The line will be mirrored of circle if is 'behind' the line.
 * @param c Center of circle
 * @param r Radius of circle
 */
void Line::extendToCircle(const Point& c, const double& r) {
  double t,d,delta;
  d = getDistance(c,t).getLength();
  if(d > r)
    return; // Line (or it's extension) is not through circle

  delta = sqrt(r*r - d*d);


  Vector vec(v);

  if(delta + t * vec.getLength() < 0) { // the direction is reversed
#ifdef LINE_SPAM    
    cout << "********* Line is mirrored! *********\n";
#endif
    vec = -vec; // mirror
    vec.setLength(-(delta + t * vec.getLength()));
  } else {
    vec.setLength(delta + t * vec.getLength());
  }
  
  v = Point(vec);
}

void Line::extendToLine(const Line& line, double& s, double &t) {
  Point cross = getIntersection(line,s,t);
  v = cross - p;
}

void Line::extendToLine(const Line& line) {
  Point cross = getIntersection(line);
  v = cross - p;
}

Point Line::getStartPoint() {
  return p;
}

Point Line::getEndPoint() {
  return p + v;
}

RS_END_NAMESPACE
