/* TerritoryControlView.cc
 *
 *  Retrieves information about players from PlayerView and
 *  calculates the 'best spot' according to teammates,
 *  opponents, positive direction (advancing)
 *  and accuracy. Each of these can be weighted to calculate
 *  the spot where to pass, clear the ball etc.
 *
 *  Currently under heavy development
 */

/* AUTHORS:
 * Mikael Brnnstrm
 * Andreas Bckert
 * Andreas Sigfridsson
 */

// $Id: TerritoryControlView.cc,v 1.9 2001/06/19 14:01:47 frehe Exp $

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <math.h>
#include "TerritoryControlView.h"
#include "ConeBlob.h"
#include "ViewManager.h"

//#define TCV_DEBUG 1

#define MAX(x, y) (x > y ? x : y)
#define MIN(x, y) (x < y ? x : y)

#define OUTX 128
#define OUTY 128

RS_BEGIN_NAMESPACE

Blob::Blob(float _depthSize, float _angleSize)
{
  depthSize = _depthSize;
  angleSize = _angleSize;
}

Blob::~Blob()
{
}

void Blob::sample(TerritoryControlView *tcv)
{
  rSize = tcv->vectorToRadiusIndex(Vector(0, depthSize));
  thSize = tcv->vectorToAngleIndex(Vector(angleSize, 0));
  blobData = new float[rSize*thSize];

  // FIXME: Add shadowing and hotspots
  for(int r = 0; r<rSize; r++)
    for(int th = 0; th<thSize; th++)
      blobData[r+ th*rSize] = getValueAt(tcv->indexToDepth(r-rSize/2),
					 tcv->indexToAngle(th-thSize/2));
//       blobData[r+ th*rSize] = getValueAt(tcv->indexToVector(r-rSize/2,
// 							    th-thSize/2));
}

TerritoryControlView::TerritoryControlView(const ViewId id,
					   const PlayerView* player_view,
					   const AgentView* agent_view,
					   int _rSize,
					   int _thSize,
					   double _zFar,
					   Blob* _teammateBlob,
					   Blob* _opponentBlob,
					   Blob* _ballBlob)
  : View(id),
    playerView(player_view),
    agentView(agent_view),
    parameterView(NULL),
    markerView(NULL),
    rSize(_rSize),
    thSize(_thSize),
    zFar(_zFar),
    teammateBlob(_teammateBlob),
    opponentBlob(_opponentBlob),
    ballBlob(_ballBlob)
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(markerView, MARKER_VIEW_ID, parameterView);
  teammatesMap = new float[rSize*thSize];
  opponentsMap = new float[rSize*thSize];
  teammatesOwner = new int[rSize*thSize];
  opponentsOwner = new int[rSize*thSize];
  teammatesOwnerContrib = new float[rSize*thSize];
  opponentsOwnerContrib = new float[rSize*thSize];
  outputMap = new float[rSize*thSize];

  teammateBlob = new ConeBlob(2, 7, 10, 45);
  opponentBlob = new ConeBlob(2, 10, 10, 45);

  if(teammateBlob) teammateBlob->sample(this);
  if(opponentBlob) opponentBlob->sample(this);
  if(ballBlob) ballBlob->sample(this);

  tunnelMap = new unsigned int[OUTX*OUTY];
  for(int y = 0; y<OUTY; y++)
    for(int x = 0; x<OUTX; x++)
    {
      Vector v = (zFar/sqrt(OUTX*OUTX/4+OUTY*OUTY/4))*(Vector)Point(x-OUTX/2, y-OUTY/2);
      int r, th;
      r = vectorToRadiusIndex(v);
      th = (vectorToAngleIndex(v) + thSize) % thSize;
//       printf("r=%d, th=%d\n", r, th);
      tunnelMap[x+y*OUTX] = (r << 16) | (th & 0xffff);
    }

  fifo_fd = 0;
}


TerritoryControlView::~TerritoryControlView()
{
  delete teammatesMap;
  delete opponentsMap;
  delete teammatesOwner;
  delete opponentsOwner;
}

void TerritoryControlView::addBlob(BlobID blobID,
				   int player,
				   int rIndex,
				   int degIndex)
{
  Blob* blob;
  float* map;
  float* ownerContrib;
  int* owner;
  switch(blobID)
  {
  case TEAMMATE:
    blob = teammateBlob;
    map = teammatesMap;
    ownerContrib = teammatesOwnerContrib;
    owner = teammatesOwner;
    break;

  case OPPONENT:
    blob = opponentBlob;
    map = opponentsMap;
    ownerContrib = opponentsOwnerContrib;
    owner = opponentsOwner;
    break;

  case BALL:
  default:
	  return;

//   case BALL:
//     blob = teammateBlob;
//     map = teammatesMap;
//     break;
  }
  int destR, destTh;
  int sourceR, sourceTh;
  int countR, countTh;

  countR = blob->rSize;
  destR = rIndex - blob->rSize/2;
  sourceR = 0;

  countTh = blob->thSize;
  destTh = degIndex - blob->thSize/2;
  sourceTh = 0;

  if(destR > rSize || destR + countR < 0)
    return;			// Blob outside of map
  if(destR < 0)
  {
    countR -= -destR;
    sourceR += -destR;
    destR = 0;
  }
  if(destR + countR > rSize)
    countR = rSize - destR;

  for(int th = 0; th<countTh; th++)
  {
    int my_destTh = (destTh+th + rSize) % rSize;
    float contrib;
    for(int r = 0; r<countR; r++)
    {
      contrib = blob->blobData[(sourceR+r) + (sourceTh+th)*blob->rSize];
      if(contrib > ownerContrib[(destR+r) + my_destTh*rSize])
      {
	ownerContrib[(destR+r) + my_destTh*rSize] = contrib;
	owner[(destR+r) + my_destTh*rSize] = player;
      }

      map[(destR+r) + my_destTh*rSize] = max(map[(destR+r) + my_destTh*rSize], contrib);
//       map[(destR+r) + my_destTh*rSize] += contrib;
    }
  }
}

void TerritoryControlView::clearMap(float* map)
{
//   for(int x=0; x<rSize; x++)
//     for(int y=0; y<thSize; y++)
//       map[x+y*rSize] = ((float)x)/y;
  memset(map, 0, rSize*thSize*sizeof(float));
}

void TerritoryControlView::clearOwner(int* map)
{
  memset(map, 0, rSize*thSize*sizeof(int));
}

void TerritoryControlView::clearOwnerContrib(float* map)
{
  memset(map, 0, rSize*thSize*sizeof(float));
}

int TerritoryControlView::vectorToRadiusIndex(Vector v)
{
  return distanceToIndex(v.getLength());
}

int TerritoryControlView::vectorToAngleIndex(Vector v)
{
  return angleToIndex(v.getAngle().getDeg());
}

int TerritoryControlView::distanceToIndex(float d)
{
  return (int)(d/zFar*rSize);
}

// Angle a given in degrees 0 <= a < 360.0
int TerritoryControlView::angleToIndex(float a)
{
  int tmp = (int)(a/360.0*thSize);
  while(tmp<0)
    tmp += thSize;
  while(tmp>thSize)
    tmp -= thSize;
  return tmp;
}

Vector TerritoryControlView::indexToVector(int rIndex, int degIndex)
{
  return Vector((double)degIndex/thSize*360.0, (double)rIndex/rSize*zFar);
}

inline float TerritoryControlView::indexToDepth(int rIndex)
{
  return (float)rIndex/rSize*zFar;
}

inline float TerritoryControlView::indexToAngle(int angIndex)
{
  float res = ((float)angIndex)/thSize*360.0;
  if(res > 180.0)
    return res-360;
  else
    return res;
}

// Weigh positions closer to the goal as advantageous
inline float TerritoryControlView::advance(int r, int th)
{
  return max(0.0,(1 + std::atan(0.3 * indexToDepth(r)-1)  / (0.5*M_PI))
            * (exp(-0.6 * pow(indexToAngle((th+goalIndex) % thSize)*M_PI/180.0, 2))));
  
//adv2 = max(0,(1 + atan(0.3 * abs(cp)-1)  / (0.5*pi)) .* (exp(-0.6 * angle(cp).^2)));
  
  //  return cos(360.0*th/thSize)*r/rSize;
//   return std::atan(3*cos(360.0*th/thSize)*r/rSize)/(M_PI/2);
}

inline float TerritoryControlView::precision(int r, int th, const float wPrecision)
{
//prec2 = max(0.0, (1.0 - 3*wPrecision*(exp(0.69*(abs(cp)/zFar).^3) - 1)));
  return max(0.0, (1.0 - 3*wPrecision*(exp(0.69*pow((float)r/rSize, 3)) - 1)));
}

// For internal use only
inline float TerritoryControlView::getValue(const float wTeammates,
				     const float wOpponents,
				     const float wAdvance,
				     const float wPrecision,
				     int r, int th)

{
  return wTeammates * teammatesMap[r+th*rSize]
    * advance(r, th) * precision(r, th, wPrecision)
    + wOpponents * opponentsMap[r+th*rSize];
}

// For external use only, is slower than the internal version
float TerritoryControlView::getValue(const float wTeammates,
                                    const float wOpponents,
                                    const float wAdvance,
                                    const float wPrecision,
                                    Vector v)

{
  if(needsUpdate)
    update();
  return getValue(wTeammates, wOpponents, wAdvance, wPrecision,
                 vectorToRadiusIndex(v), vectorToAngleIndex(v));
}

Vector TerritoryControlView::getMax(const float wTeammates,
				    const float wOpponents,
				    const float wAdvance,
				    const float wPrecision)
{
  return getMax(wTeammates, wOpponents, wAdvance, wPrecision, 0.0, 360.0);
}

Vector TerritoryControlView::getMax(const float wTeammates,
                                   const float wOpponents,
                                   const float wAdvance,
                                   const float wPrecision,
                                   float minangle, float maxangle)
{
  if(useCache
     && cachedwTeammates == wTeammates
     && cachedwOpponents == wOpponents
     && cachedwAdvance == wAdvance
     && cachedwPrecision == wPrecision)
    return cachedMax;
  else
    return getMaxNoCache(wTeammates, wOpponents, wAdvance, wPrecision, minangle, maxangle);
}

Vector TerritoryControlView::getMaxNoCache(const float wTeammates,
					   const float wOpponents,
					   const float wAdvance,
					   const float wPrecision,
					   float minangle, float maxangle)
{
  if(needsUpdate)
    update();
  useCache = 1;
  cachedwTeammates = wTeammates;
  cachedwOpponents = wOpponents;
  cachedwAdvance = wAdvance;
  cachedwPrecision = wPrecision;
  cachedMinangle = minangle;
  cachedMaxangle = maxangle;

  int minth = angleToIndex(minangle);
  int maxth = angleToIndex(maxangle);
  if(maxth < minth)
    maxth += thSize;

  for(int i=0; i<rSize*thSize; i++)
    outputMap[i] = -1.0;
  int maxTh = -1, maxR = -1;

#define MAXFLOAT 5984098234923498234.0

  float maxValue = -MAXFLOAT;
  for(int th=0; th<thSize; th++)
  {
    float firstValue = getValue(wTeammates, wOpponents, wAdvance, wPrecision, 0, th);
    float bestValue = firstValue; // For this direction only

    for(int r=0; r<rSize; r++)
    {
      float thisValue = getValue(wTeammates, wOpponents, wAdvance, wPrecision, r, th);

      if(thisValue > bestValue)
      {
	bestValue = thisValue;
      }

      if(thisValue > maxValue)
      {
	maxR = r;
	maxTh = th;
	maxValue = thisValue;
      }

      outputMap[r + th*rSize] = thisValue;
      if(thisValue < firstValue - 0.1)
	break;
    }

  }
  return (cachedMax = indexToVector(maxR, maxTh));
}

AngleDeg TerritoryControlView::getMaxMinNoCache(const float wTeammates,
                                               const float wOpponents,
                                               const float wAdvance,
                                               const float wPrecision,
                                               AngleDeg minangle = 0.0,
                                               AngleDeg maxangle = 360.0)
{
  float tmp;
  return getMaxMinNoCache(wTeammates, wOpponents, wAdvance, wPrecision,
                         tmp, minangle, maxangle);
}

AngleDeg TerritoryControlView::getMaxMinNoCache(const float wTeammates,
                                               const float wOpponents,
                                               const float wAdvance,
                                               const float wPrecision,
                                               float &maxValue,
                                               AngleDeg minangle = 0.0,
                                               AngleDeg maxangle = 360.0)
{
  if(needsUpdate)
    update();
  int minth = angleToIndex(minangle.getDeg());
  int maxth = angleToIndex(maxangle.getDeg());
  if(maxth < minth)
    maxth += thSize;

  int maxTh = -1;
  maxValue = -HUGE;
  float foo = -HUGE;
  for(int _th=minth; _th<maxth; _th++)
  {
    int th = (_th + thSize) % thSize;
    float minValue = getValue(wTeammates, wOpponents, wAdvance, wPrecision, 0, th);
    float thisMaxValue = minValue;

    for(int r=0; r<rSize; r++)
    {
      float thisValue = getValue(wTeammates, wOpponents, wAdvance, wPrecision, r, th);

      if(thisValue < minValue)
       minValue = thisValue;
      if(thisValue > thisMaxValue)
       thisMaxValue = thisValue;
    }

    const float fooThres = 0.05;
    if(minValue + fooThres > maxValue && thisMaxValue > foo)
    {
      foo = thisMaxValue;
      maxTh = th;
      maxValue = minValue;
    }
  }

  return indexToAngle(maxTh);
}

int TerritoryControlView::getOwner(const Vector &v,
				   const float wTeammates,
				   const float wOpponents,
				   const float wAdvance,
				   const float wPrecision)
{
  if(needsUpdate)
    update();
  int r = vectorToRadiusIndex(v);
  int th = vectorToAngleIndex(v);

  float teammate, opponent;
  teammate = wTeammates * teammatesMap[r + th*rSize];
  opponent = wOpponents * opponentsMap[r + th*rSize];

  if(abs(teammate) > abs(opponent))
    return teammatesOwner[r + th*rSize];
  else
    return opponentsOwner[r + th*rSize];
}

bool TerritoryControlView::init()
{
  return true;
}

#define DBG printf("debug: %d\n", __LINE__)

// Do what we usually do in myNewCycle, but will only be done on demand
void TerritoryControlView::update()
{
  clearMap(teammatesMap);
  clearMap(opponentsMap);
  clearOwner(teammatesOwner);
  clearOwner(opponentsOwner);
  clearOwnerContrib(teammatesOwnerContrib);
  clearOwnerContrib(opponentsOwnerContrib);

  VectorUCO theirGoal = VectorUCO(agentView->getAgentPosition(NOW),
				  markerView->getTheirGoal(NOW).getPosition());
  if(theirGoal.isKnown())
    goalIndex = angleToIndex(-theirGoal.getVector().getAngle().getDeg());
  else
    goalIndex = 0;

  std::vector<PlayerObject> teammates = playerView->getTeammates(NOW);
  for(unsigned int i=0; i<teammates.size(); i++)
  {
    if(teammates[i].getVector().isKnown())
    {
      Vector mate = teammates[i].getVector().getVector();
      int r, th;
      r = vectorToRadiusIndex(mate);
      th = vectorToAngleIndex(mate);
      int player = teammates[i].getUniformNumber().isKnown() ?
	teammates[i].getUniformNumber().getValue() : -1;
      addBlob(TEAMMATE, player, r, th);
    }
  }

  std::vector<PlayerObject> opponents = playerView->getOpponents(NOW);
  for(unsigned int i=0; i<opponents.size(); i++)
  {
    if(opponents[i].getVector().isKnown())
    {
      Vector mate = opponents[i].getVector().getVector();
      int r, th;
      r = vectorToRadiusIndex(mate);
      th = vectorToAngleIndex(mate);
      int player = opponents[i].getUniformNumber().isKnown() ?
	opponents[i].getUniformNumber().getValue() : -1;
      addBlob(OPPONENT, player, r, th);
    }
  }
}

void TerritoryControlView::myNewCycle()
{
  needsUpdate = 1;
  useCache = 0;

#ifdef TCV_DEBUG
//   unsigned char* buffer = new unsigned char[rSize*thSize];
//   for(int i=0; i<rSize*thSize; i++)
//     buffer[i] = (unsigned char) (255.0*teammatesMap[i]);
//   write(teammates_map_fd, buffer, rSize*thSize);
  unsigned char* buffer = new unsigned char[OUTX*OUTY];
  unsigned char grey;

//   // smeta
//   for(int th=0; th<thSize; th++)
//   {
//     float value = 0.0;
//     float new_value;
//     for(int r=0; r<rSize; r++)
//     {
//       if((new_value = opponentsMap[r+th*rSize]) > value)
// 	value = new_value;
//       opponentsMap[r+th*rSize] = value;
//     }
//   }

  int r, th;
  Vector v = getMaxNoCache(1.0, -1.0, 1.0, 1.0, 0.0, 360.0);
  r = vectorToRadiusIndex(v);
  th = (vectorToAngleIndex(v) + thSize) % thSize;
//   printf("th=%d ang=%f\n", th, v.getAngle().getDeg());
//   printf("r=%d len=%f\n", r, v.getLength());
//   int x, y;
//   x = OUTX/2 + OUTX/2*((float)r)/rSize*cos(360.0*th/thSize);
//   y = OUTY/2 + OUTY/2*((float)r)/rSize*sin(360.0*th/thSize);
  outputMap[r + th*rSize] = -1.0;
//   buffer[x + y*OUTX + OUTX*OUTY] = 0;
//   buffer[x + y*OUTX + 2*OUTX*OUTY] = 0;

  for(int i=0; i<OUTX*OUTY; i++)
  {
    unsigned int r, th;
    r = tunnelMap[i] >> 16;
    th = tunnelMap[i] & 0xffff;
    grey = (unsigned char) (MIN(255, MAX(0, 128 + 127.0*(outputMap[r + th*rSize]))));
//     grey = 128 + 127.0*(teammatesMap[r + th*rSize]
// 			- opponentsMap[r + th*rSize]);
    buffer[i] = (unsigned char) grey;
//     buffer[i+OUTX*OUTY] = (unsigned char) grey;
//     buffer[i+2*OUTX*OUTY] = (unsigned char) grey;
  }

  if(!fifo_fd)
  {
    char filename[80];
    int nummer = agentView->getAgentObject(NOW).getUniformNumber().getValue();
    sprintf(filename, "/tmp/tcv-fifo%d", nummer);
    std::cout << "(TCV) Opening fifo " << filename << std::endl;
    fifo_fd = open(filename, O_WRONLY);
  }
  write(fifo_fd, buffer, OUTX*OUTY);

//   int bytesleft = OUTX*OUTY;
//   while(bytesleft)
//     bytesleft -= write(teammates_map_fd, buffer + OUTX*OUTY - bytesleft, bytesleft);
  delete buffer;
#endif

  // Update the current time step.
//   increaseIndex();
}

void TerritoryControlView::myUpdateAfterCoachMessage(void)
{
}

void TerritoryControlView::myUpdateAfterSense(void)
{
}

void TerritoryControlView::myUpdateAfterSee(void)
{
}

void TerritoryControlView::myUpdateAfterInit(void)
{
}

void TerritoryControlView::myUpdateAfterHear(void)
{
}

void TerritoryControlView::myUpdateAfterCommand(void)
{
}

void getView(TerritoryControlView*& tmp, ViewId view_id, 
	     PlayerView* playerView, AgentView* agentView,
	     int rSize, int thSize, int zFar,
	     Blob* teammate, Blob* opponent, Blob* ball)
{
  tmp = dynamic_cast<TerritoryControlView*>(ViewManager::instance()->getView(view_id));
  if ( tmp == NULL ) {
    tmp = new TerritoryControlView(view_id, playerView, agentView,
				   rSize, thSize, zFar,
				   teammate, opponent, ball);
    ViewManager::instance()->addView(tmp, true);
  }
}

RS_END_NAMESPACE
