/* -*- C++ -*-
 * TerritoryControlView.h
 *
 *  Retrieves information about players from PlayerView and
 *  calculates the 'best spot' according to teammates,
 *  opponents, positive direction (advancing)
 *  and accuracy. Each of these can be weighted to calculate
 *  the spot where to pass, clear the ball etc.
 *
 *  Currently under heavy development
 */

/* AUTHORS:
 * Mikael Brnnstrm
 * Andreas Bckert
 * Andreas Sigfridsson
 */

// $Id: TerritoryControlView.h,v 1.7 2001/06/19 14:01:48 frehe Exp $

#ifndef TERRITORYCONTROLVIEW_H
#define TERRITORYCONTROLVIEW_H


#include "View.h"
#include "CoordVector.h"
#include "CoordPoint.h"
#include "PlayerView.h"
#include "AgentView.h"
#include "MarkerView.h"

RS_BEGIN_NAMESPACE

/**
 * @class TerritoryControlView TerritoryControlView.h
 * Retrieves information about players from PlayerView and
 * calculates the 'best spot' according to teammates,
 * opponents, positive direction (advancing)
 * and accuracy. Each of these can be weighted to calculate
 * the spot where to pass, clear the ball etc.
 *
 */

class TerritoryControlView;

class Blob {
friend class RS::TerritoryControlView;
public:
  Blob(float _depthSize, float _angleSize);
  virtual ~Blob();

  // args: rDepth, angle
  virtual float getValueAt(float, float)=0;

  float depthSize, angleSize;	// Size of sampling area
protected:
  void sample(TerritoryControlView *tcv);
  int rSize, thSize;
  float *blobData;		// Sampled area
private:
  // Hide copy-constructor and assignment.
  Blob(const Blob&);
  Blob& operator= (const Blob&);
};

enum BlobID
{
	TEAMMATE,
	OPPONENT,
	BALL
};

// -== Skippa historia? ==-

class TerritoryControlView : public View {
friend class RS::Blob;
public:
  /** Construct a view with a size of @a rSize by @a thSize pixels in the polar map
   *  that spans 360 degrees and @a zFar meters. */
  TerritoryControlView(const ViewId id,
		       const PlayerView* player_view,
		       const AgentView* agentView,
		       int rSize,
		       int thSize,
		       double zFar,
		       Blob* _teammateBlob,
		       Blob* _opponentBlob,
		       Blob* _ballBlob);

  /** Destroys the view. */
  ~TerritoryControlView();

  bool init();

  // Hm.. vi ska returnera 'owner' ocks...
  /** Find the best spot according to the weights given, searching
   *  only angles between minangle and maxangle, which are absolute
   *  degrees. If opponents are considered bad, then give a negative
   *  argument for @a wOpponents. */
  Vector getMax(const float wTeammates, const float wOpponents,
               const float wAdvance, const float wPrecision, float minangle, float maxangle);
  /** Find the best spot according to the weights given, searching
   *  only angles between minangle and maxangle, which are absolute
   *  degrees. If opponents are considered bad, then give a negative
   *  argument for @a wOpponents. */
  Vector getMax(const float wTeammates, const float wOpponents,
		const float wAdvance, const float wPrecision);
  /** Find the best spot according to the weights given. If opponents are considered
   *  bad, then give a negative argument for @a wOpponents. Bypasses the cache.
   *  Should not be neccesary to use. */
  Vector getMaxNoCache(const float wTeammates, const float wOpponents,
                       const float wAdvance, const float wPrecision,
                       float minangle, float maxangle);
  /** Find out who contributes most to the position given by the Vector @a v */
  int getOwner(const Vector &v,
	       const float wTeammates,
	       const float wOpponents,
	       const float wAdvance,
	       const float wPrecision);

  float TerritoryControlView::getValue(const float wTeammates,
                                      const float wOpponents,
                                      const float wAdvance,
                                      const float wPrecision,
                                      Vector v);

  AngleDeg TerritoryControlView::getMaxMinNoCache(const float wTeammates,
                                                 const float wOpponents,
                                                 const float wAdvance,
                                                 const float wPrecision,
                                                 AngleDeg minangle = 0.0,
                                                 AngleDeg maxangle = 360.0);

  AngleDeg TerritoryControlView::getMaxMinNoCache(const float wTeammates,
                                                 const float wOpponents,
                                                 const float wAdvance,
                                                 const float wPrecision,
                                                 float &maxValue,
                                                 AngleDeg minangle = 0.0,
                                                 AngleDeg maxangle = 360.0);


protected:
  const PlayerView* playerView;
  const AgentView* agentView;
  ParameterView* parameterView;
  MarkerView* markerView;

  int rSize, thSize;		// Size of map in pixels (distance, angle)
  double zFar;			// Maximum distance of map in meters

  Blob *teammateBlob, *opponentBlob, *ballBlob;
  float *teammatesMap, *opponentsMap;
  int *teammatesOwner, *opponentsOwner;
  float *teammatesOwnerContrib, *opponentsOwnerContrib;
  unsigned int *tunnelMap;
  float *outputMap;

  void addBlob(BlobID blobId, int player, const int rIndex, const int degIndex);
  void clearMap();

  virtual float TerritoryControlView::advance(int r, int th);
  virtual float TerritoryControlView::precision(int r, int th, float wPrecision);

  // Convert player position to index in map
  int vectorToRadiusIndex(Vector v);
  int vectorToAngleIndex(Vector v);
  int distanceToIndex(float d);
  int angleToIndex(float a);
  // Convert map index to player position vector
  Vector indexToVector(const int rIndex, const int degIndex);
  float indexToDepth(int rIndex);
  float indexToAngle(int angIndex);

  virtual void myNewCycle();
  virtual void myUpdateAfterCoachMessage();
  virtual void myUpdateAfterInit();
  virtual void myUpdateAfterSee();
  virtual void myUpdateAfterHear();
  virtual void myUpdateAfterSense();
  virtual void myUpdateAfterCommand();

  void clearMap(float* map);
  void clearOwner(int* map);
  void clearOwnerContrib(float* map);

  float cachedwTeammates, cachedwOpponents, cachedwAdvance, cachedwPrecision,
    cachedMinangle, cachedMaxangle;
  Vector cachedMax;

private:
  inline float TerritoryControlView::getValue(const float wTeammates,
                                             const float wOpponents,
                                             const float wAdvance,
                                             const float wPrecision,
                                             int r, int th);

  // Hide copy-constructor and assignment.
  TerritoryControlView(const TerritoryControlView&);
  TerritoryControlView& operator= (const TerritoryControlView&);

  void update();
  int needsUpdate;
  int useCache;

  int goalIndex;

  int fifo_fd;
};

void getView(TerritoryControlView*& tmp, ViewId view_id, 
	     PlayerView* playerView, AgentView* agentView,
	     int rSize, int thSize, int zFar,
	     Blob* teammate, Blob* opponent, Blob* ball);

RS_END_NAMESPACE
#endif // !TERRITORYCONTROLVIEW_H
