/* -*- Mode: C++ -*-
 * $Id: AngleDegrees.h,v 1.7 2000/11/08 12:56:38 olalu526 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: AngleDegrees.h
 * Unitname: Library
 * $Revision: 1.7 $
 * Created by: Fredrik Heintz 2000-08-29
 * Last modified by $Author: olalu526 $ $Date: 2000/11/08 12:56:38 $
 *
 *
 * HISTORY
 *
 * $Log: AngleDegrees.h,v $
 * Revision 1.7  2000/11/08 12:56:38  olalu526
 * Fixed a bug.
 *
 * Revision 1.6  2000/10/23 12:04:13  frehe
 * Changed the angles to be in the interval -180 < angle <= 180.
 *
 * Revision 1.5  2000/09/19 13:30:24  frehe
 * Fixed a bug in AngleRadians that converted radians from degrees to radians.
 *
 * Revision 1.4  2000/09/18 15:20:03  frehe
 * Due to a problem with namespaces the name for atan2, atan, acos and asin is changed to ATan2, ATan, ASin and ACos.
 *
 * Revision 1.3  2000/09/03 14:23:18  frehe
 * Made minor changes to improve the documentation.
 *
 * Revision 1.2  2000/09/01 03:23:50  frehe
 * Made minor changes.
 *
 * Revision 1.1.1.1  2000/08/31 13:26:04  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file AngleDegrees.h
 * Include file for the template class AngleDegrees.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.7 $
 * $Author: olalu526 $
 * $Date: 2000/11/08 12:56:38 $
 */


#ifndef ANGLEDEGREES_H
#define ANGLEDEGREES_H


#include <iostream>
#include <string>

#include "library_compiler_directives.h"
#include "math_utils.h"
#include "AngleRadians.h"


#define RS_DEG_HEAD std::string("#Deg")


RS_BEGIN_NAMESPACE


template<class T> class AngleRadians;


/**
 * @class AngleDegrees AngleDegrees.h
 * Template class for representing angles in degrees.
 * The template argument only defines what type the
 * angles should be represented by.
 * The only demands on the template argument T is that the trigonometric
 * and arithmetic functions are defined on it.
 *
 * @todo Make a single template of both AngleDegrees and AngleRadians
 *       with a second parameter which defines what the standard
 *       representation is (radians or degrees).
 *
 * @par Unit:
 *      Library
 * $Revision: 1.7 $
 * $Author: olalu526 $
 * $Date: 2000/11/08 12:56:38 $
 */
template<class T> class AngleDegrees {
public:
  /** Default constructor, the value of the angle is T(). */
  AngleDegrees();

  /** Construct and AngleDegrees with the value @a ang. */
  AngleDegrees(const T& ang);

  /** Copy constructor. */
  AngleDegrees(const AngleDegrees<T>& ang);

  /** Conversion constructor that converts an AngleRadians to an AngleDegrees. */
  AngleDegrees(const AngleRadians<T>& ang);

  
  /** Get the angle in radians. */
  T getRad() const;

  /** Get the angle in degrees. */
  T getDeg() const;

  
  //@{
  /** @name Assignment operators. */
  AngleDegrees<T>& operator=(const AngleDegrees<T>&);
  AngleDegrees<T>& operator=(const AngleRadians<T>&);
  AngleDegrees<T>& operator=(const T&);
  //@}

  
  //@{
  /** @name Arithmetic operators. */
  /** I don't think I have to define all the operators for T too,
    since the automatic type conversion will convert T to AngleDegrees<T>.
    No but the efficiency is a little bit better. */
  AngleDegrees<T> operator-() const;
  AngleDegrees<T> operator+(const AngleDegrees<T>&) const;
  AngleDegrees<T> operator+(const T&) const;
  AngleDegrees<T> operator-(const AngleDegrees<T>&) const;
  AngleDegrees<T> operator-(const T&) const;
  AngleDegrees<T> operator*(const AngleDegrees<T>&) const;
  AngleDegrees<T> operator*(const T&) const;
  AngleDegrees<T> operator/(const AngleDegrees<T>&) const;
  AngleDegrees<T> operator/(const T&) const;
  
  AngleDegrees<T>& operator+=(const AngleDegrees<T>&);
  AngleDegrees<T>& operator+=(const T&);
  AngleDegrees<T>& operator-=(const AngleDegrees<T>&);
  AngleDegrees<T>& operator-=(const T&);
  AngleDegrees<T>& operator*=(const AngleDegrees<T>&);
  AngleDegrees<T>& operator*=(const T&);
  AngleDegrees<T>& operator/=(const AngleDegrees<T>&);
  AngleDegrees<T>& operator/=(const T&);
  //@}

  
  //@{
  /** @name Increment and decrement operators. */
  AngleDegrees<T>& operator++();
  AngleDegrees<T>& operator++(int);
  AngleDegrees<T>& operator--();
  AngleDegrees<T>& operator--(int);
  //@}


  //@{
  /** Comparison operators. */
  bool operator! () const;
  bool operator==(const AngleDegrees<T>&) const;
  bool operator==(const T&) const;
  bool operator!=(const AngleDegrees<T>&) const;
  bool operator!=(const T&) const;
  bool operator> (const AngleDegrees<T>&) const;
  bool operator> (const T&) const;
  bool operator>=(const AngleDegrees<T>&) const;
  bool operator>=(const T&) const;
  bool operator< (const AngleDegrees<T>&) const;
  bool operator< (const T&) const;
  bool operator<=(const AngleDegrees<T>&) const;
  bool operator<=(const T&) const;
  //@}
  
  
  //operator T() const;

protected:
  /** Function that normalizes an angle to the interval [-180, 180]. */
  void normalize();

  /** The actual representation of the angle value. */
  T a;
};


template<class T>
inline T AngleDegrees<T>::getRad() const
{
  return deg2Rad(a);
}


template<class T>
inline T AngleDegrees<T>::getDeg() const
{
  return a;
}


template<class T>
inline AngleDegrees<T> AngleDegrees<T>::operator -() const
{
  return AngleDegrees<T>(-a);
}


template<class T>
inline AngleDegrees<T> AngleDegrees<T>::operator +(const AngleDegrees<T>& ang) const
{
  return AngleDegrees<T>(a + ang.a);
}


template<class T>
inline AngleDegrees<T> AngleDegrees<T>::operator +(const T& ang) const
{
  return AngleDegrees<T>(a + ang);
}


template<class T>
inline AngleDegrees<T> AngleDegrees<T>::operator -(const AngleDegrees<T>& ang) const
{
  return AngleDegrees<T>(a - ang.a);
}


template<class T>
inline AngleDegrees<T> AngleDegrees<T>::operator -(const T& ang) const
{
  return AngleDegrees<T>(a - ang);
}


template<class T>
inline AngleDegrees<T> AngleDegrees<T>::operator *(const AngleDegrees<T>& ang) const
{
  return AngleDegrees<T>(a * ang.a);
}


template<class T>
inline AngleDegrees<T> AngleDegrees<T>::operator *(const T& ang) const
{
  return AngleDegrees<T>(a * ang);
}


template<class T>
inline AngleDegrees<T> AngleDegrees<T>::operator /(const AngleDegrees<T>& ang) const
{
  return AngleDegrees<T>(a / ang.a);
}


template<class T>
inline AngleDegrees<T> AngleDegrees<T>::operator /(const T& ang) const
{
  return AngleDegrees<T>(a / ang);
}


template<class T>
inline AngleDegrees<T> operator +(const T& op1, const AngleDegrees<T>& op2)
{
  return op2 + op1;
}


template<class T>
inline AngleDegrees<T> operator -(const T& op1, const AngleDegrees<T>& op2)
{
  return (-op2) + op1;
}


template<class T>
inline AngleDegrees<T> operator *(const T& op1, const AngleDegrees<T>& op2)
{
  return op2 * op1;
}


template<class T>
inline AngleDegrees<T> operator /(const T& op1, const AngleDegrees<T>& op2)
{
  return AngleDegrees<T>(op1 / op2.a);
}


// assign, no range check since already in range.
template<class T>
inline AngleDegrees<T>& AngleDegrees<T>::operator =(const AngleDegrees<T>& ang)
{
  a = ang.a;
  return *this;
}


// assign, range check since @a d could have any value in the range of T.
template<class T>
inline AngleDegrees<T>& AngleDegrees<T>::operator =(const T& d)
{
  a = d;
  normalize();
  return *this;
}


template<class T>
inline AngleDegrees<T>& AngleDegrees<T>::operator ++()
{
  ++a;
  normalize();
  return *this;
}


template<class T>
inline AngleDegrees<T>& AngleDegrees<T>::operator ++(int)
{
  ++a;
  normalize();
  return *this;
}


template<class T>
inline AngleDegrees<T>& AngleDegrees<T>::operator --()
{
  --a;
  normalize();
  return *this;
}


template<class T>
inline AngleDegrees<T>& AngleDegrees<T>::operator --(int)
{
  --a;
  normalize();
  return *this;
}


template<class T>
inline AngleDegrees<T>& AngleDegrees<T>::operator +=(const AngleDegrees<T>& ang)
{
  a += ang.a;
  normalize();
  return *this;
}


template<class T>
inline AngleDegrees<T>& AngleDegrees<T>::operator +=(const T& ang)
{
  a += ang;
  normalize();
  return *this;
}


template<class T>
inline AngleDegrees<T>& AngleDegrees<T>::operator -=(const AngleDegrees<T>& ang)
{
  a -= ang.a;
  normalize();
  return *this;
}


template<class T>
inline AngleDegrees<T>& AngleDegrees<T>::operator -=(const T& ang)
{
  a -= ang;
  normalize();
  return *this;
}


template<class T>
inline AngleDegrees<T>& AngleDegrees<T>::operator *=(const AngleDegrees<T>& ang)
{
  a *= ang.a;
  normalize();
  return *this;
}


template<class T>
inline AngleDegrees<T>& AngleDegrees<T>::operator *=(const T& d)
{
  a *= d;
  normalize();
  return *this;
}


template<class T>
inline AngleDegrees<T>& AngleDegrees<T>::operator /=(const AngleDegrees<T>& ang)
{
  a /= ang.a;
  normalize();
  return *this;
}


template<class T>
inline AngleDegrees<T>& AngleDegrees<T>::operator /=(const T& d)
{
  a /= d;
  normalize();
  return *this;
}


template<class T>
inline bool AngleDegrees<T>::operator !() const
{
  return (!a);
}


template<class T>
inline bool AngleDegrees<T>::operator ==(const AngleDegrees<T>& ang) const
{
  return (a == ang.a);
}


template<class T>
inline bool AngleDegrees<T>::operator ==(const T& ang) const
{
  return (a == ang);
}


template<class T>
inline bool AngleDegrees<T>::operator !=(const AngleDegrees<T>& ang) const
{
  return (a != ang.a);
}


template<class T>
inline bool AngleDegrees<T>::operator !=(const T& ang) const
{
  return (a != ang);
}


template<class T>
inline bool AngleDegrees<T>::operator >(const AngleDegrees<T>& ang) const
{
  return (a > ang.a);
}


template<class T>
inline bool AngleDegrees<T>::operator >(const T& ang) const
{
  return (a > ang);
}


template<class T>
inline bool AngleDegrees<T>::operator >=(const AngleDegrees<T>& ang) const
{
  return (a >= ang.a);
}


template<class T>
inline bool AngleDegrees<T>::operator >=(const T& ang) const
{
  return (a >= ang);
}


template<class T>
inline bool AngleDegrees<T>::operator <(const AngleDegrees<T>& ang) const
{
  return (a < ang.a);
}


template<class T>
inline bool AngleDegrees<T>::operator <(const T& ang) const
{
  return (a < ang);
}


template<class T>
inline bool AngleDegrees<T>::operator <=(const AngleDegrees<T>& ang) const
{
  return (a <= ang.a);
}


template<class T>
inline bool AngleDegrees<T>::operator <=(const T& ang) const
{
  return (a <= ang);
}


template<class T>
inline bool operator ==(const T& op1, const AngleDegrees<T>& op2)
{
  return (op2 == op1);
}


template<class T>
inline bool operator !=(const T& op1, const AngleDegrees<T>& op2)
{
  return (op2 != op1);
}


template<class T>
inline bool operator <(const T& op1, const AngleDegrees<T>& op2)
{
  return (op2 > op1);
}


template<class T>
inline bool operator <=(const T& op1, const AngleDegrees<T>& op2)
{
  return (op2 >= op1);
}


template<class T>
inline bool operator >(const T& op1, const AngleDegrees<T>& op2) 
{
  return (op2 < op1);
}


template<class T>
inline bool operator >=(const T& op1, const AngleDegrees<T>& op2)
{
  return (op2 <= op1);
}


/*
template<class T>
inline AngleDegrees<T>::operator T() const
{
  return a;
}
*/


template<class T>
inline std::ostream& operator <<(std::ostream& os, const AngleDegrees<T>& ang)
{
#ifdef RS_DEG_HEAD
  return os << RS_DEG_HEAD << "(" << ang.getDeg() << ")";
#else
  return os << ang.getDeg();
#endif
}


template<class T>
inline std::istream& operator >>(std::istream& is, AngleDegrees<T>& ang)
{
  T inp;
  is >> inp;
  if (is) ang = AngleDegrees<T>(inp);
  return is;
}


template<class T>
inline const std::string toString(const AngleDegrees<T>& ang)
{
#ifdef RS_DEG_HEAD
  return RS_DEG_HEAD + "(" + toString(ang.getDeg()) + ")";
#else
  return toString(ang.getDeg());
#endif
}


template<class T>
AngleDegrees<T>::AngleDegrees()
  : a()
{
}


template<class T>
AngleDegrees<T>::AngleDegrees(const T& ang)
  : a(ang)
{
  normalize();
}


template<class T>
AngleDegrees<T>::AngleDegrees(const AngleDegrees<T>& ang)
  : a(ang.getDeg())
{
}


template<class T>
AngleDegrees<T>::AngleDegrees(const AngleRadians<T>& ang)
  : a(ang.getDeg())
{
}


// assign, no range check
template<class T>
AngleDegrees<T>& AngleDegrees<T>::operator =(const AngleRadians<T>& ang)
{
  a = ang.getDeg();
  return *this;
}


template<class T>
void AngleDegrees<T>::normalize()
{
  // Equal to:
  // a = Mod((a + 180), 360.0) - 180;
  // don't know which is more efficient...
  while (a <= -T(180)) a += T(360);
  while (a > T(180)) a -= T(360);
}


template<class T>
inline T sin(const AngleDegrees<T>& x) 
{
  return std::sin(x.getRad());
}


template<class T>
inline T cos(const AngleDegrees<T>& x)
{
  return std::cos(x.getRad());
}


template<class T>
inline T tan(const AngleDegrees<T>& x)
{
  return std::tan(x.getRad());
}


#ifdef RS_TRIG_IN_DEG
// Must decied wether ASin, ACos, ATan and ATan2 should return a AngleRadians<T>
// or a AngleDegrees<T>. The most efficient is to return a AngleRadians<T>, but the
// most used (in this context) is AngleDegrees<T>. Therefore the current choice
// is AngleDegrees<T>.
  
template<class T>
inline AngleDegrees<T> ASin(const T& x)
{
  return (x >= T(1)) ? AngleDegrees<T>(T(90)) : (x <= T(-1)) ? AngleDegrees<T>(T(-90)) : AngleDegrees<T>(rad2Deg(std::asin(x)));
}


template<class T>
inline AngleDegrees<T> ACos(const T& x)
{
  return (x >= T(1)) ? AngleDegrees<T>(T(0)) : (x <= T(-1)) ? AngleDegrees<T>(T(180)) : AngleDegrees<T>(rad2Deg(std::acos(x)));
}


template<class T>
inline AngleDegrees<T> ATan(const T& x)
{
  return AngleDegrees<T>(rad2Deg(std::atan(x)));
}


template<class T>
inline AngleDegrees<T> ATan2(const T& x, const T& y)
{
  return (x == T(0) && y == T(0)) ? AngleDegrees<T>(T(0)) : AngleDegrees<T>(rad2Deg(std::atan2(x,y)));
}


#endif // RS_TRIG_IN_DEG


RS_END_NAMESPACE


#endif // !ANGLEDEGREES_H
