/* -*- Mode: C++ -*-
 * $Id: AngleRadians.h,v 1.5 2000/10/23 12:04:14 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: AngleRadians.h
 * Unitname: Library
 * $Revision: 1.5 $
 * Created by: Fredrik Heintz 2000-08-30
 * Last modified by $Author: frehe $ $Date: 2000/10/23 12:04:14 $
 *
 *
 * HISTORY
 *
 * $Log: AngleRadians.h,v $
 * Revision 1.5  2000/10/23 12:04:14  frehe
 * Changed the angles to be in the interval -180 < angle <= 180.
 *
 * Revision 1.4  2000/09/19 13:30:25  frehe
 * Fixed a bug in AngleRadians that converted radians from degrees to radians.
 *
 * Revision 1.3  2000/09/03 14:23:19  frehe
 * Made minor changes to improve the documentation.
 *
 * Revision 1.2  2000/09/01 03:23:51  frehe
 * Made minor changes.
 *
 * Revision 1.1.1.1  2000/08/31 13:26:04  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file AngleRadians.h
 * Include file for the template class AngleRadians.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2000/10/23 12:04:14 $
 */


#ifndef ANGLERADIANS_H
#define ANGLERADIANS_H


//#include <iostream>
//#include <string>

#include "AngleDegrees.h"
//#include "machine_dependent_compiler_directives.h"
//#include "math_utils.h"


#define RS_RAD_HEAD std::string("#Rad")


RS_BEGIN_NAMESPACE


template<class T> class AngleDegrees;


/**
 * @class AngleRadians AngleRadians.h
 * Template class for representing angles in radians.
 * The template argument only defines what type the
 * angles should be represented by.
 * The only demands on the template argument T is that the trigonometric
 * and arithmetic functions are defined on it.
 *
 * @todo Make a single template of both AngleDegrees and AngleRadians
 *       with a second parameter which defines what the standard
 *       representation is (radians or degrees).
 *
 * @par Unit:
 *      Library
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2000/10/23 12:04:14 $
 */
template<class T> class AngleRadians {
public:
  /** Default constructor, the value of the angle is T(). */
  AngleRadians();

  /** Construct and AngleDegrees with the value @a ang. */
  AngleRadians(const T&);

  /** Copy constructor. */
  AngleRadians(const AngleRadians<T>&);

  /** Conversion constructor that converts an AngleDegrees to an AngleRadians. */
  AngleRadians(const AngleDegrees<T>&);

  
  /** Get the angle in radians. */
  T getRad() const;

  /** Get the angle in degrees. */
  T getDeg() const;

  
  //@{
  /** @name Assignment operators. */
  AngleRadians<T>& operator  =(const AngleRadians<T>&);
  AngleRadians<T>& operator  =(const AngleDegrees<T>&);
  AngleRadians<T>& operator  =(const T&);
  //@}
  
  
  //@{
  /** @name Arithmetic operators. */
  /** I don't think I have to define all the operators for T too,
    since the automatic type conversion will convert T to AngleRadians<T>.
    But the efficiency is a little bit better. */
  AngleRadians<T> operator -() const;
  AngleRadians<T> operator +(const AngleRadians<T>&) const;
  AngleRadians<T> operator +(const T&) const;
  AngleRadians<T> operator -(const AngleRadians<T>&) const;
  AngleRadians<T> operator -(const T&) const;
  AngleRadians<T> operator *(const AngleRadians<T>&) const;
  AngleRadians<T> operator *(const T&) const;
  AngleRadians<T> operator /(const AngleRadians<T>&) const;
  AngleRadians<T> operator /(const T&) const;

  AngleRadians<T>& operator +=(const AngleRadians<T>&);
  AngleRadians<T>& operator +=(const T&);
  AngleRadians<T>& operator -=(const AngleRadians<T>&);
  AngleRadians<T>& operator -=(const T&);
  AngleRadians<T>& operator *=(const AngleRadians<T>&);
  AngleRadians<T>& operator *=(const T&);
  AngleRadians<T>& operator /=(const AngleRadians<T>&);
  AngleRadians<T>& operator /=(const T&);
  //@}
  

  //@{
  /** @name Increment and decrement operators. */
  AngleRadians<T>& operator ++();
  AngleRadians<T>& operator ++(int);
  AngleRadians<T>& operator --();
  AngleRadians<T>& operator --(int);
  //@}
  
  
  //@{
  /** Comparison operators. */
  bool operator ! () const;
  bool operator ==(const AngleRadians<T>&) const;
  bool operator ==(const T&) const;
  bool operator !=(const AngleRadians<T>&) const;
  bool operator !=(const T&) const;
  bool operator > (const AngleRadians<T>&) const;
  bool operator > (const T&) const;
  bool operator >=(const AngleRadians<T>&) const;
  bool operator >=(const T&) const;
  bool operator < (const AngleRadians<T>&) const;
  bool operator < (const T&) const;
  bool operator <=(const AngleRadians<T>&) const;
  bool operator <=(const T&) const;
  //@}
  
  
  //operator T() const;

protected:
  /** Function that normalizes an angle to the interval [-PI, PI]. */
  void normalize();
  /** The actual representation of the angle value. */
  T a;
};

/*
#if CHECK_DIVISION_BY_ZERO == 1
static const AngleRadians<T> AngleRadians_DIVISION_BY_ZERO_RESULT(T_DIVISION_BY_ZERO_RESULT);
#endif CHECK_DIVISION_BY_ZERO
*/

template<class T>
inline T AngleRadians<T>::getRad() const
{
  return a;
}

template<class T>
inline T AngleRadians<T>::getDeg() const
{
  return rad2Deg(a);
}

template<class T>
inline AngleRadians<T> AngleRadians<T>::operator -() const
{
  return AngleRadians<T>(-a);
}

template<class T>
inline AngleRadians<T> AngleRadians<T>::operator +(const AngleRadians<T>& ang) const
{
  return AngleRadians<T>(a + ang.a);
}

template<class T>
inline AngleRadians<T> AngleRadians<T>::operator +(const T& ang) const
{
  return AngleRadians<T>(a + ang);
}

template<class T>
inline AngleRadians<T> AngleRadians<T>::operator -(const AngleRadians<T>& ang) const
{
  return AngleRadians<T>(a - ang.a);
}

template<class T>
inline AngleRadians<T> AngleRadians<T>::operator -(const T& ang) const
{
  return AngleRadians<T>(a - ang);
}

template<class T>
inline AngleRadians<T> AngleRadians<T>::operator *(const AngleRadians<T>& ang) const
{
  return AngleRadians<T>(a * ang.a);
}

template<class T>
inline AngleRadians<T> AngleRadians<T>::operator *(const T& ang) const
{
  return AngleRadians<T>(a * ang);
}

template<class T>
inline AngleRadians<T> AngleRadians<T>::operator /(const AngleRadians<T>& ang) const
{
#if CHECK_DIVISION_BY_ZERO == 1
  if (ang.a == T(0))
    return AngleRadians_DIVISION_BY_ZERO_RESULT;
  else
#endif
    return AngleRadians<T>(a / ang.a);
}

template<class T>
inline AngleRadians<T> AngleRadians<T>::operator /(const T& ang) const
{
#if CHECK_DIVISION_BY_ZERO == 1
  if (ang == T(0))
    return AngleRadians_DIVISION_BY_ZERO_RESULT;
  else
#endif
    return AngleRadians<T>(a / ang);
}

template<class T>
inline AngleRadians<T> operator +(const T& op1, const AngleRadians<T>& op2)
{
  return op2 + op1;
}

template<class T>
inline AngleRadians<T> operator -(const T& op1, const AngleRadians<T>& op2)
{
  return (-op2) + op1;
}

template<class T>
inline AngleRadians<T> operator *(const T& op1, const AngleRadians<T>& op2)
{
  return op2 * op1;
}

template<class T>
inline AngleRadians<T> operator /(const T& op1, const AngleRadians<T>& op2)
{
#if CHECK_DIVISION_BY_ZERO == 1
  if (op2.a == T(0))
    return AngleRadians_DIVISION_BY_ZERO_RESULT;
  else
#endif
    return AngleRadians<T>(op1 / op2.a);
}

// assign, no range check
template<class T>
inline AngleRadians<T>& AngleRadians<T>::operator =(const AngleRadians<T>& ang)
{
  a = ang.a;
  return *this;
}

// assign, range check
template<class T>
inline AngleRadians<T>& AngleRadians<T>::operator =(const T& d)
{
  a = d;
  normalize();
  return *this;
}

template<class T>
inline AngleRadians<T>& AngleRadians<T>::operator ++()
{
  ++a;
  normalize();
  return *this;
}

template<class T>
inline AngleRadians<T>& AngleRadians<T>::operator ++(int)
{
  ++a;
  normalize();
  return *this;
}

template<class T>
inline AngleRadians<T>& AngleRadians<T>::operator --()
{
  --a;
  normalize();
  return *this;
}

template<class T>
inline AngleRadians<T>& AngleRadians<T>::operator --(int)
{
  --a;
  normalize();
  return *this;
}

template<class T>
inline AngleRadians<T>& AngleRadians<T>::operator +=(const AngleRadians<T>& ang)
{
  a += ang.a;
  normalize();
  return *this;
}

template<class T>
inline AngleRadians<T>& AngleRadians<T>::operator +=(const T& ang)
{
  a += ang;
  normalize();
  return *this;
}

template<class T>
inline AngleRadians<T>& AngleRadians<T>::operator -=(const AngleRadians<T>& ang)
{
  a -= ang.a;
  normalize();
  return *this;
}

template<class T>
inline AngleRadians<T>& AngleRadians<T>::operator -=(const T& ang)
{
  a -= ang;
  normalize();
  return *this;
}

template<class T>
inline AngleRadians<T>& AngleRadians<T>::operator *=(const AngleRadians<T>& ang)
{
  a *= ang.a;
  normalize();
  return *this;
}

template<class T>
inline AngleRadians<T>& AngleRadians<T>::operator *=(const T& d)
{
  a *= d;
  normalize();
  return *this;
}

template<class T>
inline AngleRadians<T>& AngleRadians<T>::operator /=(const AngleRadians<T>& ang)
{
#if CHECK_DIVISION_BY_ZERO == 1
  if (ang.a == T(0))
    a = T_DIVISION_BY_ZERO_RESULT;
  else
#endif
    a /= ang.a;

  normalize();
  return *this;
}

template<class T>
inline AngleRadians<T>& AngleRadians<T>::operator /=(const T& d)
{
#if CHECK_DIVISION_BY_ZERO == 1
  if (d == T(0))
    a = T_DIVISION_BY_ZERO_RESULT;
  else
#endif
    a /= d;

  normalize();
  return *this;
}

template<class T>
inline bool AngleRadians<T>::operator !() const
{
  return (!a);
}

template<class T>
inline bool AngleRadians<T>::operator ==(const AngleRadians<T>& ang) const
{
  return (a == ang.a);
}

template<class T>
inline bool AngleRadians<T>::operator ==(const T& ang) const
{
  return (a == ang);
}

template<class T>
inline bool AngleRadians<T>::operator !=(const AngleRadians<T>& ang) const
{
  return (a != ang.a);
}

template<class T>
inline bool AngleRadians<T>::operator !=(const T& ang) const
{
  return (a != ang);
}

template<class T>
inline bool AngleRadians<T>::operator >(const AngleRadians<T>& ang) const
{
  return (a > ang.a);
}

template<class T>
inline bool AngleRadians<T>::operator >(const T& ang) const
{
  return (a > ang);
}

template<class T>
inline bool AngleRadians<T>::operator >=(const AngleRadians<T>& ang) const
{
  return (a >= ang.a);
}

template<class T>
inline bool AngleRadians<T>::operator >=(const T& ang) const
{
  return (a >= ang);
}

template<class T>
inline bool AngleRadians<T>::operator <(const AngleRadians<T>& ang) const
{
  return (a < ang.a);
}

template<class T>
inline bool AngleRadians<T>::operator <(const T& ang) const
{
  return (a < ang);
}

template<class T>
inline bool AngleRadians<T>::operator <=(const AngleRadians<T>& ang) const
{
  return (a <= ang.a);
}

template<class T>
inline bool AngleRadians<T>::operator <=(const T& ang) const
{
  return (a <= ang);
}

template<class T>
inline bool operator ==(const T& op1, const AngleRadians<T>& op2)
{
  return op2 == op1;
}

template<class T>
inline bool operator !=(const T& op1, const AngleRadians<T>& op2)
{
  return op2 != op1;
}

template<class T>
inline bool operator <(const T& op1, const AngleRadians<T>& op2)
{
  return op2 > op1;
}

template<class T>
inline bool operator <=(const T& op1, const AngleRadians<T>& op2)
{
  return op2 >= op1;
}

template<class T>
inline bool operator >(const T& op1, const AngleRadians<T>& op2)
{
  return op2 < op1;
}

template<class T>
inline bool operator >=(const T& op1, const AngleRadians<T>& op2)
{
  return op2 <= op1;
}

/*
template<class T>
inline AngleRadians<T>::operator T() const
{
  return a;
}
*/

template<class T>
inline std::ostream& operator <<(std::ostream& os, const AngleRadians<T>& ang)
{
#ifdef RS_RAD_HEAD
  return os << RS_RAD_HEAD << "(" << ang.getRad() << ")";
#else
  return os << ang.getRad();
#endif
}

template<class T>
inline std::istream& operator>>(std::istream& is, AngleRadians<T>& ang)
{
  T inp;
  return is >> inp;
  if (is) ang = AngleRadians<T>(inp);
}

template<class T>
inline const std::string toString(const AngleRadians<T>& ang)
{
#ifdef RS_RAD_HEAD
  return RS_RAD_HEAD + "(" + toString(ang.getRad()) + ")";
#else
  return toString(ang.getRad());
#endif
}

template<class T>
AngleRadians<T>::AngleRadians()
  : a()
{
}

template<class T>
AngleRadians<T>::AngleRadians(const T& d)
  : a(d)
{
  normalize();
}

template<class T>
AngleRadians<T>::AngleRadians(const AngleRadians<T>& ang)
  : a(ang.a)
{
}

template<class T>
AngleRadians<T>::AngleRadians(const AngleDegrees<T>& ang)
  : a(ang.getRad())
{
}

// assign, no range check
template<class T>
AngleRadians<T>& AngleRadians<T>::operator =(const AngleDegrees<T>& ang)
{
  a = ang.getRad();
  return *this;
}

template<class T>
void AngleRadians<T>::normalize()
{
  while (a <= -PI) a += TWO_PI;
  while (a > PI) a -= TWO_PI;
}

template<class T>
inline T sin(const AngleRadians<T>& x) 
{
  return std::sin(x.getRad());
}

template<class T>
inline T cos(const AngleRadians<T>& x)
{
  return std::cos(x.getRad());
}

template<class T>
inline T tan(const AngleRadians<T>& x)
{
  return std::tan(x.getRad());
}

#if RS_TRIG_IN_RAD == 1
// Must decied wether ASin, ACos, ATan and ATan2 should return a AngleRadians<T>
// or a AngleDegrees<T>. The most efficient is to return a AngleRadians<T>, but the
// most used (in this context) is AngleDegrees<T>. Therefore the current choice is
// AngleDegrees<T>.
  
template<class T>
inline AngleRadians<T> ASin(const T& x)
{
  return (x >= T(1)) ? AngleRadians<T>(T(90)) : (x <= T(-1)) ? AngleRadians<T>(T(-90)) : AngleRadians<T>(std:asin(x));
}

template<class T>
inline AngleRadians<T> ACos(const T& x)
{
  return (x >= T(1)) ? AngleRadians<T>(T(0)) : (x <= T(-1)) ? AngleRadians<T>(T(180)) : AngleRadians<T>(std:acos(x));
}

template<class T>
inline AngleRadians<T> ATan(const T& x)
{
  return AngleRadians<T>(std::atan(x));
}

template<class T>
inline AngleRadians<T> ATan2(const T& x, const T& y)
{
  return (x == T(0) && y == T(0)) ? AngleRadians<T>(T(0)) : AngleRadians<T>(std::atan2(x,y));
}

#endif // RS_TRIG_IN_RAD


RS_END_NAMESPACE


#endif // ANGLERADIANS_H
