/* -*- Mode: C++ -*-
 * $Id: Confidence.h,v 1.1 2000/09/01 01:24:37 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Confidence.h
 * Unitname: Library
 * $Revision: 1.1 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/01 01:24:37 $
 *
 *
 * HISTORY
 *
 * $Log: Confidence.h,v $
 * Revision 1.1  2000/09/01 01:24:37  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 *
 */

/**
 * @file Confidence.h
 * Include file for the class CConfidence.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2000/09/01 01:24:37 $
 */


#ifndef CONFIDENCE_H
#define CONFIDENCE_H


#include <string>
#include <iostream>

#include "machine_dependent_types.h"
#include "math_utils.h"


RS_BEGIN_NAMESPACE


/**
 * @class Confidence Confidence.h
 * Class for representing confidence values.
 *
 * <!-- @todo -->
 *
 * @par Unit:
 *      Library
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2000/09/01 01:24:37 $
 */
class Confidence
{
public:
  Confidence(const Float32 c = minConfDefault,
	     const Float32 min_c = minConfDefault,
	     const Float32 max_c = maxConfDefault);
  Confidence(const Confidence& c);

  // Should plus and minus be defined on Confidence? Probably!
  // Does it really "mean" anything? Yes, something...
  // What should be done with the limits?
  // Now I take the highest Max and lowest Min.
  // Can it be negative? Only if min_conf is negative
  // since all values should be between min_conf and max_conf.
  Confidence operator -()                   const;
  Confidence operator +(const Confidence&) const;
  Confidence operator +(const Float32)    const;
  Confidence operator -(const Confidence&) const;
  Confidence operator -(const Float32)    const;
  Confidence operator *(const Confidence&) const;
  Confidence operator *(const Float32)    const;
  Confidence operator /(const Confidence&) const;
  Confidence operator /(const Float32)    const;
  
  Confidence& operator +=(const Confidence&);
  Confidence& operator +=(const Float32);
  Confidence& operator -=(const Confidence&);
  Confidence& operator -=(const Float32);
  Confidence& operator *=(const Confidence&);
  Confidence& operator *=(const Float32);
  Confidence& operator /=(const Confidence&);
  Confidence& operator /=(const Float32);

  bool operator ==(const Confidence&) const;
  bool operator ==(const Float32)    const;
  bool operator !=(const Confidence&) const;
  bool operator !=(const Float32)    const;
  bool operator > (const Confidence&) const;
  bool operator > (const Float32)    const;
  bool operator >=(const Confidence&) const;
  bool operator >=(const Float32)    const;
  bool operator < (const Confidence&) const;
  bool operator < (const Float32)    const;
  bool operator <=(const Confidence&) const;
  bool operator <=(const Float32)    const;
  bool operator ! ()                   const;

  Confidence& setMinConfidence();
  Confidence& setMaxConfidence();
  Confidence& setConfidence(const Float32);
  Float32 getConfidence() const;

  Confidence& setMaxConfidence(const Float32);
  Float32 getMaxConfidence() const;
  Confidence& setMinConfidence(const Float32);
  Float32 getMinConfidence() const;

  static Float32 getMaxConfDefault();
  static void  setMaxConfDefault(const Float32);
  static Float32 getMinConfDefault();
  static void  setMinConfDefault(const Float32);
  
  //friend std::ostream& operator<<(std::ostream&, const Confidence&);
  friend std::istream& operator>>(std::istream&, Confidence&);
  
protected:
  /** Class variables that keep track of the default
    max confidence value. */
  static Float32 maxConfDefault;
  /** Class variables that keep track of the default
    min confidence value. */
  static Float32 minConfDefault;

  Float32 minConf;
  Float32 maxConf;
  Float32 confidence;
  
};

/*
#if CHECK_DIVISION_BY_ZERO == 1
// This constant has the default limits. Don't know if that's the correct definition.
// Since one don't want to loose the current limits, this constant
// will not be used in the code.
static const Confidence Confidence_DIVISION_BY_ZERO_RESULT(FLOAT_TYPE_DIVISION_BY_ZERO_RESULT);
#endif CHECK_DIVISION_BY_ZERO
*/

inline Confidence Confidence::operator -() const
{
  return Confidence(-confidence, minConf, maxConf);
} 

inline Confidence Confidence::operator +(const Confidence& c) const
{
  return Confidence(confidence + c.confidence, min(minConf, c.minConf), max(maxConf, c.maxConf));
} 

inline Confidence Confidence::operator +(const Float32 d) const
{
  return Confidence(confidence + d, minConf, maxConf);
} 

inline Confidence Confidence::operator -(const Confidence& c) const
{
  return Confidence(confidence - c.confidence, min(minConf, c.minConf), max(maxConf, c.maxConf));
} 

inline Confidence Confidence::operator -(const Float32 d) const
{
  return Confidence(confidence - d, minConf, maxConf);
} 

inline Confidence Confidence::operator *(const Confidence& c) const
{
  return Confidence(confidence * c.confidence, min(minConf, c.minConf), max(maxConf, c.maxConf));
} 

inline Confidence Confidence::operator *(const Float32 d) const
{
  return Confidence(confidence * d, minConf, maxConf);
} 

inline Confidence Confidence::operator /(const Confidence& c) const
{
#if CHECK_DIVISION_BY_ZERO == 1
  if (c.confidence == 0)
    return Confidence(FLOAT_DIVISION_BY_ZERO_RESULT, min(minConf, c.minConf), max(maxConf, c.maxConf));
  else
#endif
  return Confidence(confidence / c.confidence, min(minConf, c.minConf), max(maxConf, c.maxConf));
} 

inline Confidence Confidence::operator /(const Float32 d) const
{
#if CHECK_DIVISION_BY_ZERO == 1
  if (d == 0)
    return Confidence(FLOAT_DIVISION_BY_ZERO_RESULT, minConf, maxConf);
  else
#endif
    return Confidence(confidence / d, minConf, maxConf);
} 

inline Confidence& Confidence::operator +=(const Confidence& c)
{
  confidence += c.confidence;
  minConf = min(minConf, c.minConf);
  maxConf = max(minConf, max(maxConf, c.maxConf));
  confidence = minMax(minConf, confidence, maxConf);
  return *this;
}

inline Confidence& Confidence::operator +=(const Float32 d)
{
  confidence += d;
  confidence = minMax(minConf, confidence, maxConf);  
  return *this;
}

inline Confidence& Confidence::operator -=(const Confidence& c)
{
  confidence -= c.confidence;
  minConf = min(minConf, c.minConf);
  maxConf = max(minConf, max(maxConf, c.maxConf));
  confidence = minMax(minConf, confidence, maxConf);
  return *this;
}

inline Confidence& Confidence::operator -=(const Float32 d)
{
  confidence -= d;
  confidence = minMax(minConf, confidence, maxConf);  
  return *this;
}

inline Confidence& Confidence::operator *=(const Confidence& c)
{
  confidence *= c.confidence;
  minConf = min(minConf, c.minConf);
  maxConf = max(minConf, max(maxConf, c.maxConf));
  confidence = minMax(minConf, confidence, maxConf);
  return *this;
}

inline Confidence& Confidence::operator *=(const Float32 d)
{
  confidence *= d;
  confidence = minMax(minConf, confidence, maxConf);  
  return *this;
}

inline Confidence& Confidence::operator /=(const Confidence& c)
{
#if CHECK_DIVISION_BY_ZERO == 1
  if (c.confidence == 0)
    confidence = FLOAT_DIVISION_BY_ZERO_RESULT;
  else
#endif
    confidence /= c.confidence;
  
  minConf = min(minConf, c.minConf);
  maxConf = max(minConf, max(maxConf, c.maxConf));
  confidence = minMax(minConf, confidence, maxConf);
  return *this;
}

inline Confidence& Confidence::operator /=(const Float32 d)
{
#if CHECK_DIVISION_BY_ZERO == 1
  if (d == 0)
    confidence = FLOAT_DIVISION_BY_ZERO_RESULT;
  else
#endif
    confidence /= d;
  
  confidence = minMax(minConf, confidence, maxConf);  
  return *this;
}

inline bool Confidence::operator ==(const Confidence& c) const
{
  return (confidence == c.confidence);
}

inline bool Confidence::operator ==(const Float32 d) const
{
  return (confidence == d);
}

inline bool Confidence::operator !=(const Confidence& c) const
{
  return (confidence != c.confidence);
}

inline bool Confidence::operator !=(const Float32 d) const
{
  return (confidence != d);
}

inline bool Confidence::operator > (const Confidence& c) const
{
  return (confidence > c.confidence);
}

inline bool Confidence::operator > (const Float32 d) const
{
  return (confidence > d);
}

inline bool Confidence::operator >=(const Confidence& c) const
{
  return (confidence >= c.confidence);
}

inline bool Confidence::operator >=(const Float32 d) const
{
  return (confidence >= d);
}

inline bool Confidence::operator < (const Confidence& c) const
{
  return (confidence < c.confidence);
}

inline bool Confidence::operator < (const Float32 d) const
{
  return (confidence < d);
}

inline bool Confidence::operator <=(const Confidence& c) const
{
  return (confidence <= c.confidence);
}

inline bool Confidence::operator <=(const Float32 d) const
{
  return (confidence <= d);
}

inline bool Confidence::operator ! () const
{
  return (confidence != 0);
}

inline Confidence& Confidence::setMinConfidence()
{
  confidence = minConf;
  return *this;
}

inline Confidence& Confidence::setMaxConfidence()
{
  confidence = maxConf;
  return *this;
}

inline Confidence& Confidence::setConfidence(const Float32 c)
{
  confidence = minMax(minConf, c, maxConf);
  return *this;
}

inline Float32 Confidence::getConfidence() const
{
  return confidence;
}

inline Confidence& Confidence::setMaxConfidence(const Float32 max_c)
{
  maxConf = max(max_c, minConf);
  confidence = min(confidence, maxConf);
  return *this;
}

inline Float32 Confidence::getMaxConfidence() const
{
  return maxConf;
}


inline Confidence& Confidence::setMinConfidence(const Float32 min_c)
{
  minConf = min(min_c, maxConf);
  confidence = max(confidence, minConf);
  return *this;
}

inline Float32 Confidence::getMinConfidence() const
{
  return minConf;
}

inline Float32 Confidence::getMaxConfDefault() 
{
  return maxConfDefault;
}

inline void Confidence::setMaxConfDefault(const Float32 max_c_def)
{
  maxConfDefault = max(minConfDefault, max_c_def);
}

inline Float32 Confidence::getMinConfDefault()
{
  return minConfDefault;
}

inline void Confidence::setMinConfDefault(const Float32 min_c_def)
{
  minConfDefault = min(maxConfDefault, min_c_def);
}
  
inline std::ostream& operator<<(std::ostream& os, const Confidence& c)
{
#ifdef CONFIDENCE_HEAD
  return os << CONFIDENCE_HEAD << "(" << c.getConfidence() << ")";
#else
  return os << c.getConfidence();
#endif
}

inline std::istream& operator>>(std::istream& is, Confidence& c)
{
  is >> c.confidence;
  return is;
}


RS_END_NAMESPACE


#endif // !CONFIDENCE_H
