/* -*- Mode: C++ -*-
 * $Id: CoordPoint.h,v 1.5 2000/09/20 15:17:41 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: CoordPoint.h
 * Unitname: Library
 * $Revision: 1.5 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/20 15:17:41 $
 *
 *
 * HISTORY
 *
 * $Log: CoordPoint.h,v $
 * Revision 1.5  2000/09/20 15:17:41  frehe
 * Added documentation.
 *
 * Revision 1.4  2000/09/18 15:20:04  frehe
 * Due to a problem with namespaces the name for atan2, atan, acos and asin is changed to ATan2, ATan, ASin and ACos.
 *
 * Revision 1.3  2000/09/04 15:21:48  frehe
 * Changed the name of the trigonometric functions to the same as in the standard math library (sin, cos, tan, asin, acos, atan, and atan2).
 *
 * Revision 1.2  2000/09/01 03:23:51  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/01 01:24:37  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 *
 */

/**
 * @file CoordPoint.h
 * Include file for the template CoordPoint.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2000/09/20 15:17:41 $
 */


#ifndef COORDPOINT_H
#define COORDPOINT_H


#include <iostream>

#include "math_utils.h"
#include "AngleDegrees.h"
#include "CoordVector.h"


/** String to use as the header when outputing a CoordPoint. */
#define POINT_HEAD std::string("#Point")


RS_BEGIN_NAMESPACE


template <class T> class CoordVector;


/**
 * @class CoordPoint CoordPoint.h
 * Template class for representing a cartesian two dimensional
 * coordinate. The X and Y values are of type T.
 *
 * <!-- @todo -->
 *
 * @par Unit:
 *      Library
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2000/09/20 15:17:41 $
 */
template<class T> class CoordPoint {
public:
  CoordPoint();
  CoordPoint(const T&, const T&);
  CoordPoint(const CoordPoint<T>&);
  CoordPoint(const CoordVector<T>&, const CoordPoint<T>&);
  CoordPoint(const CoordPoint<T>&, const CoordVector<T>&);
  explicit CoordPoint(const CoordVector<T>&);

  CoordPoint<T> operator - () const;
  CoordPoint<T> operator - (const CoordPoint<T>&) const;
  CoordPoint<T> operator - (const CoordVector<T>&) const;
  CoordPoint<T> operator + (const CoordPoint<T>&) const;
  CoordPoint<T> operator + (const CoordVector<T>&) const;
  CoordPoint<T> operator * (const T&) const;
  CoordPoint<T> operator * (const CoordPoint<T>&) const;
  CoordPoint<T> operator / (const T&) const;
  CoordPoint<T> operator / (const CoordPoint<T>&) const;

  CoordPoint<T>& operator  =(const CoordPoint<T>&);
  CoordPoint<T>& operator  =(const CoordVector<T>&);

  CoordPoint<T>& operator -=(const CoordPoint<T>&);
  CoordPoint<T>& operator -=(const CoordVector<T>&);
  CoordPoint<T>& operator +=(const CoordPoint<T>&);
  CoordPoint<T>& operator +=(const CoordVector<T>&);
  CoordPoint<T>& operator *=(const T&);
  CoordPoint<T>& operator *=(const CoordPoint<T>&);
  CoordPoint<T>& operator /=(const T&);
  CoordPoint<T>& operator /=(const CoordPoint<T>&);
  
  bool operator ! () const;
  bool operator ==(const CoordPoint<T>&) const;
  bool operator !=(const CoordPoint<T>&) const;

  /* Define these in a separate file. They need not be methods.
  // These use the coordinate system of the soccer field
  bool InFront(const CoordPoint<T>&) const;  // Equal to LeftOf
  bool Behind(const CoordPoint<T>&)  const;   // Equal to RightOf
  // These use standar coordinate system
  bool LeftOf(const CoordPoint<T>&)  const;
  bool RightOf(const CoordPoint<T>&) const;
  boolbove(const CoordPoint<T>&)     const;
  bool Below(const CoordPoint<T>&)   const;
  */
  
  T getX() const;
  T getY() const;
  CoordPoint<T>& addX(const T& pt_x);
  CoordPoint<T>& addY(const T& pt_y);
  CoordPoint<T>& addXY(const T& pt_x, const T& pt_y);
  CoordPoint<T>& setX(const T& pt_x);
  CoordPoint<T>& setY(const T& pt_y);
  CoordPoint<T>& set (const T& pt_x, const T& pt_y);
  
  CoordVector<T> getVectorTo(const CoordPoint<T>& pt) const;
  AngleDegrees<T> getDirectionTo(const CoordPoint<T>& pt) const;
  T getDistanceTo(const CoordPoint<T>& pt) const;
  
protected:
  T x;
  T y;  
};


// Extend this to be the point on the line between pt1 and pt2 
// with a certain ratio 1:1 is the middle point?
// CoordPoint<T> MiddlePoint(const CoordPoint<T>&, const CoordPoint<T>&);
template<class T>
inline CoordPoint<T> middlePoint(const CoordPoint<T>& pt1,
				 const CoordPoint<T>& pt2)
{
  return pt1 + CoordPoint<T>((pt2.getX() - pt1.getX())/2,
			     (pt2.getY() - pt1.getY())/2);
}

template<class T>
inline CoordPoint<T> CoordPoint<T>::operator -() const
{
  return CoordPoint<T>(-x,-y);
}

template<class T>
inline CoordPoint<T> CoordPoint<T>::operator -(const CoordPoint<T>& pt) const
{
  return CoordPoint<T>(x - pt.x, y - pt.y);
}

template<class T>
inline CoordPoint<T> CoordPoint<T>::operator *(const T& a) const
{
  return CoordPoint<T>(x * a, y * a);
}

template<class T>
inline CoordPoint<T> CoordPoint<T>::operator *(const CoordPoint<T>& pt) const
{
  return CoordPoint<T>(x * pt.x, y * pt.y);
}

template<class T>
inline CoordPoint<T> operator *(const T& a, const CoordPoint<T>& pt)
{
  return pt * a;
}

template<class T>
inline CoordPoint<T> CoordPoint<T>::operator /(const T& a) const
{
#if CHECK_DIVISION_BY_ZERO == 1
  if (a == T(0))
    return COORDPOINT_DIVISION_BY_ZERO_RESULT;
  else
#endif
    return CoordPoint<T>(x / a, y / a);
}

template<class T>
inline CoordPoint<T> CoordPoint<T>::operator /(const CoordPoint<T>& pt) const
{
#if CHECK_DIVISION_BY_ZERO == 1
  if ( pt.x == T(0) ) {
    if ( pt.y == T(0) ) {
      return CoordPoint_DIVISION_BY_ZERO_RESULT;
    } else {
      return CoordPoint<T>(T_DIVISION_BY_ZERO_RESULT, y / pt.y);
    }
  } else {
    if ( pt.y == T(0) ) {
      return CoordPoint<T>(x / pt.x, T_DIVISION_BY_ZERO_RESULT);
    } else {
#endif
      return CoordPoint<T>(x / pt.x, y / pt.y);
#if CHECK_DIVISION_BY_ZERO == 1      
    }
  }
#endif
}

template<class T>
inline CoordPoint<T> CoordPoint<T>::operator +(const CoordPoint<T>& pt) const
{
  return CoordPoint<T>(x + pt.x, y + pt.y);
}

template<class T>
inline CoordPoint<T>& CoordPoint<T>::operator-=(const CoordPoint<T>& pt)
{
  x -= pt.x;
  y -= pt.y;
  return *this;
}

template<class T>
inline CoordPoint<T>& CoordPoint<T>::operator-=(const CoordVector<T>& vec)
{
  *this = *this - vec;
  return *this;
}

template<class T>
inline CoordPoint<T>& CoordPoint<T>::operator+=(const CoordPoint<T>& pt)
{
  x += pt.x;
  y += pt.y;
  return *this;
}

template<class T>
inline CoordPoint<T>& CoordPoint<T>::operator+=(const CoordVector<T>& vec)
{
  *this = *this + vec;
  return *this;
}

template<class T>
inline CoordPoint<T>& CoordPoint<T>::operator*=(const T& a)
{
  x *= a;
  y *= a;
  return *this;
}

template<class T>
inline CoordPoint<T>& CoordPoint<T>::operator*=(const CoordPoint<T>& pt)
{
  x *= pt.x;
  y *= pt.y;
  return *this;
}

template<class T>
inline CoordPoint<T>& CoordPoint<T>::operator/=(const T& a)
{
#if CHECK_DIVISION_BY_ZERO == 1
  if ( a == T(0)) {
    x = T_DIVISION_BY_ZERO_RESULT;
    y = T_DIVISION_BY_ZERO_RESULT;
  } else {
#endif
    x /= a;
    y /= a;
#if CHECK_DIVISION_BY_ZERO == 1
  }
#endif
  return *this;
}

template<class T>
inline CoordPoint<T>& CoordPoint<T>::operator/=(const CoordPoint<T>& pt)
{
#if CHECK_DIVISION_BY_ZERO == 1
  if ( pt.x == T(0) )
    x = T_DIVISION_BY_ZERO_RESULT;
  else
#endif
    x /= pt.x;
  
#if CHECK_DIVISION_BY_ZERO == 1
  if ( pt.y == T(0) )
    y = T_DIVISION_BY_ZERO_RESULT;
  else
#endif
    y /= pt.y;

  return *this;
}

template<class T>
CoordPoint<T>& CoordPoint<T>::operator =(const CoordPoint<T>& pt)
{
  if ( this != &pt ) {
    y = pt.y;
    x = pt.x;
  }
  return *this;
}

template<class T>
inline CoordPoint<T>& CoordPoint<T>::operator =(const CoordVector<T>& vec)
{
  *this = CoordPoint(vec);
  return *this;
}

template<class T>
inline bool CoordPoint<T>::operator !() const
{
  return (!y) && (!x);
}

template<class T>
inline bool CoordPoint<T>::operator ==(const CoordPoint<T>& pt) const
{
  return (y == pt.y) && (x == pt.x);
}

template<class T>
inline bool CoordPoint<T>::operator !=(const CoordPoint<T>& pt) const
{
  return (y != pt.y) || (x != pt.x);
}

/*
inline bool CoordPoint<T>::InFront(const CoordPoint<T>& pt) const
{
  return (x > pt.x);
}

inline bool CoordPoint<T>::Behind(const CoordPoint<T>& pt) const
{
  return (x < pt.x);
}
*/

/*
inline bool CoordPoint<T>::LeftOf(const CoordPoint<T>& pt) const
{
  return (x > pt.x);
}

inline bool CoordPoint<T>::RightOf(const CoordPoint<T>& pt) const
{
  return (x < pt.x);
}

inline bool CoordPoint<T>::Above(const CoordPoint<T>& pt) const
{
  return (y < pt.y);
}

inline bool CoordPoint<T>::Below(const CoordPoint<T>& pt) const
{
  return (y > pt.y);
}
*/

template<class T>
inline CoordPoint<T>& CoordPoint<T>::addX (const T& pt_x)
{
  x += pt_x;
  return *this;
}

template<class T>
inline CoordPoint<T>& CoordPoint<T>::addY (const T& pt_y)
{
  y += pt_y;
  return *this;
}

template<class T>
inline CoordPoint<T>& CoordPoint<T>::addXY (const T& pt_x, const T& pt_y)
{
  x += pt_x;
  y += pt_y;
  return *this;
}

template<class T>
inline T CoordPoint<T>::getX() const
{
  return x;
}

template<class T>
inline T CoordPoint<T>::getY() const
{
  return y;
}

template<class T>
inline CoordPoint<T>& CoordPoint<T>::setX(const T& pt_x)
{
  x = pt_x;
  return *this;
}

template<class T>
inline CoordPoint<T>& CoordPoint<T>::setY(const T& pt_y)
{
  y = pt_y;
  return *this;
}

template<class T>
inline CoordPoint<T>& CoordPoint<T>::set(const T& pt_x, const T& pt_y)
{
  x = pt_x;
  y = pt_y;
  return *this;
}

template<class T>
inline AngleDegrees<T> CoordPoint<T>::getDirectionTo(const CoordPoint<T>& pt) const
{
  return ATan2((pt.y - y), (pt.x - x));
}

template<class T>
inline T CoordPoint<T>::getDistanceTo(const CoordPoint<T>& pt) const
{
  return sqrt(sqr(pt.x - x) + sqr(pt.y - y));
}

template<class T>
inline std::ostream& operator<<(std::ostream& os, const CoordPoint<T>& pt)
{
#ifdef POINT_HEAD
  return os << POINT_HEAD << "(" << pt.getX() << ", " << pt.getY() << ")";
#else
  return os << "(" << pt.getX() << ", " << pt.getY() << ")";
#endif
}

template<class T>
inline const std::string ToString(const CoordPoint<T>& pt)
{
#ifdef POINT_HEAD
  return POINT_HEAD + "(" + ToString(pt.getX()) + ", " + ToString(pt.getY()) + ")";
#else
  return "(" + ToString(pt.getX()) + ", " + ToString(pt.getY()) + ")";
#endif
}

template<class T>
CoordPoint<T>::CoordPoint()
  : x(),
    y()
{
}

template<class T>
CoordPoint<T>::CoordPoint(const T& pt_x, const T& pt_y)
  : x(pt_x),
    y(pt_y)
{
}

template<class T>
CoordPoint<T>::CoordPoint(const CoordPoint<T>& pt)
  : x(pt.x),
    y(pt.y)
{
}

template<class T>
CoordPoint<T>::CoordPoint(const CoordVector<T>& v, const CoordPoint<T>& pt)
{
  *this = v + pt;
}

template<class T>
CoordPoint<T>::CoordPoint(const CoordPoint<T>& pt, const CoordVector<T>& v)
{
  *this = v + pt;
}

template<class T>
CoordPoint<T>::CoordPoint(const CoordVector<T>& v)
{
  x = v.getLength() * cos(v.getAngle());
  y = v.getLength() * sin(v.getAngle());
}

template<class T>
CoordPoint<T> CoordPoint<T>::operator-(const CoordVector<T>& v) const
{
  return (-v) + *this;
}

template<class T>
CoordPoint<T> CoordPoint<T>::operator+(const CoordVector<T>& v) const
{
  return v+*this;
}

template<class T>
CoordVector<T> CoordPoint<T>::getVectorTo(const CoordPoint<T>& pt) const
{
  return CoordVector<T>(*this, pt);
}

template<class T>
std::istream& operator>>(std::istream& is, CoordPoint<T>& pt)
{
  /* Input formats: f is a T
	(f,f) = (f,f)
  */

  T pt_x;
  T pt_y;
  char c = 0;

  is >> c;
  if (c == '(') {
    is >> pt_x >> c;
    if (c == ',') is >> pt_y >> c;
    if (c != ')') is.clear(std::ios::badbit); //error...;
  }
  else {
    is.clear(std::ios::badbit); //error...;
  }

  if (is) pt = CoordPoint<T>(pt_x, pt_y);

  return is;
}


RS_END_NAMESPACE


#endif // COORDPOINT_H
