/* -*- Mode: C++ -*-
 * $Id: CoordPointUnknown.h,v 1.4 2000/09/18 15:20:05 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: CoordPointUnknown.h
 * Unitname: Library
 * $Revision: 1.4 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/18 15:20:05 $
 *
 *
 * HISTORY
 *
 * $Log: CoordPointUnknown.h,v $
 * Revision 1.4  2000/09/18 15:20:05  frehe
 * Due to a problem with namespaces the name for atan2, atan, acos and asin is changed to ATan2, ATan, ASin and ACos.
 *
 * Revision 1.3  2000/09/04 15:21:49  frehe
 * Changed the name of the trigonometric functions to the same as in the standard math library (sin, cos, tan, asin, acos, atan, and atan2).
 *
 * Revision 1.2  2000/09/01 03:23:52  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/01 01:24:38  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 *
 */

/**
 * @file CoordPointUnknown.h
 * Include file for the template CoordPointUnknown.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2000/09/18 15:20:05 $
 */


#ifndef COORDPOINTUNKNOWN_H
#define COORDPOINTUNKNOWN_H


#include <iostream>

#include "math_utils.h"
#include "AngleDegrees.h"
#include "WithUnknown.h"
#include "CoordPoint.h"
#include "CoordVector.h"
#include "CoordVectorUnknown.h"
#include "UnknownValueException.h"
#include "rs_throw.h"


#define POINTUNKNOWN_HEAD std::string("#PointUnknown")


RS_BEGIN_NAMESPACE


template<class T> class CoordVectorUnknown;


/**
 * @class CoordPointUnknown CoordPointUnknown.h
 * Template class for representing a cartesian two dimensional
 * coordinate that can be unknown.
 * The X and Y values are of type WithUnknown<T>.
 *
 * <!-- @todo -->
 *
 * @par Unit:
 *      Library
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2000/09/18 15:20:05 $
 */
template<class T> class CoordPointUnknown {
public:
  CoordPointUnknown();
  CoordPointUnknown(const T&, const T&);
  CoordPointUnknown(const WithUnknown<T>&, const WithUnknown<T>&);
  CoordPointUnknown(const CoordPoint<T>&);
  CoordPointUnknown(const CoordPointUnknown<T>&);
  CoordPointUnknown(const CoordVectorUnknown<T>&, const CoordPointUnknown<T>&);
  CoordPointUnknown(const CoordPointUnknown<T>&, const CoordVectorUnknown<T>&);
  explicit CoordPointUnknown(const CoordVectorUnknown<T>&);

  bool isKnown() const;
  bool isUnknown() const;
  CoordPointUnknown<T>& makeUnknown();
  CoordPoint<T> getPoint() const throw(UnknownValueException);
  
  CoordPointUnknown<T> operator - () const;
  CoordPointUnknown<T> operator - (const CoordPointUnknown<T>&) const;
  CoordPointUnknown<T> operator - (const CoordVectorUnknown<T>&) const;
  CoordPointUnknown<T> operator + (const CoordPointUnknown<T>&) const;
  CoordPointUnknown<T> operator + (const CoordVectorUnknown<T>&) const;
  CoordPointUnknown<T> operator * (const WithUnknown<T>&) const;
  CoordPointUnknown<T> operator * (const CoordPointUnknown<T>&) const;
  CoordPointUnknown<T> operator / (const WithUnknown<T>&) const;
  CoordPointUnknown<T> operator / (const CoordPointUnknown<T>&) const;

  CoordPointUnknown<T>& operator =(const CoordPointUnknown<T>&);
  CoordPointUnknown<T>& operator =(const CoordVectorUnknown<T>&);

  CoordPointUnknown<T>& operator -=(const CoordPointUnknown<T>&);
  CoordPointUnknown<T>& operator -=(const CoordVectorUnknown<T>&);
  CoordPointUnknown<T>& operator +=(const CoordPointUnknown<T>&);
  CoordPointUnknown<T>& operator +=(const CoordVectorUnknown<T>&);
  CoordPointUnknown<T>& operator *=(const WithUnknown<T>&);
  CoordPointUnknown<T>& operator *=(const CoordPointUnknown<T>&);
  CoordPointUnknown<T>& operator /=(const WithUnknown<T>&);
  CoordPointUnknown<T>& operator /=(const CoordPointUnknown<T>&);

  bool operator ! () const;
  bool operator ==(const CoordPointUnknown<T>&) const;
  bool operator !=(const CoordPointUnknown<T>&) const;

  /* Define these in a separate file. They need not be methods.
  // These use the coordinate system of the soccer field
  bool InFront(const CoordPointUnknown<T>&) const;  // Equal to LeftOf
  bool Behind(const CoordPointUnknown<T>&)  const;   // Equal to RightOf
  // These use standar coordinate system
  bool LeftOf(const CoordPointUnknown<T>&)  const;
  bool RightOf(const CoordPointUnknown<T>&) const;
  boolbove(const CoordPointUnknown<T>&)     const;
  bool Below(const CoordPointUnknown<T>&)   const;
  */
  
  WithUnknown<T> getX() const;
  WithUnknown<T> getY() const;
  CoordPointUnknown<T>& addX(const WithUnknown<T>& pt_x);
  CoordPointUnknown<T>& addY(const WithUnknown<T>& pt_y);
  CoordPointUnknown<T>& addXY(const WithUnknown<T>& pt_x,
			      const WithUnknown<T>& pt_y);
  CoordPointUnknown<T>& setX(const WithUnknown<T>& pt_x);
  CoordPointUnknown<T>& setY(const WithUnknown<T>& pt_y);
  CoordPointUnknown<T>& set (const WithUnknown<T>& pt_x, const WithUnknown<T>& pt_y);
  
  CoordVectorUnknown<T> getVectorTo(const CoordPointUnknown<T>& pt) const;
  AngleDegreesUnknown<T> getDirectionTo(const CoordPointUnknown<T>& pt) const;
  WithUnknown<T> getDistanceTo(const CoordPointUnknown<T>& pt) const;
  
protected:
  WithUnknown<T> x;
  WithUnknown<T> y;  
};

/*
#if CHECK_DIVisION_BY_ZERO == 1
static const CoordPointUnknown<T> CoordPointUnknown<T>_DIVisION_BY_ZERO_RESULT(T_DIVisION_BY_ZERO_RESULT,
						   T_DIVisION_BY_ZERO_RESULT);
#endif CHECK_DIVisION_BY_ZERO
*/

// Extend this to be the point on the line between pt1 and pt2 
// with a certain ratio 1:1 is the middle point?
template<class T>
inline CoordPointUnknown<T> middlePoint(const CoordPointUnknown<T>& pt1,
					const CoordPointUnknown<T>& pt2)
{
  return pt1 + CoordPointUnknown<T>((pt2.getX() - pt1.getX())/2,
				    (pt2.getY() - pt1.getY())/2);
}

template<class T>
inline bool CoordPointUnknown<T>::isKnown() const
{
  return x.isKnown() && y.isKnown();
}

template<class T>
inline bool CoordPointUnknown<T>::isUnknown() const
{
  return x.isUnknown() && y.isUnknown();
}

template<class T>
CoordPointUnknown<T>& CoordPointUnknown<T>::makeUnknown()
{
  x.makeUnknown();
  y.makeUnknown();
  return *this;
}

template <class T> 
inline CoordPoint<T> CoordPointUnknown<T>::getPoint() const throw(UnknownValueException)
{
  if ( isUnknown() ) {
    RS_THROW(UnknownValueException, "Tried to take the value of an unknown point");
    return CoordPoint<T>();
  } else
    return CoordPoint<T>(x.getValue(), y.getValue());
}

template<class T>
inline CoordPointUnknown<T> CoordPointUnknown<T>::operator -() const
{
  return CoordPointUnknown<T>(-x,-y);
}

template<class T>
inline CoordPointUnknown<T> CoordPointUnknown<T>::operator -(const CoordPointUnknown<T>& pt) const
{
  return CoordPointUnknown<T>(x - pt.x, y - pt.y);
}

template<class T>
inline CoordPointUnknown<T> operator -(const CoordPoint<T>& pt1,
				       const CoordPointUnknown<T>& pt2)
{
  return (-pt2) + pt1;
}

template<class T>
inline CoordPointUnknown<T> CoordPointUnknown<T>::operator *(const WithUnknown<T>& a) const
{
  return CoordPointUnknown<T>(x * a, y * a);
}

template<class T>
inline CoordPointUnknown<T> CoordPointUnknown<T>::operator *(const CoordPointUnknown<T>& pt) const
{
  return CoordPointUnknown<T>(x * pt.x, y * pt.y);
}

template<class T>
inline CoordPointUnknown<T> operator *(const T& a,
				       const CoordPointUnknown<T>& pt)
{
  return pt * a;
}

template<class T>
inline CoordPointUnknown<T> operator *(const WithUnknown<T>& a,
				       const CoordPointUnknown<T>& pt)
{
  return pt * a;
}

template<class T>
inline CoordPointUnknown<T> operator *(const CoordPoint<T>& pt1,
				       const CoordPointUnknown<T>& pt2)
{
  return pt2 * pt1;
}

template<class T>
inline CoordPointUnknown<T> CoordPointUnknown<T>::operator /(const WithUnknown<T>& a) const
{
#if CHECK_DIVisION_BY_ZERO == 1
  if (a == T(0))
    return CoordPointUnknown_DIVisION_BY_ZERO_RESULT;
  else
#endif
    return CoordPointUnknown<T>(x / a, y / a);
}

template<class T>
inline CoordPointUnknown<T> CoordPointUnknown<T>::operator /(const CoordPointUnknown<T>& pt) const
{
#if CHECK_DIVisION_BY_ZERO == 1
  if ( pt.x == T(0) ) {
    if ( pt.y == T(0) ) {
      return CoordPointUnknown_DIVisION_BY_ZERO_RESULT;
    } else {
      return CoordPointUnknown<T>(T_DIVisION_BY_ZERO_RESULT, y / pt.y);
    }
  } else {
    if ( pt.y == T(0) ) {
      return CoordPointUnknown<T>(x / pt.x, T_DIVisION_BY_ZERO_RESULT);
    } else {
#endif
      return CoordPointUnknown<T>(x / pt.x, y / pt.y);
#if CHECK_DIVisION_BY_ZERO == 1      
    }
  }
#endif
}

template<class T>
CoordPointUnknown<T> operator /(const CoordPoint<T>& pt1,
				       const CoordPointUnknown<T>& pt2)
{
  if ( pt2.isUnknown() )
    return CoordPointUnknown<T>();
  else
    return pt1 / pt2.getPoint();
}

template<class T>
inline CoordPointUnknown<T> CoordPointUnknown<T>::operator +(const CoordPointUnknown<T>& pt) const
{
  return CoordPointUnknown<T>(x + pt.x, y + pt.y);
}

template<class T>
inline CoordPointUnknown<T> operator +(const CoordPoint<T>& pt1,
				       const CoordPointUnknown<T>& pt2)
{
  return pt2 + pt1;
}

template<class T>
inline CoordPointUnknown<T>& CoordPointUnknown<T>::operator-=(const CoordPointUnknown<T>& pt)
{
  x -= pt.x;
  y -= pt.y;
  return *this;
}

template<class T>
inline CoordPointUnknown<T>& CoordPointUnknown<T>::operator-=(const CoordVectorUnknown<T>& vec)
{
  *this = *this - vec;
  return *this;
}

template<class T>
inline CoordPointUnknown<T>& CoordPointUnknown<T>::operator+=(const CoordVectorUnknown<T>& vec)
{
  *this = *this + vec;
  return *this;
}

template<class T>
inline CoordPointUnknown<T>& CoordPointUnknown<T>::operator+=(const CoordPointUnknown<T>& pt)
{
  x += pt.x;
  y += pt.y;
  return *this;
}

template<class T>
inline CoordPointUnknown<T>& CoordPointUnknown<T>::operator*=(const WithUnknown<T>& a)
{
  x *= a;
  y *= a;
  return *this;
}

template<class T>
inline CoordPointUnknown<T>& CoordPointUnknown<T>::operator*=(const CoordPointUnknown<T>& pt)
{
  x *= pt.x;
  y *= pt.y;
  return *this;
}

template<class T>
CoordPointUnknown<T>& CoordPointUnknown<T>::operator/=(const WithUnknown<T>& a)
{
#if CHECK_DIVisION_BY_ZERO == 1
  if ( a == T(0)) {
    x = T_DIVisION_BY_ZERO_RESULT;
    y = T_DIVisION_BY_ZERO_RESULT;
  } else {
#endif
    x /= a;
    y /= a;
#if CHECK_DIVisION_BY_ZERO == 1
  }
#endif
  return *this;
}

template<class T>
inline CoordPointUnknown<T>& CoordPointUnknown<T>::operator/=(const CoordPointUnknown<T>& pt)
{
#if CHECK_DIVisION_BY_ZERO == 1
  if ( pt.x == T(0) )
    x = T_DIVisION_BY_ZERO_RESULT;
  else
#endif
    x /= pt.x;
  
#if CHECK_DIVisION_BY_ZERO == 1
  if ( pt.y == T(0) )
    y = T_DIVisION_BY_ZERO_RESULT;
  else
#endif
    y /= pt.y;

  return *this;
}

template<class T>
CoordPointUnknown<T>& CoordPointUnknown<T>::operator =(const CoordPointUnknown<T>& pt)
{
  if ( this != &pt ) {
    y = pt.y;
    x = pt.x;
  }
  return *this;
}

template<class T>
CoordPointUnknown<T>& CoordPointUnknown<T>::operator =(const CoordVectorUnknown<T>& vec)
{
  *this = CoordPointUnknown<T>(vec);
  return *this;
}

template<class T>
inline bool CoordPointUnknown<T>::operator !() const
{
  return (!y) && (!x);
}

template<class T>
inline bool CoordPointUnknown<T>::operator ==(const CoordPointUnknown<T>& pt) const
{
  return (y == pt.y) && (x == pt.x);
}

template<class T>
inline bool operator ==(const CoordPoint<T>& pt1,
			const CoordPointUnknown<T>& pt2)
{
  return pt2 == pt1;
}

template<class T>
inline bool CoordPointUnknown<T>::operator !=(const CoordPointUnknown<T>& pt) const
{
  return (y != pt.y) || (x != pt.x);
}

template<class T>
inline bool operator !=(const CoordPoint<T>& pt1,
			const CoordPointUnknown<T>& pt2)
{
  return pt2 != pt1;
}

/*
inline bool CoordPointUnknown<T>::InFront(const CoordPointUnknown<T>& pt) const
{
  return (x > pt.x);
}

inline bool CoordPointUnknown<T>::Behind(const CoordPointUnknown<T>& pt) const
{
  return (x < pt.x);
}
*/

/*
inline bool CoordPointUnknown<T>::LeftOf(const CoordPointUnknown<T>& pt) const
{
  return (x > pt.x);
}

inline bool CoordPointUnknown<T>::RightOf(const CoordPointUnknown<T>& pt) const
{
  return (x < pt.x);
}

inline bool CoordPointUnknown<T>::Above(const CoordPointUnknown<T>& pt) const
{
  return (y < pt.y);
}

inline bool CoordPointUnknown<T>::Below(const CoordPointUnknown<T>& pt) const
{
  return (y > pt.y);
}
*/

template<class T>
inline CoordPointUnknown<T>& CoordPointUnknown<T>::addX (const WithUnknown<T>& pt_x)
{
  x += pt_x;
  return *this;
}

template<class T>
inline CoordPointUnknown<T>& CoordPointUnknown<T>::addY (const WithUnknown<T>& pt_y)
{
  y += pt_y;
  return *this;
}

template<class T>
inline CoordPointUnknown<T>& CoordPointUnknown<T>::addXY (const WithUnknown<T>& pt_x,
							  const WithUnknown<T>& pt_y)
{
  x += pt_x;
  y += pt_y;
  return *this;
}

template<class T>
inline WithUnknown<T> CoordPointUnknown<T>::getX() const
{
  return x;
}

template<class T>
inline WithUnknown<T> CoordPointUnknown<T>::getY() const
{
  return y;
}

template<class T>
inline CoordPointUnknown<T>& CoordPointUnknown<T>::setX(const WithUnknown<T>& pt_x)
{
  x = pt_x;
  return *this;
}

template<class T>
inline CoordPointUnknown<T>& CoordPointUnknown<T>::setY(const WithUnknown<T>& pt_y)
{
  y = pt_y;
  return *this;
}

template<class T>
inline CoordPointUnknown<T>& CoordPointUnknown<T>::set(const WithUnknown<T>& pt_x, const WithUnknown<T>& pt_y)
{
  x = pt_x;
  y = pt_y;
  return *this;
}

template<class T>
AngleDegreesUnknown<T> CoordPointUnknown<T>::getDirectionTo(const CoordPointUnknown<T>& pt) const
{
  if ( pt.isUnknown() )
    return AngleDegreesUnknown<T>();
  else
    return ATan2((pt.y - y), (pt.x - x));
}

template<class T>
inline WithUnknown<T> CoordPointUnknown<T>::getDistanceTo(const CoordPointUnknown<T>& pt) const
{
  return sqrt(sqr(pt.x - x) + sqr(pt.y - y));
}

template<class T>
inline std::ostream& operator<<(std::ostream& os, const CoordPointUnknown<T>& pt)
{
#ifdef POINTUNKNOWN_HEAD
  return os << POINTUNKNOWN_HEAD << "(" << pt.getX() << ", " << pt.getY() << ")";
#else
  return os << "(" << pt.getX() << ", " << pt.getY() << ")";
#endif
}

template<class T>
inline const std::string toString(const CoordPointUnknown<T>& pt)
{
#ifdef POINTUNKNOWN_HEAD
  return POINTUNKNOWN_HEAD + "(" + toString(pt.getX()) + ", " + toString(pt.getY()) + ")";
#else
  return "(" + toString(pt.getX()) + ", " + toString(pt.getY()) + ")";
#endif
}

template<class T>
CoordPointUnknown<T>::CoordPointUnknown()
{
}

template<class T>
CoordPointUnknown<T>::CoordPointUnknown(const WithUnknown<T>& pt_x, const WithUnknown<T>& pt_y)
  : x(pt_x),
    y(pt_y)
{
}

template<class T>
CoordPointUnknown<T>::CoordPointUnknown(const T& pt_x, const T& pt_y)
  : x(pt_x),
    y(pt_y)
{
}

template<class T>
CoordPointUnknown<T>::CoordPointUnknown(const CoordPoint<T>& pt)
  : x(pt.getX()),
    y(pt.getY())
{
}

template<class T>
CoordPointUnknown<T>::CoordPointUnknown(const CoordPointUnknown<T>& pt)
  : x(pt.x),
    y(pt.y)
{
}

template<class T>
CoordPointUnknown<T>::CoordPointUnknown(const CoordVectorUnknown<T>& v, const CoordPointUnknown<T>& pt)
{
  *this = v + pt;
}

template<class T>
CoordPointUnknown<T>::CoordPointUnknown(const CoordPointUnknown<T>& pt, const CoordVectorUnknown<T>& v)
{
  *this = v + pt;
}

template<class T>
CoordPointUnknown<T>::CoordPointUnknown(const CoordVectorUnknown<T>& v)
{
  x = v.getLength() * cos(v.getAngle());
  y = v.getLength() * sin(v.getAngle());
}

template<class T>
CoordPointUnknown<T> CoordPointUnknown<T>::operator-(const CoordVectorUnknown<T>& v) const
{
  return (-v) + *this;
}

template<class T>
CoordPointUnknown<T> CoordPointUnknown<T>::operator+(const CoordVectorUnknown<T>& v) const
{
  return v+*this;
}

template<class T>
CoordVectorUnknown<T> CoordPointUnknown<T>::getVectorTo(const CoordPointUnknown<T>& pt) const
{
  return CoordVectorUnknown<T>(*this, pt);
}

template<class T>
std::istream& operator>>(std::istream& is, CoordPointUnknown<T>& pt)
{
  /* Input formats: f is a T
	(f,f) = (f,f)
  */

  WithUnknown<T> pt_x;
  WithUnknown<T> pt_y;
  char c = 0;

  is >> c;
  if (c == '(') {
    is >> pt_x >> c;
    if (c == ',') is >> pt_y >> c;
    if (c != ')') is.clear(std::ios::badbit); //error...;
  }
  else {
    is.clear(std::ios::badbit); //error...;
  }

  if (is) pt = CoordPointUnknown<T>(pt_x, pt_y);

  return is;
}


RS_END_NAMESPACE


#endif // COORDPOINTUNKNOWN_H
