/* -*- Mode: C++ -*-
 * $Id: CoordVector.h,v 1.5 2000/09/20 15:12:52 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: CoordVector.h
 * Unitname: Library
 * $Revision: 1.5 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/20 15:12:52 $
 *
 *
 * HISTORY
 *
 * $Log: CoordVector.h,v $
 * Revision 1.5  2000/09/20 15:12:52  frehe
 * Added documentation.
 *
 * Revision 1.4  2000/09/18 15:20:05  frehe
 * Due to a problem with namespaces the name for atan2, atan, acos and asin is changed to ATan2, ATan, ASin and ACos.
 *
 * Revision 1.3  2000/09/04 15:21:49  frehe
 * Changed the name of the trigonometric functions to the same as in the standard math library (sin, cos, tan, asin, acos, atan, and atan2).
 *
 * Revision 1.2  2000/09/01 03:23:52  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/01 01:24:38  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 *
 */

/**
 * @file CoordVector.h
 * Include file for the template CoordVector.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2000/09/20 15:12:52 $
 */


#ifndef COORDVECTOR_H
#define COORDVECTOR_H


#include <iostream>
#include <string>

#include "AngleDegrees.h"
#include "CoordPoint.h"


/** String to use as the header when outputing a Vector. */
#define VECTOR_HEAD std::string("#Vector")


RS_BEGIN_NAMESPACE


template<class T> class CoordPoint;

/**
 * @class CoordVector CoordVector.h
 * Template class for representing a two dimensional mathematical vector.
 * The distance and direction are of type T.
 * T is the base type for the angle: AngleDegrees<T> and the distance T.
 *
 * <!-- @todo -->
 *
 * @par Unit:
 *      Library
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2000/09/20 15:12:52 $
 */
template<class T> class CoordVector
{
public:
  CoordVector();
  CoordVector(const T&, const T&);
  CoordVector(const AngleDegrees<T>&, const T&);
  CoordVector(const CoordVector&);
  CoordVector(const CoordPoint<T>&, const CoordPoint<T>&);
  explicit CoordVector(const CoordPoint<T>&);

  CoordVector<T> operator - () const;
  CoordVector<T> operator + (const CoordVector<T>&) const; 
  CoordPoint<T>  operator + (const CoordPoint<T>&) const;
  CoordVector<T> operator - (const CoordVector<T>&) const;
  CoordVector<T> operator * (const T&) const;
  CoordVector<T> operator / (const T&) const;

  CoordVector<T>& operator  =(const CoordVector<T>&);
  CoordVector<T>& operator  =(const CoordPoint<T>&);
  
  CoordVector<T>& operator -=(const CoordVector<T>&);
  CoordVector<T>& operator +=(const CoordVector<T>&);
  CoordVector<T>& operator *=(const T&);
  CoordVector<T>& operator /=(const T&);

  // Is the vector the zero vector?
  bool     operator ! ()                 const;
  // Are the vectors equal or not.
  bool     operator ==(const CoordVector<T>&)   const;
  bool     operator !=(const CoordVector<T>&)   const;
  // These compares the length of the vectors
  bool     operator < (const CoordVector<T>&)   const;
  bool     operator <=(const CoordVector<T>&)   const;
  bool     operator > (const CoordVector<T>&)   const;
  bool     operator >=(const CoordVector<T>&)   const;

  // Shoule I include these? Instead of the above operators?
  // bool SameLength
  // bool Longer
  // bool Shorter

  // Adds to the angle resp. to the distance of the vector.
  CoordVector<T>& rotate(const AngleDegrees<T>&);
  // Bad name! I should have thought this was multiplication not addition.
  CoordVector<T>& Scale (const T&);

  T getLength() const;
  CoordVector<T>& setLength(const T&);
  AngleDegrees<T> getAngle() const;
  CoordVector<T>& setAngle(const AngleDegrees<T>&);

protected:
  AngleDegrees<T> dir; // The direction of the vector
  T dist; // The distance/length of the vector

};

/*
#if CHECK_DIVISION_BY_ZERO == 1
static const CoordVector<T> CoordVector<T>_DIVISION_BY_ZERO_RESULT(T_DIVISION_BY_ZERO_RESULT,
						     T_DIVISION_BY_ZERO_RESULT);
#endif CHECK_DIVISION_BY_ZERO
*/

template <class T>
inline CoordVector<T> CoordVector<T>::operator -() const
{
  return CoordVector<T>(dir-AngleDegrees<T>(180), dist);
}

template <class T>
inline CoordVector<T> CoordVector<T>::operator -(const CoordVector<T>& v) const
{
  return *this+CoordVector<T>(v.dir-AngleDegrees<T>(180), v.dist);
}

template <class T>
inline CoordVector<T> CoordVector<T>::operator *(const T& a) const
{
  return CoordVector<T>(dir, dist * a);
}

template <class T>
inline CoordVector<T> operator *(const T& a, const CoordVector<T>& vec)
{
  return vec * a;
}

template <class T>
inline CoordVector<T> CoordVector<T>::operator /(const T& a) const
{
#if CHECK_DIVISION_BY_ZERO == 1
  if (a == T(0))
    return COORDVECTOR_DIVISION_BY_ZERO_RESULT;
  else
#endif
    return CoordVector<T>(dir, dist / a);
}

template <class T>
inline CoordVector<T>& CoordVector<T>::operator-=(const CoordVector<T>& v)
{
  return *this = *this - v;
}

template <class T>
inline CoordVector<T>& CoordVector<T>::operator+=(const CoordVector<T>& v)
{
  return *this = *this + v;
}

template <class T>
inline CoordVector<T>& CoordVector<T>::operator*=(const T& a)
{
  dist *= a;
  return *this;
}

template <class T>
inline CoordVector<T>& CoordVector<T>::operator/=(const T& a)
{
#if CHECK_DIVISION_BY_ZERO == 1
  if (a == T(0))
    dist = T_DIVISION_BY_ZERO_RESULT;
  else
#endif
    dist /= a;
  
  return *this;
}

template <class T>
inline bool CoordVector<T>::operator ==(const CoordVector<T>& v) const
{
  return (dist == v.dist) && (dir == v.dir);
}

template <class T>
inline bool CoordVector<T>::operator !=(const CoordVector<T>& v) const
{
  return (dist != v.dist) || (dir != v.dir);
}

template <class T>
inline bool CoordVector<T>::operator < (const CoordVector<T>& v) const
{
  return (dist <  v.dist);
}

template <class T>
inline bool CoordVector<T>::operator <=(const CoordVector<T>& v) const
{
  return (dist <= v.dist);
}

template <class T>
inline bool CoordVector<T>::operator > (const CoordVector<T>& v) const
{
  return (dist >  v.dist);
}

template <class T>
inline bool CoordVector<T>::operator >=(const CoordVector<T>& v) const
{
  return (dist >= v.dist);
}

template <class T>
inline bool CoordVector<T>::operator !() const
{
  return (!dist) && (!dir);
}

template <class T>
inline CoordVector<T>& CoordVector<T>::rotate(const AngleDegrees<T>& vdir)
{
  dir += vdir;
  return *this;
}

template <class T>
inline CoordVector<T>& CoordVector<T>::Scale(const T& vdist)
{
  dist += vdist;
  return *this;
}

template <class T>
inline T CoordVector<T>::getLength() const
{
  return dist;
}

template <class T>
inline AngleDegrees<T> CoordVector<T>::getAngle() const
{
  return dir;
}

template <class T>
inline std::ostream& operator<<(std::ostream& os, const CoordVector<T>& v)
{
#ifdef VECTOR_HEAD
  return os << VECTOR_HEAD << "(" << v.getAngle() << ", " << v.getLength() << ")";
#else
  return os << "(" << v.getAngle() << ", " << v.getLength() << ")";
#endif
}

template <class T>
inline const std::string ToString(const CoordVector<T>& v)
{
#ifdef VECTOR_HEAD
  return VECTOR_HEAD + "(" + ToString(v.getAngle()) + ", " + ToString(v.getLength()) + ")";
#else
  return "(" + ToString(v.getAngle()) + ", " + ToString(v.getLength()) + ")";
#endif
}

template <class T>
CoordVector<T>::CoordVector()
  : dir(),
    dist()
{
}

template <class T>
CoordVector<T>::CoordVector(const T& vdir, const T& vdist)
  : dir(vdir),
    dist(vdist)
{
  if (dist < T(0)) {
    dir -= AngleDegrees<T>(180);
    dist = -dist;
  }
}

template <class T>
CoordVector<T>::CoordVector(const AngleDegrees<T>& vdir, const T& vdist)
  : dir(vdir),
    dist(vdist)
{
  if (dist < T(0)) {
    dir -= AngleDegrees<T>(180);
    dist = -dist;
  }
}

template <class T>
CoordVector<T>::CoordVector(const CoordVector<T>& v)
  : dir(v.dir),
    dist(v.dist)
{
}

template <class T>
CoordVector<T>::CoordVector(const CoordPoint<T>& pt1, const CoordPoint<T>& pt2)
{
  if (pt2 == pt1) {
    dir = AngleDegrees<T>(0);
    dist = T(0);
  } else {
    dir = ATan2(pt2.getY()-pt1.getY(), pt2.getX()-pt1.getX());
    dist = sqrt(sqr(pt1.getX()-pt2.getX()) + sqr(pt1.getY()-pt2.getY()));
  }
}

template <class T>
CoordVector<T>::CoordVector(const CoordPoint<T>& pt)
{
  if ( pt.getX() == T(0) &&
       pt.getY() == T(0) ) {
    dir = AngleDegrees<T>(0);
    dist = T(0);
  } else {
    dir = ATan2(pt.getY(), pt.getX());
    dist = sqrt(sqr(pt.getX()) + sqr(pt.getY()));
  }
}

template <class T>
CoordVector<T>& CoordVector<T>::operator =(const CoordVector<T>& v)
{
  if (this != &v) {
    dist = v.dist;
    dir = v.dir;
  }

  return *this;
}

template <class T>
inline CoordVector<T>& CoordVector<T>::operator =(const CoordPoint<T>& pt)
{
  *this = CoordVector<T>(pt);
  return *this;
}

template <class T>
CoordVector<T> CoordVector<T>::operator +(const CoordVector<T>& v) const
{
  return CoordVector<T>(CoordPoint<T>(v) + CoordPoint<T>(*this));
}

template <class T>
CoordPoint<T> CoordVector<T>::operator+(const CoordPoint<T>& pt1) const
{
  return CoordPoint<T>(dist*cos(dir), dist*sin(dir)) + pt1;
}
 
template <class T>
inline CoordVector<T>& CoordVector<T>::setLength(const T& len)
{
  dist = len;
  return *this;
}

template <class T>
inline CoordVector<T>& CoordVector<T>::setAngle(const AngleDegrees<T>& ang)
{
  dir = ang;
  return *this;
}

template <class T>
std::istream& operator>>(std::istream& is, CoordVector<T>& v)
{
  /* Input formats: f is a floating-point number
        f     = (0,f)
	(f)   = (0,f)
	(f,f) = (f,f)
  */

  T dist;
  AngleDegrees<T> dir;
  char c = 0;

  is >> c;
  if (c == '(') {
    is >> dir >> c;
    if (c == ',') is >> dist >> c;
    if (c != ')') is.clear(std::ios::badbit); //error...;
  }
  else {
    is.putback(c);
    is >> dist;
  }

  if (dist == T(0)) {
    dist = T(dir);
    dir = AngleDegrees<T>(0);
  }
  
  if (is) v = CoordVector<T>(dir, dist);

  return is;

}


RS_END_NAMESPACE


#endif  // !COORDVECTOR_H
