/* -*- Mode: C++ -*-
 * $Id: CoordVectorUnknown.h,v 1.4 2000/09/18 15:20:06 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: CoordVectorUnknown.h
 * Unitname: Library
 * $Revision: 1.4 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/18 15:20:06 $
 *
 *
 * HISTORY
 *
 * $Log: CoordVectorUnknown.h,v $
 * Revision 1.4  2000/09/18 15:20:06  frehe
 * Due to a problem with namespaces the name for atan2, atan, acos and asin is changed to ATan2, ATan, ASin and ACos.
 *
 * Revision 1.3  2000/09/04 15:21:50  frehe
 * Changed the name of the trigonometric functions to the same as in the standard math library (sin, cos, tan, asin, acos, atan, and atan2).
 *
 * Revision 1.2  2000/09/01 03:23:53  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/01 01:24:39  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 *
 */

/**
 * @file CoordVectorUnknown.h
 * Include file for the template CoordVectorUnknown.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2000/09/18 15:20:06 $
 */


#ifndef COORDVECTORUNKNOWN_H
#define COORDVECTORUNKNOWN_H


#include <iostream>

#include "math_utils.h"
#include "CoordVector.h"
#include "WithUnknown.h"
#include "AngleDegreesUnknown.h"
#include "CoordPointUnknown.h"
#include "UnknownValueException.h"
#include "rs_throw.h"


#define VECTORUNKNOWN_HEAD std::string("#VectorUnknown")


RS_BEGIN_NAMESPACE


template<class T> class CoordPointUnknown;


/**
 * @class CoordVectorUnknown CoordVectorUnknown.h
 * Template class for representing a two dimensional mathematical vector
 * that can be unknown.
 * The distance and direction are of type WithUnknown<T>.
 * T is the base type for the angle: AngleDegreesUnknown<T> and the
 * distance WithUnknown<T>.
 *
 * <!-- @todo -->
 *
 * @par Unit:
 *      Library
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2000/09/18 15:20:06 $
 */
template<class T> class CoordVectorUnknown
{
public:
  CoordVectorUnknown();
  CoordVectorUnknown(const T&, const T&);
  CoordVectorUnknown(const AngleDegrees<T>&, const T&);
  CoordVectorUnknown(const AngleDegreesUnknown<T>&,
		     const T&);
  CoordVectorUnknown(const AngleDegreesUnknown<T>&,
		     const WithUnknown<T>&);
  CoordVectorUnknown(const CoordVectorUnknown<T>&);
  CoordVectorUnknown(const CoordVector<T>&);
  CoordVectorUnknown(const CoordPointUnknown<T>&,
		     const CoordPointUnknown<T>&);
  explicit CoordVectorUnknown(const CoordPointUnknown<T>&);

  bool isKnown() const;
  bool isUnknown() const;
  CoordVectorUnknown<T>& makeUnknown();
  CoordVector<T> getVector() const throw(UnknownValueException);
  
  CoordVectorUnknown<T> operator -() const;
  CoordVectorUnknown<T> operator -(const CoordVectorUnknown<T>&) const;
  CoordVectorUnknown<T> operator +(const CoordVectorUnknown<T>&) const; 
  CoordPointUnknown<T>  operator +(const CoordPointUnknown<T>&) const;
  CoordVectorUnknown<T> operator *(const WithUnknown<T>&) const;
  CoordVectorUnknown<T> operator /(const WithUnknown<T>&) const;

  CoordVectorUnknown<T>& operator =(const CoordVectorUnknown<T>&);
  CoordVectorUnknown<T>& operator =(const CoordPointUnknown<T>&);  
  
  CoordVectorUnknown<T>& operator -=(const CoordVectorUnknown<T>&);
  CoordVectorUnknown<T>& operator +=(const CoordVectorUnknown<T>&);
  CoordVectorUnknown<T>& operator *=(const WithUnknown<T>&);
  CoordVectorUnknown<T>& operator /=(const WithUnknown<T>&);

  // is the vector the zero vector?
  bool     operator ! ()                 const;
  bool     operator ==(const CoordVectorUnknown<T>&)   const;
  bool     operator !=(const CoordVectorUnknown<T>&)   const;
  // These compares the length of the vectors
  bool     operator < (const CoordVectorUnknown<T>&)   const;
  bool     operator <=(const CoordVectorUnknown<T>&)   const;
  bool     operator > (const CoordVectorUnknown<T>&)   const;
  bool     operator >=(const CoordVectorUnknown<T>&)   const;

  // Shoule I include these? Instead of the above operators?
  // bool SameLength
  // bool Longer
  // bool Shorter

  // Adds to the angle resp. to the distance of the vector.
  CoordVectorUnknown<T>& rotate(const AngleDegreesUnknown<T>&);
  // Bad name! I should have thought this was multiplication not addition.
  CoordVectorUnknown<T>& scale (const WithUnknown<T>&);

  WithUnknown<T> getLength() const;
  CoordVectorUnknown<T>& setLength(const WithUnknown<T>&);
  AngleDegreesUnknown<T> getAngle() const;
  CoordVectorUnknown<T>& setAngle(const AngleDegreesUnknown<T>&);

protected:
  AngleDegreesUnknown<T> dir; /* The direction of the vector */
  WithUnknown<T> dist; /* The distance/length of the vector */
};

template <class T>
CoordVectorUnknown<T>::CoordVectorUnknown()
  : dir(),
    dist()
{
}

template <class T>
CoordVectorUnknown<T>::CoordVectorUnknown(const T& vdir,
					  const T& vdist)
  : dir(vdir),
    dist(vdist)
{
  if (dist < WithUnknown<T>(0)) {
    dir -= AngleDegreesUnknown<T>(180);
    dist = -dist;
  }
}

template <class T>
CoordVectorUnknown<T>::CoordVectorUnknown(const AngleDegrees<T>& vdir,
					  const T& vdist)
  : dir(vdir),
    dist(vdist)
{
  if (dist < WithUnknown<T>(0)) {
    dir -= AngleDegreesUnknown<T>(180);
    dist = -dist;
  }
}

template <class T>
CoordVectorUnknown<T>::CoordVectorUnknown(const AngleDegreesUnknown<T>& vdir,
					  const T& vdist)
  : dir(vdir),
    dist(vdist)
{
  if (dist < WithUnknown<T>(0)) {
    dir -= AngleDegreesUnknown<T>(180);
    dist = -dist;
  }
}

template <class T>
CoordVectorUnknown<T>::CoordVectorUnknown(const AngleDegreesUnknown<T>& vdir,
					  const WithUnknown<T>& vdist)
  : dir(vdir),
    dist(vdist)
{
  if (dist < WithUnknown<T>(0)) {
    dir -= AngleDegreesUnknown<T>(180);
    dist = -dist;
  }
}

template <class T>
CoordVectorUnknown<T>::CoordVectorUnknown(const CoordVector<T>& v)
  : dir(v.getAngle()),
    dist(v.getLength())
{
}

template <class T>
CoordVectorUnknown<T>::CoordVectorUnknown(const CoordVectorUnknown<T>& v)
  : dir(v.dir),
    dist(v.dist)
{
}

template <class T>
CoordVectorUnknown<T>::CoordVectorUnknown(const CoordPointUnknown<T>& pt1,
					  const CoordPointUnknown<T>& pt2)
{
  if (pt2 == pt1) {
    dir = AngleDegreesUnknown<T>(0);
    dist = WithUnknown<T>(0);
  } else {
    dir = ATan2(pt2.getY()-pt1.getY(), pt2.getX()-pt1.getX());
    dist = sqrt(sqr(pt1.getX()-pt2.getX()) + sqr(pt1.getY()-pt2.getY()));
  }
}

template <class T>
CoordVectorUnknown<T>::CoordVectorUnknown(const CoordPointUnknown<T>& pt)
{
  if ( pt.getX() == WithUnknown<T>(0) &&
       pt.getY() == WithUnknown<T>(0) ) {
    dir = AngleDegreesUnknown<T>(0);
    dist = WithUnknown<T>(0);
  } else {
    dir = ATan2(pt.getY(), pt.getX());
    dist = sqrt(sqr(pt.getX()) + sqr(pt.getY()));
  }
}

template<class T>
inline bool CoordVectorUnknown<T>::isKnown() const
{
  return dir.isKnown() && dist.isKnown();
}

template<class T>
inline bool CoordVectorUnknown<T>::isUnknown() const
{
  return dir.isUnknown() && dist.isUnknown();
}

template<class T>
CoordVectorUnknown<T>& CoordVectorUnknown<T>::makeUnknown()
{
  dir.makeUnknown();
  dist.makeUnknown();
  return *this;
}

template <class T> 
inline CoordVector<T> CoordVectorUnknown<T>::getVector() const throw(UnknownValueException)
{
  if ( isUnknown() ) {
    RS_THROW(UnknownValueException, "Tried to take the value of an unknown vector");
    return CoordVector<T>();
  } else
    return CoordVector<T>(dir.getAngle(), dist.getValue());
}

template <class T>
inline CoordVectorUnknown<T> CoordVectorUnknown<T>::operator -() const
{
  return CoordVectorUnknown<T>(dir-AngleDegrees<T>(180), dist);
}

template <class T>
inline CoordVectorUnknown<T> CoordVectorUnknown<T>::operator -(const CoordVectorUnknown<T>& v) const
{
  return *this+CoordVectorUnknown<T>(v.dir-AngleDegrees<T>(180), v.dist);
}

template <class T>
inline CoordVectorUnknown<T> operator -(const CoordVector<T>& v1,
					const CoordVectorUnknown<T>& v2)
{
  return (-v2) + v1;
}

template <class T>
inline CoordPointUnknown<T> operator -(const CoordPoint<T>& v1,
				       const CoordVectorUnknown<T>& v2)
{
  return (-v2) + v1;
}

template <class T>
inline CoordVectorUnknown<T> CoordVectorUnknown<T>::operator +(const CoordVectorUnknown<T>& v) const
{
  return CoordVectorUnknown<T>(CoordPointUnknown<T>(v) + CoordPointUnknown<T>(*this));
}

template <class T>
inline CoordPointUnknown<T> CoordVectorUnknown<T>::operator+(const CoordPointUnknown<T>& pt1) const
{
  return CoordPointUnknown<T>(dist*cos(dir), dist*sin(dir)) + pt1;
}
 
template <class T>
inline CoordVectorUnknown<T> operator +(const CoordVector<T>& v1,
					const CoordVectorUnknown<T>& v2)
{
  return v2 + v1;
}

template<class T>
inline CoordPointUnknown<T> operator +(const CoordPoint<T>& pt1,
				       const CoordVectorUnknown<T>& pt2)
{
  return pt2 + pt1;
}

template <class T>
inline CoordVectorUnknown<T> CoordVectorUnknown<T>::operator *(const WithUnknown<T>& a) const
{
  return CoordVectorUnknown<T>(dir, dist * a);
}

template <class T>
inline CoordVectorUnknown<T> operator *(const T& a,
					const CoordVectorUnknown<T>& vec)
{
  return vec * a;
}

template <class T>
inline CoordVectorUnknown<T> operator *(const WithUnknown<T>& a,
					const CoordVectorUnknown<T>& vec)
{
  return vec * a;
}

template <class T>
inline CoordVectorUnknown<T> CoordVectorUnknown<T>::operator /(const WithUnknown<T>& a) const
{
#if CHECK_DIVisION_BY_ZERO == 1
  if (a == T(0))
    return COORDVECTOR_DIVisION_BY_ZERO_RESULT;
  else
#endif
    return CoordVectorUnknown<T>(dir, dist / a);
}

template <class T>
CoordVectorUnknown<T>& CoordVectorUnknown<T>::operator =(const CoordVectorUnknown<T>& v)
{
  if (this != &v) {
    dist = v.dist;
    dir = v.dir;
  }

  return *this;
}

template <class T>
inline CoordVectorUnknown<T>& CoordVectorUnknown<T>::operator =(const CoordPointUnknown<T>& pt)
{
  *this = CoordVectorUnknown<T>(pt);
  return *this;
}

template <class T>
inline CoordVectorUnknown<T>& CoordVectorUnknown<T>::operator-=(const CoordVectorUnknown<T>& v)
{
  return *this = *this - v;
}

template <class T>
inline CoordVectorUnknown<T>& CoordVectorUnknown<T>::operator+=(const CoordVectorUnknown<T>& v)
{
  return *this = *this + v;
}

template <class T>
inline CoordVectorUnknown<T>& CoordVectorUnknown<T>::operator*=(const WithUnknown<T>& a)
{
  dist *= a;
  return *this;
}

template <class T>
inline CoordVectorUnknown<T>& CoordVectorUnknown<T>::operator/=(const WithUnknown<T>& a)
{
#if CHECK_DIVisION_BY_ZERO == 1
  if (a == T(0))
    dist = T_DIVisION_BY_ZERO_RESULT;
  else
#endif
    dist /= a;
  
  return *this;
}

template <class T>
inline bool CoordVectorUnknown<T>::operator ==(const CoordVectorUnknown<T>& v) const
{
  return (dist == v.dist) && (dir == v.dir);
}

template<class T>
inline bool operator==(const CoordVector<T>& v1,
		       const CoordVectorUnknown<T>& v2)
{
  return v2 == v1;
}

template <class T>
inline bool CoordVectorUnknown<T>::operator !=(const CoordVectorUnknown<T>& v) const
{
  return (dist != v.dist) || (dir != v.dir);
}

template<class T>
inline bool operator!=(const CoordVector<T>& v1,
		       const CoordVectorUnknown<T>& v2)
{
  return v2 != v1;
}

template <class T>
inline bool CoordVectorUnknown<T>::operator < (const CoordVectorUnknown<T>& v) const
{
  return (dist <  v.dist);
}

template<class T>
inline bool operator<(const CoordVector<T>& v1,
		      const CoordVectorUnknown<T>& v2)
{
  return v2 > v1;
}

template <class T>
inline bool CoordVectorUnknown<T>::operator <=(const CoordVectorUnknown<T>& v) const
{
  return (dist <= v.dist);
}

template<class T>
inline bool operator<=(const CoordVector<T>& v1,
		       const CoordVectorUnknown<T>& v2)
{
  return v2 >= v1;
}

template <class T>
inline bool CoordVectorUnknown<T>::operator > (const CoordVectorUnknown<T>& v) const
{
  return (dist >  v.dist);
}

template<class T>
inline bool operator>(const CoordVector<T>& v1,
		       const CoordVectorUnknown<T>& v2)
{
  return v2 < v1;
}

template <class T>
inline bool CoordVectorUnknown<T>::operator >=(const CoordVectorUnknown<T>& v) const
{
  return (dist >= v.dist);
}

template<class T>
inline bool operator>=(const CoordVector<T>& v1,
		       const CoordVectorUnknown<T>& v2)
{
  return v2 <= v1;
}

template <class T>
inline bool CoordVectorUnknown<T>::operator !() const
{
  return (!dist) && (!dir);
}

template<class T>
inline WithUnknown<T> CoordVectorUnknown<T>::getLength() const
{
  return dist;
}

template<class T>
inline CoordVectorUnknown<T>& CoordVectorUnknown<T>::setLength(const WithUnknown<T>& d)
{
  dist = d;
  return *this;
}

template<class T>
inline AngleDegreesUnknown<T> CoordVectorUnknown<T>::getAngle() const
{
  return dir;
}

template<class T>
inline CoordVectorUnknown<T>& CoordVectorUnknown<T>::setAngle(const AngleDegreesUnknown<T>& a)
{
  dir = a;
  return *this;
}

template <class T>
inline CoordVectorUnknown<T>& CoordVectorUnknown<T>::rotate(const AngleDegreesUnknown<T>& vdir)
{
  dir += vdir;
  return *this;
}

template <class T>
inline CoordVectorUnknown<T>& CoordVectorUnknown<T>::scale(const WithUnknown<T>& vdist)
{
  dist += vdist;
  return *this;
}

template <class T>
inline std::ostream& operator<<(std::ostream& os, const CoordVectorUnknown<T>& v)
{
#ifdef VECTORUNKNOWN_HEAD
  return os << VECTORUNKNOWN_HEAD << "(" << v.getAngle() << ", "
	    << v.getLength() << ")";
#else
  return os << "(" << v.getAngle() << ", " << v.getLength() << ")";
#endif
}


RS_END_NAMESPACE


#endif // COORDVECTORUNKNOWN_H
