/* -*- Mode: C++ -*-
 * $Id: LogStream.h,v 1.3 2001/06/19 14:01:49 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: LogStream.h
 * Unitname: Library
 * $Revision: 1.3 $
 * Created by: Fredrik Heintz 2000-08-23
 * Last modified by $Author: frehe $ $Date: 2001/06/19 14:01:49 $
 *
 *
 * HISTORY
 *
 * $Log: LogStream.h,v $
 * Revision 1.3  2001/06/19 14:01:49  frehe
 * Made minor changes to compile RoboSoc with gcc-3.0
 *
 * Revision 1.2  2000/09/05 11:07:41  frehe
 * Changed the log area constants from NAME to LA_NAME.
 *
 * Revision 1.1.1.1  2000/08/31 13:26:04  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 * Refactored CLogStream into LogStream 1.1 2000/08/23 10:30:12
 */

/**
 * @file LogStream.h
 * Include file for the class LogStream.
 * Should be used when printing log messages.
 * Implements the singleton pattern.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.3 $
 * $Author: frehe $
 * $Date: 2001/06/19 14:01:49 $
 */


#ifndef LOGSTREAM_H
#define LOGSTREAM_H


// Not needed! #include <iostream>
#include <stack>
#include <map>

#include "library_compiler_directives.h"
#include "LogArea.h"
// Not needed! #include "LogAreaName.h"
#include "LogStreamManip.h"


RS_BEGIN_NAMESPACE


/**
 * @class LogStream LogStream.h
 * Used to output logging info to a stream.
 * The class is an implementation of the singleton pattern.
 *
 * <!-- @invariant The class invariant is... -->
 *
 * @todo The logging to different streams is not very good.
 *       A better solution is probably to associate a list
 *       of pointers to stream objects that can be iterated
 *       over and the message is sent to each of them.
 *       The stream object has to know whether it should really
 *       log the message or not. In this way as many streams as
 *       wanted can be processed at the same time and the interface
 *       would become much more uniform.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.3 $
 * $Author: frehe $
 * $Date: 2001/06/19 14:01:49 $
 */
class LogStream {
public:
  /** LogStream is a singleton, which means that there is only
    one instance of this class in the system. This instance
    can only be accessed with the LogStream::instance() method,
    which returns a pointer to the only instance of the class. */
  static LogStream* instance();

  
  /** The destructor. */
  ~LogStream();

  
  /** Set the current area of logging. */
  LogStream& setCurrentArea(const LogAreaName& area);

  /** Get the current area of logging. */
  LogAreaName getCurrentArea() const;

  /** Set the current area, temporarily, of logging.
    To return to the current area use ResetTemporaryArea().
    What should happen if SetCurrentArea is called after SetTemporaryArea? */
  LogStream& setTemporaryArea(const LogAreaName& area);

  /** Resets the current area to the area used before the
    temporary area. */
  LogStream& resetTemporaryArea();

  
  /** Stop printing log info for an area. */
  LogStream& stopLogFileArea(const LogAreaName& area);

  /** Start printing log info for an area. */
  LogStream& startLogFileArea(const LogAreaName& area);

  /** Stop printing log info for an area. */
  LogStream& stopLogScreenArea(const LogAreaName& area);

  /** Start printing log info for an area. */
  LogStream& startLogScreenArea(const LogAreaName& area);

  
  /** Set the stream used by a certain area. */
  void setAreaFileStream(const LogAreaName& area, std::ostream* os = NULL);

  /** Get the stream currently used by a certain area. */
  std::ostream* getAreaFileStream(const LogAreaName& area);

  /** Set the stream used by a certain area. */
  void setAreaScreenStream(const LogAreaName& area, std::ostream* os = NULL);

  /** Get the stream currently used by a certain area. */
  std::ostream* getAreaScreenStream(const LogAreaName& area);

  
  /** Add a new area node to the log area tree.
   If the area already exist then it will be overwritten. */
  void addLogArea(const LogAreaName& area,
		  const LogAreaName& parent_area = LA_ROBOSOC,
		  const bool do_log_file = true,
		  std::ostream* file_log_stream = NULL, 
		  const bool do_log_screen = true,
		  std::ostream* screen_log_stream = NULL); 
  
  /* Print using the old fashion C style, which is faster than
    iostreams used in C++.  But can only print lines with less than
    1023 chars. */
  //int Printf(char const *fmt, ...) const;

  /** This operator<< is used to call functions by using the
    << operator. Ex. log_stream << resetTemporaryArea; */
  LogStream& operator<<(LogStream& (*fn)(LogStream&));

  /** This operator<< is used to call functions defined on ostreams by
    using the LogStream << operator. Ex. log_stream << std::endl; */
  LogStream& operator<<(std::ostream& (*fn)(std::ostream&));
  
  /** This operator<< is used to call functions taking an log area name argument
    by using the << operator. Ex. log_stream << setTemporaryArea(skill_area); */
  LogStream& operator<<(const LogStreamManip& m);

  /** This template makes it possible to easily output anything to
    a logstream that you can output to a standard ostream. */
  template <typename T> LogStream& operator<< (T thing) {
    if ( doFileLog && fileLogOut != NULL) {
      *fileLogOut << thing;
      fileLogOut->flush();
    }
    if ( doScreenLog && screenLogOut != NULL) {
      *screenLogOut << thing;
      screenLogOut->flush();
    }
    return *this;
  }

private:
  /** @c true iff the log message should be sent to a file. */
  bool doFileLog;
  /** A pointer to a ostream object that is used to output messages to a file. */
  std::ostream* fileLogOut;
  /** @c true iff the log message should be sent to the screen. */
  bool doScreenLog;
  /** A pointer to the ostream object used to output messages to the screen. */
  std::ostream* screenLogOut;

  /** The current area. */
  LogAreaName currentArea;
  /** A stack of temporary areas, used to nest temporary areas. */
  std::stack<LogAreaName> temporaryAreas; 

  /** An associative array of log area names and log area objects. */
  std::map<LogAreaName, LogArea> logAreas;  

  
  /** Create it using an ostream for the output to the a file,
    and an ostream for output to the screen.
    The class doesn't own them so the caller must make
    sure that the streams exist, are closed and freed. */
  LogStream(const bool do_file_log = true, std::ostream* file_os = NULL,
	    const bool do_screen_log = true, std::ostream* screen_os = NULL);
  

  /** The copy constructor. Hidden since the class is a singleton. */
  LogStream(const LogStream&);
  /** The assignment operator. Hidden since the class is a singleton. */
  LogStream& operator= (const LogStream&);

  /** A static pointer to the only instance of the class. */
  static LogStream* onlyInstance;
};


#ifndef DOXYGEN_SHOULD_SKIP_THIS

//@{
/** @name Functions callable from the << operator. */
/** Functions that should be callable from the << operator. */
LogStream& resetTemporaryArea(LogStream& ls);
LogStreamManip setCurrentArea(const LogAreaName& area);
LogStreamManip setTemporaryArea(const LogAreaName& area);
LogStreamManip stopLogFileArea(const LogAreaName& area);
LogStreamManip startLogFileArea(const LogAreaName& area);
LogStreamManip stopLogScreenArea(const LogAreaName& area);
LogStreamManip startLogScreenArea(const LogAreaName& area);
LogStream& _SetCurrentArea(LogStream& ls, const LogAreaName& area);
LogStream& _SetTemporaryArea(LogStream& ls, const LogAreaName& area);
LogStream& _StopLogFileArea(LogStream& ls, const LogAreaName& area);
LogStream& _StartLogFileArea(LogStream& ls, const LogAreaName& area);
LogStream& _StopLogScreenArea(LogStream& ls, const LogAreaName& area);
LogStream& _StartLogScreenArea(LogStream& ls, const LogAreaName& area);
//@}

#endif // DOXYGEN_SHOULD_SKIP_THIS


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "LogStream.icc"
#endif


#endif // !LOGSTREAM_H
