/* -*- Mode: C++ -*-
 * $Id: ObjectCollection.h,v 1.4 2001/01/09 12:04:39 olalu526 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: ObjectCollection.h
 * Unitname: Library
 * $Revision: 1.4 $
 * Created by: Fredrik Heintz 2000-08-29
 * Last modified by $Author: olalu526 $ $Date: 2001/01/09 12:04:39 $
 *
 *
 * HISTORY
 *
 * $Log: ObjectCollection.h,v $
 * Revision 1.4  2001/01/09 12:04:39  olalu526
 * Added setLength to the ObjectCollection to make the alternative view structure possible.
 *
 * Revision 1.3  2000/10/18 23:05:20  frehe
 * Removed the illegal default constructor.
 *
 * Revision 1.2  2000/09/01 03:23:53  frehe
 * Made minor changes.
 *
 * Revision 1.1.1.1  2000/08/31 13:26:04  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file ObjectCollection.h
 * Include file for the template ObjectCollection.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.4 $
 * $Author: olalu526 $
 * $Date: 2001/01/09 12:04:39 $
 */


#ifndef OBJECTCOLLECTION_H
#define OBJECTCOLLECTION_H


#include <vector>


/** If defined the range checks will be made on all accesses to
  the ObjectCollection instances. */
#define RS_DO_RANGE_CHECK


RS_BEGIN_NAMESPACE


/**
 * @class ObjectCollection ObjectCollection.h
 * Used to store a collection of objects.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.4 $
 * $Author: olalu526 $
 * $Date: 2001/01/09 12:04:39 $
 */
template<class T>
class ObjectCollection {
public:
  /** Init the collection with @a len number of objects with the value @a def_val,
    which will also be the default value after a reset of an element. */
  ObjectCollection(const UInt16 len, const T& def_val = T());
  /** Copy constructor. */
  ObjectCollection(const ObjectCollection<T>& obj);

  /** Assignement operator. */
  ObjectCollection<T>& operator= (const ObjectCollection<T>& obj);

  /** Access operator that retrieves the @a elt element in the collection.
   The retrieved value can be assigned to since it is a reference. */
  T& operator[](const UInt16 elt);
  /** Constant access operator, retrieves the @a elt element in the collection. */
  const T& operator[](const UInt16 elt) const;

  /** Returns the number of elements in the collection. */
  UInt16 getLength() const;
  void setLength(const UInt16);

  /** Resets the @a elt element in the collection to its default value. */
  void reset(const UInt16 elt);
  /** Resets all elements in the collection to their default value. */
  void resetAll();

  /** Copies the element at position @a from to the position @a to in the collection. */
  void copy(const UInt16 from, const UInt16 to);
  /** Copies the element at position @a from in the collection @a col to the position
    @a to in this collection. */
  void copyFrom(const ObjectCollection<T>& col,
		const UInt16 from, const UInt16 to);

private:
  /** The number of elements in the collection. */
  UInt16 length;
  /** The default value for the collection. */
  T defaultValue;
  /** The actual data stored in the collection. */
  std::vector<T> data;
};


template<class T>
inline T& ObjectCollection<T>::operator[](const UInt16 elt)
{
#ifdef RS_DO_RANGE_CHECK
  if ( elt < length )
#endif
    return data[elt];
#ifdef RS_DO_RANGE_CHECK
  else
    return defaultValue;
#endif
}


template<class T>
inline const T& ObjectCollection<T>::operator[](const UInt16 elt) const
{
#ifdef RS_DO_RANGE_CHECK
  if ( elt < length )
#endif
    return data[elt];
#ifdef RS_DO_RANGE_CHECK
  else
    return defaultValue;
#endif
}


template<class T>
UInt16 ObjectCollection<T>::getLength() const
{
  return length;
}

template<class T>
void ObjectCollection<T>::setLength(const UInt16 l)
{
  for (int i = length; i < l; i++) {
    data[i] = defaultValue;
  }
  length = l;
}

template<class T>
inline void ObjectCollection<T>::reset(const UInt16 elt)
{
#ifdef RS_DO_RANGE_CHECK
  if ( elt < length )
#endif
    data[elt] = defaultValue;
}


template<class T>
inline void ObjectCollection<T>::copy(const UInt16 from, const UInt16 to)
{
#ifdef RS_DO_RANGE_CHECK
  if ( from < length && to < length )
#endif
    data[to] = data[from];
}


template<class T>
inline void ObjectCollection<T>::copyFrom(const ObjectCollection<T>& col,
					  const UInt16 from, const UInt16 to)
{
#ifdef RS_DO_RANGE_CHECK
  if ( to < length )
#endif
    data[to] = col[from];
}


template<class T>
ObjectCollection<T>::ObjectCollection(const UInt16 len, const T& def_val)
  : length(len),
    defaultValue(def_val),
    data(length, defaultValue)
{
}


template<class T>
ObjectCollection<T>::ObjectCollection(const ObjectCollection<T>& obj)
  : length(obj.length),
    defaultValue(obj.defaultValue),
    data(obj.data)
{
}


template<class T>
ObjectCollection<T>& ObjectCollection<T>::operator= (const ObjectCollection<T>& obj)
{
  if ( this != &obj ) {
    data = obj.data;
    length = obj.length;
    defaultValue = obj.defaultValue;
  }

  return *this;
}


template<class T>
void ObjectCollection<T>::resetAll()
{
  for (std::vector<T>::iterator i = data.begin(); i != data.end(); ++i)
    *i = defaultValue;
}


RS_END_NAMESPACE


#endif // !OBJECTCOLLECTION_H
