/* -*- Mode: C++ -*-
 * $Id: Plan.h,v 1.4 2000/10/25 19:34:08 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Plan.h
 * Unitname: Library
 * $Revision: 1.4 $
 * Created by: Fredrik Heintz 2000-09-03
 * Last modified by $Author: frehe $ $Date: 2000/10/25 19:34:08 $
 *
 *
 * HISTORY
 *
 * $Log: Plan.h,v $
 * Revision 1.4  2000/10/25 19:34:08  frehe
 * Fixed Plan to check for empty steps and plans.
 *
 * Revision 1.3  2000/09/05 11:07:06  frehe
 * Improved the default constructor.
 *
 * Revision 1.2  2000/09/03 20:05:51  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/03 13:35:56  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file Plan.h
 * Include file for the class Plan.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2000/10/25 19:34:08 $
 */


#ifndef PLAN_H
#define PLAN_H


#include <deque>

#include "Step.h"


RS_BEGIN_NAMESPACE


/**
 * @class Plan Plan.h
 * Used to represent a plan. A Plan is a sequence of Step.
 * The steps in the plans are stored in a deque (dubbel ended queu),
 * in order to have fast insertion at the back of the sequence, fast
 * deletion at the top of the sequence and at the same time have random
 * access to all element, for algortihms to work on plans as sequences.
 *
 * <!--@pre The preconditions for constructing the class are ...
 * @post The postconditions after constructing the class are...
 * @invariant A Plan contains at least one command. -->
 *
 * @todo The current implementation of Step and Plan will probably
 *       lead to a lot of copying of commands. This might be a problem.
 *       If it is the plan should take step pointers as arguments and
 *       own them.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2000/10/25 19:34:08 $
 */
class Plan {
public:
  /** The default constructor. If @a step is empty then create an empty
    plan, othwerwise a plan containing the step @a step. */
  Plan(const Step& step = Step());

  /** The copy constructor.
    @param obj is the object to be copied.
    @post The constructed object is an exact copy of @a obj. */
  Plan(const Plan& obj);

  /** The destructor. */
  virtual ~Plan();

  
  /** The assignment operator.
    @param obj is the object to be assigned from.
    @return A reference to the the newly assigned object.
    @post The object is an exact copy of @a obj. */
  Plan& operator=(const Plan& obj);


  /** Add a step to the plan. If the step is empty then the
      plan is still empty. */
  Plan& addStep(const Step& step);

  
  /** Get the first step in the plan.
      If the plan is empty then an empty step is returned. */
  Step getNextStep() const;

  /** Remove the first step in the plan. */
  Plan& removeNextStep();


  /** Get the last step in the plan.
      If the plan is empty then an empty step is returned. */
  Step getLastStep() const;

  /** Remove the last step in the plan. */
  Plan& removeLastStep();

  
  /** Test if the plan contains no steps.
    @returns @c true iff the plan is empty. */
  bool empty() const;

  /** Removes all the steps stored in the plan. */
  Plan& makeEmpty();


protected:
  /** Test the class invariant.
    This function implements a test of the class invariant.
    It should be tested after each constructor, before each destructor,
    and at the end of methods that change the state of the object.
    @return @c true iff the class invariant is satisfied, otherwise @c false. */
  bool invariant() const;

  
private:
  /** Store the sequence of steps. */
  std::deque<Step> stepSequence;
};


RS_END_NAMESPACE

  
/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "Plan.icc"
#endif


#endif // !PLAN_H
