/* -*- Mode: C++ -*-
 * $Id: RSException.h,v 1.1.1.1 2000/08/31 13:26:04 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: RSException.h
 * Unitname: Library
 * $Revision: 1.1.1.1 $
 * Created by: Fredrik Heintz 2000-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/08/31 13:26:04 $
 *
 *
 * HISTORY
 *
 * $Log: RSException.h,v $
 * Revision 1.1.1.1  2000/08/31 13:26:04  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file RSException.h
 * Include file for the class RSException.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.1.1.1 $
 * $Author: frehe $
 * $Date: 2000/08/31 13:26:04 $
 */


#ifndef RSEXCEPTION_H
#define RSEXCEPTION_H


#include <exception>
#include <string>
#include "library_compiler_directives.h"
#include "machine_dependent_types.h"


RS_BEGIN_NAMESPACE


/**
 * @class RSException RSException.h
 * Class used as the base class of all exceptions defined in
 * RoboSoc.
 * It inherits from std::exception which means that if the
 * user tries to catch std::exception she will also catch
 * RSException and all its child classes.
 * Non of its methods are allowed to throw exceptions since
 * then we could end up with a endless recursive throwing
 * of exceptions.
 *
 * <!-- @invariant The class invariant is... -->
 *
 * <!-- @todo -->
 *
 * @par Unit:
 *      Library
 * $Revision: 1.1.1.1 $
 * $Author: frehe $
 * $Date: 2000/08/31 13:26:04 $
 */
class RSException : public std::exception {
public:
  /** Constructor for RSEcception. No default value
    is allowed for the description string @a descr,
    since all exceptions should have a description.
    But it is possible for sub classes to have
    default constructors if they provide the description
    instead. */
  RSException(const std::string& descr,
	      const std::string& file_n = "",
	      const UInt32 line_no = 0,
	      const std::string& nme = "RSException") throw();
  /** Copy constructor. */
  RSException(const RSException&) throw();
  /** A virtual destructor. */
  virtual ~RSException () throw();

  /** Assignment operator. */
  RSException& operator= (const RSException&) throw();

  /** Find out what the exception is about, returns a C++ string. */
  virtual const std::string What() const throw();
  /** Find out what the exception is about, returns a C string. */
  virtual const char* what() const throw();

  /** Change the description of an exception. */
  void setDescription(const std::string& descr) throw();
  /** Get the description of an exception. */
  const std::string& getDescription() throw();

  /** Change the name of an exception. */
  void setName(const std::string& nme) throw();
  /** Get the name of an exception. */
  const std::string& getName() throw();

  /** Change the filename of an exception, usually it represent the
    file where the exception was thrown. */
  void setFileName(const std::string& f_name) throw();
  /** Get the filename of the exception. */
  const std::string& getFileName() throw();

  /** Change the line number of an exception, usually it represent the
    line number where the exception was thrown. */
  void setLineNumber(const UInt32 line_no) throw();
  /** Get the line number of the exception. */
  UInt32 getLineNumber() throw();  

private:
  /** The description string of the exception. */
  std::string description;
  /** The name string of the exception. */
  std::string name;
  /** The file name of the exception. */
  std::string fileName;
  /** The line number of the exception. */
  UInt32 lineNumber;
};


/** Overloaded ostream::operator<< to output RSExceptions. */
std::ostream& operator <<(std::ostream& os, const RSException& excp);


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "RSException.icc"
#endif


#endif // !RSEXCEPTION_H

