/* -*- Mode: C++ -*-
 * $Id: Step.h,v 1.5 2000/10/25 19:33:36 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Step.h
 * Unitname: Library
 * $Revision: 1.5 $
 * Created by: Fredrik Heintz 2000-09-03
 * Last modified by $Author: frehe $ $Date: 2000/10/25 19:33:36 $
 *
 *
 * HISTORY
 *
 * $Log: Step.h,v $
 * Revision 1.5  2000/10/25 19:33:36  frehe
 * Fixed Step to check for empty steps and NULL commands.
 *
 * Revision 1.4  2000/09/05 21:07:25  frehe
 * Fixed a bug in the copying of steps.
 *
 * Revision 1.3  2000/09/05 11:07:06  frehe
 * Improved the default constructor.
 *
 * Revision 1.2  2000/09/03 14:23:19  frehe
 * Made minor changes to improve the documentation.
 *
 * Revision 1.1  2000/09/03 13:35:57  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file Step.h
 * Include file for the class Step.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2000/10/25 19:33:36 $
 */


#ifndef STEP_H
#define STEP_H


#include <deque>

//#include "machine_dependent_compiler_directives.h"
#include "Command.h"


RS_BEGIN_NAMESPACE


/**
 * @class Step Step.h
 * Used to represent a step in a plan. A step is a set of commands
 * that should and could be executed by the soccer server and
 * the agent in a single simulator cycle.
 * The commands in the steps are stored as a queue, i.e. the first
 * command to be added to the step will be executed first.
 *
 * <!--@pre The preconditions for constructing the class are ...
 * @post The postconditions after constructing the class are...
 * @invariant A step contains at least one command.
 *
 * @todo Things that should be done... -->
 *
 * @par Unit:
 *      Library
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2000/10/25 19:33:36 $
 */
class Step {
public:
  /** The default constructor. If @a command is NULL the create an empty
    step, otherwise create a step with the @a command command. */
  Step(const Command* const command = NULL);

  /** The copy constructor.
      @param obj is the object to be copied.
      @post The constructed object is an exact copy of @a obj. */
  Step(const Step& obj);

  /** The destructor. */
  virtual ~Step();

  
  /** The assignment operator.
      @param obj is the object to be assigned from.
      @return A reference to the the newly assigned object.
      @post The object is an exact copy of @a obj. */
  Step& operator=(const Step& obj);


  /** Add a command to the step, the command added will be a
      copy of @a command. No tests to see if it's legal to add
      the command is done. If the command is NULL it will not
      be added. */
  Step& addCommand(const Command* const command);

  
  /** Get a copy of the first command in the sequence.
    Remember to delete the command object otherwise you will
    have a memory leak.
    If the step is empty then NULL is returned. */
  Command* getNextCommand() const;

  /** Remove the first command in the sequence, if the step is
      not empty. */
  Step& removeNextCommand();

  
  /** Test if the step contains no commands.
    @returns @c true iff the step is empty. */
  bool empty() const;

  /** Removes all the commands stored in the step. */
  Step& makeEmpty();


protected:
  /** Test the class invariant.
    This function implements a test of the class invariant.
    It should be tested after each constructor, before each destructor,
    and at the end of methods that change the state of the object.
    @return @c true iff the class invariant is satisfied, otherwise @c false. */
  bool invariant() const;

  
private:
  /** Store the sequence of commands. */
  std::deque<Command*> commandSequence;
};


RS_END_NAMESPACE

  
/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "Step.icc"
#endif


#endif // !STEP_H
