/* -*- Mode: C++ -*-
 * $Id: WithConfidence.h,v 1.2 2000/09/01 03:23:54 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: WithConfidence.h
 * Unitname: Library
 * $Revision: 1.2 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/01 03:23:54 $
 *
 *
 * HISTORY
 *
 * $Log: WithConfidence.h,v $
 * Revision 1.2  2000/09/01 03:23:54  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/01 01:24:42  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 *
 */

/**
 * @file WithConfidence.h
 * Include file for the template WithConfidence.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2000/09/01 03:23:54 $
 */


#ifndef WITHCONFIDENCE_H
#define WITHCONFIDENCE_H


#include <string>
#include <iostream>

#include "math_utils.h"
#include "Confidence.h"
#include "string_utils.h"


/** If is defined then pessimistic observations will be used. */
#define RS_PESSIMISTIC_OBSERVATIONS


RS_BEGIN_NAMESPACE


/**
 * @class WithConfidence WithConfidence.h
 * Template class for representing basic types that can have a confidence
 * value. The template argument only defines what type should be annotated
 * with a confidence value.
 * The only demands on the template argument T is that the arithmetic
 * functions are defined on it.
 *
 * There are (at least) two ways of defining operators with confidence.
 * Either you take the highest confidence (optimistic) or
 * you take the lowest confidence (pessimistic).
 * A more realistic approach is to take the product of the confidences!
 *
 * Another problem is what the initial confidence should be.
 * Currently it's 0, but it could also be 1 (depending on what the
 * user thinks about assignment, if they are true or if they are
 * just some new value).
 * A third option is to keep the old value, but I think that is the
 * least natural.
 *
 * <!-- @todo -->
 *
 * @par Unit:
 *      Library
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2000/09/01 03:23:54 $
 */
template <class T> class WithConfidence
{
public:
  WithConfidence();
  WithConfidence(const T& val, const Confidence& Confidence = Confidence(),
		 const bool print_confidence = true);
  WithConfidence(const WithConfidence<T>&);

  T getValue() const;

  Confidence getConfidence() const;
  void setConfidence(const Confidence& Confidence);

  bool printConfidence() const;
  void setPrintConfidence(const bool print_confidence);
  
  WithConfidence<T>& operator= (const WithConfidence<T>& op);
  WithConfidence<T>& operator= (const T& op);

  //@{
  /** @name Operators on WithConfidence<T> */
  WithConfidence<T>& operator++ ();
  WithConfidence<T>& operator++ (int);
  WithConfidence<T>& operator-- ();
  WithConfidence<T>& operator-- (int);
  WithConfidence<T>& operator+= (const WithConfidence<T>& op);
  WithConfidence<T>& operator-= (const WithConfidence<T>& op);
  WithConfidence<T>& operator*= (const WithConfidence<T>& op);
  WithConfidence<T>& operator/= (const WithConfidence<T>& op);
  //WithConfidence<T>& operator%= (const WithConfidence<T>& op);

  WithConfidence<T> operator- () const;
  WithConfidence<T> operator+ (const WithConfidence<T>& op) const;
  WithConfidence<T> operator- (const WithConfidence<T>& op) const;
  WithConfidence<T> operator* (const WithConfidence<T>& op) const;
  WithConfidence<T> operator/ (const WithConfidence<T>& op) const;
  //WithConfidence<T> operator% (const WithConfidence<T>& op) const;
  //@}
  

  //@{
  /** @name Comparision operators on WithConfidence<T>
    Confidence doesn't effect anything in these comparissions so:
    !obj == true iff !obj.value
    obj1 > obj2 iff obj1.value > obj2.value */
  bool operator!  () const;
  bool operator== (const WithConfidence<T>& op) const;
  //bool operator== (const T& op) const;
  bool operator!= (const WithConfidence<T>& op) const;
  //bool operator!= (const T& op) const;
  bool operator<  (const WithConfidence<T>& op) const;
  //bool operator<  (const T& op) const;
  bool operator<= (const WithConfidence<T>& op) const;
  //bool operator<= (const T& op) const;
  bool operator>  (const WithConfidence<T>& op) const;
  //bool operator>  (const T& op) const;
  bool operator>= (const WithConfidence<T>& op) const;
  //bool operator>= (const T& op) const;
  //@}
  
protected:
  T value;
  Confidence confidence;
  bool doPrintConfidence;
};

template <class T> 
WithConfidence<T>::WithConfidence()
  : value(),
    confidence(),
    doPrintConfidence(true)
{
}

template <class T> 
WithConfidence<T>::WithConfidence(const T& val,
				  const Confidence& confidence,
				  const bool print_confidence)
  : value(val),
    confidence(confidence),
    doPrintConfidence(print_confidence)
{
}

template <class T> 
WithConfidence<T>::WithConfidence(const WithConfidence<T>& op)
  : value(op.value),
    confidence(op.confidence),
    doPrintConfidence(op.doPrintConfidence)
{
}

template <class T>
inline T WithConfidence<T>::getValue() const
{
  return value;
}

template <class T> 
inline Confidence WithConfidence<T>::getConfidence() const
{
  return confidence;
}

template <class T> 
inline void WithConfidence<T>::setConfidence(const Confidence& conf)
{
  confidence = conf;
}

template <class T>
inline bool WithConfidence<T>::printConfidence() const
{
  return doPrintConfidence;
}

template <class T>
inline void WithConfidence<T>::setPrintConfidence(const bool print_conf)
{
  doPrintConfidence = print_conf;
}

template <class T>
WithConfidence<T>& WithConfidence<T>::operator= (const WithConfidence<T>& op)
{
  if ( &op != this ) {
    value = op.value;
    confidence = op.confidence;
    doPrintConfidence = op.doPrintConfidence;
  }

  return *this;
}

template <class T>
WithConfidence<T>& WithConfidence<T>::operator= (const T& op)
{
  value = op;
  confidence = Confidence();
  doPrintConfidence = true;
  return *this;
}

template <class T>
inline WithConfidence<T>& WithConfidence<T>::operator++ ()
{
  ++value;
  return *this;
}

template <class T>
inline WithConfidence<T>& WithConfidence<T>::operator++ (int)
{
  return ++(*this);
}

template <class T>
inline WithConfidence<T>& WithConfidence<T>::operator-- ()
{
  --value;
  return *this;
}

template <class T>
inline WithConfidence<T>& WithConfidence<T>::operator-- (int)
{
  return --(*this);
}

template <class T>
WithConfidence<T>& WithConfidence<T>::operator+= (const WithConfidence<T>& op)
{
  value += op.value;
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  confidence = min(confidence, op.confidence);
#else
  confidence = max(confidence, op.confidence);
#endif
  return *this;
}

template <class T>
WithConfidence<T>& WithConfidence<T>::operator-= (const WithConfidence<T>& op)
{
  value -= op.value;
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  confidence = min(confidence, op.confidence);
#else
  confidence = max(confidence, op.confidence);
#endif
  return *this;
}

template <class T>
WithConfidence<T>& WithConfidence<T>::operator*= (const WithConfidence<T>& op)
{
  value *= op.value;
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  confidence = min(confidence, op.confidence);
#else
  confidence = max(confidence, op.confidence);
#endif
  return *this;
}

template <class T>
WithConfidence<T>& WithConfidence<T>::operator/= (const WithConfidence<T>& op)
{
  value /= op.value;
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  confidence = min(confidence, op.confidence);
#else
  confidence = max(confidence, op.confidence);
#endif
  return *this;
}

template<class T>
inline WithConfidence<T> WithConfidence<T>::operator- () const
{
  return WithConfidence<T>(-value, confidence);
}

template<class T>
inline WithConfidence<T> WithConfidence<T>::operator+ (const WithConfidence<T>& op) const
{
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  return WithConfidence<T>(value + op.value,
			   min(confidence, op.confidence));
#else
  return WithConfidence<T>(value + op.value,
			   max(confidence, op.confidence));
#endif
}

template<class T>
inline WithConfidence<T> WithConfidence<T>::operator- (const WithConfidence<T>& op) const
{
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  return WithConfidence<T>(value - op.value,
			   min(confidence, op.confidence));
#else
  return WithConfidence<T>(value - op.value,
			   max(confidence, op.confidence));
#endif
}

template<class T>
inline WithConfidence<T> WithConfidence<T>::operator* (const WithConfidence<T>& op) const
{
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  return WithConfidence<T>(value * op.value,
			   min(confidence, op.confidence));
#else
  return WithConfidence<T>(value * op.value,
			   max(confidence, op.confidence));
#endif
}

template<class T>
inline WithConfidence<T> WithConfidence<T>::operator/ (const WithConfidence<T>& op) const
{
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  return WithConfidence<T>(value / op.value,
			   min(confidence, op.confidence));
#else
  return WithConfidence<T>(value / op.value,
			   max(confidence, op.confidence));
#endif
}

template<class T>
inline bool WithConfidence<T>::operator!() const
{
  return (!value );
}

template<class T>
inline bool WithConfidence<T>::operator==(const WithConfidence<T>& op) const
{
  return ( value == op.value );
}

template<class T>
inline bool WithConfidence<T>::operator!=(const WithConfidence<T>& op) const
{
  return !( *this == op );
} 

template<class T>
inline bool WithConfidence<T>::operator<(const WithConfidence<T>& op) const
{ 
  return (value < op.value );
} 

template<class T>
inline bool WithConfidence<T>::operator<=(const WithConfidence<T>& op) const
{
  return !( *this > op );
}

template<class T>
inline bool WithConfidence<T>::operator>(const WithConfidence<T>& op) const
{ 
  return (value > op.value );
} 

template<class T>
inline bool WithConfidence<T>::operator>=(const WithConfidence<T>& op) const
{
  return !( *this < op );
}

template<class T>
inline WithConfidence<T> operator+ (const T& op1, const WithConfidence<T>& op2)
{
  return op2 + op1;
}

template<class T>
inline WithConfidence<T> operator- (const T& op1, const WithConfidence<T>& op2)
{
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  return WithConfidence<T>(op1 - op2.getValue(), min(Confidence(), op2.confidence));
#else
  return WithConfidence<T>(op1 - op2.getValue(), max(Confidence(), op2.confidence));
#endif
}

template<class T>
inline WithConfidence<T> operator* (const T& op1, const WithConfidence<T>& op2)
{
  return op2 * op1;
}

template<class T>
inline WithConfidence<T> operator/ (const T& op1, const WithConfidence<T>& op2)
{
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  return WithConfidence<T>(op1 / op2.getValue(), min(Confidence(), op2.confidence));
#else
  return WithConfidence<T>(op1 / op2.getValue(), max(Confidence(), op2.confidence));
#endif
}

template<class T>
inline bool operator==(const T& op1, const WithConfidence<T>& op2)
{
  return op2 == op1;
}

template<class T>
inline bool operator!=(const T& op1, const WithConfidence<T>& op2)
{
  return op2 != op1;
}

template<class T>
inline bool operator<(const T& op1, const WithConfidence<T>& op2)
{
  return op2 > op1;
}

template<class T>
inline bool operator<=(const T& op1, const WithConfidence<T>& op2)
{
  return op2 >= op1;
}

template<class T>
inline bool operator>(const T& op1, const WithConfidence<T>& op2)
{
  return op2 < op1;
}

template<class T>
inline bool operator>=(const T& op1, const WithConfidence<T>& op2)
{
  return op2 <= op1;
}

template<class T>
std::ostream& operator<<(std::ostream& os, const WithConfidence<T>& obj)
{
  if (obj.printConfidence())
    return os << obj.getValue() << " " << obj.getConfidence();
  else
    return os << obj.getValue();
}

/*
template<class T>
inline std::istream& operator>>(std::istream& is, WithConfidence<T>& obj)
{
  // Should add something so that the user can enter an observation time too.
  return is >> obj.getValue();
}
*/

template <class T> 
const std::string toString(const WithConfidence<T>& obj)
{
  if (obj.printConfidence())
    return toString(obj.getValue()) + " " + toString(obj.getConfidence());
  else
    return toString(obj.getValue());
}

template<class T>
WithConfidence<T> sqrt(const WithConfidence<T>& op)
{
  return WithConfidence<T>(sqrt(op.getValue()));
}

/*
#ifdef RS_TRIG_IN_DEG
#include "AngleDeg.h"

template<class Time>
WithConfidence<WithObservationTime<WithUnknown< AngleDeg<T> >, Time> > ASin(const WithConfidence<WithObservationTime<WithUnknown<WithValue<T>>, Time>>& op)
{
  return WithConfidence<WithObservationTime<WithUnknown< AngleDeg<T> >, Time>>(ASin(WithObservationTime<WithUnknown<WithValue<T> >, Time>(op)));
}

template<class Time>
WithConfidence<WithObservationTime<WithUnknown< AngleDeg<T> >, Time> > ACos(const WithConfidence<WithObservationTime<WithUnknown<WithValue<T>>, Time>>& op)
{
  return WithConfidence<WithObservationTime<WithUnknown< AngleDeg<T> >, Time>>(ACos(WithObservationTime<WithUnknown<WithValue<T> >, Time>(op)));
}

template<class Time>
WithConfidence<WithObservationTime<WithUnknown< AngleDeg<T> >, Time>> ATan(const WithConfidence<WithObservationTime<WithUnknown<WithValue<T>>, Time>>& op)
{
  return WithConfidence<WithObservationTime<WithUnknown< AngleDeg<T> >, Time>>(ATan(WithObservationTime<WithUnknown<WithValue<T> >, Time>(op)));
}

template<class Time>
WithConfidence<WithObservationTime<WithUnknown< AngleDeg<T> >, Time>>
ATan2(const WithConfidence<WithObservationTime<WithUnknown<WithValue<T>>, Time>>& op1,
      const WithConfidence<WithObservationTime<WithUnknown<WithValue<T>>, Time>>& op2)
{
  return WithConfidence<WithObservationTime<WithUnknown< AngleDeg<T> >, Time>>(ATan2(WithObservationTime<WithUnknown<WithValue<T> >, Time>(op1),
													      WithObservationTime<WithUnknown<WithValue<T> >, Time>(op2)));
}

#endif // RS_TRIG_IN_DEG
*/

/*
#ifdef RS_TRIG_IN_DEG
#include "AngleDeg.h"

template<class T>
WithConfidence< AngleDeg<T> > ASin(const WithConfidence<T>& op)
{
  return WithConfidence< AngleDeg<T>>(ASin(op.getValue()));
}

template<class T>
WithConfidence< AngleDeg<T> > ACos(const WithConfidence<T>& op)
{
  return WithConfidence< AngleDeg<T>>(ACos(op.getValue()));
}

template<class T>
WithConfidence< AngleDeg<T>> ATan(const WithConfidence<T>& op)
{
  return WithConfidence< AngleDeg<T>>(ATan(op.getValue()));
}

template<class T>
WithConfidence< AngleDeg<T>> ATan2(const WithConfidence<T>& op1, const WithConfidence<T>& op2)
{
  return WithConfidence< AngleDeg<T>>(ATan2(T(op1), T(op2)));
}

#endif // RS_TRIG_IN_DEG

#ifdef RS_TRIG_IN_RAD
#include "AngleRadians.h"

template<class T>
WithConfidence< AngleRadians<T>> ASin(const WithConfidence<T>& op)
{
  return WithConfidence< AngleRadians<T>>(ASin(op.getValue()));
}

template<class T>
WithConfidence< AngleRadians<T>> ACos(const WithConfidence<T>& op)
{
  return WithConfidence< AngleRadians<T>>(ACos(op.getValue()));
}

template<class T>
WithConfidence< AngleRadians<T>> ATan(const WithConfidence<T>& op)
{
  return WithConfidence< AngleRadians<T>>(ATan(op.getValue()));
}

template<class T>
WithConfidence< AngleRadians<T>> ATan2(const WithConfidence<T>& op1, const WithConfidence<T>& op2)
{
  return WithConfidence< AngleRadians<T>>(ATan2(op1.getValue(), op2.getValue()));
}

#endif // RS_TRIG_IN_RAD
*/


RS_END_NAMESPACE


#endif // !WITHCONFIDENCE_H
