/* -*- Mode: C++ -*-
 * $Id: WithObservationTime.h,v 1.2 2000/09/01 03:23:54 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: WithObservationTime.h
 * Unitname: Library
 * $Revision: 1.2 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/01 03:23:54 $
 *
 *
 * HISTORY
 *
 * $Log: WithObservationTime.h,v $
 * Revision 1.2  2000/09/01 03:23:54  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/01 01:24:43  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 *
 */

/**
 * @file WithObservationTime.h
 * Include file for the template WithObservationTime.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2000/09/01 03:23:54 $
 */


#ifndef WITHOBSERVATIONTIME_H
#define WITHOBSERVATIONTIME_H


#include <string>
#include <iostream>

#include "library_types.h"
#include "math_utils.h"
#include "string_utils.h"


/** If is defined then pessimistic observations will be used. */
#define RS_PESSIMISTIC_OBSERVATIONS


RS_BEGIN_NAMESPACE


/**
 * @class WithObservationTime WithObservationTime.h
 * Template class for representing basic types that can have a confidence
 * value. The template argument only defines what type should be annotated
 * with a confidence value.
 * The only demands on the template argument T is that the arithmetic
 * functions are defined on it.
 *
 * There are (at least) two ways of defining operators with
 * observation time. Either you take the latest observation (optimistic) or
 * you take the earliest observation (pessimistic).
 *
 * Another problem is the default value of observation times.
 * There are three options:
 * 1. Set it to the default value of Time.
 * 2. Set it to Time::currentTime (which then must be defined).
 * 3. Keep the current time.
 * Currently the first option is used.
 *
 * <!-- @todo -->
 *
 * @par Unit:
 *      Library
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2000/09/01 03:23:54 $
 */
template <class T, class Time> class WithObservationTime
{
public:
  WithObservationTime();
  WithObservationTime(const T& val, const Time& obs_time = Time(),
		      const bool print_obs_time = true);
  WithObservationTime(const WithObservationTime<T, Time>&);

  T getValue() const;
  Time getObservationTime() const;
  void setObservationTime(const Time& obs_time);
  T getCyclesSince(const Time& cur_time) const;

  bool printObservationTime() const;
  void setPrintObservationTime(const bool print_obs_time);
  
  WithObservationTime<T, Time>& operator= (const WithObservationTime<T, Time>& op);
  WithObservationTime<T, Time>& operator= (const T& op);

  WithObservationTime<T, Time>& operator++ ();
  WithObservationTime<T, Time>& operator++ (int);
  WithObservationTime<T, Time>& operator-- ();
  WithObservationTime<T, Time>& operator-- (int);
  WithObservationTime<T, Time>& operator+= (const WithObservationTime<T, Time>& op);
  WithObservationTime<T, Time>& operator-= (const WithObservationTime<T, Time>& op);
  WithObservationTime<T, Time>& operator*= (const WithObservationTime<T, Time>& op);
  WithObservationTime<T, Time>& operator/= (const WithObservationTime<T, Time>& op);
  //WithObservationTime<T, Time>& operator%= (const WithObservationTime<T, Time>& op);

  WithObservationTime<T, Time> operator- () const;
  WithObservationTime<T, Time> operator+ (const WithObservationTime<T, Time>& op) const;
  WithObservationTime<T, Time> operator- (const WithObservationTime<T, Time>& op) const;
  WithObservationTime<T, Time> operator* (const WithObservationTime<T, Time>& op) const;
  WithObservationTime<T, Time> operator/ (const WithObservationTime<T, Time>& op) const;
  //WithObservationTime<T, Time> operator% (const WithObservationTime<T, Time>& op) const;

  // ObservationTime doesn't effect anything in these comparissions so:
  // !obj == true iff !obj.value
  // obj1 > obj2 iff obj1.value > obj2.value
  bool operator!  () const;
  bool operator== (const WithObservationTime<T, Time>& op) const;
  bool operator!= (const WithObservationTime<T, Time>& op) const;
  bool operator<  (const WithObservationTime<T, Time>& op) const;
  bool operator<= (const WithObservationTime<T, Time>& op) const;
  bool operator>  (const WithObservationTime<T, Time>& op) const;
  bool operator>= (const WithObservationTime<T, Time>& op) const;
  
private:
  T value;
  Time observationTime;
  bool doPrintObservationTime;
};

template <class T, class Time> 
WithObservationTime<T, Time>::WithObservationTime()
  : value(),
    observationTime(),
    doPrintObservationTime(true)
{
}

template <class T, class Time> 
WithObservationTime<T, Time>::WithObservationTime(const T& val,
						  const Time& obs_time,
						  const bool print_obs_time)
  : value(val),
    observationTime(obs_time),
    doPrintObservationTime(print_obs_time)
{
}

template <class T, class Time> 
WithObservationTime<T, Time>::WithObservationTime(const WithObservationTime<T, Time>& op)
  : value(op.value),
    observationTime(op.observationTime),
    doPrintObservationTime(op.doPrintObservationTime)
{
}

template <class T, class Time>
inline T WithObservationTime<T, Time>::getValue() const
{
  return value;
}

template <class T, class Time> 
inline Time WithObservationTime<T, Time>::getObservationTime() const
{
  return observationTime;
}

template <class T, class Time> 
inline void WithObservationTime<T, Time>::setObservationTime(const Time& obs_time)
{
  observationTime = obs_time;
}

template <class T, class Time> 
inline T WithObservationTime<T, Time>::getCyclesSince(const Time& cur_time) const
{
  return observationTime - cur_time;
}

template <class T, class Time>
inline bool WithObservationTime<T, Time>::printObservationTime() const
{
  return doPrintObservationTime;
}

template <class T, class Time>
inline void WithObservationTime<T, Time>::setPrintObservationTime(const bool print_obs_time)
{
  doPrintObservationTime = print_obs_time;
}

template <class T, class Time>
WithObservationTime<T, Time>& WithObservationTime<T, Time>::operator= (const WithObservationTime<T, Time>& op)
{
  if ( &op != this ) {
    value = op.value;
    observationTime = op.observationTime;
    doPrintObservationTime = op.doPrintObservationTime;
  }

  return *this;
}

template <class T, class Time>
WithObservationTime<T, Time>& WithObservationTime<T, Time>::operator= (const T& op)
{
  value = op;
  observationTime = Time();
  doPrintObservationTime = true;

  return *this;
}

template <class T, class Time>
inline WithObservationTime<T, Time>& WithObservationTime<T, Time>::operator++ ()
{
  ++value;
  return *this;
}

template <class T, class Time>
inline WithObservationTime<T, Time>& WithObservationTime<T, Time>::operator++ (int)
{
  return ++(*this);
}

template <class T, class Time>
inline WithObservationTime<T, Time>& WithObservationTime<T, Time>::operator-- ()
{
  --value;
  return *this;
}

template <class T, class Time>
inline WithObservationTime<T, Time>& WithObservationTime<T, Time>::operator-- (int)
{
  return --(*this);
}

template <class T, class Time>
WithObservationTime<T, Time>& WithObservationTime<T, Time>::operator+= (const WithObservationTime<T, Time>& op)
{
  value += op.value;
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  observationTime = Min(observationTime, op.observationTime);
#else
  observationTime = Max(observationTime, op.observationTime);
#endif
  return *this;
}

template <class T, class Time>
WithObservationTime<T, Time>& WithObservationTime<T, Time>::operator-= (const WithObservationTime<T, Time>& op)
{
  value -= op.value;
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  observationTime = Min(observationTime, op.observationTime);
#else
  observationTime = Max(observationTime, op.observationTime);
#endif
  return *this;
}

template <class T, class Time>
WithObservationTime<T, Time>& WithObservationTime<T, Time>::operator*= (const WithObservationTime<T, Time>& op)
{
  value *= op.value;
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  observationTime = Min(observationTime, op.observationTime);
#else
  observationTime = Max(observationTime, op.observationTime);
#endif
  return *this;
}

template <class T, class Time>
WithObservationTime<T, Time>& WithObservationTime<T, Time>::operator/= (const WithObservationTime<T, Time>& op)
{
  value /= op.value;
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  observationTime = Min(observationTime, op.observationTime);
#else
  observationTime = Max(observationTime, op.observationTime);
#endif
  return *this;
}

template<class T, class Time>
inline WithObservationTime<T, Time> WithObservationTime<T, Time>::operator- () const
{
  return WithObservationTime<T, Time>(-value, observationTime);
}

template<class T, class Time>
inline WithObservationTime<T, Time> WithObservationTime<T, Time>::operator+ (const WithObservationTime<T, Time>& op) const
{
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  return WithObservationTime<T, Time>(value + op.value,
				Min(observationTime, op.observationTime));
#else
  return WithObservationTime<T, Time>(value + op.value,
				Max(observationTime, op.observationTime));
#endif
}

template<class T, class Time>
inline WithObservationTime<T, Time> WithObservationTime<T, Time>::operator- (const WithObservationTime<T, Time>& op) const
{
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  return WithObservationTime<T, Time>(value - op.value,
				Min(observationTime, op.observationTime));
#else
  return WithObservationTime<T, Time>(value - op.value,
				Max(observationTime, op.observationTime));
#endif
}

template<class T, class Time>
inline WithObservationTime<T, Time> WithObservationTime<T, Time>::operator* (const WithObservationTime<T, Time>& op) const
{
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  return WithObservationTime<T, Time>(value * op.value,
				Min(observationTime, op.observationTime));
#else
  return WithObservationTime<T, Time>(value * op.value,
				Max(observationTime, op.observationTime));
#endif
}

template<class T, class Time>
inline WithObservationTime<T, Time> WithObservationTime<T, Time>::operator/ (const WithObservationTime<T, Time>& op) const
{
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  return WithObservationTime<T, Time>(value / op.value,
				Min(observationTime, op.observationTime));
#else
  return WithObservationTime<T, Time>(value / op.value,
				Max(observationTime, op.observationTime));
#endif
}

template<class T, class Time>
inline bool WithObservationTime<T, Time>::operator!() const
{
  return ( !value );
}

template<class T, class Time>
inline bool WithObservationTime<T, Time>::operator==(const WithObservationTime<T, Time>& op) const
{
  return ( value == op.value );
}

template<class T, class Time>
inline bool WithObservationTime<T, Time>::operator!=(const WithObservationTime<T, Time>& op) const
{
  return !( *this == op );
}

template<class T, class Time>
inline bool WithObservationTime<T, Time>::operator<(const WithObservationTime<T, Time>& op) const
{
  return ( value < op.value );
}

template<class T, class Time>
inline bool WithObservationTime<T, Time>::operator<=(const WithObservationTime<T, Time>& op) const
{
  return !( *this > op );
}

template<class T, class Time>
inline bool WithObservationTime<T, Time>::operator>(const WithObservationTime<T, Time>& op) const
{
  return ( value > op.value );
}

template<class T, class Time>
inline bool WithObservationTime<T, Time>::operator>=(const WithObservationTime<T, Time>& op) const
{
  return !( *this < op );
}

template<class T, class Time>
inline WithObservationTime<T, Time> operator+ (const T& op1, const WithObservationTime<T, Time>& op2)
{
  return op2 + op1;
}

template<class T, class Time>
inline WithObservationTime<T, Time> operator- (const T& op1, const WithObservationTime<T, Time>& op2)
{
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  return WithObservationTime<T, Time>(op1 - op2.getValue(),
				Min(Time(), op2.getObservationTime()));
#else
  return WithObservationTime<T, Time>(op1 - op2.getValue(),
				Max(Time(), op2.observationTime()));
#endif
}

template<class T, class Time>
inline WithObservationTime<T, Time> operator* (const T& op1, const WithObservationTime<T, Time>& op2)
{
  return op2 * op1;
}

template<class T, class Time>
inline WithObservationTime<T, Time> operator/ (const T& op1, const WithObservationTime<T, Time>& op2)
{
#ifdef RS_PESSIMISTIC_OBSERVATIONS
  return WithObservationTime<T, Time>(op1 / op2.getValue(),
				Min(Time(), op2.getObservationTime()));
#else
  return WithObservationTime<T, Time>(op1 / op2.getValue(),
				Max(Time(), op2.observationTime()));
#endif
}

template<class T, class Time>
inline bool operator==(const T& op1, const WithObservationTime<T, Time>& op2)
{
  return op2 == op1;
}

template<class T, class Time>
inline bool operator!=(const T& op1, const WithObservationTime<T, Time>& op2)
{
  return op2 != op1;
}

template<class T, class Time>
inline bool operator<(const T& op1, const WithObservationTime<T, Time>& op2)
{
  return op2 > op1;
}

template<class T, class Time>
inline bool operator<=(const T& op1, const WithObservationTime<T, Time>& op2)
{
  return op2 >= op1;
}

template<class T, class Time>
inline bool operator>(const T& op1, const WithObservationTime<T, Time>& op2)
{
  return op2 < op1;
}

template<class T, class Time>
inline bool operator>=(const T& op1, const WithObservationTime<T, Time>& op2)
{
  return op2 <= op1;
}

template<class T, class Time>
std::ostream& operator<<(std::ostream& os, const WithObservationTime<T, Time>& obj)
{
  if (obj.printObservationTime())
    return os << obj.getValue() << " observed at " << obj.getObservationTime();
  else
    return os << obj.getValue();
}

/*
template<class T, class Time>
inline std::istream& operator>>(std::istream& is, WithObservationTime<T, Time>& obj)
{
  // Should add something so that the user can enter an observation time too.
  return is >> obj.getValue();
}
*/

template <class T, class Time> 
const std::string toString(const WithObservationTime<T, Time>& obj)
{
  if (obj.printObservationTime())
    return toString(obj.getValue()) + " observed at " + toString(obj.getObservationTime());
  else
    return toString(obj.getValue());
}

template<class T, class Time>
WithObservationTime<T, Time> sqrt(const WithObservationTime<T, Time>& op)
{
  return WithObservationTime<T, Time>(sqrt(op.getValue()));
}

/*
#ifdef RS_TRIG_IN_DEG
#include "AngleDeg.h"

template<class Time>
WithObservationTime<WithUnknown<AngleDeg<T>>, Time >
ASin(const WithObservationTime<WithUnknown<WithValue<T>>, Time>& op)
{
  return WithObservationTime<WithUnknown< AngleDeg<T> >, Time>(ASin(WithUnknown<WithValue<T>>(op)));
}

template<class Time>
WithObservationTime<WithUnknown<AngleDeg<T>>, Time >
ACos(const WithObservationTime<WithUnknown<WithValue<T>>, Time>& op)
{
  return WithObservationTime<WithUnknown< AngleDeg<T> >, Time>(ACos(WithUnknown<WithValue<T>>(op)));
}

template<class Time>
WithObservationTime<WithUnknown< AngleDeg<T> >, Time>
ATan(const WithObservationTime<WithUnknown<WithValue<T>>, Time>& op)
{
  return WithObservationTime<WithUnknown< AngleDeg<T> >, Time>(ATan(WithUnknown<WithValue<T>>(op)));
}

template<class Time>
WithObservationTime<WithUnknown< AngleDeg<T> >, Time>
ATan2(const WithObservationTime<WithUnknown<WithValue<T>>, Time>& op1,
      const WithObservationTime<WithUnknown<WithValue<T>>, Time>& op2)
{
  return WithObservationTime<WithUnknown< AngleDeg<T> >, Time>(ATan2(WithUnknown<WithValue<T>>(op1), WithUnknown<WithValue<T>>(op2)));
}

#endif // RS_TRIG_IN_DEG == 1
*/

/*
#ifdef RS_TRIG_IN_DEG
#include "AngleDeg.h"

template<class T, class Time>
WithObservationTime< AngleDeg<T>, Time > ASin(const WithObservationTime<T, Time>& op)
{
  return WithObservationTime< AngleDeg<T>, Time>(ASin(op.getValue()));
}

template<class T, class Time>
WithObservationTime< AngleDeg<T>, Time > ACos(const WithObservationTime<T, Time>& op)
{
  return WithObservationTime< AngleDeg<T>, Time>(ACos(op.getValue()));
}

template<class T, class Time>
WithObservationTime< AngleDeg<T>, Time> ATan(const WithObservationTime<T, Time>& op)
{
  return WithObservationTime< AngleDeg<T>, Time>(ATan(op.getValue()));
}

template<class T, class Time>
WithObservationTime< AngleDeg<T>, Time> ATan2(const WithObservationTime<T, Time>& op1, const WithObservationTime<T, Time>& op2)
{
  return WithObservationTime< AngleDeg<T>, Time>(ATan2(T(op1), T(op2)));
}

#endif // RS_TRIG_IN_DEG

#ifdef RS_TRIG_IN_RAD
#include "AngleRadians.h"

template<class T, class Time>
WithObservationTime< AngleRadians<T>, Time> ASin(const WithObservationTime<T, Time>& op)
{
  return WithObservationTime< AngleRadians<T>, Time>(ASin(op.getValue()));
}

template<class T, class Time>
WithObservationTime< AngleRadians<T>, Time> ACos(const WithObservationTime<T, Time>& op)
{
  return WithObservationTime< AngleRadians<T>, Time>(ACos(op.getValue()));
}

template<class T, class Time>
WithObservationTime< AngleRadians<T>, Time> ATan(const WithObservationTime<T, Time>& op)
{
  return WithObservationTime< AngleRadians<T>, Time>(ATan(op.getValue()));
}

template<class T, class Time>
WithObservationTime< AngleRadians<T>, Time> ATan2(const WithObservationTime<T, Time>& op1, const WithObservationTime<T, Time>& op2)
{
  return WithObservationTime< AngleRadians<T>, Time>(ATan2(op1.getValue(), op2.getValue()));
}

#endif // RS_TRIG_IN_RAD
*/


RS_END_NAMESPACE


#endif // !WITHOBSERVATIONTIME_H
