/* -*- Mode: C++ -*-
 * $Id: WithUnknown.h,v 1.6 2000/10/11 18:40:09 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: WithUnknown.h
 * Unitname: Library
 * $Revision: 1.6 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/10/11 18:40:09 $
 *
 *
 * HISTORY
 *
 * $Log: WithUnknown.h,v $
 * Revision 1.6  2000/10/11 18:40:09  frehe
 * Tried to fix a segmentation fault in WithUnknown, but without success.
 *
 * Revision 1.5  2000/10/08 17:33:14  frehe
 * Remade the sqrt template to a single function and also added a test to see that the argument is positive.
 *
 * Revision 1.4  2000/09/03 20:05:52  frehe
 * Made minor changes.
 *
 * Revision 1.3  2000/09/03 14:23:20  frehe
 * Made minor changes to improve the documentation.
 *
 * Revision 1.2  2000/09/01 03:23:55  frehe
 * Made minor changes.
 *
 * Revision 1.1  2000/09/01 01:24:43  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 *
 */

/**
 * @file WithUnknown.h
 * Include file for the template WithUnknown.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.6 $
 * $Author: frehe $
 * $Date: 2000/10/11 18:40:09 $
 */


#ifndef WITHUNKNOWN_H
#define WITHUNKNOWN_H


#include <string>
#include <iostream>

#include "math_utils.h"
#include "UnknownValueException.h"
#include "rs_throw.h"


RS_BEGIN_NAMESPACE


/**
 * @class WithUnknown WithUnknown.h
 * Template class for representing basic types that can be unknown.
 * The template argument only defines what type should be annotated
 * with a possibility to be unknown.
 * The only demands on the template argument T is that the arithmetic
 * functions are defined on it.
 *
 * <!-- @todo -->
 *
 * @par Unit:
 *      Library
 * $Revision: 1.6 $
 * $Author: frehe $
 * $Date: 2000/10/11 18:40:09 $
 */
template <class T> class WithUnknown
{
public:
  /** The default constructor will set the value to unknown. */
  WithUnknown();
  WithUnknown(const T&);
  WithUnknown(const WithUnknown<T>&);

  WithUnknown<T>& makeUnknown();
  bool isKnown() const;
  bool isUnknown() const;
  T getValue() const throw(RSException);

  const std::string getTypeString() const;

  WithUnknown<T>& operator= (const T& op);
  WithUnknown<T>& operator= (const WithUnknown<T>& op);

  //@{
  /** @name Operators on WithUnknown<T>
    The method operator@ (const T& op) is not necessary
    but should make the program more efficient, since there
    is no need for a type conversion between T and WithUnknown<T>.
    The knowledge that op can't be unknown can also be used to
    increase the efficency. */
  WithUnknown<T>& operator++ ();
  WithUnknown<T>& operator++ (int);
  WithUnknown<T>& operator-- ();
  WithUnknown<T>& operator-- (int);
  WithUnknown<T>& operator+= (const WithUnknown<T>& op);
  //WithUnknown<T>& operator+= (const T& op);
  WithUnknown<T>& operator-= (const WithUnknown<T>& op);
  //WithUnknown<T>& operator-= (const T& op);
  WithUnknown<T>& operator*= (const WithUnknown<T>& op);
  //WithUnknown<T>& operator*= (const T& op);
  WithUnknown<T>& operator/= (const WithUnknown<T>& op);
  //WithUnknown<T>& operator/= (const T& op);
  //WithUnknown<T>& operator%= (const WithUnknown<T>& op);
  //WithUnknown<T>& operator%= (const T& op);

  WithUnknown<T> operator- () const;
  WithUnknown<T> operator+ (const WithUnknown<T>& op) const;
  //WithUnknown<T> operator+ (const T& op) const;
  WithUnknown<T> operator- (const WithUnknown<T>& op) const;
  //WithUnknown<T> operator- (const T& op) const;
  WithUnknown<T> operator* (const WithUnknown<T>& op) const;
  //WithUnknown<T> operator* (const T& op) const;
  WithUnknown<T> operator/ (const WithUnknown<T>& op) const;
  //WithUnknown<T> operator/ (const T& op) const;
  //WithUnknown<T> operator% (const WithUnknown<T>& op) const;
  //WithUnknown<T> operator% (const T& op) const;
  //@}
  

  //@{
  /** @name Comparision operators on WithUnknown<T>
    !obj == true iff obj is unknown or !obj.value
    unknown is the largest number so:
    obj1 > obj2 iff obj1 is unknown and obj2 is known or
                    obj1.value > obj2.value */
  bool operator!  () const;
  bool operator== (const WithUnknown<T>& op) const;
  //bool operator== (const T& op) const;
  bool operator!= (const WithUnknown<T>& op) const;
  //bool operator!= (const T& op) const;
  bool operator<  (const WithUnknown<T>& op) const;
  //bool operator<  (const T& op) const;
  bool operator<= (const WithUnknown<T>& op) const;
  //bool operator<= (const T& op) const;
  bool operator>  (const WithUnknown<T>& op) const;
  //bool operator>  (const T& op) const;
  bool operator>= (const WithUnknown<T>& op) const;
  //bool operator>= (const T& op) const;
  //@}

  /** A string describing the base type. It is used
    when outputting an unknown value of this type. */
  static const std::string typeString;

private:
  T value;
  bool unknown;
};

template <class T> 
WithUnknown<T>::WithUnknown()
  : value(),
    unknown(true)
{
}

template <class T> 
WithUnknown<T>::WithUnknown(const T& val)
  : value(val),
    unknown(false)
{
}

template <class T> 
WithUnknown<T>::WithUnknown(const WithUnknown<T>& op)
  : value(op.value),
    unknown(op.unknown)
{
}

template <class T> 
inline WithUnknown<T>& WithUnknown<T>::makeUnknown()
{
  unknown = true;
  return *this;
}

template <class T> 
inline bool WithUnknown<T>::isUnknown() const
{
  return unknown;
}

template <class T> 
inline bool WithUnknown<T>::isKnown() const
{
  return !unknown;
}

template <class T> 
inline T WithUnknown<T>::getValue() const throw(RSException)
{
  if ( unknown ) {
    RS_THROW(UnknownValueException,
	     "Tried to take the value of an unknown " + WithUnknown<T>::typeString);
  return T(); // numeric_limits<T>::max() - 1;
  } else
    return value;
}

template <class T>
const std::string WithUnknown<T>::getTypeString() const
{
  return typeString;
}

template <class T>
WithUnknown<T>& WithUnknown<T>::operator= (const WithUnknown<T>& op)
{
  if ( &op != this ) {
    value = op.value;
    unknown = op.unknown;
  }

  return *this;
}

template <class T>
WithUnknown<T>& WithUnknown<T>::operator= (const T& val) 
{ 
  value = val;
  unknown = false;

  return *this;
}

template <class T>
WithUnknown<T>& WithUnknown<T>::operator++ ()
{
  if ( !unknown )
    ++value;

  return *this;
}

template <class T>
inline WithUnknown<T>& WithUnknown<T>::operator++ (int)
{
  return ++(*this);
}

template <class T>
WithUnknown<T>& WithUnknown<T>::operator-- ()
{
  if ( !unknown )
    --value;

  return *this;
}

template <class T>
inline WithUnknown<T>& WithUnknown<T>::operator-- (int)
{
  return --(*this);
}

template <class T>
WithUnknown<T>& WithUnknown<T>::operator+= (const WithUnknown<T>& op)
{
  if ( !unknown &&
       !op.unknown )
    value += op.value;

  return *this;
}

template <class T>
WithUnknown<T>& WithUnknown<T>::operator-= (const WithUnknown<T>& op)
{
  if ( !unknown &&
       !op.unknown )
    value -= op.value;

  return *this;
}

template <class T>
WithUnknown<T>& WithUnknown<T>::operator*= (const WithUnknown<T>& op)
{
  if ( !unknown &&
       !op.unknown )
    value *= op.value;

  return *this;
}

template <class T>
WithUnknown<T>& WithUnknown<T>::operator/= (const WithUnknown<T>& op)
{
  if ( !unknown &&
       !op.unknown )
    value /= op.value;

  return *this;
}

/*
template <class T>
WithUnknown<T>& WithUnknown<T>::operator+= (const T& op)
{
  if ( !unknown )
    value += op;

  return *this;
}

template <class T>
WithUnknown<T>& WithUnknown<T>::operator-= (const T& op)
{
  if ( !unknown )
    value -= op;

  return *this;
}

template <class T>
WithUnknown<T>& WithUnknown<T>::operator*= (const T& op)
{
  if ( !unknown )
    value *= op;

  return *this;
}

template <class T>
WithUnknown<T>& WithUnknown<T>::operator/= (const T& op)
{
  if ( !unknown )
    value /= op;

  return *this;
}
*/

template<class T>
WithUnknown<T> WithUnknown<T>::operator- () const
{
  if ( unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(-value);
}

template<class T>
WithUnknown<T> WithUnknown<T>::operator+ (const WithUnknown<T>& op) const
{
  if ( unknown ||
       op.unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(value + op.value);
}

template<class T>
WithUnknown<T> WithUnknown<T>::operator- (const WithUnknown<T>& op) const
{
  if ( unknown ||
       op.unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(value - op.value);
}

template<class T>
WithUnknown<T> WithUnknown<T>::operator* (const WithUnknown<T>& op) const
{
  if ( unknown ||
       op.unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(value * op.value);
}

template<class T>
WithUnknown<T> WithUnknown<T>::operator/ (const WithUnknown<T>& op) const
{
  if ( unknown ||
       op.unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(value / op.value);
}

/*
template<class T>
WithUnknown<T> WithUnknown<T>::operator+ (const T& op) const
{
  if ( unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(value + op);
}

template<class T>
WithUnknown<T> WithUnknown<T>::operator- (const T& op) const
{
  if ( unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(value - op);
}

template<class T>
WithUnknown<T> WithUnknown<T>::operator* (const T& op) const
{
  if ( unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(value * op);
}

template<class T>
WithUnknown<T> WithUnknown<T>::operator/ (const T& op) const
{
  if ( unknown )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(value / op);
}
*/

template<class T>
bool WithUnknown<T>::operator!() const
{
  if ( unknown )
    return true;
  else
    return ( !value );
}

template<class T>
bool WithUnknown<T>::operator==(const WithUnknown<T>& op) const
{
  if ( unknown && op.unknown )
    return true;
  else if ( !unknown && !op.unknown )
    return ( value == op.value );
  else
    return false;
}

template<class T>
bool WithUnknown<T>::operator!=(const WithUnknown<T>& op) const
{
  return !( *this == op );
}

template<class T>
bool WithUnknown<T>::operator<(const WithUnknown<T>& op) const
{
  if ( !unknown && !op.unknown )
    return ( value < op.value );
  else if ( !unknown && op.unknown )
    return true;
  else
    return false;
}

template<class T>
inline bool WithUnknown<T>::operator<=(const WithUnknown<T>& op) const
{
  return !( *this > op );
}

template<class T>
bool WithUnknown<T>::operator>(const WithUnknown<T>& op) const
{
  if ( !unknown && !op.unknown )
    return ( value > op.value );
  else if ( unknown && !op.unknown )
    return true;
  else
    return false;
}

template<class T>
bool WithUnknown<T>::operator>=(const WithUnknown<T>& op) const
{
  return !( *this < op );
}

/*
template<class T>
bool WithUnknown<T>::operator==(const T& op) const
{
  if ( !unknown )
    return ( value == op );
  else
    return false;
}

template<class T>
inline bool WithUnknown<T>::operator!=(const T& op) const
{
  return !( *this == op );
}

template<class T>
bool WithUnknown<T>::operator<(const T& op) const
{
  if ( !unknown )
    return ( value < op );
  else
    return false;
}

template<class T>
inline bool WithUnknown<T>::operator<=(const T& op) const
{
  return !( *this > op );
}

template<class T>
bool WithUnknown<T>::operator>(const T& op) const
{
  if ( !unknown )
    return ( value > op );
  else
    return true;
}

template<class T>
inline bool WithUnknown<T>::operator>=(const T& op) const
{
  return !( *this < op );
}
*/

template<class T>
inline WithUnknown<T> operator- (const T& op1,
				 const WithUnknown<T>& op2)
{
  if ( op2.isUnknown() )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(op1 - op2.getValue());
}

template<class T>
inline WithUnknown<T> operator+ (const T& op1,
				 const WithUnknown<T>& op2)
{
  return op2 + op1;
}

template<class T>
inline WithUnknown<T> operator* (const T& op1,
				 const WithUnknown<T>& op2)
{
  return op2 * op1;
}

template<class T>
WithUnknown<T> operator/ (const T& op1, const WithUnknown<T>& op2)
{
  if ( op2.isUnknown() )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(op1.getValue() / op2.getValue());
}

template<class T>
inline bool operator==(const T& op1, const WithUnknown<T>& op2)
{
  return op2 == op1;
}

template<class T>
inline bool operator!=(const T& op1, const WithUnknown<T>& op2)
{
  return op2 != op1;
}

template<class T>
inline bool operator<(const T& op1, const WithUnknown<T>& op2)
{
  return op2 > op1;
}

template<class T>
inline bool operator<=(const T& op1, const WithUnknown<T>& op2)
{
  return op2 >= op1;
}

template<class T>
inline bool operator>(const T& op1, const WithUnknown<T>& op2)
{
  return op2 < op1;
}

template<class T>
inline bool operator>=(const T& op1, const WithUnknown<T>& op2)
{
  return op2 <= op1;
}

template<class T>
std::ostream& operator<<(std::ostream& os, const WithUnknown<T>& v)
{
  if ( v.isUnknown() )
    return os << "UNKNOWN"; //"(" << v.getTypeString() << ")";
  else
    return os << v.getValue();
}

/*
template<class T>
inline std::istream& operator>>(std::istream& is, WithUnknown<T>& v)
{
  // Should add something so that the user can enter an unknown value too.
  return is >> value;
}
*/

template<class T>
const std::string toString(const WithUnknown<T>& obj)
{
  if ( obj.isUnknown() )
    return "UNKNOWN(" + obj.getTypeString() + ")";
  else
    return toString(obj.getValue());
}

template<class T>
WithUnknown<T> min(const WithUnknown<T>& op1,
		   const WithUnknown<T>& op2)
{
  if ( op1.isUnknown() ||
       op2.isUnknown() )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(min(op1.getValue(), op2.getValue()));
}

template<class T>
WithUnknown<T> max(const WithUnknown<T>& op1,
		   const WithUnknown<T>& op2)
{
  if ( op1.isUnknown() ||
       op2.isUnknown() )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(max(op1.getValue(), op2.getValue()));
}

template<class T>
WithUnknown<T> sqrt(const WithUnknown<T>& op)
{
  if ( op.isUnknown() )
    return WithUnknown<T>();
  else
    return WithUnknown<T>(sqrt(op.getValue()));
}


RS_END_NAMESPACE


#endif // !WITHUNKNOWN_H
