/* -*- Mode: C++ -*-
 * $Id: library_compiler_directives.h,v 1.2 2000/09/03 13:01:05 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: library_compiler_directives.h
 * Unitname: Library
 * $Revision: 1.2 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/03 13:01:05 $
 *
 *
 * HISTORY
 *
 * $Log: library_compiler_directives.h,v $
 * Revision 1.2  2000/09/03 13:01:05  frehe
 * Added guards for eiffel-style design by contract checks RS_USE_INVARIANT, RS_USE_REQUIRE, and RS_USE_ENSURE.
 *
 * Revision 1.1.1.1  2000/08/31 13:26:03  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file library_compiler_directives.h
 * Include file with the compiler directives for the RoboSoc library.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.2 $
 * $Author: frehe $
 * $Date: 2000/09/03 13:01:05 $
 */


#ifndef LIBRARY_COMPILER_DIRECTIVES_H
#define LIBRARY_COMPILER_DIRECTIVES_H


#include "machine_dependent_compiler_directives.h"


/** @def RS_TRIG_IN_DEG
  If defined then trigonometric functions will return angles in degrees. */
#define RS_TRIG_IN_DEG
/** @def RS_TRIG_IN_RAD
 If defined then trigonometric functions will return angles in radians. */
#define RS_TRIG_IN_RAD

#ifndef RS_TRIG_IN_DEG
#  define RS_TRIG_IN_RAD
#else
#  undef RS_TRIG_IN_RAD
#endif


//#define CHECK_DIVISION_BY_ZERO 1
// The constants for DIVISION_BY_ZERO_RESULT is defined in types.h or
// the definition of the type/class.


/** @def RS_USE_ASSERT
  If defined then assertions will be check by the RS_ASSERT( cond, msg )
  macro. If the assertion fails then an action will be taken depending
  on the other compiler directives. The default is to call the external
  function rs_assert(const std::string& cond, const std::string& msg,
  const std::string& file, int line), which must be defined by the user
  if used.*/
#define RS_USE_ASSERT

/** @def RS_LOG_ASSERT
  If defined then the action taken when an assertion fails is to send
  a message to the LogStream area ASSERT. */
#define RS_LOG_ASSERT

/** @def RS_THROW_ASSERT
  If defined and RS_LOG_ASSERT is not defined then the action when an
  assertion fails is to throw and AssertionFailed( cond, msg ) exception. */
#define RS_THROW_ASSERT
#ifdef RS_LOG_ASSERT
#  undef RS_THROW_ASSERT
#endif


/** @def RS_USE_INVARIANT
  If defined then invariants will be check by the RS_INVARIANT( cond )
  and RS_CHECK_INVARIANT macros. They are implemented using RS_ASSERT,
  so the result if the test fails depends on its definition. */
#define RS_USE_INVARIANT

/** @def RS_USE_REQUIRE
  If defined then preconditions will be check by the RS_REQUIRE( cond )
  macro. It is implemented using RS_ASSERT, so the result if the test fails
  depends on its definition. */
#define RS_USE_REQUIRE

/** @def RS_USE_ENSURE
  If defined then postconditions will be check by the RS_ENSURE( cond )
  macro. It is implemented using RS_ASSERT, so the result if the test fails
  depends on its definition. */
#define RS_USE_ENSURE


/** @def RS_USE_EXCEPTION
  If defined then exceptions raised with the RS_THROW ( Exception, msg )
  macro will be thrown. Otherwise they will either be logged to the log
  stream, if RS_LOG_EXCEPTION is defined, or reported through the extern
  function void rs_throw(const std::string& exception, const std::string& msg,
  const std::string& file, int line), which will have to be defined by the
  user if used. */
#define RS_USE_EXCEPTION

/** @def RS_LOG_EXCEPTION
  If defined and RS_USE_EXCEPTION is not defined then exceptions
  that are thrown will instead be logged to the log stream. */
#define RS_LOG_EXCEPTION


/** @def RS_USE_ERROR
  If defined then RS_WARNING( str ), RS_ERROR( str ), and RS_FATAL( str )
  statements will be reported. The standard way of reporting the error
  information is through a external functions called rs_warning, rs_error,
  and rs_fatal, all with the signatur void rs_error(const std::string& str,
  const std::string& file, int line). Each functions has to be defined by
  the user if it is used.
  If not defined then no code is generated at all for RS_WARNING, RS_ERROR, or
  RS_FATAL statements. */
#define RS_USE_ERROR

/** If defined then errors that are reported will be sent to the log stream.
 If RS_LOG_ERROR is defined but RS_USE_LOG is not defined then all errors will
 be disregarded. */
#define RS_LOG_ERROR


/** @def RS_USE_DEBUG
  If defined then RS_DEBUG( str ) statements will be reported. The standard
  way of reporting the debug information is through a external function
  called void rs_debug(const std::string& str, const std::string& file, int line),
  which the user has to define if it is used.
  If not defined then no code is generated at all for RS_DEBUG statements. */
#define RS_USE_DEBUG

/** If defined then debug information will be sent to the log stream instead of
  to the rs_debug function.
  If RS_LOG_DEBUG is defined but RS_USE_LOG is not defined then all debug
  information will be disregarded. */
#define RS_LOG_DEBUG


/** @def RS_USE_LOG
  If defined then RS_LOG( area, str ) statements will be sent to the log stream.
  If not defined then no code is generated at all for RS_LOG statements. */
#define RS_USE_LOG


#endif // !LIBRARY_COMPILER_DIRECTIVES_H
