/* -*- Mode: C++ -*-
 * $Id: math_utils.h,v 1.6 2000/10/08 17:33:14 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: math_utils.h
 * Unitname: Library
 * $Revision: 1.6 $
 * Created by: Fredrik Heintz 2000-08-29
 * Last modified by $Author: frehe $ $Date: 2000/10/08 17:33:14 $
 *
 *
 * HISTORY
 *
 * $Log: math_utils.h,v $
 * Revision 1.6  2000/10/08 17:33:14  frehe
 * Remade the sqrt template to a single function and also added a test to see that the argument is positive.
 *
 * Revision 1.5  2000/09/18 15:20:07  frehe
 * Due to a problem with namespaces the name for atan2, atan, acos and asin is changed to ATan2, ATan, ASin and ACos.
 *
 * Revision 1.4  2000/09/05 21:08:19  frehe
 * Made sin, cos, tan, atan2, atan, acos, asin and sqrt call the standard function.
 *
 * Revision 1.3  2000/09/04 15:21:51  frehe
 * Changed the name of the trigonometric functions to the same as in the standard math library (sin, cos, tan, asin, acos, atan, and atan2).
 *
 * Revision 1.2  2000/09/01 01:24:44  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 * Revision 1.1.1.1  2000/08/31 13:26:04  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file math_utils.h
 * Include file for the RoboSoc math utilities.
 * This file contain some useful mathematical functions.
 * Many of them are already implemented by C++.
 * Some should be faster, like pow...
 *
 * @par Unit:
 *      Library
 * $Revision: 1.6 $
 * $Author: frehe $
 * $Date: 2000/10/08 17:33:14 $
 */


#ifndef MATH_UTILS_H
#define MATH_UTILS_H


#include <cmath>

#include "machine_dependent_types.h"
#include "library_compiler_directives.h"


RS_BEGIN_NAMESPACE


/** Constant representing Pi. */
static const Float64 PI = 3.14159265358979323846;

/** Constant representing 2*Pi. */
static const Float64 TWO_PI = PI + PI;


/** Constant used to convert between degrees and radians fast. */
static const Float64 DEG2RAD = PI / 180.0;
/** Constant used to convert between radians and degrees fast. */
static const Float64 RAD2DEG = 180.0 / PI;


/** Constant representing the epsilon of Float32. */
static const Float32 FLOAT32EPSILON = 0.0001;

/** Constant representing the epsilon of Float32. */
static const Float64 FLOAT64EPSILON = 0.00001;


/** Converts an angle in degrees to an angle in radians. */
Float64 deg2Rad(const Float64 deg);

/** Converts an angle in radians to an angle in degrees. */
Float64 rad2Deg(const Float64 rad);


/** Template for returning the maximum of two values. */
template<class T>
inline T max(const T& v1, const T& v2)
{ return v1 > v2 ? v1 : v2; }  

/** Template for returning the minimum of two values. */
template<class T>
inline T min(const T& v1, const T& v2)
{ return v1 < v2 ? v1 : v2; }

/* Template for taking min(max(@a min_val, @a val), @a max_val),
   which is the same as returning the value between @a min_val
   and @a max_val that is closest to @a val. */
template<class T>
inline T minMax(const T& min_val, const T& val, const T& max_val)
{ return min(max(min_val, val), max_val); }


/** Template for calculating the square of @a v. */
template<class T>
inline T sqr(const T& v) 
{ return v * v; }


/** Template for calculating the absolute value of @a v. */
template<class T>
inline T abs(const T& v)
{ return ((v < T(0)) ? -v : v); }

/** Template for returning the sign of @a v. If it returns -1
  then @a v is negative, if it returns 1 then @a v is positive. */
template<class T>
inline Int32 sign(const T& v) 
{ return ((v < T(0)) ? -1 : 1); }


/** Function returning the integer part of division of two floats. */
Int32 div(const Float64 d1, const Float64 d2);

/** Function returning the integer part of division of two integers. */
Int32 div(const Int32 i1, const Int32 i2);


/** Function for calculating the signed modulus of two
  floating point numbers. */
Float64 sMod(const Float64 d1, const Float64 d2);

/** Function for calculating the signed modulus of two integers. */
Int32 sMod(const Int32 i1, const Int32 i2);

/** Function for calculating the unsigned modulus of two
  floating point numbers. */
Float64 mod(const Float64 d1, const Float64 d2);

/** Function for calculating the unsigned modulus of two integers. */
Int32 mod(const Int32 i1, const Int32 i2);


/** Calculate the square root of @a v. */
Float64 sqrt(const Float64& v);


/** Return x to the power of y.
  It works but is not very efficient. */
Int32 power(const Int32 x, const UInt32 y);

/** Return x to the power of y.
  It works but is not very efficient. */
Float64 power(const Float64 x, const UInt32 y);

/** Return x to the power of y.
  It works but is not very efficient. */
Float64 power(const Float64 x, const Float64 y);


/** Calculates the arcsin of @a op. If RS_TRIG_IN_DEG is defined then
 the return value is in degrees, otherwise in radians. */
//Float64 asin(const Float64 op);
/** Calculates the arccos of @a op. If RS_TRIG_IN_DEG is defined then
 the return value is in degrees, otherwise in radians. */
//Float64 acos(const Float64 op);
/** Calculates the arctan of @a op. If RS_TRIG_IN_DEG is defined then
 the return value is in degrees, otherwise in radians. */
//Float64 atan(const Float64 op);
/** Calculates the arctan of @a op1/@a op2. If RS_TRIG_IN_DEG is defined then
 the return value is in degrees, otherwise in radians. */
//Float64 atan2(const Float64 op1, const Float64 op2);

/** Calculates the sin of @a op. If RS_TRIG_IN_DEG is defined then
 op is assumed to be in degrees, otherwise in radians. */
Float64 sin(const Float64 op);
/** Calculates the cos of @a op. If RS_TRIG_IN_DEG is defined then
 op is assumed to be in degrees, otherwise in radians. */
Float64 cos(const Float64 op);
/** Calculates the tan of @a op. If RS_TRIG_IN_DEG is defined then
 op is assumed to be in degrees, otherwise in radians. */
Float64 tan(const Float64 op);


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "math_utils.icc"
#endif


#endif // !MATH_UTILS_H
