/* -*- Mode: C++ -*-
 * $Id: rs_assert.h,v 1.5 2000/09/05 21:09:24 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: rs_assert.h
 * Unitname: Library
 * $Revision: 1.5 $
 * Created by: Fredrik Heintz 2000-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/05 21:09:24 $
 *
 *
 * HISTORY
 *
 * $Log: rs_assert.h,v $
 * Revision 1.5  2000/09/05 21:09:24  frehe
 * Fixed a bug in the assert macro.
 *
 * Revision 1.4  2000/09/05 11:07:41  frehe
 * Changed the log area constants from NAME to LA_NAME.
 *
 * Revision 1.3  2000/09/04 15:21:11  frehe
 * Made minor changes.
 *
 * Revision 1.2  2000/09/03 13:00:22  frehe
 * Added eiffel-style design by contract checks, RS_CHECK_INVARIANT, RS_INVARIANT, RS_REQUIRE, and RS_ENSURE.
 *
 * Revision 1.1.1.1  2000/08/31 13:26:04  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file rs_assert.h
 * Include file for the RoboSoc RS_ASSERT assertion macro.
 *
 * @todo Add a macro RS_ASSERT_AREA that will assert only
 *       a certain area and also send its error messages to that area.
 * @par Unit:
 *      Library
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2000/09/05 21:09:24 $
 */


#ifndef RS_ASSERT_H
#define RS_ASSERT_H


#include "library_compiler_directives.h"


/** @def RS_ASSERT( cond, msg )
  If RS_USE_ASSERT is defined then RS_ASSERT( cond, msg ) will assert that
  @a cond is true. If it is false then an action will be taken with the
  message @a msg.
  The action taken depends on other compiler directives. The standard action
  is to call the external function void rs_assert(const std::string& cond,
  const std::string& msg, const std::string& file, int line), which must be
  defined by the user if used.
  If RS_LOG_ASSERT is defined then a RS_LOG( ASSERT, cond + msg ) logging
  message will be generated.
  Else if RS_THROW_ASSERT is defined then an AssertionFailed(cond, msg)
  exception will be thrown.
  If RS_USE_ASSERT is not defined then nothing is produced, use this if you
  need to produce fast code. */
#ifdef RS_USE_ASSERT

#  ifdef RS_LOG_ASSERT

#    include "rs_log.h"
#    define RS_ASSERT( cond, msg ) \
       if ( !(cond) ) { \
			RS_LOG( RS_NAMESPACE::LA_ASSERT, "Assertion " \
				<< #cond << " failed: " << msg ); \
		    } else {}

#  elif defined(RS_THROW_ASSERT)

#    include "rs_throw.h"
#    include "AssertionFailed.h"
#    define RS_ASSERT( cond, msg ) \
       if ( !(cond) ) { \
			RS_THROW( RS_NAMESPACE::AssertionFailed, \
				  "Assertion" #cond " failed: " + msg ); \
		    } else {}

#  else // !RS_LOG_ASSERT && !RS_THROW_ASSERT

#    include <string>
     RS_BEGIN_NAMESPACE
     extern void rs_assert(const std::string& cond,
			   const std::string& msg,
			   const std::string& file,
			   int line);
     RS_END_NAMESPACE
#    define RS_ASSERT( cond, msg ) \
       if ( !(cond) ) { \
			RS_NAMESPACE::rs_assert(#cond, msg, \
			__FILE__, __LINE__); \
		    } else {}
     
#  endif // RS_LOG_ASSERT

#else // !RS_USE_ASSERT

#  define RS_ASSERT( cond, msg )

#endif // RS_USE_ASSERT


/** @def RS_CHECK_INVARIANT
  If RS_USE_INVARIANT is defined then RS_CHECK_INVARIANT will check
  the class invariant is satisfied by calling the method bool invariant(),
  which has to be defined for the class. The test is made by the RS_ASSERT
  macro. So all assertions can be turned off by not defining RS_USE_ASSERT.
  The idea for this macro comes from the Nana library and Eiffel. */
/** @def RS_INVARIANT( cond )
  If RS_USE_INVARIANT is defined then RS_INVARIANT( cond ) will check
  invariant @a cond. The test is made by the RS_ASSERT macro. So all
  assertions can be turned off by not defining RS_USE_ASSERT.
  The idea for this macro comes from the Nana library and Eiffel. */
#ifdef RS_USE_INVARIANT

#  define RS_CHECK_INVARIANT RS_ASSERT( invariant(), "Invariant check failed!" )
#  define RS_INVARIANT( cond ) RS_ASSERT( cond, "Invariant check failed!" )
     
#else

#  define RS_CHECK_INVARIANT
#  define RS_INVARIANT
     
#endif

     
/** @def RS_REQUIRE( cond )
  If RS_USE_REQUIRE is defined then RS_REQUIRE( cond ) will check
  the precondition @a cond. The test is made by the RS_ASSERT macro. So
  all assertions can be turned off by not defining RS_USE_ASSERT.
  The idea for this macro comes from the Nana library and Eiffel. */
#ifdef RS_USE_REQUIRE

#  define RS_REQUIRE( cond ) RS_ASSERT( cond, "Precondition check failed!" )
     
#else

#  define RS_REQUIRE
     
#endif

     
/** @def RS_ENSURE( cond )
  If RS_USE_ENSURE is defined then RS_ENSURE( cond ) will check
  the postcondition @a cond. The test is made by the RS_ASSERT macro. So
  all assertions can be turned off by not defining RS_USE_ASSERT.
  The idea for this macro comes from the Nana library and Eiffel. */
#ifdef RS_USE_ENSURE

#  define RS_ENSURE( cond ) RS_ASSERT( cond, "Postcondition check failed!" )
     
#else

#  define RS_ENSURE
     
#endif


#endif // !RS_ASSERT_H
