/* -*- Mode: C++ -*-
 * $Id: LogStream.cc,v 1.3 2001/06/19 14:01:50 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: LogStream.cc
 * Unitname: Library
 * $Revision: 1.3 $
 * Created by: Fredrik Heintz 2000-xx-xx
 * Last modified by $Author: frehe $ $Date: 2001/06/19 14:01:50 $
 *
 *
 * HISTORY
 *
 * $Log: LogStream.cc,v $
 * Revision 1.3  2001/06/19 14:01:50  frehe
 * Made minor changes to compile RoboSoc with gcc-3.0
 *
 * Revision 1.2  2000/09/05 11:08:16  frehe
 * Changed the log area constants from NAME to LA_NAME.
 *
 * Revision 1.1.1.1  2000/08/31 13:26:03  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file LogStream.cc
 * Implementation of the class LogStream.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.3 $
 * $Author: frehe $
 * $Date: 2001/06/19 14:01:50 $
 */


//#include <cstdio>
//#include <cstdarg>
//#include <string>

#include "LogStream.h"


RS_BEGIN_NAMESPACE


LogStream* LogStream::onlyInstance = NULL;


LogStream::LogStream(const bool do_file_log, std::ostream* file_os,
		     const bool do_screen_log, std::ostream* screen_os)
  : doFileLog(do_file_log),
    fileLogOut(file_os),
    doScreenLog(do_screen_log),
    screenLogOut(screen_os),
    currentArea(LA_ROBOSOC),
    temporaryAreas(),
    logAreas()
{
  // Create the top node in the log area tree
  logAreas[LA_ROBOSOC] = LogArea(LA_TOP, doFileLog, fileLogOut,
				 doScreenLog, screenLogOut);
}


LogStream::~LogStream()
{
}


LogStream* LogStream::instance()
{
  if ( onlyInstance == NULL )
    onlyInstance = new LogStream;

  return onlyInstance;
}

LogStream& LogStream::setCurrentArea(const LogAreaName& area)
{
  currentArea = area;
  if ( logAreas.find(currentArea) == logAreas.end() ) {
    doFileLog = logAreas[LA_ROBOSOC].logFile();
    fileLogOut = logAreas[LA_ROBOSOC].getFileLogStream();
    doScreenLog = logAreas[LA_ROBOSOC].logScreen();
    screenLogOut = logAreas[LA_ROBOSOC].getScreenLogStream();
  } else {
    // Find out if the current area should be printed or
    // not. By setting the areas parent do_log to false
    // non of its children will be printed.
    LogAreaName tmp_area = area;
    doFileLog = true;
    while ( doFileLog && tmp_area != LA_TOP ) {
      doFileLog = logAreas[tmp_area].logFile();
      tmp_area = logAreas[tmp_area].getParentArea();
    }

    // Find out if the current area should be printed or
    // not. By setting the areas parent do_log to false
    // non of its children will be printed.
    tmp_area = area;
    doScreenLog = true;
    while ( doScreenLog && tmp_area != LA_TOP ) {
      doScreenLog = logAreas[tmp_area].logScreen();
      tmp_area = logAreas[tmp_area].getParentArea();
    }

    // Find the ostream that should be used.
    // First look in the current area, if its log stream
    // is NULL try its parent and so on.
    tmp_area = area;
    fileLogOut = NULL;
    while ( fileLogOut == NULL && tmp_area != LA_TOP ) {
      fileLogOut = logAreas[tmp_area].getFileLogStream();
      tmp_area = logAreas[tmp_area].getParentArea();
    }

    // Find the ostream that should be used.
    // First look in the current area, if its log stream
    // is NULL try its parent and so on.
    tmp_area = area;
    screenLogOut = NULL;
    while ( screenLogOut == NULL && tmp_area != LA_TOP ) {
      screenLogOut = logAreas[tmp_area].getScreenLogStream();
      tmp_area = logAreas[tmp_area].getParentArea();
    }
  }
  
  return *this;
}


LogStream& LogStream::setTemporaryArea(const LogAreaName& area)
{
  temporaryAreas.push(currentArea);
  setCurrentArea(area);
  return *this;
}


LogStream& LogStream::resetTemporaryArea()
{
  setCurrentArea(temporaryAreas.top());
  temporaryAreas.pop();
  return *this;
}


LogStream& LogStream::stopLogFileArea(const LogAreaName& area)
{
  logAreas[area].setLogFile(false);
  if ( area == currentArea )
    setCurrentArea(area);

  return *this;
}


LogStream& LogStream::startLogFileArea(const LogAreaName& area)
{
  logAreas[area].setLogFile(true);
  if ( area == currentArea )
    setCurrentArea(area);

  return *this;
}


LogStream& LogStream::stopLogScreenArea(const LogAreaName& area)
{
  logAreas[area].setLogScreen(false);
  if ( area == currentArea )
    setCurrentArea(area);

  return *this;
}


LogStream& LogStream::startLogScreenArea(const LogAreaName& area)
{
  logAreas[area].setLogScreen(true);
  if ( area == currentArea )
    setCurrentArea(area);

  return *this;
}


void LogStream::setAreaFileStream(const LogAreaName& area, std::ostream* os)
{
  logAreas[area].setFileLogStream(os);
  if ( currentArea == area )
    setCurrentArea(area);
}


std::ostream* LogStream::getAreaFileStream(const LogAreaName& area)
{
  return logAreas[area].getFileLogStream();
}


void LogStream::setAreaScreenStream(const LogAreaName& area, std::ostream* os)
{
  logAreas[area].setScreenLogStream(os);
  if ( currentArea == area )
    setCurrentArea(area);
}


std::ostream* LogStream::getAreaScreenStream(const LogAreaName& area)
{
  return logAreas[area].getScreenLogStream();
}


void LogStream::addLogArea(const LogAreaName& area,
			   const LogAreaName& parent_area,
			   const bool do_file_log,
			   std::ostream* log_file_stream,
			   const bool do_screen_log,
			   std::ostream* log_screen_stream)
{
  logAreas[area] = LogArea(parent_area, do_file_log, log_file_stream,
			   do_screen_log, log_screen_stream);
  if ( currentArea == area )
    setCurrentArea(area);
}


LogStream& LogStream::operator<<(std::ostream& (*fn)(std::ostream&))
{
  if ( doFileLog && fileLogOut != NULL) {
    fn(*fileLogOut);
  }
  if ( doScreenLog && screenLogOut != NULL) {
    fn(*screenLogOut);
  }
  return *this;
}


/*
int LogStream::Printf(char const *fmt, ...) const
{
  static char buf[1024];
  int ret;
  va_list ap;

  va_start(ap, fmt);
  ret = vsnprintf(buf, 1024, fmt, ap);
  va_end(ap);

  if(ret == 0)
    return ret;

  if (do_log) {
    log_out->write(buf, ret);
    log_out->flush();
  }
  if (do_log) {
    log_out->write(buf, ret);
  }

  return 0;
}
*/


#ifndef DOXYGEN_SHOULD_SKIP_THIS

LogStream& _SetCurrentArea(LogStream& ls, const LogAreaName& area)
{
  return ls.setCurrentArea(area);
}


LogStream& _SetTemporaryArea(LogStream& ls, const LogAreaName& area)
{
  return ls.setTemporaryArea(area);
}


LogStream& _StopLogFileArea(LogStream& ls, const LogAreaName& area)
{
  return ls.stopLogFileArea(area);
}


LogStream& _StartLogFileArea(LogStream& ls, const LogAreaName& area)
{
  return ls.startLogFileArea(area);
}


LogStream& _StopLogScreenArea(LogStream& ls, const LogAreaName& area)
{
  return ls.stopLogScreenArea(area);
}


LogStream& _StartLogScreenArea(LogStream& ls, const LogAreaName& area)
{
  return ls.startLogScreenArea(area);
}

#endif // DOXYGEN_SHOULD_SKIP_THIS


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "LogStream.icc"
#endif
