/* -*- Mode: C++ -*-
 * $Id: Socket.cc,v 1.1 2000/09/01 11:43:42 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Socket.cc
 * Unitname: Library
 * $Revision: 1.1 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/01 11:43:42 $
 *
 *
 * HISTORY
 *
 * $Log: Socket.cc,v $
 * Revision 1.1  2000/09/01 11:43:42  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file Socket.cc
 * Implementation file for the class Socket.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2000/09/01 11:43:42 $
 */


#include "Socket.h"
#include "rs_error.h"
#include "rs_throw.h"


RS_BEGIN_NAMESPACE


Socket::Socket(const std::string& address,
	       const UInt16 p,
	       const UInt32 max_msg_size,
	       const SocketType sock_type)
  throw()
  : addr(address),
    port(p),
    maxMessageSize(max_msg_size),
    socketType(sock_type),
    connected(false),
    sockFD(-1),
    theErrorMessage(SOCKET_NO_ERROR),
    hostEnt(NULL),
    addrPtr(NULL)//,
    //clientAddr(),
    //serverAddr()
{
}

Socket::Socket(const Socket& obj)
  throw()
  : addr(obj.addr),
    port(obj.port),
    maxMessageSize(obj.maxMessageSize),
    socketType(obj.socketType),
    connected(obj.connected),
    sockFD(obj.sockFD),
    theErrorMessage(obj.theErrorMessage),
    hostEnt(obj.hostEnt),
    addrPtr(obj.addrPtr),
    clientAddr(obj.clientAddr),
    serverAddr(obj.serverAddr)
{
}

Socket::~Socket()
  throw(RSException)
{
  if (connected)
    close();
}

//int Socket::setFD(const int fd)
//{
//  int oldfd = sockFD;
//  sockFD = fd;
//  return oldfd;
//}

int Socket::getFlags()
  throw(RSException)
{
  int tmp_flags = fcntl(sockFD, F_GETFL, 0);
  if (tmp_flags < 0) {
    setErrorMessage(errno);
    RS_THROW(SocketException, "Socket [getFlags]: " + errorMessage(errno));
    return -1;
  } else
    return tmp_flags;
}

bool Socket::setFlags(const int flags)
  throw(RSException)
{
  if (fcntl(sockFD, F_SETFL, flags) < 0) {
    setErrorMessage(errno);
    RS_THROW(SocketException, "Socket [setFlags]: " + errorMessage(errno));
    return false;
  } else
    return true;
}

int Socket::getOwner()
  throw(RSException)
{
  int tmp_owner = fcntl(sockFD, F_GETOWN, 0);
  if (tmp_owner < 0) {
    setErrorMessage(errno);
    RS_THROW(SocketException, "Socket [getOwner]: " + errorMessage(errno));
    return -1;
  } else
    return tmp_owner;
}

bool Socket::setOwner(const int owner)
  throw(RSException)
{
  if (fcntl(sockFD, F_SETOWN, owner) < 0) {
    setErrorMessage(errno);
    RS_THROW(SocketException, "Socket [setOwner]: " + errorMessage(errno));
    return false;
  } else
    return true;
}

bool Socket::open()
  throw(RSException)
{
  // I am pretty sure that this code will NOT work with a numerical address...
  if ((hostEnt = (struct hostent *)gethostbyname(addr.c_str())) == NULL) {
    /* Check if numeric address */
    if ((int)inet_addr(addr.c_str()) == -1) {
      RS_ERROR("Socket [open]: Illegal address " + addr);
      theErrorMessage = SOCKET_ILLEGAL_ADDRESS;
      return false;
    }
  } else {
    addrPtr = (struct in_addr *)*hostEnt->h_addr_list;
    addr = inet_ntoa(*addrPtr);
  }

  switch (socketType) {
  case SOCKET_TYPE_TCP :
    if ((sockFD = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      RS_ERROR("Socket [open]: Couldn't open TCP socket " + errorMessage(errno));
      theErrorMessage = SOCKET_OPEN_FAILED;
      return false;
    }
    break;
  case SOCKET_TYPE_UDP :
    if ((sockFD = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
      RS_ERROR("Socket [open]: Couldn't open UDP socket " + errorMessage(errno));
      theErrorMessage = SOCKET_OPEN_FAILED;
      return false;
    }
    break;
  default:
    RS_ERROR("Socket [open]: Unknown socket type");
    theErrorMessage = SOCKET_UNKNOWN_TYPE;
    return false;
  }

  /*
   *  Bind any local address.
   */
  memset((char *) &clientAddr, 0, sizeof(clientAddr));
  clientAddr.sin_family      = AF_INET;
  // since INADDR_ANY is an ugly macro I can't just add RS_NAMESPACE::
  // since then the macro will expand into something unknown
  // for the compiler, since that's in RS!
  clientAddr.sin_addr.s_addr = htonl(INADDR_ANY);
  clientAddr.sin_port	      = htons(0);

  if (bind(sockFD, (struct sockaddr *) &clientAddr,
	   sizeof(clientAddr)) < 0) {
    RS_ERROR("Socket [open]: " + errorMessage(errno));
    theErrorMessage = SOCKET_BIND_FAILED;
    return false;  /* Can't bind local address */
  }

  /*
   *  Fill in the address of the server.
   */
  memset((char *) &serverAddr, 0, sizeof(serverAddr));
  serverAddr.sin_family      = AF_INET;
  serverAddr.sin_addr.s_addr = inet_addr(addr.c_str());
  serverAddr.sin_port	      = htons(port);

  connected = true;
  
  return true;
}  

bool Socket::close()
  throw(RSException)
{
  if ( connected ) {
    if (::close(sockFD) < 0)
      {
	RS_ERROR("Socket [close]: " + errorMessage(errno));
	theErrorMessage = SOCKET_CLOSE_FAILED;
	return false;
      }
    
    sockFD = -1;
    connected = false;
  }

  return true;
}


bool Socket::writeData(const std::string& data, const Int32 size)
  throw(RSException)
{

  if (!connected) {
    theErrorMessage = SOCKET_NO_CONNECTION;
    return false;
  }
  
  if (data == "") {
    RS_ERROR("Socket [write]: Tried to write an empty string");
    theErrorMessage = SOCKET_NO_DATA;
    return false;
  }

  //  b = write(sockFD, data, size);  // Old code, don't really know the difference!
 
  int b = sendto(sockFD, data.c_str(), size, 0,
		 (struct sockaddr *)&serverAddr, sizeof(serverAddr));

  if (b < 0) {
    RS_ERROR("Socket [write]: " + errorMessage(errno));
    theErrorMessage = SOCKET_WRITE_FAILED;
    return false;
  } else if ( b != size ) {
    RS_ERROR("Socket [write]: Not all of the data came through");
    theErrorMessage = SOCKET_WRITE_FAILED;
    return false;
  }
  
  return true;
}

bool Socket::readData(char* const data, Int32* size)
  throw(RSException)
{

  /* OLD CODE 
  
  if ((*size = read(sockFD, data, maxdata)) < 1)
    {
      if ((errno == EWOULDBLOCK) || (*size==0))
	{
	  *size=0;
	  return SOCKET_NODATA;
	}
      else
	{
	  perror("Socket [read]");
	  return SOCKET_ERROR;
	}
    }
  */

  if (!connected) {
    theErrorMessage = SOCKET_NO_CONNECTION;
    return false;
  }

  socklen_t servlen;
  struct sockaddr_in servAddr;
  
  servlen = sizeof(servAddr);
  *size = recvfrom(sockFD, data, maxMessageSize, 0,
		   (struct sockaddr *)&servAddr, &servlen);

  if (*size < 0) {
    if ( *size == -1 && errno == EWOULDBLOCK) {
      //RS_WARNING("Socket [read]: Socket marked non-blocking and operation would block");
      data[0] = '\0';      
      theErrorMessage = SOCKET_NO_DATA;
      return false;
    } else {
      RS_ERROR("Socket [read]: " + errorMessage(errno));
      theErrorMessage = SOCKET_READ_FAILED;
      return false;
    }
  } else {
    data[*size] = '\0';

    serverAddr.sin_port = servAddr.sin_port;

    if (*size == 0) {
      RS_WARNING("Socket [read]: Read an empty string");
      theErrorMessage = SOCKET_NO_DATA;
      return false;
    } else
      return true;
  }
}

void Socket::setErrorMessage(int error_no)
  throw()
{
  switch (error_no) {
  case EBADF :
    theErrorMessage = SOCKET_FD_ERROR;
    break;
  case EINVAL :
    theErrorMessage = SOCKET_INVALID_ARGUMENT;
    break;
  case EIO :
    theErrorMessage = SOCKET_IO_ERROR;
    break;
#ifdef ENOLINK
  case ENOLINK :
    theErrorMessage = SOCKET_NO_LINK;
    break;
#endif
  case EOVERFLOW :
    theErrorMessage = SOCKET_OVERFLOW;
    break;
  default :
    theErrorMessage = SOCKET_UNKNOWN_ERROR;
  }
}

const std::string Socket::errorMessage(int error_no) const
  throw()
{
  switch (error_no) {
  case EBADF :
    return "Bad file descriptor";
    break;
  case EINVAL :
    return "Invalid argument";
    break;
  case EIO :
    return "I/O error";
    break;
#ifdef ENOLINK
  case ENOLINK :
    return "No link to file descriptor";
    break;
#endif
  case EOVERFLOW :
    return "Return value overflow";
    break;
  case EADDRINUSE :
    return "Address already in use";
    break;
  case EADDRNOTAVAIL :
    return "Address not available";
    break;
  case EACCES :
    return "Don't have access to the address";
    break;
  case ENOTSOCK :
    return "File descriptor is no socket";
    break;
#ifdef ENOSR
  case ENOSR :
    return "Insufficient STREAMS resources to complete request";
    break;
#endif
  case EMFILE :
    return "File descriptor table is full";
    break;
  case ENOMEM :
    return "No memory";
    break;
  case EPROTONOSUPPORT :
    return "Protocol is not supported";
    break;
  case EINTR :
    return "Function was interrupted by a signal";
    break;
  case ENOSPC :
    return "No free space remaining";
    break;
  case EMSGSIZE :
    return "Message too long to be sent atomically";
    break;
  case EWOULDBLOCK :
    return "Socket marked non-blocking and operation would block";
    break;
  default :
    return "Unknown error";
  }
}


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "Socket.icc"
#endif
