/* -*- Mode: C++ -*-
 * $Id: Step.cc,v 1.5 2001/06/19 14:01:50 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Step.cc
 * Unitname: Library
 * $Revision: 1.5 $
 * Created by: Fredrik Heintz 2000-09-03
 * Last modified by $Author: frehe $ $Date: 2001/06/19 14:01:50 $
 *
 *
 * HISTORY
 *
 * $Log: Step.cc,v $
 * Revision 1.5  2001/06/19 14:01:50  frehe
 * Made minor changes to compile RoboSoc with gcc-3.0
 *
 * Revision 1.4  2000/10/25 19:33:42  frehe
 * Fixed Step to check for empty steps and NULL commands.
 *
 * Revision 1.3  2000/09/05 21:09:39  frehe
 * Fixed a bug in the copying of steps.
 *
 * Revision 1.2  2000/09/05 11:08:05  frehe
 * Improved the default constructor.
 *
 * Revision 1.1  2000/09/03 13:35:58  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file Step.cc
 * Implementation file for the class Step.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2001/06/19 14:01:50 $
 */


#include <algorithm>
#include <functional>
#include "Step.h"
#include "ServerCommand.h"
#include "rs_assert.h"


RS_BEGIN_NAMESPACE


struct add_command : public std::binary_function<Step*, const Command*, void> {
  void operator()(Step* step, const Command* const command) const
  {
    step->addCommand(command);
  }
};


// --- CONSTRUCTORS ---

Step::Step(const Command* const command)
  : commandSequence()
{
  if ( command != NULL )
    addCommand(command);
  RS_CHECK_INVARIANT;
}


Step::Step(const Step& obj)
  : commandSequence()
{
  std::for_each(obj.commandSequence.begin(), obj.commandSequence.end(),
		  bind1st(add_command(), this));
  RS_CHECK_INVARIANT;
}


// --- DESTRUCTOR ---

Step::~Step()
{
  RS_CHECK_INVARIANT;
  // Clean up the old data
  makeEmpty();
}


// --- OPERATORS ---

Step& Step::operator=(const Step& obj)
{
  RS_CHECK_INVARIANT;
  if ( &obj != this ) {
    // Clean up the old data
    makeEmpty();
    // Copy the new data from obj
    //commandSequence = obj.commandSequence;
    for_each(obj.commandSequence.begin(),
	     obj.commandSequence.end(),
	     bind1st(add_command(), this));
  }
  RS_CHECK_INVARIANT;
  return *this;
}


// --- OTHER METHODS ---

Command* Step::getNextCommand() const
{
  RS_CHECK_INVARIANT;
  Command* cmd = NULL;
  if ( !empty() ) {
    cmd = commandSequence.front()->clone();
  }
  RS_CHECK_INVARIANT;
  return cmd;
}


Step& Step::removeNextCommand()
{
  RS_CHECK_INVARIANT;
  if ( !empty() ) {
    Command* cmd = commandSequence.front();
    commandSequence.pop_front();
    delete cmd;
  }
  RS_CHECK_INVARIANT;
  return *this;
}


Step& Step::makeEmpty()
{
  RS_CHECK_INVARIANT;
  while ( !empty() ) {
    removeNextCommand();
  }
  RS_CHECK_INVARIANT;
  return *this;
}


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "step.icc"
#endif
