/* -*- Mode: C++ -*-
 * $Id: Time.cc,v 1.1 2000/09/01 01:24:47 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Time.cc
 * Unitname: Library
 * $Revision: 1.1 $
 * Created by: Fredrik Heintz 2000-xx-xx
 * Last modified by $Author: frehe $ $Date: 2000/09/01 01:24:47 $
 *
 *
 * HISTORY
 *
 * $Log: Time.cc,v $
 * Revision 1.1  2000/09/01 01:24:47  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 * Revision 1.1.1.1  2000/08/31 13:26:04  frehe
 * Importet the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file Time.cc
 * Implementation of the class Time.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.1 $
 * $Author: frehe $
 * $Date: 2000/09/01 01:24:47 $
 */


#include "Time.h"
#include "string_utils.h"
#include "rs_error.h"


#ifdef RS_WARNING
#  define TIME_WARNING( str ) RS_WARNING( std::string("Time: ") + str )
#else
#  define TIME_WARNING( str )
#endif

#ifdef RS_ERROR
#  define TIME_ERROR( str ) RS_ERROR( std::string("Time: ") + str )
#else
#  define TIME_ERROR( str )
#endif


RS_BEGIN_NAMESPACE


bool Time::clockStopped = false;
Time Time::currentTime = Time(0,0);
Time Time::lastStopClockTime = Time(0,0);
Time Time::secondLastStopClockTime = Time(0,0);


Time::Time(const UInt16 vt, const UInt16 vs)
  : t(vt),
    s(vs)
{
}


Time::Time(const Time& obj)
  : t(obj.t),
    s(obj.s)
{
}


// What to do about the connection between Time and
// the actual starting and stopping of the game?
// Does Time have to depend on something else?
// Another alternative (from RoboLog) is to define currentTime,
// lastStopClockTime and secondLastStopClockTime
// as global or class variables. Not pretty, but solves the problem...
// The current solution is that Time is separate from the rest of
// the program, but that CBasicIP and CTheGameObject have references
// to the Time class variables.
// The update of the time and tracking of the clock will be done in
// CBasicIP.

Time Time::operator-(const UInt16 a) const
{ 
  if ( s==0 && t-a > lastStopClockTime.t ) // default case
    return Time(t-a,0);

  if ( s>0 ) { 
    if ( a <= s ) 
      return Time(t,s-a); // Just take off from stopped time 
    else { // a > s 
      Time new_time = Time( t-(a-s),0 );
      // take off from stopped time, then server time 
      if ( new_time < lastStopClockTime ) 
	TIME_WARNING("can't be sure of this subtraction (1) " + toString(*this)
		      + " - " + toString(a));
      return new_time;
    }
  } else { // t-a <= lastStopClockTime.t 
    UInt16 stopped_time =  a - (t - lastStopClockTime.t);  
    if ( stopped_time > lastStopClockTime.s ) {
      if ( !lastStopClockTime.t )
	// if == 0, OK--account for players starting at different times 
	return Time(0,0);
      UInt16 tmp = lastStopClockTime.t - (stopped_time - lastStopClockTime.s);
      if ( tmp <= secondLastStopClockTime.t ) 
	TIME_WARNING("can't be sure of this subtraction (2) " + toString(*this) + " - "
		      + toString(a) + "  " + toString(lastStopClockTime));
      return Time( tmp, 0 );
    }
    return Time( lastStopClockTime.t, lastStopClockTime.s - stopped_time );
  }
}


UInt16 Time::operator -(const Time& a) const
{
  if ( s==0 ) {
    if ( a.t < lastStopClockTime.t ) {
      if ( a.t <= secondLastStopClockTime.t )
	TIME_WARNING("Can't be sure of this subtraction (3) " + toString(*this)
		      + " - " + toString(a));
      return (t - a.t) + lastStopClockTime.s;
    }
    if ( a.t > lastStopClockTime.t )
      return (t - a.t);
    if ( a.t == lastStopClockTime.t )
      return (t - a.t) + (lastStopClockTime.s - a.s);
  } else if ( s > 0 ) {
    if ( a.t < secondLastStopClockTime.t )  // If they're equal, it's OK 
      TIME_WARNING("Can't be sure of this subtraction (4) " + toString(*this)
		  + " - " + toString(a));
    if ( a.t < lastStopClockTime.t )
      return lastStopClockTime.s + (t - a.t);
    else if ( a.t == lastStopClockTime.t && a.t != t )
      // a is during the last stopped interval 
      return ( s + ( lastStopClockTime.s - a.s ) ) + (t - a.t);
    return (s - a.s) + (t - a.t);
  } else // s<0 
    TIME_ERROR("s shouldn't be <0");
  return 0;
}


Time Time::operator +(const UInt16 a) const
{
  
  if ( s==0 && t > lastStopClockTime.t && t+a < currentTime.t )
    // default case 
    return Time(t+a,s);

  Time new_time;
  
  if ( s==0 ) {
    if ( lastStopClockTime.t > t ) { // Could've missed one already 
      TIME_WARNING("Can't be sure of this addition (1) " + toString(*this)
		    + " + " + toString(a) + "  "
		    + toString(lastStopClockTime));
      new_time = Time(t+a,0);
    }
    if ( t+a < currentTime.t )
      new_time = Time( t+a,0 );
    else  // t+a >= currentTime.t 
      new_time = Time( currentTime.t, a-(currentTime.t-t) );
  } else if ( s>0 ) {
    if ( t == currentTime.t ) // clock still stopped 
      new_time = Time( t,s+a );
    else {
      if ( lastStopClockTime.t != t ) // Could've missed one already 
	TIME_WARNING("Can't be sure of this addition (2)" + toString(*this)
		      + " + "  + toString(a));
      new_time = Time ( t+(a-(lastStopClockTime.s - s)),0 );
    }
  } else // s<0 
    TIME_ERROR("s shouldn't be <0");

  if ( new_time > currentTime ) // clock might stop 
    TIME_WARNING("Can't be sure of this addition (3)" + toString(*this)
		  + " + " + toString(a));

  return new_time;
}


Time& Time::operator=(const UInt16 a)
{
  t = a;
  s = 0;
  return *this;
}


Time& Time::operator=(const Time& a)
{
  if ( this != &a ) {
    t = a.t;
    s = a.s;
  }
  return *this;
}


std::istream& operator>>(std::istream& is, Time& ct)
{
  /* Input formats: i is an UInt16
     i     = (i,0)
     (i)   = (i,0)
     (i,i) = (i,i)
     */

  UInt16 t = 0;
  UInt16 s = 0;
  char c = 0;

  is >> c;
  if (c == '(') {
    is >> t >> c;
    if (c == ',') is >> s >> c;
    if (c != ')') is.clear(std::ios::badbit); //error...;
  }
  else {
    is.putback(c);
    is >> t;
  }

  if (is) ct = Time(t, s);

  return is;
}


const std::string toString(const Time& obj)
{
#ifdef TIME_HEAD
  return TIME_HEAD + "(" + toString(obj.getT()) + ", " + toString(obj.getS()) + ")";
#else
  return "(" + toString(obj.getT()) + ", " + ::toString(obj.getS()) + ")";
#endif
}


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "Time.icc"
#endif
