/**
 * @file SkillDependencies.h
 *
 * This file will be used to keep track
 * of the current dependencies between the skills
 * in the RoboSoc system.
 *
 */


/** Base class for all skills. */
class Skill {};


/** Base class for all with arguments. */
class SkillWithArguments : public Skill {};

/** Base class for all skills using a direction
    as a parameter. */
class SkillWithDirection : public SkillWithArguments {};


/** Base class for all skills using a position
    as a parameter. */
class SkillWithPosition : public SkillWithArguments {};


/** Base class for all skills using a speed
    as a parameter. */
class SkillWithSpeed : public SkillWithArguments {};


/** Base class for all skills using a vector
    as a parameter. */
class SkillWithVector : public SkillWithArguments {};


/** Base class for all skills using a zone
    (circular) as parameter. */
class SkillWithZone : public SkillWithArguments {
  MoveToPositionFacingBall gotop;
};

/** Skill to turn the agent to a direction. */
class TurnToDirection : public SkillWithDirection {
};


/** Skill to move the agent to a position Point(X, Y). */
class MoveToPosition : public SkillWithPosition {
  TurnToDirection turnToDirection;
};


/** Skill to move the agent to a position Point(X, Y) while
    avoiding obstacles. */
class MoveToPositionAvoidingObstacles : public SkillWithPosition {
  MoveToPosition moveToPosition;
};


/** Skill used by an agent to defend a zone. */
class ZoneDefence : public SkillWithZone {
  ZoneDefenceDefender zoneDefenceDefender;
  ZoneDefenceToDefend zoneDefenceToDefend;
  ZoneDefenceRest zoneDefenceRest;
  ZoneDefencePosition zoneDefencePosition;
};

/** Skill to defend a zone from inside it. */
class ZoneDefenceDefender : public SkillWithZone {
  InterceptBallStamina interceptball;
  GotoBall gotoball;
  FindBall findball;
  TurnDirection turn;
};

/** Skill to place the agent in a good position for
    defending a zone. */
class ZoneDefencePosition : public SkillWithZone {
  MoveToPositionFacingBall moveToPosition;
}

/** Skill to prepare a player for defending a zone,
    ie move to it :) */
class ZoneDefenceToDefend : public SkillWithZone {
  ZoneDefencePosition positionate;
}

/** Let a defender rest if there is nothing to defend
    in the zone. Will keep an eye on the ball and stuff
    like that */
class ZoneDefenceRest : public SkillWithZone {
}

/** Skill used to get to the ball as fast as possible. */
class InterceptBall : public Skill {
};


/** Skill used to get to the ball as fast as possible not going below
    a certain level of stamina. */
class InterceptBallStamina : public SkillWithArguments {
  MoveToPositionFacingBall   moveToPositionFacingBall;
};

/** skill used to put the ball in the opponent goal. */
class Score : public Skill {
};


/** Skill used to take the ball to a position Point(X, Y) without
    thinking about other players. */
class MoveToPositionWithBall : public SkillWithPosition {
};


/** Skill used to take the ball to a position Point(X, Y) and
    avoiding other players. */
class DribbleToPosition : public SkillWithPosition {
  GotoBall gotoBall;
};


/** Skill used to take the ball in a direction without
    taking other players in consideration. */
class MoveInDirectionWithBall : public SkillWithDirection {
};


/** Skill used to take the ball in a direction and
    avoiding other players. */
class DribbleInDirection : public SkillWithDirection {
  MoveInDirectionWithBall moveInDirectionWithBall;
};


/** Skill used by a player to cut off another player's
    pass angles. */
class CutOffPassAngle : public Skill {
};


/** Skill used by a player to cut off another player's
    score angles. */
class CutOffScoreAngle : public Skill {
};


/** Skill used by a goal keeper to catch the ball. */
class CatchBall : public Skill {
};


/** Skill used by a player to pass the ball to another
    player. */
class Pass : public Skill {
};


/** Skill used by a player to find out where the ball is. */
class FindBall : public Skill {
};


/** Skill used by a player to face a direction. */
class FaceDirection : public SkillWithDirection {
};


/** Skill used by a player to dash until it reach a
    certain speed. */
class DashToSpeed : public SkillWithSpeed {
};


/** Skill used by a player to dash until it reach a
    certain speed vector. */
class DashToSpeedVector : public SkillWithVector {
};


/** Skill used by a player to kick the ball in a direction. */
class KickBallInDirection : public SkillWithDirection {
};


/** Skill used by a player to kick the ball to a position Point(X,Y). */
class KickBallToPosition : public SkillWithPosition {
};


/** Skill used by a player to kick the ball to a certain speed vector. */
class KickBallToSpeedVector : public SkillWithVector {
};

class TurnDirection : public SkillWithArguments {};

class FaceDirection : public SkillWithArguments {};

class DashToSpeedVectorFacingDirection : public SkillWithArguments {
  TurnDirection turn;
  FaceDirection face;
};

class MoveToPositionFacingDirection : public DashToSpeedVectorFacingDirection {
};

class MoveToPositionFacingDirectionAvoidingObstacles : MoveToPositionFacingDirection {
  MoveToPositionFacingDirection subgoto;
};

class MoveToPositionFacingBall : public MoveToPositionFacingDirectionAvoidingObstacles {};
